@@ -11,7 +11,7 @@
 #define SHELL_Q '"'
 #define DL_DFLT_NAME "libpari.dll"
 
-#define PARIVERSION "GP/PARI CALCULATOR Version 2.2.2 (alpha)"
+#define PARIVERSION "GP/PARI CALCULATOR Version 2.2.3 (alpha)"
 #ifdef __cplusplus
 #  define PARIINFO_PRE "C++"
 #else
@@ -28,7 +28,7 @@
 #endif
 #define PARIINFO PARIINFO_PRE/**/PARIINFO_POST
 
-#define PARI_BYTE_ORDER    1234
+#define PARI_DOUBLE_FORMAT 1
 #ifdef _MSC_VER /* MSVC inline directive */
 #  define INLINE __inline
 #endif
@@ -65,8 +65,8 @@
 /* Timings: Don't use times because of the HZ / CLK_TCK confusion. */
 #  define USE_FTIME 1
 #  define HAS_STRFTIME
- /* comment this one if not on NT ??? */
-#  define HAS_OPENDIR
+/* This might work on NT ??? */
+/* #  define HAS_OPENDIR */
 #endif
 
 #endif
