@@ -1,4 +1,4 @@
-/* $Id: arith1.c,v 1.69 2002/01/09 19:13:44 karim Exp $
+/* $Id: arith1.c,v 1.98 2002/06/08 14:02:28 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -144,7 +144,7 @@ gassoc_proto(GEN f(GEN,GEN), GEN x, GEN y)
   int tx=typ(x);
   if (!y)
   {
-    ulong av=avma;
+    gpmem_t av = avma;
     if (tx!=t_VEC && tx!=t_COL)
       err(typeer,"association");
     return gerepileupto(av,divide_conquer_prod(x,f));
@@ -161,7 +161,8 @@ gassoc_proto(GEN f(GEN,GEN), GEN x, GEN y)
 GEN
 order(GEN x)
 {
-  long av=avma,av1,i,e;
+  gpmem_t av = avma,av1;
+  long i,e;
   GEN o,m,p;
 
   if (typ(x) != t_INTMOD || !gcmp1(mppgcd((GEN) x[1],(GEN) x[2])))
@@ -196,14 +197,15 @@ ggener(GEN m)
 GEN
 gener(GEN m)
 {
-  long av=avma,av1,k,i,e;
+  gpmem_t av=avma,av1;
+  long k,i,e;
   GEN x,t,q,p;
 
   if (typ(m) != t_INT) err(arither1);
   e = signe(m);
   if (!e) err(talker,"zero modulus in znprimroot");
-  if (is_pm1(m)) { avma=av; return gmodulss(0,1); }
-  if (e<0) m = absi(m);
+  if (is_pm1(m)) return gmodulss(0,1);
+  if (e < 0) m = absi(m);
 
   e = mod4(m);
   if (e == 0) /* m = 0 mod 4 */
@@ -220,33 +222,54 @@ gener(GEN m)
 
   t=decomp(m); if (lg(t[1]) != 2) err(generer);
   p=gcoeff(t,1,1); e=itos(gcoeff(t,1,2)); q=subis(p,1);
-  if (e>=2)
+  if (e >= 2)
   {
-    x = (GEN) gener(p)[2];
+    x = (GEN)gener(p)[2];
     if (gcmp1(powmodulo(x,q, sqri(p)))) x = addii(x,p);
     av1=avma; return gerepile(av,av1,gmodulcp(x,m));
   }
 
   p=(GEN)decomp(q)[1]; k=lg(p)-1; x=stoi(1);
+  for (i=1; i<=k; i++) p[i] = (long)diviiexact(q, (GEN)p[i]);
   for(;;)
   {
     x[2]++;
     if (gcmp1(mppgcd(m,x)))
     {
       for (i=k; i; i--)
-	if (gcmp1(powmodulo(x, divii(q,(GEN)p[i]), m))) break;
+	if (gcmp1(powmodulo(x, (GEN)p[i], m))) break;
       if (!i) break;
     }
   }
   av1=avma; return gerepile(av,av1,gmodulcp(x,m));
 }
 
+/* assume p odd SMALL prime */
+ulong
+u_gener(ulong p)
+{
+  const gpmem_t av = avma;
+  const long q = p - 1;
+  const GEN L = (GEN)decomp(stoi(q))[1];
+  const int k = lg(L) - 1;
+  int i,x;
+
+  for (x=2;;x++)
+    if (x % p)
+    {
+      for (i=k; i; i--)
+	if (powuumod(x, q/itos((GEN)L[i]), p) == 1) break;
+      if (!i) break;
+    }
+  avma = av; return x;
+}
+
 GEN
 znstar(GEN n)
 {
   GEN p1,z,q,u,v,d,list,ep,h,gen,moduli,p,a;
   long i,j,nbp,sizeh;
-  ulong av;
+  gpmem_t av;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n))
@@ -339,7 +362,8 @@ gracine(GEN a)
 static GEN
 racine_r(GEN a, long l)
 {
-  long av,s;
+  gpmem_t av;
+  long s;
   GEN x,y,z;
 
   if (l <= 4) return utoi(mpsqrtl(a));
@@ -355,20 +379,21 @@ racine_r(GEN a, long l)
     y = x; x = z;
   }
   while (cmpii(x,y) < 0);
-  avma = (long)y;
+  avma = (gpmem_t)y;
   return gerepileuptoint(av,y);
 }
 
 GEN
 racine_i(GEN a, int roundup)
 {
-  long k,m,l = lgefint(a), av = avma;
+  gpmem_t av = avma;
+  long k,m,l = lgefint(a);
   GEN x = racine_r(a, l);
   if (roundup && signe(x))
   {
     m = modBIL(x);
     k = (m * m != a[l-1] || !egalii(sqri(x),a));
-    avma = (long)x;
+    avma = (gpmem_t)x;
     if (k) x = gerepileuptoint(av, addis(x,1));
   }
   return x;
@@ -433,7 +458,7 @@ ucarrecomplet(ulong A)
 long
 carrecomplet(GEN x, GEN *pt)
 {
-  long av;
+  gpmem_t av;
   GEN y;
 
   switch(signe(x))
@@ -451,14 +476,15 @@ carrecomplet(GEN x, GEN *pt)
   if (!carremod((ulong)smodis(x, 64*63*65*11))) return 0;
   av=avma; y = racine(x);
   if (!egalii(sqri(y),x)) { avma=av; return 0; }
-  if (pt) { *pt = y; avma=(long)y; } else avma=av; 
+  if (pt) { *pt = y; avma=(gpmem_t)y; } else avma=av; 
   return 1;
 }
 
 static GEN
 polcarrecomplet(GEN x, GEN *pt)
 {
-  long i,l,av,av2;
+  gpmem_t av,av2;
+  long i,l;
   GEN y,a,b;
 
   if (!signe(x)) return gun;
@@ -517,8 +543,9 @@ gcarrecomplet(GEN x, GEN *pt)
 GEN
 gcarreparfait(GEN x)
 {
+  gpmem_t av;
   GEN p1,a,p;
-  long tx=typ(x),l,i,av,v;
+  long tx=typ(x),l,i,v;
 
   switch(tx)
   {
@@ -637,10 +664,10 @@ gkronecker(GEN x, GEN y)
 long
 kronecker(GEN x, GEN y)
 {
+  const gpmem_t av = avma;
   GEN z;
-  long av,r,s=1;
+  long r,s=1;
 
-  av=avma;
   switch (signe(y))
   {
     case -1: y=negi(y); if (signe(x)<0) s= -1; break;
@@ -681,9 +708,9 @@ gkrogs(GEN x, long y)
 long
 krogs(GEN x, long y)
 {
-  long av,r,s=1,x1,z;
+  const gpmem_t av = avma;
+  long r,s=1,x1,z;
 
-  av=avma;
   if (y<=0)
   {
     if (y) { y = -y; if (signe(x)<0) s = -1; }
@@ -717,7 +744,8 @@ krogs(GEN x, long y)
 long
 krosg(long s, GEN x)
 {
-  long av = avma, y = kronecker(stoi(s),x);
+  gpmem_t av = avma;
+  long y = kronecker(stoi(s),x);
   avma = av; return y;
 }
 
@@ -766,7 +794,8 @@ hil0(GEN x, GEN y, GEN p)
 long
 hil(GEN x, GEN y, GEN p)
 {
-  long a,b,av,tx,ty,z;
+  gpmem_t av;
+  long a,b,tx,ty,z;
   GEN p1,p2,u,v;
 
   if (gcmp0(x) || gcmp0(y)) return 0;
@@ -863,16 +892,28 @@ hil(GEN x, GEN y, GEN p)
 
 #define sqrmod(x,p) (resii(sqri(x),p))
 
-/* Assume p is prime. return NULL if (a,p) = -1 */
+static GEN ffsqrtmod(GEN a, GEN p);
+
+/* Tonelli-Shanks. Assume p is prime and return NULL if (a,p) = -1. */
 GEN
 mpsqrtmod(GEN a, GEN p)
 {
-  long av = avma, av1,lim,i,k,e;
+  gpmem_t av = avma, av1,lim;
+  long i,k,e;
   GEN p1,q,v,y,w,m;
 
   if (typ(a) != t_INT || typ(p) != t_INT) err(arither1);
   if (signe(p) <= 0 || is_pm1(p)) err(talker,"not a prime in mpsqrtmod");
   p1 = addsi(-1,p); e = vali(p1);
+  
+  /* If `e' is "too big", use Cipolla algorithm ! [GTL] */
+  if (e*(e-1) > 20 + 8 * bit_accuracy(lgefint(p)))
+  {
+    v = ffsqrtmod(a,p);
+    if (!v) { avma = av; return NULL; }
+    return gerepileuptoint(av,v);
+  }
+
   if (e == 0) /* p = 2 */
   {
     avma = av;
@@ -930,6 +971,113 @@ mpsqrtmod(GEN a, GEN p)
   return gerepileuptoint(av, v);
 }
 
+/* Cipolla's algorithm is better when e = v_2(p-1) is "too big".
+ * Otherwise, is a constant times worse than the above one.
+ * For p = 3 (mod 4), is about 3 times worse, and in average
+ * is about 2 or 2.5 times worse.
+ * 
+ * But try both algorithms for e.g. S(n)=(2^n+3)^2-8 with
+ * n = 750, 771, 779, 790, 874, 1176, 1728, 2604, etc.
+ *
+ * If X^2 = t^2 - a  is not a square in F_p, then
+ * 
+ *   (t+X)^(p+1) = (t+X)(t-X) = a
+ *    
+ * so we get sqrt(a) in F_p^2 by
+ * 
+ *   sqrt(a) = (t+X)^((p+1)/2)
+ *    
+ * If (a|p)=1, then sqrt(a) is in F_p.
+ *
+ * cf: LNCS 2286, pp 430-434 (2002)  [Gonzalo Tornaria] */
+static GEN  
+ffsqrtmod(GEN a, GEN p)
+{
+  gpmem_t av = avma, av1, lim;
+  long e, t, man, k, nb;
+  GEN q, n, u, v, d, d2, b, u2, v2, qptr;
+
+  if (kronecker(a, p) < 0) return NULL;
+  
+  av1 = avma;
+  for(t=1; ; t++)
+  {
+    n = subsi(t*t, a);
+    if (kronecker(n, p) < 0) break;
+    avma = av1;
+  }
+
+  u = stoi(t); v = gun; /* u+vX = t+X */
+  e = vali(addsi(-1,p)); q = shifti(p, -e);
+  /* p = 2^e q + 1  and  (p+1)/2 = 2^(e-1)q + 1 */
+
+  /* Compute u+vX := (t+X)^q */
+  av1 = avma; lim = stack_lim(av1, 1); 
+  qptr = q+2; man = *qptr;
+  k = 1+bfffo(man); man<<=k; k=BITS_IN_LONG-k;
+  for (nb=lgefint(q)-2;nb; nb--)
+  {
+    for (; k; man<<=1, k--)
+    {
+      if (man < 0)
+      { /* u+vX := (u+vX)^2 (t+X) */
+        d = addii(u, mulsi(t,v));
+        d2 = sqri(d);
+        b = resii(mulii(a,v), p);
+        u = modii(subii(mulsi(t,d2), mulii(b,addii(u,d))), p);
+        v = modii(subii(d2, mulii(b,v)), p);
+      }
+      else
+      { /* u+vX := (u+vX)^2 */
+        u2 = sqri(u);
+        v2 = sqri(v);
+        v = modii(subii(sqri(addii(v,u)), addii(u2,v2)), p);
+        u = modii(addii(u2, mulii(v2,n)), p);
+      }
+    }
+    
+    if (low_stack(lim, stack_lim(av1, 1)))
+    {
+       GEN *gptr[2]; gptr[0]=&u; gptr[1]=&v;
+       if (DEBUGMEM>1) err(warnmem, "ffsqrtmod");
+       gerepilemany(av1,gptr,2);
+    }
+    
+    man = *++qptr; k = BITS_IN_LONG;
+  }
+
+  while (--e)
+  { /* u+vX := (u+vX)^2 */
+    u2 = sqri(u);
+    v2 = sqri(v);
+    v = modii(subii(sqri(addii(v,u)), addii(u2,v2)), p);
+    u = modii(addii(u2, mulii(v2,n)), p);
+
+    if (low_stack(lim, stack_lim(av1, 1)))
+    {
+       GEN *gptr[2]; gptr[0]=&u; gptr[1]=&v;
+       if (DEBUGMEM>1) err(warnmem, "ffsqrtmod");
+       gerepilemany(av1,gptr,2);
+    }
+  }
+     
+  /* Now u+vX = (t+X)^(2^(e-1)q); thus
+   * 
+   *   (u+vX)(t+X) = sqrt(a) + 0 X
+   *    
+   * Whence,
+   * 
+   *   sqrt(a) = (u+vt)t - v*a
+   *   0       = (u+vt) 
+   *    
+   * Thus a square root is v*a */
+     
+  v = modii(mulii(v,a), p);
+
+  u = subii(p,v); if (cmpii(v,u) > 0) v = u;
+  return gerepileuptoint(av,v);
+}
+ 
 /*******************************************************************/
 /*                                                                 */
 /*                       n-th ROOT MODULO p                        */
@@ -949,13 +1097,13 @@ mpsqrtmod(GEN a, GEN p)
 static GEN
 mplgenmod(GEN l, long e, GEN r,GEN p,GEN *zeta)
 {
-  ulong av1;
+  const gpmem_t av1 = avma;
   GEN m,m1;
   long k,i; 
-  av1 = avma;
   for (k=1; ; k++)
   {
     m1 = m = powmodulo(stoi(k+1),r,p);
+    if (gcmp1(m)) {avma=av1; continue;}
     for (i=1; i<e; i++)
       if (gcmp1(m=powmodulo(m,l,p))) break;
     if (i==e) break;
@@ -979,11 +1127,11 @@ GEN Fp_shanks(GEN x,GEN g0,GEN p, GEN q);
 static GEN
 mpsqrtlmod(GEN a, GEN l, GEN p, GEN q,long e, GEN r, GEN y, GEN m)
 {
-  ulong av = avma, tetpil,lim;
+  gpmem_t av = avma, tetpil,lim;
   long k;
   GEN p1,u1,u2,v,w,z,dl;
   /* y contient un generateur de (Z/pZ)^* eleve a la puis (p-1)/(l^e) */
-  bezout(r,l,&u1,&u2);
+  (void)bezout(r,l,&u1,&u2);
   v=powmodulo(a,u2,p);
   w=powmodulo(a,modii(mulii(negi(u1),r),q),p);
   lim = stack_lim(av,1);
@@ -1034,7 +1182,7 @@ If a=0 ,return 0 and if zetan is not NULL zetan is set to gun
 GEN 
 mpsqrtnmod(GEN a, GEN n, GEN p, GEN *zetan)
 {
-  ulong ltop=avma,lbot=0,av1,lim;
+  gpmem_t ltop=avma,lbot=0,av1,lim;
   GEN m,u1,u2;
   GEN q,z;
   GEN *gptr[2];
@@ -1202,7 +1350,8 @@ ulong mppgcd_resiu(GEN y, ulong x);
 GEN
 mppgcd(GEN a, GEN b)
 {
-  long av,v,w;
+  long v, w;
+  gpmem_t av;
   GEN t, p1;
 
   if (typ(a) != t_INT || typ(b) != t_INT) err(arither1);
@@ -1223,7 +1372,7 @@ mppgcd(GEN a, GEN b)
   }
 
   /* larger than gcd: "avma=av" gerepile (erasing t) is valid */
-  av = avma; new_chunk(lgefint(b)); /* HACK */
+  av = avma; (void)new_chunk(lgefint(b)); /* HACK */
   t = resii(a,b);
   if (!signe(t)) { avma=av; return absi(b); }
 
@@ -1256,7 +1405,7 @@ mppgcd(GEN a, GEN b)
     }
   }
   {
-    long r[] = {evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
+    long r[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
     r[2] = (long) ugcd((ulong)b[2], (ulong)a[2]);
     avma = av; return shifti(r,v);
   }
@@ -1265,7 +1414,7 @@ mppgcd(GEN a, GEN b)
 GEN
 mpppcm(GEN x, GEN y)
 {
-  ulong av;
+  gpmem_t av;
   GEN p1,p2;
   if (typ(x) != t_INT || typ(y) != t_INT) err(arither1);
   if (!signe(x)) return gzero;
@@ -1311,8 +1460,8 @@ GEN chinese(GEN x, GEN y)
 GEN
 chinois(GEN x, GEN y)
 {
-  ulong av,tetpil, tx = typ(x);
-  long i,lx,vx;
+  gpmem_t av,tetpil;
+  long i,lx,vx, tx = typ(x);
   GEN z,p1,p2,d,u,v;
 
   if (gegal(x,y)) return gcopy(x);
@@ -1358,7 +1507,7 @@ chinois(GEN x, GEN y)
 GEN
 chinois_int_coprime(GEN x2, GEN y2, GEN x1, GEN y1, GEN z1)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN ax,p1;
   (void)new_chunk((lgefint(z1)<<1)+lgefint(x1)+lgefint(y1)); /* HACK */
   ax = mulii(mpinvmod(x1,y1), x1);
@@ -1386,7 +1535,9 @@ mpinvmod(GEN a, GEN m)
 /**                    MODULAR EXPONENTIATION                       **/
 /**                                                                 **/
 /*********************************************************************/
-GEN resiimul(GEN x, GEN y);
+extern ulong invrev(ulong b);
+extern GEN resiimul(GEN x, GEN y);
+
 static GEN _resii(GEN x, GEN y) { return resii(x,y); }
 
 /* Montgomery reduction */
@@ -1396,8 +1547,6 @@ typedef struct {
   ulong inv; /* inv = -N^(-1) mod B, */
 } montdata;
 
-extern ulong invrev(ulong b);
-
 void
 init_montdata(GEN N, montdata *s)
 {
@@ -1409,10 +1558,8 @@ GEN
 init_remainder(GEN M)
 {
   GEN sM = cgetg(3, t_VEC);
-  GEN Mr = cgetr(lgefint(M) + 1);
-  affir(M, Mr);
   sM[1] = (long)M;
-  sM[2] = (long)linv(Mr);
+  sM[2] = (long)linv( itor(M, lgefint(M) + 1) ); /* 1. / M */
   return sM;
 }
 
@@ -1486,7 +1633,7 @@ _sqr(void *data, GEN x)
 GEN
 powmodulo(GEN A, GEN k, GEN N)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long t,s, lN;
   int base_is_2, use_montgomery;
   GEN y;
@@ -1553,84 +1700,115 @@ powmodulo(GEN A, GEN k, GEN N)
 /**                                                                 **/
 /*********************************************************************/
 GEN
-gnextprime(GEN n)
-{
-  return garith_proto(nextprime,n,0);
-}
+gnextprime(GEN n) { return garith_proto(nextprime,n,0); }
 
 GEN
-gprecprime(GEN n)
-{
-  return garith_proto(precprime,n,0);
-}
+gprecprime(GEN n) { return garith_proto(precprime,n,0); }
 
 GEN
 gisprime(GEN x, long flag)
 {
   switch (flag)
   {
-  case 0:
-    return arith_proto(isprime,x,1);
-  case 1:
-    return garith_proto2gs(plisprime,x,0);
-  case 2:
-    return garith_proto2gs(plisprime,x,1);
+    case 0: return arith_proto(isprime,x,1);
+    case 1: return garith_proto2gs(plisprime,x,1);
+    case 2: return arith_proto(isprimeAPRCL,x,1);
   }
   err(flagerr,"gisprime");
   return 0;
 }
 
 long
-isprime(GEN x)
+isprimeSelfridge(GEN x) { return (plisprime(x,0)==gun); }
+
+/* assume x BSW pseudoprime. Check whether it's small enough to be certified
+ * prime */
+int
+BSW_isprime_small(GEN x)
 {
-  return millerrabin(x,10);
+  long l = lgefint(x);
+  if (l < 4) return 1;
+  if (l == 4)
+  {
+    gpmem_t av = avma;
+    long t = cmpii(x, u2toi(0x918UL, 0x4e72a000UL)); /* 10^13 */
+    avma = av;
+    if (t < 0) return 1;
+  }
+  return 0;
 }
 
-GEN
-gispsp(GEN x)
+/* assume x a BSW pseudoprime */
+int
+BSW_isprime(GEN x)
 {
-  return arith_proto(ispsp,x,1);
+  gpmem_t av = avma;
+  long l, res;
+  GEN F, p;
+ 
+  if (BSW_isprime_small(x)) return 1;
+  F = (GEN)auxdecomp(subis(x,1), 0)[1];
+  l = lg(F); p = (GEN)F[l-1];
+  if (BSW_psp(p))
+  { /* smooth */
+    GEN z = cgetg(3, t_VEC);
+    z[1] = (long)x;
+    z[2] = (long)F; res = isprimeSelfridge(z);
+  }
+  else
+    res = isprimeAPRCL(x);
+  avma = av; return res;
 }
 
 long
-ispsp(GEN x)
+isprime(GEN x)
 {
-  return millerrabin(x,1);
+  return BSW_psp(x) && BSW_isprime(x);
 }
 
 GEN
-gmillerrabin(GEN n, long k)
+gispseudoprime(GEN x, long flag)
 {
-  return arith_proto2gs(millerrabin,n,k);
+  if (flag == 0) return arith_proto(BSW_psp,x,1);
+  return gmillerrabin(x, flag);
 }
 
+long
+ispseudoprime(GEN x, long flag)
+{
+  if (flag == 0) return BSW_psp(x);
+  return millerrabin(x, flag);
+}
+
+GEN
+gispsp(GEN x) { return arith_proto(ispsp,x,1); }
+
+long
+ispsp(GEN x) { return millerrabin(x,1); }
+
+GEN
+gmillerrabin(GEN n, long k) { return arith_proto2gs(millerrabin,n,k); }
+
 /*********************************************************************/
 /**                                                                 **/
 /**                    FUNDAMENTAL DISCRIMINANTS                    **/
 /**                                                                 **/
 /*********************************************************************/
-
-/* gissquarefree, issquarefree moved into arith2.c with the other
-   basic arithmetic functions (issquarefree is the square of the
-   Moebius mu function, after all...) --GN */
-
 GEN
-gisfundamental(GEN x)
-{
-  return arith_proto(isfundamental,x,1);
-}
+gisfundamental(GEN x) { return arith_proto(isfundamental,x,1); }
 
 long
 isfundamental(GEN x)
 {
-  long av,r;
+  long r;
   GEN p1;
 
   if (gcmp0(x)) return 0;
   r=mod4(x);
   if (!r)
   {
-    av=avma; p1=shifti(x,-2);
+    const gpmem_t av = avma;
+    p1=shifti(x,-2);
     r=mod4(p1); if (!r) return 0;
     if (signe(x)<0) r=4-r;
     r = (r==1)? 0: issquarefree(p1);
@@ -1643,17 +1821,18 @@ isfundamental(GEN x)
 GEN
 quaddisc(GEN x)
 {
-  long av=avma,tetpil=avma,i,r,tx=typ(x);
+  const gpmem_t av = avma;
+  long i,r,tx=typ(x);
   GEN p1,p2,f,s;
 
   if (tx != t_INT && ! is_frac_t(tx)) err(arither1);
   f=factor(x); p1=(GEN)f[1]; p2=(GEN)f[2];
-  s=gun;
+  s = gun;
   for (i=1; i<lg(p1); i++)
-    if ( odd(mael(p2,i,2)) ) { tetpil=avma; s=gmul(s,(GEN)p1[i]); }
+    if (odd(mael(p2,i,2))) s = gmul(s,(GEN)p1[i]);
   r=mod4(s); if (gsigne(x)<0) r=4-r;
-  if (r>1) { tetpil=avma; s=shifti(s,2); }
-  return gerepile(av,tetpil,s);
+  if (r>1) s = shifti(s,2);
+  return gerepileuptoint(av, s);
 }
 
 /*********************************************************************/
@@ -1661,12 +1840,11 @@ quaddisc(GEN x)
 /**                              FACTORIAL                          **/
 /**                                                                 **/
 /*********************************************************************/
-GEN muluu(ulong x, ulong y);
-
 GEN
 mpfact(long n)
 {
-  long lx,k,l, av = avma;
+  const gpmem_t av = avma;
+  long lx,k,l;
   GEN x;
 
   if (n<2)
@@ -1694,10 +1872,10 @@ mpfact(long n)
 GEN
 mpfactr(long n, long prec)
 {
-  long av = avma, lim,k;
-  GEN f = cgetr(prec);
+  gpmem_t av = avma, lim;
+  long k;
+  GEN f = realun(prec);
 
-  affsr(1,f);
   if (n<2)
   {
     if (n<0) err(facter);
@@ -1725,36 +1903,37 @@ mpfactr(long n, long prec)
 void
 lucas(long n, GEN *ln, GEN *ln1)
 {
-  long taille,av;
+  gpmem_t av;
+  long taille;
   GEN z,t;
 
-  if (!n) { *ln=stoi(2); *ln1=stoi(1); return; }
+  if (!n) { *ln = stoi(2); *ln1 = stoi(1); return; }
 
-  taille=(long)(pariC3*(1+labs(n))+3);
-  *ln=cgeti(taille); *ln1=cgeti(taille);
-  av=avma; lucas(n/2,&z,&t);
+  taille = 3 + (long)(pariC3 * (1+labs(n)));
+  *ln = cgeti(taille);
+  *ln1= cgeti(taille);
+  av = avma; lucas(n/2, &z, &t);
   switch(n % 4)
   {
     case -3:
-      addsiz(2,sqri(z),*ln1);
-      subiiz(addsi(1,mulii(z,t)),*ln1,*ln); break;
-    case -2:
-      addsiz(2,sqri(z),*ln); addsiz(1,mulii(z,t),*ln1); break;
+      addsiz(2,sqri(z), *ln1);
+      subiiz(addsi(1,mulii(z,t)),*ln1, *ln); break;
     case -1:
-      subisz(sqri(z),2,*ln1);
-      subiiz(subis(mulii(z,t),1),*ln1,*ln); break;
-    case  0: subisz(sqri(z),2,*ln); subisz(mulii(z,t),1,*ln1); break;
-    case  1: subisz(mulii(z,t),1,*ln); addsiz(2,sqri(t),*ln1); break;
-    case  2: addsiz(2,sqri(z),*ln); addsiz(1,mulii(z,t),*ln1); break;
-    case  3: addsiz(1,mulii(z,t),*ln); subisz(sqri(t),2,*ln1);
+      subisz(sqri(z),2, *ln1);
+      subiiz(subis(mulii(z,t),1),*ln1, *ln); break;
+    case  0: subisz(sqri(z),2,    *ln); subisz(mulii(z,t),1, *ln1); break;
+    case  1: subisz(mulii(z,t),1, *ln); addsiz(2,sqri(t),    *ln1); break;
+    case -2:
+    case  2: addsiz(2,sqri(z),    *ln); addsiz(1,mulii(z,t), *ln1); break;
+    case  3: addsiz(1,mulii(z,t), *ln); subisz(sqri(t),2,    *ln1);
   }
-  avma=av;
+  avma = av;
 }
 
 GEN
 fibo(long n)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN ln,ln1;
 
   lucas(n-1,&ln,&ln1);
@@ -1766,149 +1945,141 @@ fibo(long n)
 /*                      CONTINUED FRACTIONS                        */
 /*                                                                 */
 /*******************************************************************/
-static GEN sfcont2(GEN b, GEN x, long k);
-
-GEN
-gcf(GEN x)
+static GEN
+icopy_lg(GEN x, long l)
 {
-  return sfcont(x,x,0);
-}
-
-GEN
-gcf2(GEN b, GEN x)
-{
-  return contfrac0(x,b,0);
-}
-
-GEN
-gboundcf(GEN x, long k)
-{
-  return sfcont(x,x,k);
-}
-
-GEN
-contfrac0(GEN x, GEN b, long flag)
-{
-  long lb,tb,i;
+  long lx = lgefint(x);
   GEN y;
-
-  if (!b || gcmp0(b)) return sfcont(x,x,flag);
-  tb = typ(b);
-  if (tb == t_INT) return sfcont(x,x,itos(b));
-  if (! is_matvec_t(tb)) err(typeer,"contfrac0");
-
-  lb=lg(b); if (lb==1) return cgetg(1,t_VEC);
-  if (tb != t_MAT) return sfcont2(b,x,flag);
-  if (lg(b[1])==1) return sfcont(x,x,flag);
-  y = (GEN) gpmalloc(lb*sizeof(long));
-  for (i=1; i<lb; i++) y[i]=coeff(b,1,i);
-  x = sfcont2(y,x,flag); free(y); return x;
+  
+  if (lx >= l) return icopy(x);
+  y = cgeti(l); affii(x, y); return y;
 }
 
-GEN
-sfcont(GEN x, GEN x1, long k)
+/* if y != NULL, stop as soon as partial quotients differ from y */
+static GEN
+Qsfcont(GEN x, GEN y, long k)
 {
-  long  av,lx=lg(x),tx=typ(x),e,i,j,l,l1,lx1,tetpil,f;
-  GEN   y,p1,p2,p3,p4,yp;
+  GEN  z, p1, p2, p3;
+  long i, l, ly;
+
+  ly = lgefint(x[2]);
+  l = 3 + (long) ((ly-2) / pariC3);
+  if (k > 0 && ++k > 0 && l > k) l = k; /* beware overflow */
+  if ((ulong)l > LGBITS) l = LGBITS;
+
+  p1 = (GEN)x[1];
+  p2 = (GEN)x[2];
+  z = cgetg(l,t_VEC);
+  l--;
+  if (y) {
+    gpmem_t av = avma;
+    if (l >= lg(y)) l = lg(y)-1;
+    for (i = 1; i <= l; i++)
+    {
+      z[i] = y[i];
+      p3 = p2; if (!gcmp1((GEN)z[i])) p3 = mulii((GEN)z[i], p2);
+      p3 = subii(p1, p3);
+      if (signe(p3) < 0)
+      { /* partial quotient too large */
+        p3 = addii(p3, p2);
+        if (signe(p3) >= 0) i++; /* by 1 */
+        break;
+      }
+      if (cmpii(p3, p2) >= 0)
+      { /* partial quotient too small */
+        p3 = subii(p3, p2);
+        if (cmpii(p3, p2) < 0) i++; /* by 1 */
+        break;
+      }
+      if ((i & 0xff) == 0) gerepileall(av, 2, &p2, &p3);
+      p1 = p2; p2 = p3;
+    }
+  } else {
+    p1 = icopy_lg(p1, ly);
+    p2 = icopy(p2);
+    for (i = 1; i <= l; i++)
+    {
+      z[i] = (long)truedvmdii(p1,p2,&p3);
+      if (p3 == gzero) { i++; break; }
+      affii(p3, p1); cgiv(p3); p3 = p1;
+      p1 = p2; p2 = p3;
+    }
+  }
+  i--;
+  if (i > 2 && gcmp1((GEN)z[i]))
+  {
+    cgiv((GEN)z[i]); --i;
+    addsiz(1,(GEN)z[i], (GEN)z[i]);
+  }
+  setlg(z,i+1); return z;
+}
+
+static GEN
+sfcont(GEN x, long k)
+{
+  gpmem_t av;
+  long lx,tx=typ(x),e,i,l;
+  GEN  y,p1,p2,p3;
 
   if (is_scalar_t(tx))
   {
-    if (gcmp0(x)) { y=cgetg(2,t_VEC); y[1]=zero; return y; }
+    if (gcmp0(x)) return _vec(gzero);
     switch(tx)
     {
-      case t_INT:
-        y=cgetg(2,t_VEC); y[1]=lcopy(x); return y;
+      case t_INT: y = cgetg(2,t_VEC); y[1] = (long)icopy(x); return y;
       case t_REAL:
-        l=avma; p1=cgetg(3,t_FRACN);
+        av = avma; lx = lg(x);
 	p2 = rcopy(x); settyp(p2,t_INT); setlgefint(p2,lx);
-	p1[1] = (long) p2; e = bit_accuracy(lx)-1-expo(x);
-	if (e<0) err(talker,"integral part not significant in scfont");
-
-	l1 = (e>>TWOPOTBITS_IN_LONG)+3; p2=cgeti(l1);
-	p2[1]= evalsigne(1)|evallgefint(l1);
-	p2[2]= (1L<<(e&(BITS_IN_LONG-1)));
-	for (i=3; i<l1; i++) p2[i]=0;
-	p1[2]=(long) p2; p3=cgetg(3,t_FRACN);
-	p3[2]=lcopy(p2);
-	p3[1]=laddsi(signe(x),(GEN)p1[1]);
-	p1=sfcont(p1,p1,k); tetpil=avma;
-	return gerepile(l,tetpil,sfcont(p3,p1,k));
+        e = bit_accuracy(lx)-1-expo(x);
+	if (e < 0) err(talker,"integral part not significant in scfont");
+        p1 = cgetg(3, t_FRACN);
+	p1[1] = (long)p2;
+	p1[2] = lshifti(gun, e);
+       
+        p3 = cgetg(3, t_FRACN);
+	p3[1] = laddsi(signe(x), p2);
+	p3[2] = p1[2];
+	p1 = Qsfcont(p1,NULL,k);
+	return gerepilecopy(av, Qsfcont(p3,p1,k));
 
       case t_FRAC: case t_FRACN:
-        l=avma; lx1=lgefint(x[2]);
-	l1 = (long) ((double) BYTES_IN_LONG/4.0*46.093443*(lx1-2)+3);
-        if (k>0 && ++k > 0 && l1 > k) l1 = k; /* beware overflow */
-	if ((ulong)l1>LGBITS) l1=LGBITS;
-	if (lgefint(x[1]) >= lx1) p1=icopy((GEN)x[1]);
-	else affii((GEN)x[1],p1=cgeti(lx1));
-	p2=icopy((GEN)x[2]); lx1=lg(x1);
-	y=cgetg(l1,t_VEC); f=(x!=x1); i=0;
-	while (!gcmp0(p2) && i<=l1-2)
-	{
-	  i++; y[i]=ldvmdii(p1,p2,&p3);
-	  if (signe(p3)>=0) affii(p3,p1);
-          else
-	  {
-	    p4=addii(p3,p2); affii(p4,p1); cgiv(p4);
-	    y[1]=laddsi(-1,(GEN)y[1]);
-	  }
-	  cgiv(p3); p4=p1; p1=p2; p2=p4;
-	  if (f)
-          {
-	    if (i>=lx1) { i--; break; }
-            if (!egalii((GEN)y[i],(GEN)x1[i]))
-            {
-              av=avma;
-              if (gcmp1(absi(subii((GEN)x1[i],(GEN)y[i]))))
-              {
-                if (i>=lx1-1 || !gcmp1((GEN)x1[i+1]))
-                  affii((GEN)x1[i],(GEN)y[i]);
-              }
-              else i--;
-              avma=av; break;
-            }
-          }
-	}
-	if (i>=2 && gcmp1((GEN)y[i]))
-	{
-	  cgiv((GEN)y[i]); --i; cgiv((GEN)y[i]);
-	  y[i]=laddsi(1,(GEN)y[i]);
-	}
-	setlg(y,i+1); return gerepileupto(l, y);
+        av = avma;
+        return gerepileupto(av, Qsfcont(x, NULL, k));
     }
     err(typeer,"sfcont");
   }
 
   switch(tx)
   {
-    case t_POL:
-      y=cgetg(2,t_VEC); y[1]=lcopy(x); break;
+    case t_POL: return _veccopy(x);
     case t_SER:
-      av=avma; p1=gtrunc(x); tetpil=avma;
-      y=gerepile(av,tetpil,sfcont(p1,p1,k)); break;
+      av = avma; p1 = gtrunc(x);
+      return gerepileupto(av,sfcont(p1,k));
     case t_RFRAC:
     case t_RFRACN:
-      av=avma; l1=lgef(x[1]); if (lgef(x[2])>l1) l1=lgef(x[2]);
-      if (k>0 && l1>k+1) l1=k+1;
-      yp=cgetg(l1,t_VEC); p1=(GEN)x[1]; i=0; p2=(GEN)x[2];
-      while (!gcmp0(p2) && i<=l1-2)
+      av = avma;
+      l = typ(x[1]) == t_POL? lgef(x[1]): 3;
+      if (lgef(x[2]) > l) l = lgef(x[2]);
+      if (k > 0 && l > k+1) l = k+1;
+      y = cgetg(l,t_VEC);
+      p1 = (GEN)x[1];
+      p2 = (GEN)x[2];
+      for (i=1; i<l; i++)
       {
-	i++; yp[i]=ldivres(p1,p2,&p3);
-	p1=p2; p2=p3;
+	y[i] = ldivres(p1,p2,&p3);
+        if (gcmp0(p3)) { i++; break; }
+	p1 = p2; p2 = p3;
       }
-      tetpil=avma; y=cgetg(i+1,t_VEC);
-      for (j=1; j<=i; j++) y[j]=lcopy((GEN)yp[j]);
-      y=gerepile(av,tetpil,y); break;
-    default: err(typeer,"sfcont");
-      return NULL; /* not reached */
+      setlg(y, i); return gerepilecopy(av, y);
   }
-  return y;
+  err(typeer,"sfcont");
+  return NULL; /* not reached */
 }
 
 static GEN
 sfcont2(GEN b, GEN x, long k)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long l1 = lg(b), tx = typ(x), i;
   GEN y,p1;
 
@@ -1942,10 +2113,49 @@ sfcont2(GEN b, GEN x, long k)
   return gerepilecopy(av,y);
 }
 
+
+GEN
+gcf(GEN x)
+{
+  return sfcont(x,0);
+}
+
+GEN
+gcf2(GEN b, GEN x)
+{
+  return contfrac0(x,b,0);
+}
+
+GEN
+gboundcf(GEN x, long k)
+{
+  return sfcont(x,k);
+}
+
+GEN
+contfrac0(GEN x, GEN b, long flag)
+{
+  long lb,tb,i;
+  GEN y;
+
+  if (!b || gcmp0(b)) return sfcont(x,flag);
+  tb = typ(b);
+  if (tb == t_INT) return sfcont(x,itos(b));
+  if (! is_matvec_t(tb)) err(typeer,"contfrac0");
+
+  lb=lg(b); if (lb==1) return cgetg(1,t_VEC);
+  if (tb != t_MAT) return sfcont2(b,x,flag);
+  if (lg(b[1])==1) return sfcont(x,flag);
+  y = (GEN) gpmalloc(lb*sizeof(long));
+  for (i=1; i<lb; i++) y[i]=coeff(b,1,i);
+  x = sfcont2(y,x,flag); free(y); return x;
+}
+
 GEN
 pnqn(GEN x)
 {
-  long av=avma,tetpil,lx,ly,tx=typ(x),i;
+  gpmem_t av=avma,tetpil;
+  long lx,ly,tx=typ(x),i;
   GEN y,p0,p1,q0,q1,a,b,p2,q2;
 
   if (! is_matvec_t(tx)) err(typeer,"pnqn");
@@ -1995,7 +2205,7 @@ bestappr_mod(GEN x, GEN A, GEN B)
   {
     case t_INTMOD:
     {
-      ulong av = avma;
+      gpmem_t av = avma;
       GEN a,b,d, t = cgetg(3, t_FRAC);
       if (! ratlift((GEN)x[2], (GEN)x[1], &a,&b,A,B)) return NULL;
       if (is_pm1(b)) return icopy_av(a, (GEN)av);
@@ -2024,7 +2234,7 @@ bestappr_mod(GEN x, GEN A, GEN B)
 GEN
 bestappr(GEN x, GEN k)
 {
-  ulong av = avma, tetpil;
+  gpmem_t av = avma, tetpil;
   long tx,tk=typ(k),lx,i,e;
   GEN p0,p1,p,q0,q1,q,a,y;
 
@@ -2076,7 +2286,7 @@ bestappr(GEN x, GEN k)
 GEN
 bestappr0(GEN x, GEN a, GEN b)
 {
-  ulong av;
+  gpmem_t av;
   GEN t;
   if (!b) return bestappr(x,a);
   av = avma;
@@ -2123,7 +2333,8 @@ update_f(GEN f, GEN a)
 GEN
 fundunit(GEN x)
 {
-  long av = avma, av2,tetpil,lim,r,flp,flq;
+  gpmem_t av = avma, av2, lim;
+  long r,flp,flq;
   GEN pol,y,a,u,v,u1,v1,sqd,f;
 
   if (typ(x) != t_INT) err(arither1);
@@ -2152,9 +2363,9 @@ fundunit(GEN x)
   pol = quadpoly(x); y = get_quad(f,pol,r);
   if (!flq) v1 = y; else { update_f(f,a); v1 = get_quad(f,pol,r); }
 
-  u1=gconj(y); tetpil=avma; y=gdiv(v1,u1);
-  if (signe(y[3]) < 0) { tetpil=avma; y=gneg(y); }
-  return gerepile(av,tetpil,y);
+  u1=gconj(y); y=gdiv(v1,u1);
+  if (signe(y[3]) < 0) y = gneg(y);
+  return gerepileupto(av, y);
 }
 
 GEN
@@ -2166,7 +2377,8 @@ gregula(GEN x, long prec)
 GEN
 regula(GEN x, long prec)
 {
-  long av = avma,av2,lim,r,fl,rexp;
+  gpmem_t av = avma,av2,lim;
+  long r,fl,rexp;
   GEN reg,rsqd,y,u,v,u1,v1, sqd = racine(x);
 
   if (signe(x)<=0) err(arither2);
@@ -2175,7 +2387,7 @@ regula(GEN x, long prec)
   rsqd = gsqrt(x,prec);
   if (egalii(sqri(sqd),x)) err(talker,"square argument in regula");
 
-  rexp=0; reg=cgetr(prec); affsr(2,reg);
+  rexp=0; reg = stor(2, prec);
   av2=avma; lim = stack_lim(av2,2);
   u = stoi(r); v = gdeux;
   for(;;)
@@ -2199,7 +2411,7 @@ regula(GEN x, long prec)
   y = mplog(divri(reg,v));
   if (rexp)
   {
-    u1 = mulsr(rexp, glog(gdeux, prec));
+    u1 = mulsr(rexp, mplog2(prec));
     setexpo(u1, expo(u1)+1);
     y = addrr(y,u1);
   }
@@ -2270,7 +2482,7 @@ end_classno(GEN h, GEN hin, GEN forms, long lform)
   q = ground(gdiv(hin,h)); /* approximate order of G/H */
   for (i=1; i < lform; i++)
   {
-    long av = avma; 
+    gpmem_t av = avma; 
     fg = powgi((GEN)forms[i], h);
     fh = powgi(fg, q);
     a = (GEN)fh[1];
@@ -2301,22 +2513,24 @@ conductor_part(GEN x, GEN *ptD, GEN *ptreg, GEN *ptfa)
   GEN e,p,H,d,D,fa,reg;
 
   fa = auxdecomp(absi(x),1);
-  e = (GEN)fa[2]; fa = (GEN)fa[1];
+  e = gtovecsmall((GEN)fa[2]);
+  fa = (GEN)fa[1];
   n = lg(fa); d = gun;
   for (i=1; i<n; i++)
-    if (mod2((GEN)e[i])) d = mulii(d,(GEN)fa[i]);
-  fl2 = (mod4(x)==0); /* 4 | x */
+    if (e[i] & 1) d = mulii(d,(GEN)fa[i]);
   if (mod4(d) == 2-s) fl2 = 0;
   else
   {
-    if (!fl2) err(funder2,"classno2");
+    fl2 = (mod4(x)==0);
+    if (!fl2) err(funder2,"classno");
     d = shifti(d,2);
   }
   H = gun; D = (s<0)? negi(d): d; /* d = abs(D) */
   /* f \prod_{p|f}  [ 1 - (D/p) p^-1 ] */
   for (i=1; i<n; i++)
   {
-    k = itos((GEN)e[i]); p = (GEN)fa[i];
+    p = (GEN)fa[i];
+    k = e[i];
     if (fl2 && i==1) k -= 2; /* p = 2 */
     if (k >= 2)
     {
@@ -2375,8 +2589,8 @@ two_rank(GEN x)
 GEN
 classno(GEN x)
 {
-  long av = avma, av2,c,lforms,k,l,i,j,j1,j2,lim,com,s, forms[MAXFORM];
-  long r2;
+  gpmem_t av = avma, av2, lim;
+  long r2,c,lforms,k,l,i,j,j1,j2,com,s, forms[MAXFORM];
   GEN a,b,count,index,tabla,tablb,hash,p1,p2,hin,h,f,fh,fg,ftest;
   GEN Hf,D,fa;
   byteptr p = diffptr;
@@ -2472,7 +2686,7 @@ classno(GEN x)
 GEN
 classno2(GEN x)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long n,i,k,s = signe(x);
   GEN p1,p2,p3,p4,p5,p7,Hf,Pi,reg,logd,d,D;
 
