@@ -1,4 +1,4 @@
-/* $Id: arith2.c,v 1.34 2001/12/05 20:48:01 karim Exp $
+/* $Id: arith2.c,v 1.49 2002/05/16 11:40:38 bill Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -161,7 +161,7 @@ initprimes0(ulong maxnum, long *lenp, long *lastp)
   asize = ARENA_IN_ROOTS * rootnum;	/* Make % overhead negligeable. */
   if (asize < PRIME_ARENA) asize = PRIME_ARENA;
   if (asize > remains) asize = remains + 1;/* + room for a sentinel byte */
-  alloced = (avma - bot < (ulong)asize>>1); /* enough room on the stack ? */
+  alloced = (avma - bot < (size_t)asize>>1); /* enough room on the stack ? */
   if (alloced)
     p = (byteptr) gpmalloc(asize);
   else
@@ -274,7 +274,7 @@ cleanprimetab(void)
 GEN
 addprimes(GEN p)
 {
-  ulong av;
+  gpmem_t av;
   long i,k,tx,lp;
   GEN L;
 
@@ -282,7 +282,7 @@ addprimes(GEN p)
   tx = typ(p);
   if (is_vec_t(tx))
   {
-    for (i=1; i < lg(p); i++) addprimes((GEN) p[i]);
+    for (i=1; i < lg(p); i++) (void)addprimes((GEN) p[i]);
     return primetab;
   }
   if (tx != t_INT) err(typeer,"addprime");
@@ -305,7 +305,7 @@ addprimes(GEN p)
   }
   primetab = (GEN) gprealloc(primetab, (lp+1)*sizeof(long));
   primetab[i] = lclone(p); setlg(primetab, lp+1);
-  if (k > 1) { cleanprimetab(); setlg(L,k); addprimes(L); }
+  if (k > 1) { cleanprimetab(); setlg(L,k); (void)addprimes(L); }
   avma = av; return primetab;
 }
 
@@ -341,7 +341,7 @@ removeprimes(GEN prime)
     }
     else
     {
-      for (i=1; i < lg(prime); i++) removeprime((GEN) prime[i]);
+      for (i=1; i < lg(prime); i++) (void)removeprime((GEN) prime[i]);
     }
     return primetab;
   }
@@ -362,7 +362,11 @@ removeprimes(GEN prime)
 /* Some overkill removed from this  (15 spsp for an integer < 2^32 ?!).
    Should really revert to isprime() once the new primality tester is ready
    --GN */
+#if 0
 #define pseudoprime(p) millerrabin(p,3*lgefint(p))
+#else /* remove further overkill :-)  --KB */
+#define pseudoprime(p) BSW_psp(p)
+#endif
 
 /* where to stop trial dividing in factorization */
 
@@ -409,44 +413,46 @@ GEN
 auxdecomp1(GEN n, long (*ifac_break)(GEN n, GEN pairs, GEN here, GEN state),
 		  GEN state, long all, long hint)
 {
-  long pp[] = { evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3),2,0 };
-  long nb = 0,i,k,lim1,av,lp;
+  gpmem_t av;
+  long pp[] = { evaltyp(t_INT)|_evallg(4), 0,0,0 };
+  long nb = 0,i,k,lp,p,lim1;
   byteptr d=diffptr+1;
 
   if (typ(n) != t_INT) err(arither1);
   i=signe(n); if (!i) err(arither2);
-  cgetg(3,t_MAT);
-  if (i<0) { stoi(-1); stoi(1); nb++; }
+  (void)cgetg(3,t_MAT);
+  if (i<0) { (void)stoi(-1); (void)stoi(1); nb++; }
   if (is_pm1(n)) return aux_end(NULL,nb);
 
   n = gclone(n);  setsigne(n,1);
   i = vali(n);
   if (i)
   {
-    stoi(2); stoi(i); nb++;
+    (void)stoi(2); (void)stoi(i); nb++;
     av=avma; affii(shifti(n,-i), n); avma=av;
   }
   if (is_pm1(n)) return aux_end(n,nb);
   lim1 = tridiv_bound(n, all);
 
   /* trial division */
-  while (*d && pp[2] <= lim1)
+  p = 2;
+  while (*d && p <= lim1)
   {
-    pp[2] += *d++;
-    if (mpdivis(n,pp,n))
+    p += *d++;
+    if (mpdivisis(n,p,n))
     {
-      nb++; k=1; while (mpdivis(n,pp,n)) k++;
-      icopy(pp); stoi(k);
+      nb++; k=1; while (mpdivisis(n,p,n)) k++;
+      (void)utoi(p); (void)stoi(k);
       if (is_pm1(n)) return aux_end(n,nb);
     }
   }
 
   /* pp = square of biggest p tried so far */
-  av=avma; setlg(pp,4); affii(sqri(pp),pp); avma=av;
+  av=avma; affii(muluu(p,p),pp); avma=av;
   if (cmpii(pp,n) > 0)
   {
     nb++;
-    icopy(n); stoi(1); return aux_end(n,nb);
+    (void)icopy(n); (void)stoi(1); return aux_end(n,nb);
   }
 
   /* trial divide by the "special primes" (usually huge composites...) */
@@ -455,7 +461,7 @@ auxdecomp1(GEN n, long (*ifac_break)(GEN n, GEN pairs, GEN here, GEN state),
     if (mpdivis(n,(GEN) primetab[i],n))
     {
       nb++; k=1; while (mpdivis(n,(GEN) primetab[i],n)) k++;
-      icopy((GEN) primetab[i]); stoi(k);
+      (void)icopy((GEN) primetab[i]); (void)stoi(k);
       if (is_pm1(n)) return aux_end(n,nb);
     }
 
@@ -463,7 +469,7 @@ auxdecomp1(GEN n, long (*ifac_break)(GEN n, GEN pairs, GEN here, GEN state),
   if ((k && cmpii(pp,n) > 0) || (all==1 && pseudoprime(n)))
   {
     nb++;
-    icopy(n); stoi(1); return aux_end(n,nb);
+    (void)icopy(n); (void)stoi(1); return aux_end(n,nb);
   }
 
   /* now we have a large composite.  Use hint as is if all==1 */
@@ -485,7 +491,7 @@ auxdecomp1(GEN n, long (*ifac_break)(GEN n, GEN pairs, GEN here, GEN state),
 static long
 ifac_break_limit(GEN n, GEN pairs/*unused*/, GEN here, GEN state)
 {
-  ulong ltop = avma;
+  gpmem_t ltop = avma;
   int res;
   if (!here) /* initial call */
    /* Small prime have been removed since start, n is the new unfactored part.
@@ -558,7 +564,7 @@ GEN
 boundfact(GEN n, long lim)
 {
   GEN p1,p2,p3,p4,p5,y;
-  long av = avma,tetpil;
+  gpmem_t av = avma,tetpil;
 
   if (lim<=1) lim=0;
   switch(typ(n))
@@ -585,6 +591,50 @@ boundfact(GEN n, long lim)
   err(arither1);
   return NULL; /* not reached */
 }
+/***********************************************************************/
+/**                                                                   **/
+/**                    SIMPLE FACTORISATIONS                          **/
+/**                                                                   **/
+/***********************************************************************/
+
+/*Factorize n and output [fp,fe] where
+ * fp and fe are vecsmall and n=prod{fp[i]^fe[i]}
+ */
+
+GEN
+decomp_small(long n)
+{
+  gpmem_t ltop=avma;
+  GEN F = factor(stoi(n));
+  long i, l=lg(F[1]);
+  GEN f=cgetg(3,t_VEC);
+  GEN fp=cgetg(l,t_VECSMALL);
+  GEN fe=cgetg(l,t_VECSMALL);
+  f[1] = (long) fp;
+  f[2] = (long) fe;
+  for(i = 1; i < l; i++)
+  {
+    fp[i]=itos(gcoeff(F,i,1));
+    fe[i]=itos(gcoeff(F,i,2));
+  }
+  return gerepileupto(ltop,f);
+}
+
+/*Return the primary factors of a small integer as a vecsmall*/
+GEN
+decomp_primary_small(long n)
+{
+  gpmem_t ltop=avma;
+  GEN F = factor(stoi(n));
+  GEN fc = cgetg(lg(F[1]), t_VECSMALL);
+  gpmem_t av=avma;
+  long i;
+  for (i = 1; i < lg(fc); i++)
+    fc[i] = itos(powgi(gmael(F,1,i), gmael(F,2,i)));
+  avma=av;
+  return gerepileupto(ltop,fc);
+}
+
 
 /***********************************************************************/
 /**                                                                   **/
@@ -627,8 +677,9 @@ long
 mu(GEN n)
 {
   byteptr d = diffptr+1;	/* point at 3 - 2 */
-  ulong p,lim1, av = avma;
-  long s, v;
+  gpmem_t av = avma;
+  ulong p;
+  long s, v, lim1;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -668,8 +719,9 @@ gissquarefree(GEN x)
 long
 issquarefree(GEN x)
 {
-  ulong av = avma,lim1,p;
-  long tx;
+  ulong p;
+  gpmem_t av = avma;
+  long tx, lim1;
   GEN d;
 
   if (gcmp0(x)) return 0;
@@ -713,8 +765,8 @@ long
 omega(GEN n)
 {
   byteptr d=diffptr+1;
-  ulong p, lim1, av = avma;
-  long nb,v;
+  gpmem_t av = avma;
+  long p,nb,v, lim1;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -750,8 +802,9 @@ long
 bigomega(GEN n)
 {
   byteptr d=diffptr+1;
-  ulong av = avma, p,lim1; 
-  long nb,v;
+  ulong p;
+  gpmem_t av = avma;
+  long nb,v, lim1;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -786,8 +839,9 @@ phi(GEN n)
 {
   byteptr d = diffptr+1;
   GEN m;
-  ulong av = avma, lim1, p;
-  long v;
+  ulong p;
+  gpmem_t av = avma;
+  long v, lim1;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -828,8 +882,9 @@ numbdiv(GEN n)
 {
   byteptr d=diffptr+1;
   GEN m;
-  long l, v;
-  ulong av = avma, p,lim1;
+  long l, v, lim1;
+  ulong p;
+  gpmem_t av = avma;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -866,8 +921,9 @@ sumdiv(GEN n)
 {
   byteptr d=diffptr+1;
   GEN m,m1;
-  ulong av=avma,lim1,p;
-  long v;
+  ulong p;
+  gpmem_t av=avma;
+  long v, lim1;
 
   if (typ(n) != t_INT) err(arither1);
   if (!signe(n)) err(arither2);
@@ -908,8 +964,9 @@ sumdivk(GEN n, long k)
 {
   byteptr d=diffptr+1;
   GEN n1,m,m1,pk;
-  ulong av = avma, p, lim1;
-  long k1,v;
+  ulong p;
+  gpmem_t av = avma;
+  long k1,v, lim1;
 
   if (!k) return numbdiv(n);
   if (k==1) return sumdiv(n);
@@ -958,7 +1015,8 @@ sumdivk(GEN n, long k)
 GEN
 divisors(GEN n)
 {
-  long tetpil,av=avma,i,j,l;
+  gpmem_t tetpil,av=avma;
+  long i,j,l;
   GEN *d,*t,*t1,*t2,*t3, nbdiv,e;
   
   if (typ(n) != t_MAT || lg(n) != 3) n = auxdecomp(n,1);
@@ -982,72 +1040,78 @@ divisors(GEN n)
   tetpil=avma; return gerepile(av,tetpil,sort((GEN)t));
 }
 
-GEN
-core(GEN n)
+static GEN
+_core(GEN n, int all)
 {
-  long av=avma,tetpil,i;
-  GEN fa,p1,p2,res=gun;
+  gpmem_t av = avma;
+  long i;
+  GEN fa,p1,p2,c = gun;
 
-  fa=auxdecomp(n,1); p1=(GEN)fa[1]; p2=(GEN)fa[2];
+  fa = auxdecomp(n,all);
+  p1 = (GEN)fa[1];
+  p2 = (GEN)fa[2];
   for (i=1; i<lg(p1); i++)
-    if (mod2((GEN)p2[i])) res = mulii(res,(GEN)p1[i]);
-  tetpil=avma; return gerepile(av,tetpil,icopy(res));
+    if (mod2((GEN)p2[i])) c = mulii(c,(GEN)p1[i]);
+  return gerepileuptoint(av, c);
 }
 
-GEN
-core2(GEN n)
+static GEN
+_core2(GEN n, int all)
 {
-  long av=avma,tetpil,i;
+  gpmem_t av = avma;
+  long i;
+  GEN fa,p1,p2,e,c=gun,f=gun,y;
 
-  GEN fa,p1,p2,p3,res=gun,res2=gun,y;
-
-  fa=auxdecomp(n,1); p1=(GEN)fa[1]; p2=(GEN)fa[2];
+  fa = auxdecomp(n,all);
+  p1 = (GEN)fa[1];
+  p2 = (GEN)fa[2];
   for (i=1; i<lg(p1); i++)
   {
-    p3=(GEN)p2[i];
-    if (mod2(p3)) res=mulii(res,(GEN)p1[i]);
-    if (!gcmp1(p3)) res2=mulii(res2,gpui((GEN)p1[i],shifti(p3,-1),0));
+    e = (GEN)p2[i];
+    if (mod2(e))   c = mulii(c, (GEN)p1[i]);
+    if (!gcmp1(e)) f = mulii(f, powgi((GEN)p1[i], shifti(e,-1)));
   }
-  tetpil=avma; y=cgetg(3,t_VEC);
-  y[1]=(long)icopy(res); y[2]=(long)icopy(res2);
-  return gerepile(av,tetpil,y);
+  y = cgetg(3,t_VEC);
+  y[1] = (long)c;
+  y[2] = (long)f; return gerepilecopy(av, y);
 }
 
+GEN core(GEN n)        { return _core(n,1); }
+GEN corepartial(GEN n) { return _core(n,0); }
+GEN core2(GEN n)       { return _core2(n,1); }
+GEN core2partial(GEN n){ return _core2(n,0); }
+
 GEN
-core0(GEN n,long flag)
-{
-  return flag? core2(n): core(n);
-}
+core0(GEN n,long flag) { return flag? core2(n): core(n); }
+
+static long
+_mod4(GEN c) { long r = mod4(c); if (signe(c) < 0) r = 4-r; return r; }
 
 GEN
 coredisc(GEN n)
 {
-  long av=avma,tetpil,r;
-  GEN p1=core(n);
-
-  r=mod4(p1); if (signe(p1)<0) r = 4-r;
-  if (r==1 || r==4) return p1;
-  tetpil=avma; return gerepile(av,tetpil,shifti(p1,2));
+  gpmem_t av = avma;
+  GEN c = core(n);
+  long r = _mod4(c);
+  if (r==1 || r==4) return c;
+  return gerepileuptoint(av, shifti(c,2));
 }
 
 GEN
 coredisc2(GEN n)
 {
-  long av=avma,tetpil,r;
-  GEN y,p1,p2=core2(n);
-
-  p1=(GEN)p2[1]; r=mod4(p1); if (signe(p1)<0) r=4-r;
-  if (r==1 || r==4) return p2;
-  tetpil=avma; y=cgetg(3,t_VEC);
-  y[1]=lshifti(p1,2); y[2]=lmul2n((GEN)p2[2],-1);
-  return gerepile(av,tetpil,y);
+  gpmem_t av = avma;
+  GEN y = core2(n);
+  GEN c = (GEN)y[1], f = (GEN)y[2];
+  long r = _mod4(c);
+  if (r==1 || r==4) return y;
+  y = cgetg(3,t_VEC);
+  y[1] = lshifti(c,2);
+  y[2] = lmul2n(f,-1); return gerepileupto(av, y);
 }
 
 GEN
-coredisc0(GEN n,long flag)
-{
-  return flag? coredisc2(n): coredisc(n);
-}
+coredisc0(GEN n,long flag) { return flag? coredisc2(n): coredisc(n); }
 
 /***********************************************************************/
 /**                                                                   **/
@@ -1069,7 +1133,7 @@ qf_create(GEN x, GEN y, GEN z, long s)
   if (typ(x)!=t_INT || typ(y)!=t_INT || typ(z)!=t_INT) err(typeer,"Qfb");
   if (!s)
   {
-    long av=avma; s = signe(qf_disc(x,y,z)); avma=av;
+    gpmem_t av=avma; s = signe(qf_disc(x,y,z)); avma=av;
     if (!s) err(talker,"zero discriminant in Qfb");
   }
   t = (s>0)? cgetg(5,t_QFR): cgetg(4,t_QFI);
@@ -1159,7 +1223,8 @@ comp_gen(GEN z,GEN x,GEN y)
 static GEN
 compimag0(GEN x, GEN y, int raw)
 {
-  ulong tx = typ(x), av = avma;
+  gpmem_t av = avma;
+  long tx = typ(x);
   GEN z;
 
   if (typ(y) != tx || tx!=t_QFI) err(typeer,"composition");
@@ -1172,7 +1237,8 @@ compimag0(GEN x, GEN y, int raw)
 static GEN
 compreal0(GEN x, GEN y, int raw)
 {
-  ulong tx = typ(x), av = avma;
+  gpmem_t av = avma;
+  long tx = typ(x);
   GEN z;
 
   if (typ(y) != tx || tx!=t_QFR) err(typeer,"composition");
@@ -1203,7 +1269,7 @@ compraw(GEN x, GEN y)
 GEN
 sqcompimag0(GEN x, long raw)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN z = cgetg(4,t_QFI);
 
   if (typ(x)!=t_QFI) err(typeer,"composition");
@@ -1215,7 +1281,7 @@ sqcompimag0(GEN x, long raw)
 static GEN
 sqcompreal0(GEN x, long raw)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN z = cgetg(5,t_QFR);
 
   if (typ(x)!=t_QFR) err(typeer,"composition");
@@ -1240,7 +1306,7 @@ static GEN
 real_unit_form_by_disc(GEN D, long prec)
 {
   GEN y = cgetg(5,t_QFR), isqrtD;
-  long av = avma;
+  gpmem_t av = avma;
 
   if (typ(D) != t_INT || signe(D) <= 0) err(typeer,"real_unit_form_by_disc");
   switch(mod4(D))
@@ -1253,14 +1319,15 @@ real_unit_form_by_disc(GEN D, long prec)
   if (mod2(D) != mod2(isqrtD))
     isqrtD = gerepileuptoint(av, addsi(-1,isqrtD));
   y[2] = (long)isqrtD; av = avma;
-  y[3] = (long)gerepileuptoint(av, shifti(subii(sqri(isqrtD),D),-2));
+  y[3] = lpileuptoint(av, shifti(subii(sqri(isqrtD),D),-2));
   y[4] = (long)realzero(prec); return y;
 }
 
 GEN
 real_unit_form(GEN x)
 {
-  long av = avma, prec;
+  gpmem_t av = avma;
+  long prec;
   GEN D;
   if (typ(x) != t_QFR) err(typeer,"real_unit_form");
   prec = precision((GEN)x[4]);
@@ -1287,8 +1354,8 @@ imag_unit_form_by_disc(GEN D)
   y[3] = lshifti(D,-2); setsigne(y[3],1);
   if (isodd)
   {
-    long av = avma;
-    y[3] = (long)gerepileuptoint(av, addis((GEN)y[3],1));
+    gpmem_t av = avma;
+    y[3] = lpileuptoint(av, addis((GEN)y[3],1));
   }
   return y;
 }
@@ -1297,20 +1364,21 @@ GEN
 imag_unit_form(GEN x)
 {
   GEN p1,p2, y = cgetg(4,t_QFI);
-  long av;
+  gpmem_t av;
   if (typ(x) != t_QFI) err(typeer,"imag_unit_form");
   y[1] = un;
   y[2] = mpodd((GEN)x[2])? un: zero;
   av = avma; p1 = mulii((GEN)x[1],(GEN)x[3]);
   p2 = shifti(sqri((GEN)x[2]),-2);
-  y[3] = (long)gerepileuptoint(av, subii(p1,p2));
+  y[3] = lpileuptoint(av, subii(p1,p2));
   return y;
 }
 
 GEN
 powrealraw(GEN x, long n)
 {
-  long av = avma, m;
+  gpmem_t av = avma;
+  long m;
   GEN y;
 
   if (typ(x) != t_QFR)
@@ -1333,7 +1401,8 @@ powrealraw(GEN x, long n)
 GEN
 powimagraw(GEN x, long n)
 {
-  long av = avma, m;
+  gpmem_t av = avma;
+  long m;
   GEN y;
 
   if (typ(x) != t_QFI)
@@ -1364,7 +1433,8 @@ powraw(GEN x, long n)
 GEN
 nucomp(GEN x, GEN y, GEN l)
 {
-  long av=avma,tetpil,cz;
+  gpmem_t av=avma,tetpil;
+  long cz;
   GEN a,a1,a2,b,b2,d,d1,e,g,n,p1,p2,p3,q1,q2,q3,q4,s,t2,t3,u,u1,v,v1,v2,v3,z;
 
   if (x==y) return nudupl(x,l);
@@ -1428,7 +1498,8 @@ nucomp(GEN x, GEN y, GEN l)
 GEN
 nudupl(GEN x, GEN l)
 {
-  long av=avma,tetpil,cz;
+  gpmem_t av=avma,tetpil;
+  long cz;
   GEN u,v,d,d1,p1,a,b,c,b2,z,v2,v3,t2,t3,e,g;
 
   if (typ(x) != t_QFI)
@@ -1468,7 +1539,8 @@ nudupl(GEN x, GEN l)
 GEN
 nupow(GEN x, GEN n)
 {
-  long av,tetpil,i,j;
+  gpmem_t av,tetpil;
+  long i,j;
   unsigned long m;
   GEN y,l;
 
@@ -1622,7 +1694,7 @@ decodeform(GEN x, GEN d0)
     p1 = shiftr(p1,-e);
     p2 = addis(mulsi(EXP220,p2), e);
     p1 = mplog(p1);
-    p1 = mpadd(p1, mulir(p2, glog(gdeux, lg(d0))));
+    p1 = mpadd(p1, mulir(p2, mplog2(lg(d0))));
   }
   else
   { /* to avoid loss of precision */
@@ -1668,7 +1740,8 @@ redrealform5(GEN x, GEN D, GEN sqrtD, GEN isqrtD)
 static GEN
 redreal0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
 {
-  long av = avma, prec;
+  gpmem_t av = avma;
+  long prec;
   GEN d0;
 
   if (typ(x) != t_QFR) err(talker,"not a real quadratic form in redreal");
@@ -1706,7 +1779,7 @@ redreal0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
 GEN
 comprealform5(GEN x, GEN y, GEN D, GEN sqrtD, GEN isqrtD)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN z = cgetg(6,t_VEC); comp_gen(z,x,y);
   if (x == y)
   {
@@ -1726,7 +1799,8 @@ comprealform5(GEN x, GEN y, GEN D, GEN sqrtD, GEN isqrtD)
 GEN
 powrealform(GEN x, GEN n)
 {
-  long av = avma, i,m;
+  gpmem_t av = avma;
+  long i,m;
   GEN y,D,sqrtD,isqrtD,d0;
 
   if (typ(x) != t_QFR)
@@ -1758,7 +1832,8 @@ powrealform(GEN x, GEN n)
 GEN
 redimag(GEN x)
 {
-  long av=avma, fl;
+  gpmem_t av=avma;
+  long fl;
   do x = rhoimag0(x, &fl); while (fl == 0);
   x = gerepilecopy(av,x);
   if (fl == 2) setsigne(x[2], -signe(x[2]));
@@ -1793,7 +1868,7 @@ GEN
 qfbred0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
 {
   long tx=typ(x),fl;
-  ulong av;
+  gpmem_t av;
 
   if (!is_qf_t(tx)) err(talker,"not a quadratic form in qfbred");
   if (!D) D = qf_disc(x,NULL,NULL);
@@ -1818,7 +1893,8 @@ qfbred0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
 GEN
 primeform(GEN x, GEN p, long prec)
 {
-  long av,tetpil,s=signe(x);
+  gpmem_t av,tetpil;
+  long s=signe(x);
   GEN y,b;
 
   if (typ(x) != t_INT || !s) err(arither1);
@@ -1977,7 +2053,7 @@ bittest_many(GEN x, GEN gn, long c)
 	  two_adic = 1;			/* treat x 2-adically... */
   }
   if (n < 0) {
-      long oa, na;
+      gpmem_t oa, na;
 
       if (n + c <= 0)
 	  return gzero;
@@ -2042,7 +2118,7 @@ bittest_many(GEN x, GEN gn, long c)
   pnew = p;
   while ((pnew < p + l_res - 2) && !*pnew) /* Skip 0 words */
       pnew++;
-  avma = (long)(pnew - 2);
+  avma = (gpmem_t)(pnew - 2);
   if (pnew >= p - 2 + l_res)
       return gzero;
   l_res -= (pnew - p);
@@ -2329,9 +2405,9 @@ ibitnegimply(GEN x, GEN y)
 }
 
 static GEN
-inegate_inplace(GEN z, long ltop)
+inegate_inplace(GEN z, gpmem_t ltop)
 {
-  long lbot, o;
+  long o;
 
   /* Negate z */
   o = incdec(z, 1);			/* Can overflow z... */
@@ -2341,16 +2417,14 @@ inegate_inplace(GEN z, long ltop)
   else if (lgefint(z) == 2)
       setsigne(z, 0);      
   incdec(z,-1);			/* Restore a normalized value */
-  lbot = avma;
-  z = gsub(z,gun);
-  return gerepile(ltop,lbot,z);
+  return gerepileupto(ltop, subis(z,1));
 }
 
 GEN
 gbitor(GEN x, GEN y)
 {
+  gpmem_t ltop;
   long sx,sy;
-  long ltop;
   GEN z;
 
   if (typ(x) != t_INT || typ(y) != t_INT)
@@ -2381,8 +2455,8 @@ gbitor(GEN x, GEN y)
 GEN
 gbitand(GEN x, GEN y)
 {
+  gpmem_t ltop;
   long sx,sy;
-  long ltop;
   GEN z;
 
   if (typ(x) != t_INT || typ(y) != t_INT)
@@ -2415,8 +2489,8 @@ gbitand(GEN x, GEN y)
 GEN
 gbitxor(GEN x, GEN y)
 {
+  gpmem_t ltop;
   long sx,sy;
-  long ltop;
   GEN z;
 
   if (typ(x) != t_INT || typ(y) != t_INT)
@@ -2448,8 +2522,8 @@ gbitxor(GEN x, GEN y)
 GEN
 gbitnegimply(GEN x, GEN y)		/* x & ~y */
 {
+  gpmem_t ltop;
   long sx,sy;
-  long ltop;
   GEN z;
 
   if (typ(x) != t_INT || typ(y) != t_INT)
