@@ -1,4 +1,4 @@
-/* $Id: base5.c,v 1.18 2001/12/06 17:30:28 karim Exp $
+/* $Id: base5.c,v 1.25 2002/05/07 21:19:20 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -20,7 +20,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 /*                                                                 */
 /*******************************************************************/
 #include "pari.h"
-GEN mat_to_vecpol(GEN x, long v);
+extern GEN mul_content(GEN cx, GEN cy);
 
 GEN
 matbasistoalg(GEN nf,GEN x)
@@ -45,7 +45,7 @@ GEN
 matalgtobasis(GEN nf,GEN x)
 {
   long i,j,lx,li;
-  GEN p1,z;
+  GEN p1,z,c;
 
   if (typ(x)!=t_MAT)
     err(talker,"argument must be a matrix in matalgtobasis");
@@ -55,7 +55,12 @@ matalgtobasis(GEN nf,GEN x)
   for (j=1; j<lx; j++)
   {
     p1=cgetg(li,t_COL); z[j]=(long)p1;
-    for (i=1; i<li; i++) p1[i]=(long)algtobasis(nf,gcoeff(x,i,j));
+    for (i=1; i<li; i++)
+    {
+      c = gcoeff(x,i,j);
+      c = typ(c)==t_COL? gcopy(c): algtobasis(nf,c);
+      p1[i]=(long)c;
+    }
   }
   return z;
 }
@@ -111,38 +116,21 @@ rnfmakematrices(GEN rnf)
 GEN
 rnfinitalg(GEN nf,GEN pol,long prec)
 {
-  ulong av = avma;
-  long m,n,r1,r2,vnf,i,j,k,vpol,v1,r1j,r2j,lfac,degabs;
+  gpmem_t av = avma;
+  long m,n,r1,r2,vnf,i,j,k,vpol,r1j,r2j,lfac,degabs;
   GEN RES,sig,rac,p1,p2,liftpol,delta,RAC,ro,p3,bas;
-  GEN f,f2,fac,fac1,fac2,id,p4,p5;
+  GEN f,f2,fac,fac1,fac2,id,p4;
 
   if (typ(pol)!=t_POL) err(notpoler,"rnfinitalg");
   nf=checknf(nf); n=degpol(pol); vpol=varn(pol);
-  vnf=0;
-  for (i=0; i<=n; i++)
-  {
-    long tp1;
+  pol = fix_relative_pol(nf,pol,0);
+  vnf = varn(nf[1]);
 
-    p1=(GEN)pol[i+2];
-    tp1=typ(p1);
-    if (! is_const_t(tp1))
-    {
-      if (tp1!=t_POLMOD) err(typeer,"rnfinitalg");
-      p1 = checknfelt_mod(nf, p1, "rnfinitalg");
-      if (! is_const_t(typ(p1)))
-      {
-	v1=varn(p1);
-	if (vnf && vnf!=v1) err(talker,"different variables in rnfinitalg");
-	if (!vnf) vnf=v1;
-      }
-    }
-  }
-  if (!vnf) vnf=varn(nf[1]);
-  if (vpol>=vnf)
+  if (vpol >= vnf)
     err(talker,"main variable must be of higher priority in rnfinitalg");
   RES=cgetg(12,t_VEC);
   RES[1]=(long)pol;
-  m=degpol(nf[1]); degabs=n*m;
+  m = degpol(nf[1]); degabs=n*m;
   r1 = nf_get_r1(nf); r2 = (m-r1) >> 1;
   sig=cgetg(r1+r2+1,t_VEC); RES[2]=(long)sig;
   rac=(GEN)nf[6]; liftpol=lift(pol);
@@ -179,10 +167,11 @@ rnfinitalg(GEN nf,GEN pol,long prec)
   RES[7]=(long)bas;
   RES[8]=linvmat(p2);
 
-  f2=idealdiv(nf,discsr(pol),(GEN)p1[3]);
-  fac=idealfactor(nf,f2);
-  fac1=(GEN)fac[1]; fac2=(GEN)fac[2]; lfac=lg(fac1)-1;
-  f=idmat(m);
+  f2 = idealdiv(nf, discsr(pol), (GEN)p1[3]);
+  fac = idealfactor(nf,f2);
+  fac1 = (GEN)fac[1];
+  fac2 = (GEN)fac[2]; lfac = lg(fac1)-1;
+  f = idmat(m);
   for (i=1; i<=lfac; i++)
   {
     if (mpodd((GEN)fac2[i])) err(bugparier,"rnfinitalg (odd exponent)");
@@ -198,26 +187,26 @@ rnfinitalg(GEN nf,GEN pol,long prec)
   p4=cgetg(degabs+1,t_MAT);
   for (i=1; i<=n; i++)
   { /* removing denominators speeds up multiplication */
-    GEN cop3,com, om = rnfelementreltoabs(RES,gmael(bas,1,i));
+    GEN c, cop3,com, om = rnfelementreltoabs(RES,gmael(bas,1,i));
 
     if (DEBUGLEVEL>1) msgtimer("i = %ld",i);
-    com = content(om); om = gdiv(om,com);
+    om = primitive_part(om, &com);
     id=gmael(bas,2,i);
     for (j=1; j<=m; j++)
     {
-      p5=cgetg(degabs+1,t_COL); p4[(i-1)*m+j]=(long)p5;
-      p1=gmul((GEN)nf[7],(GEN)id[j]);
-      p3 = gsubst(p1,varn(nf[1]), (GEN)p2[2]);
-      cop3 = content(p3); p3 = gdiv(p3,cop3);
-      p3 = gmul(gmul(com,cop3), lift_intern(gmul(om,p3)));
+      p1 = gmul((GEN)nf[7],(GEN)id[j]);
+      p3 = gsubst(p1, vnf, (GEN)p2[2]);
+      p3 = primitive_part(p3, &cop3);
+      c = mul_content(cop3, com);
 
-      for (k=1; k<lgef(p3)-1; k++) p5[k]=p3[k+1];
-      for (   ; k<=degabs;    k++) p5[k]=zero;
+      p3 = lift_intern(gmul(om,p3));
+      if (c) p3 = gmul(c,p3);
+      p4[(i-1)*m+j] = (long)pol_to_vec(p3, degabs);
     }
   }
   if (DEBUGLEVEL>1) msgtimer("p4");
-  p3 = denom(p4); 
-  p4 = hnfmodid(gmul(p3,p4), p3);
+  p4 = Q_remove_denom(p4, &p3);
+  if (p3) p4 = hnfmodid(p4, p3); else p4 = idmat(degabs);
   if (DEBUGLEVEL>1) msgtimer("hnfmod");
   for (j=degabs-1; j>0; j--)
     if (cmpis(gcoeff(p4,j,j),2) > 0)
@@ -228,7 +217,7 @@ rnfinitalg(GEN nf,GEN pol,long prec)
           for (i=1; i<=j; i++)
             coeff(p4,i,k)=lsubii(gcoeff(p4,i,k),gcoeff(p4,i,j));
     }
-  p4 = gdiv(p4,p3);
+  if (p3) p4 = gdiv(p4,p3);
   p2[4]=(long)mat_to_vecpol(p4,vpol);
   p2[5]=linvmat(p4);
   return gerepilecopy(av,RES);
@@ -237,7 +226,8 @@ rnfinitalg(GEN nf,GEN pol,long prec)
 GEN
 rnfbasistoalg(GEN rnf,GEN x)
 {
-  long tx=typ(x),lx=lg(x),av=avma,tetpil,i,n;
+  long tx=typ(x), lx=lg(x), i, n;
+  gpmem_t av=avma, tetpil;
   GEN p1,z,nf;
 
   checkrnf(rnf); nf=(GEN)rnf[10];
@@ -284,7 +274,8 @@ extern GEN mulmat_pol(GEN A, GEN x);
 GEN
 rnfalgtobasis(GEN rnf,GEN x)
 {
-  long av=avma,tx=typ(x), i,lx;
+  long tx=typ(x), i, lx;
+  gpmem_t av=avma;
   GEN z;
 
   checkrnf(rnf);
@@ -300,7 +291,7 @@ rnfalgtobasis(GEN rnf,GEN x)
 	err(talker,"not the same number field in rnfalgtobasis");
       x = lift_to_pol(x); /* fall through */
     case t_POL:
-    { /* cf algtobasis_intern */
+    { /* cf algtobasis_i */
       GEN P = (GEN)rnf[1];
       long N = degpol(P);
       if (degpol(x) >= N) x = gres(x,P);
@@ -314,7 +305,8 @@ rnfalgtobasis(GEN rnf,GEN x)
 GEN
 rnfelementreltoabs(GEN rnf,GEN x)
 {
-  long av=avma,tx,i,lx,va,tp3;
+  long tx, i, lx, va, tp3;
+  gpmem_t av=avma;
   GEN z,p1,p2,p3,polabs,teta,alpha,s,k;
 
   checkrnf(rnf); tx=typ(x); lx=lg(x); va=varn((GEN)rnf[1]);
@@ -358,11 +350,22 @@ rnfelementreltoabs(GEN rnf,GEN x)
   }
 }
 
+/* assume x,T,pol t_POL. T defines base field, pol defines rnf over T.
+ * x an absolute element of the extension */
+GEN 
+eltabstorel(GEN x, GEN T, GEN pol, GEN k)
+{
+  GEN theta = gmodulcp(gadd(polx[varn(pol)],
+                            gmul(k, gmodulcp(polx[varn(T)],T))), pol);
+  return poleval(x, theta);
+}
+
 GEN
 rnfelementabstorel(GEN rnf,GEN x)
 {
-  long av=avma,tx,i,lx;
-  GEN z,p1,s,tetap,k,nf;
+  long tx, i, lx;
+  gpmem_t av=avma;
+  GEN z,k,nf,T;
 
   checkrnf(rnf); tx=typ(x); lx=lg(x);
   switch(tx)
@@ -373,14 +376,11 @@ rnfelementabstorel(GEN rnf,GEN x)
       return z;
 
     case t_POLMOD:
-      x=lift_to_pol(x); /* fall through */
+      x = lift_to_pol(x); /* fall through */
     case t_POL:
-      p1=(GEN)rnf[11]; k=(GEN)p1[3]; nf=(GEN)rnf[10];
-      tetap=gmodulcp(gadd(polx[varn(rnf[1])],
-	    gmul(k,gmodulcp(polx[varn(nf[1])],(GEN)nf[1]))),(GEN)rnf[1]);
-      s=gzero;
-      for (i=lgef(x)-1; i>1; i--) s=gadd((GEN)x[i],gmul(tetap,s));
-      return gerepileupto(av,s);
+      k  = (GEN)rnf[11]; k = (GEN)k[3];
+      nf = (GEN)rnf[10]; T = (GEN)nf[1];
+      return gerepileupto(av, eltabstorel(x,T,(GEN)rnf[1],k));
 
     default: return gcopy(x);
   }
@@ -414,7 +414,7 @@ rnfelementup(GEN rnf,GEN x)
 GEN
 rnfelementdown(GEN rnf,GEN x)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long i,lx,tx;
   GEN p1,z;
 
@@ -445,14 +445,14 @@ rnfelementdown(GEN rnf,GEN x)
 static GEN
 rnfprincipaltohermite(GEN rnf,GEN x)
 {
-  long av=avma,tetpil;
+  gpmem_t av=avma, tetpil;
   GEN nf,bas,bas1,p1,z;
 
   x=rnfbasistoalg(rnf,x); nf=(GEN)rnf[10];
   bas=(GEN)rnf[7]; bas1=(GEN)bas[1];
   p1=rnfalgtobasis(rnf,gmul(x,gmodulcp(bas1,(GEN)rnf[1])));
   z=cgetg(3,t_VEC); z[2]=bas[2];
-  settyp(p1,t_MAT); z[1]=(long)matalgtobasis(nf,p1);
+  settyp(p1,t_MAT); z[1]=(long)p1;
 
   tetpil=avma;
   return gerepile(av,tetpil,nfhermite(nf,z));
@@ -461,7 +461,8 @@ rnfprincipaltohermite(GEN rnf,GEN x)
 GEN
 rnfidealhermite(GEN rnf,GEN x)
 {
-  long tx=typ(x),lx=lg(x),av=avma,tetpil,i,j,n,m;
+  long tx=typ(x), lx=lg(x), i, j, n, m;
+  gpmem_t av=avma, tetpil;
   GEN z,p1,p2,x1,x2,x1j,nf,bas,unnf,zeronf;
 
   checkrnf(rnf);
@@ -538,7 +539,8 @@ rnfidealhermite(GEN rnf,GEN x)
 GEN
 rnfidealnormrel(GEN rnf,GEN id)
 {
-  long av=avma,i,n;
+  long i, n;
+  gpmem_t av=avma;
   GEN z,id2,nf;
 
   checkrnf(rnf);
@@ -552,7 +554,8 @@ rnfidealnormrel(GEN rnf,GEN id)
 GEN
 rnfidealnormabs(GEN rnf,GEN id)
 {
-  long av=avma,i,n;
+  long i, n;
+  gpmem_t av=avma;
   GEN z,id2;
 
   checkrnf(rnf);
@@ -565,7 +568,8 @@ rnfidealnormabs(GEN rnf,GEN id)
 GEN
 rnfidealreltoabs(GEN rnf,GEN x)
 {
-  long av=avma,i,j,k,n,m;
+  long i, j, k, n, m;
+  gpmem_t av=avma;
   GEN nf,basinv,om,id,p1,p2,p3,p4,p5,c;
 
   checkrnf(rnf);
@@ -597,7 +601,8 @@ rnfidealreltoabs(GEN rnf,GEN x)
 GEN
 rnfidealabstorel(GEN rnf,GEN x)
 {
-  long av=avma,tetpil,n,m,j,k;
+  long n, m, j, k;
+  gpmem_t av=avma, tetpil;
   GEN nf,basabs,ma,xj,p1,p2,id;
 
   checkrnf(rnf); n=degpol(rnf[1]); nf=(GEN)rnf[10]; m=degpol(nf[1]);
@@ -613,7 +618,6 @@ rnfidealabstorel(GEN rnf,GEN x)
       p2[k+1]=(long)truecoeff(xj,k);
   }
   ma=gmul((GEN)rnf[8],ma);
-  ma=matalgtobasis(nf,ma);
   p1=cgetg(n*m+1,t_VEC); id=idmat(m);
   for (j=1; j<=n*m; j++) p1[j]=(long)id;
   p2=cgetg(3,t_VEC); p2[1]=(long)ma; p2[2]=(long)p1;
@@ -623,7 +627,7 @@ rnfidealabstorel(GEN rnf,GEN x)
 GEN
 rnfidealdown(GEN rnf,GEN x)
 {
-  long av=avma;
+  gpmem_t av=avma;
 
   checkrnf(rnf); x=rnfidealhermite(rnf,x);
   return gerepilecopy(av,gmael(x,2,1));
@@ -633,7 +637,8 @@ rnfidealdown(GEN rnf,GEN x)
 GEN
 rnfidealup(GEN rnf,GEN x)
 {
-  long av=avma,tetpil,i,n,m;
+  long i, n, m;
+  gpmem_t av=avma, tetpil;
   GEN nf,bas,bas2,p1,p2,zeronf,unnf;
 
   checkrnf(rnf);
@@ -651,7 +656,8 @@ rnfidealup(GEN rnf,GEN x)
 GEN
 rnfidealtwoelement(GEN rnf,GEN x)
 {
-  long av=avma,tetpil,j;
+  long j;
+  gpmem_t av=avma, tetpil;
   GEN p1,p2,p3,res,polabs,nfabs,z;
 
   checkrnf(rnf);
@@ -670,7 +676,8 @@ rnfidealtwoelement(GEN rnf,GEN x)
 GEN
 rnfidealmul(GEN rnf,GEN x,GEN y) /* x et y sous HNF relative uniquement */
 {
-  long av=avma,tetpil,i,j,n;
+  long i, j, n;
+  gpmem_t av=avma, tetpil;
   GEN z,nf,x1,x2,p1,p2,p3,p4,p5,res;
 
   z=rnfidealtwoelement(rnf,y);
@@ -691,136 +698,3 @@ rnfidealmul(GEN rnf,GEN x,GEN y) /* x et y sous HNF relative uniquement */
   res[2]=(long)concatsp(x2,x2);
   tetpil=avma; return gerepile(av,tetpil,nfhermite(nf,res));
 }
-
-/*********************************************************************/
-/**                                                                 **/
-/**         LIBRARY FOR POLYNOMIALS WITH COEFFS. IN Z_K/P	    **/
-/**  An element in Z_K/P is a t_COL with degree(nf[1]) components.  **/
-/**  These are integers modulo the prime p under prime ideal P      **/
-/**  (only f(P/p) elements are non zero). These components are      **/
-/**  given on the integer basis of K.                               **/
-/**                                                                 **/
-/*********************************************************************/
-
-/* K.B: What follows is not meant to work (yet?) */
-
-GEN
-polnfmulscal(GEN nf,GEN s,GEN x)
-{
-  long i,lx=lgef(x);
-  GEN z;
-
-  if (lx<3) return gcopy(x);
-  if (gcmp0(s))
-  {
-    z=cgetg(2,t_POL); z[1]=evallgef(2) | evalvarn(varn(x));
-    return z;
-  }
-  z=cgetg(lx,t_POL); z[1]=x[1];
-  for (i=2; i<lx; i++) z[i]=(long)element_mul(nf,s,(GEN)x[i]);
-  return z;
-}
-
-GEN
-polnfmul(GEN nf, GEN x, GEN y)
-{
-  ulong av;
-  long m,i,d,imin,imax,lx,ly,lz;
-  GEN p1,z,zeronf;
-
-  if (gcmp0(x) || gcmp0(y)) return zeropol(varn(x));
-  m=degpol(nf[1]); av=avma;
-  lx=degpol(x); ly=degpol(y); lz=lx+ly;
-  zeronf=gscalcol_i(gzero,m);
-  z=cgetg(lz+3,t_POL);
-  z[1] = evallgef(lz+3) | evalvarn(x) | evalsigne(1);
-  for (d=0; d<=lz; d++)
-  {
-    p1=zeronf; imin=max(0,d-ly); imax=min(d,lx);
-    for (i=imin; i<=imax; i++)
-      p1=gadd(p1,element_mul(nf,(GEN)x[i+2],(GEN)y[d-i+2]));
-    z[d+2]=(long)p1;
-  }
-  return gerepilecopy(av,z);
-}
-
-/* division euclidienne */
-GEN
-polnfdeuc(GEN nf, GEN x, GEN y, GEN *ptr)
-{
-  long av=avma,m,i,d,tx,lx,ly,lz,fl;
-  GEN z,unnf,lcy,r;
-  GEN *gptr[2];
-
-  if (gcmp0(y)) err(talker,"division by zero in polnfdiv");
-  lx=lgef(x); ly=lgef(y); lz=lx-ly;
-  if (gcmp0(x) || lz<0) { *ptr=gcopy(x); return zeropol(varn(x)); }
-
-  m=degpol(nf[1]); unnf=gscalcol_i(gun,m);
-  x=dummycopy(x); y=dummycopy(y);
-  for (i=2; i<lx; i++)
-  {
-    tx=typ(x[i]);
-    if (is_intreal_t(tx) || tx == t_INTMOD || is_frac_t(tx))
-      x[i]=lmul((GEN)x[i],unnf);
-  }
-  for (i=2; i<ly; i++)
-  {
-    tx=typ(y[i]);
-    if (is_intreal_t(tx) || tx == t_INTMOD || is_frac_t(tx))
-      y[i]=lmul((GEN)y[i],unnf);
-  }
-
-  lz += 3;
-  z=cgetg(lz,t_POL); z[1]=evallgef(lz) | evalvarn(x) | evalsigne(1);
-  lcy=(GEN)y[ly-1];
-  if (gegal(lift(lcy),unnf)) fl=0;
-  else
-  {
-    fl=1; y=polnfmulscal(nf,element_inv(nf,lcy),y);
-  }
-  for (d=lz-1; d>=2; d--)
-  {
-    z[d]=x[d+ly-3];
-    for (i=d; i<d+ly-3; i++)
-      x[i]=lsub((GEN)x[i],element_mul(nf,(GEN)z[d],(GEN)y[i-d-2]));
-  }
-  if (fl) z=polnfmulscal(nf,lcy,z);
-
-  for(;;)
-  {
-    if (!gcmp0((GEN)x[d]))
-    {
-      r=cgetg(d,t_POL);
-      r[1] = evallgef(d) | evalvarn(varn(x)) | evalsigne(1);
-      for (i=2; i<d; i++) r[i]=x[i];
-      break;
-    }
-    if (d==2) { r = zeropol(varn(x)); break; }
-    d--;
-  }
-  *ptr=r; gptr[0]=ptr; gptr[1]=&z;
-  gerepilemany(av,gptr,2); return z;
-}
-
-GEN
-polnfpow(GEN nf,GEN x,GEN k)
-{
-  long s,av=avma,m;
-  GEN y,z;
-
-  m=degpol(nf[1]);
-  if (typ(k)!=t_INT) err(talker,"not an integer exponent in nfpow");
-  s=signe(k); if (s<0) err(impl,"polnfpow for negative exponents");
-
-  z=x; y=cgetg(3,t_POL);
-  y[1] = evallgef(3) | evalvarn(varn(x)) | evalsigne(1);
-  y[2] = (long)gscalcol_i(gun,m);
-  for(;;)
-  {
-    if (mpodd(k)) y=polnfmul(nf,z,y);
-    k=shifti(k,-1);
-    if (!signe(k)) { cgiv(k); return gerepileupto(av,y); }
-    z=polnfmul(nf,z,z);
-  }
-}
