@@ -1,4 +1,4 @@
-/* $Id: buch1.c,v 1.37 2001/11/22 22:38:08 karim Exp $
+/* $Id: buch1.c,v 1.46 2002/06/08 14:02:29 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -45,7 +45,13 @@ static GEN  **powsubfactorbase,vperm,subfactorbase,Disc,sqrtD,isqrtD;
 
 GEN buchquad(GEN D, double c, double c2, long RELSUP0, long flag, long prec);
 extern GEN roots_to_pol_intern(GEN L, GEN a, long v, int plus);
-extern GEN colreducemodmat(GEN x, GEN y, GEN *Q);
+extern GEN colreducemodHNF(GEN x, GEN y, GEN *Q);
+extern GEN quadhilbertreal(GEN D, GEN flag, long prec);
+extern void comp_gen(GEN z,GEN x,GEN y);
+extern GEN codeform5(GEN x, long prec);
+extern GEN comprealform5(GEN x, GEN y, GEN D, GEN sqrtD, GEN isqrtD);
+extern GEN redrealform5(GEN x, GEN D, GEN sqrtD, GEN isqrtD);
+extern GEN rhoreal_aux(GEN x, GEN D, GEN sqrtD, GEN isqrtD);
 
 GEN
 quadclassunit0(GEN x, long flag, GEN data, long prec)
@@ -209,11 +215,12 @@ gpq(GEN form, GEN p, GEN q, long e, GEN sqd, GEN u, long prec)
 static GEN
 quadhilbertimag(GEN D, GEN flag)
 {
-  long av=avma,h,i,e,prec;
+  long h, i, e, prec;
+  gpmem_t av=avma;
   GEN z,L,P,p,q,qfp,qfq,up,uq,u;
   int raw = ((typ(flag)==t_INT && signe(flag)));
 
-  if (DEBUGLEVEL>=2) timer2();
+  if (DEBUGLEVEL>=2) (void)timer2();
   if (gcmpgs(D,-11) >= 0) return polx[0];
   L = getallforms(D,&h,&z);
   if (DEBUGLEVEL>=2) msgtimer("class number = %ld",h);
@@ -236,7 +243,8 @@ quadhilbertimag(GEN D, GEN flag)
   prec = raw? DEFAULTPREC: 3;
   for(;;)
   {
-    long av0 = avma, ex, exmax = 0;
+    long ex, exmax = 0;
+    gpmem_t av0 = avma;
     GEN lead, sqd = gsqrt(negi(D),prec);
     P = cgetg(h+1,t_VEC); 
     for (i=1; i<=h; i++)
@@ -270,8 +278,6 @@ quadhilbertimag(GEN D, GEN flag)
   return gerepileupto(av,P);
 }
 
-GEN quadhilbertreal(GEN D, GEN flag, long prec);
-
 GEN
 quadhilbert(GEN D, GEN flag, long prec)
 {
@@ -418,14 +424,6 @@ get_prec(GEN P, long prec)
 }
 
 /* AUXILLIARY ROUTINES FOR QUADRAYSIGMA */
-GEN
-PiI2(long prec)
-{
-  GEN z = cgetg(3,t_COMPLEX);
-  GEN x = mppi(prec); setexpo(x,2);
-  z[1] = zero;
-  z[2] = (long)x; return z; /* = 2I Pi */
-}
 
 /* Compute data for ellphist */
 static GEN
@@ -474,7 +472,8 @@ computeth2(GEN om, GEN la, long prec)
 static GEN
 computeP2(GEN bnr, GEN la, int raw, long prec)
 {
-  long av=avma, av2, clrayno,i, first = 1;
+  long clrayno, i, first = 1;
+  gpmem_t av=avma, av2;
   GEN listray,P0,P,f,lanum, nf = checknf(bnr);
   
   f = gmael3(bnr,2,1,1);
@@ -542,7 +541,8 @@ galoisapplypol(GEN nf, GEN s, GEN x)
 static GEN
 findquad(GEN a, GEN x, GEN p)
 {
-  long tu,tv, av = avma;
+  long tu, tv;
+  gpmem_t av = avma;
   GEN u,v;
   if (typ(x) == t_POLMOD) x = (GEN)x[2];
   if (typ(a) == t_POLMOD) a = (GEN)a[2];
@@ -644,14 +644,15 @@ treatspecialsigma(GEN nf, GEN gf, int raw, long prec)
     return NULL;
   }
 
-  p1 = gcoeff(gf,1,1);
+  p1 = gcoeff(gf,1,1); /* integer > 0 */
   p2 = gcoeff(gf,2,2);
   if (gcmp1(p2)) { fl = 0; tryf = p1; }
   else {
     if (Ds % 16 != 8 || !egalii(content(gf),gdeux)) return NULL;
     fl = 1; tryf = shifti(p1,-1);
   }
-  if (cmpis(tryf, 3) <= 0 || !gcmp0(resii(D, tryf)) || !isprime(tryf))
+  /* tryf integer > 0 */
+  if (cmpis(tryf, 3) <= 0 || signe(resii(D, tryf)) || !isprime(tryf))
     return NULL;
 
   i = itos(tryf); if (fl) i *= 4;
@@ -700,7 +701,7 @@ quadrayimagsigma(GEN bnr, int raw, long prec)
   f2 = 2 * itos(gcoeff(f,1,1));
   u = getallrootsof1(bnf); lu = lg(u);
   for (i=1; i<lu; i++)
-    u[i] = (long)colreducemodmat((GEN)u[i], f, NULL); /* roots of 1, mod f */
+    u[i] = (long)colreducemodHNF((GEN)u[i], f, NULL); /* roots of 1, mod f */
   if (DEBUGLEVEL>1)
     fprintferr("quadray: looking for [a,b] != unit mod 2f\n[a,b] = ");
   for (a=0; a<f2; a++)
@@ -711,7 +712,7 @@ quadrayimagsigma(GEN bnr, int raw, long prec)
       if (DEBUGLEVEL>1) fprintferr("[%ld,%ld] ",a,b);
 
       labas = algtobasis(nf, la);
-      lamodf = colreducemodmat(labas, f, NULL);
+      lamodf = colreducemodHNF(labas, f, NULL);
       for (i=1; i<lu; i++)
         if (gegal(lamodf, (GEN)u[i])) break;
       if (i < lu) continue; /* la = unit mod f */
@@ -730,7 +731,8 @@ GEN
 quadray(GEN D, GEN f, GEN flag, long prec)
 {
   GEN bnr,y,p1,pol,bnf,lambda;
-  long av = avma, raw;
+  long raw;
+  gpmem_t av = avma;
 
   if (!flag) flag = gzero;
   if (typ(flag)==t_INT) lambda = NULL;
@@ -780,12 +782,6 @@ quadray(GEN D, GEN f, GEN flag, long prec)
 /*  Routines related to binary quadratic forms (for internal use)  */
 /*                                                                 */
 /*******************************************************************/
-extern void comp_gen(GEN z,GEN x,GEN y);
-extern GEN codeform5(GEN x, long prec);
-extern GEN comprealform5(GEN x, GEN y, GEN D, GEN sqrtD, GEN isqrtD);
-extern GEN redrealform5(GEN x, GEN D, GEN sqrtD, GEN isqrtD);
-extern GEN rhoreal_aux(GEN x, GEN D, GEN sqrtD, GEN isqrtD);
-
 #define rhorealform(x) rhoreal_aux(x,Disc,sqrtD,isqrtD)
 #define redrealform(x) gcopy(fix_signs(redrealform5(x,Disc,sqrtD,isqrtD)))
 #define comprealform(x,y) fix_signs(comprealform5(x,y,Disc,sqrtD,isqrtD))
@@ -794,7 +790,8 @@ extern GEN rhoreal_aux(GEN x, GEN D, GEN sqrtD, GEN isqrtD);
 static GEN 
 rhoreal_pow(GEN x, long n)
 {
-  long i, av = avma, lim = stack_lim(av,1);
+  long i;
+  gpmem_t av = avma, lim = stack_lim(av, 1);
   for (i=1; i<=n; i++)
   {
     x = rhorealform(x);
@@ -823,7 +820,7 @@ fix_signs(GEN x)
 static GEN
 redrealprimeform5(GEN Disc, long p)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN y = primeform(Disc,stoi(p),PRECREG);
   y = codeform5(y,PRECREG);
   return gerepileupto(av, redrealform(y));
@@ -832,7 +829,7 @@ redrealprimeform5(GEN Disc, long p)
 static GEN
 redrealprimeform(GEN Disc, long p)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN y = primeform(Disc,stoi(p),PRECREG);
   return gerepileupto(av, redrealform(y));
 }
@@ -840,7 +837,7 @@ redrealprimeform(GEN Disc, long p)
 static GEN
 comprealform3(GEN x, GEN y)
 {
-  long av = avma;
+  gpmem_t av = avma;
   GEN z = cgetg(4,t_VEC); comp_gen(z,x,y);
   return gerepileupto(av, redrealform(z));
 }
@@ -864,7 +861,7 @@ initrealform5(long *ex)
 static void
 buch_init(void)
 {
-  if (DEBUGLEVEL) timer2();
+  if (DEBUGLEVEL) (void)timer2();
   primfact  = new_chunk(100);
   primfact1 = new_chunk(100);
   exprimfact  = new_chunk(100);
@@ -886,7 +883,8 @@ check_bach(double cbach, double B)
 static long
 factorisequad(GEN f, long kcz, long limp)
 {
-  long i,p,k,av,lo;
+  long i, p, k, lo;
+  gpmem_t av;
   GEN q,r, x = (GEN)f[1];
 
   if (is_pm1(x)) { primfact[0]=0; return 1; }
@@ -959,7 +957,8 @@ badmod8(GEN x)
 static void
 factorbasequad(GEN Disc, long n2, long n)
 {
-  long i,p,bad, av = avma;
+  long i, p, bad;
+  gpmem_t av = avma;
   byteptr d=diffptr;
 
   numfactorbase = (long*) gpmalloc(sizeof(long)*(n2+1));
@@ -1100,7 +1099,8 @@ desalloc(long **mat)
 static GEN
 Lval1(GEN Disc)
 {
-  long av=avma, p;
+  long p;
+  gpmem_t av=avma;
   GEN y=realun(DEFAULTPREC);
   byteptr d=diffptr;
 
@@ -1138,7 +1138,8 @@ get_clgp(GEN Disc, GEN W, GEN *ptmet, long prec)
 static GEN
 extra_relations(long LIMC, long *ex, long nlze, GEN extramatc)
 {
-  long av,fpc,p,ep,i,j,k,nlze2, *col, *colg, s = 0, extrarel = nlze+2;
+  long fpc, p, ep, i, j, k, nlze2, *col, *colg, s = 0, extrarel = nlze+2;
+  gpmem_t av;
   long MAXRELSUP = min(50,4*KC);
   GEN p1,form, extramat = cgetg(extrarel+1,t_MAT);
 
@@ -1219,7 +1220,8 @@ extra_relations(long LIMC, long *ex, long nlze, GEN extramatc)
 static GEN
 imag_random_form(long current, long *ex)
 {
-  long av = avma,i;
+  long i;
+  gpmem_t av = avma;
   GEN form,pc;
 
   for(;;)
@@ -1240,7 +1242,8 @@ static void
 imag_relations(long lim, long s, long LIMC, long *ex, long **mat)
 {
   static long nbtest;
-  long av = avma, i,j,pp,fpc,b1,b2,ep,current, first = (s==0);
+  long i, j, pp, fpc, b1, b2, ep, current, first = (s==0);
+  gpmem_t av = avma;
   long *col,*fpd,*oldfact,*oldexp;
   GEN pc,form,form1;
 
@@ -1277,7 +1280,7 @@ imag_relations(long lim, long s, long LIMC, long *ex, long **mat)
       if (DEBUGLEVEL) { fprintferr(" %ld",s); flusherr(); }
       oldfact = primfact; oldexp = exprimfact;
       primfact = primfact1; exprimfact = exprimfact1;
-      factorisequad(form1,KC,LIMC);
+      (void)factorisequad(form1,KC,LIMC);
 
       if (b1==b2)
       {
@@ -1334,7 +1337,8 @@ imag_relations(long lim, long s, long LIMC, long *ex, long **mat)
 static int
 imag_be_honest(long *ex)
 {
-  long p,fpc, s = KC, nbtest = 0, av = avma;
+  long p, fpc, s = KC, nbtest = 0;
+  gpmem_t av = avma;
   GEN form;
 
   while (s<KC2)
@@ -1360,7 +1364,8 @@ imag_be_honest(long *ex)
 static GEN
 real_random_form(long *ex)
 {
-  long av = avma, i;
+  long i;
+  gpmem_t av = avma;
   GEN p1,form = NULL;
 
   for(;;)
@@ -1384,8 +1389,9 @@ real_relations(long lim, long s, long LIMC, long *ex, long **mat, GEN glog2,
                GEN vecexpo)
 {
   static long nbtest;
-  long av = avma,av1,i,j,p,fpc,b1,b2,ep,current, first = (s==0);
-  long *col,*fpd,*oldfact,*oldexp,limstack;
+  long i, j, p, fpc, b1, b2, ep, current, first = (s==0);
+  gpmem_t av = avma, av1, limstack;
+  long *col,*fpd,*oldfact,*oldexp;
   long findecycle,nbrhocumule,nbrho;
   GEN p1,p2,form,form0,form1,form2;
 
@@ -1484,7 +1490,7 @@ real_relations(long lim, long s, long LIMC, long *ex, long **mat, GEN glog2,
         s++; col = mat[s]; if (DEBUGLEVEL) fprintferr(" %ld",s);
         oldfact = primfact; oldexp = exprimfact;
         primfact = primfact1; exprimfact = exprimfact1;
-        factorisequad(form2,KC,LIMC);
+        (void)factorisequad(form2,KC,LIMC);
         if (b1==b2)
         {
           for (i=1; i<=lgsub; i++)
@@ -1559,7 +1565,8 @@ real_relations(long lim, long s, long LIMC, long *ex, long **mat, GEN glog2,
 static int
 real_be_honest(long *ex)
 {
-  long p,fpc,s = KC,nbtest = 0,av = avma;
+  long p, fpc, s = KC, nbtest = 0;
+  gpmem_t av = avma;
   GEN p1,form,form0;
 
   while (s<KC2)
@@ -1602,18 +1609,18 @@ gcdreal(GEN a,GEN b,long *pte)
   if (typ(a)==t_INT)
   {
     if (typ(b)==t_INT) return mppgcd(a,b);
-    k1=cgetr(lg(b)); affir(a,k1); a=k1;
+    a = itor(a, lg(b));
   }
   else if (typ(b)==t_INT)
   {
-    k1=cgetr(lg(a)); affir(b,k1); b=k1;
+    b = itor(b, lg(a));
   }
   if (expo(a)<-5) return absr(b);
   if (expo(b)<-5) return absr(a);
   a=absr(a); b=absr(b);
   while (expo(b) >= -5  && signe(b))
   {
-    k1=gcvtoi(divrr(a,b),&e);
+    k1 = gcvtoi(divrr(a,b),&e);
     if (e > 0) return NULL;
     r=subrr(a,mulir(k1,b)); a=b; b=r;
   }
@@ -1644,7 +1651,7 @@ get_reg(GEN matc, long sreg)
 GEN
 buchquad(GEN D, double cbach, double cbach2, long RELSUP0, long flag, long prec)
 {
-  ulong av0 = avma, av;
+  gpmem_t av0 = avma, av;
   long KCCO,KCCOPRO,i,j,s, *ex,**mat;
   long extrarel,nrelsup,nreldep,LIMC,LIMC2,cp,nbram,nlze;
   GEN p1,h,W,met,res,basecl,dr,pdep,C,B,extramat,extraC;
@@ -1677,7 +1684,9 @@ buchquad(GEN D, double cbach, double cbach2, long RELSUP0, long flag, long prec)
   if (!isfundamental(Disc))
     err(warner,"not a fundamental discriminant in quadclassunit");
   buch_init(); RELSUP = RELSUP0;
-  dr=cgetr(3); affir(Disc,dr); drc=fabs(rtodbl(dr)); LOGD=log(drc);
+  dr = itor(Disc, 3);
+  drc = fabs(rtodbl(dr));
+  LOGD = log(drc);
 
   lim = sqrt(drc);
   resc = dbltor(lim); Res = Lval1(Disc);
@@ -1703,7 +1712,7 @@ INCREASE: avma = av; cbach = check_bach(cbach,6.);
   if (PRECREG)
   {
     PRECREG = max(prec+1, MEDDEFAULTPREC + 2*(expi(Disc)>>TWOPOTBITS_IN_LONG));
-    glog2  = glog(gdeux,PRECREG);
+    glog2  = mplog2(PRECREG);
     sqrtD  = gsqrt(Disc,PRECREG);
     isqrtD = gfloor(sqrtD);
   }
