@@ -1,4 +1,4 @@
-/* $Id: buch2.c,v 1.120 2001/12/18 21:41:26 karim Exp $
+/* $Id: buch2.c,v 1.150 2002/06/07 21:13:26 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -21,6 +21,10 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 /*******************************************************************/
 #include "pari.h"
 #include "parinf.h"
+extern GEN nfbasic_to_nf(nfbasic_t *T, GEN ro, long prec);
+extern GEN get_nfindex(GEN bas);
+extern GEN sqred1_from_QR(GEN x, long prec);
+extern GEN computeGtwist(GEN nf, GEN vdir);
 extern GEN famat_to_arch(GEN nf, GEN fa, long prec);
 extern GEN to_famat_all(GEN x, GEN y);
 extern int addcolumntomatrix(GEN V, GEN invp, GEN L);
@@ -28,17 +32,14 @@ extern double check_bach(double cbach, double B);
 extern GEN gmul_mat_smallvec(GEN x, GEN y);
 extern GEN gmul_mati_smallvec(GEN x, GEN y);
 extern GEN get_arch_real(GEN nf,GEN x,GEN *emb,long prec);
-extern GEN get_roots(GEN x,long r1,long ru,long prec);
-extern void get_nf_matrices(GEN nf, long small);
-extern long int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *t, long v);
+extern GEN get_roots(GEN x,long r1,long prec);
+extern long int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *t);
 extern GEN init_idele(GEN nf);
 extern GEN norm_by_embed(long r1, GEN x);
 extern void minim_alloc(long n,double ***q,long **x,double **y,double **z,double **v);
-extern GEN idealmulpowprime(GEN nf, GEN x, GEN vp, GEN n);
 extern GEN arch_mul(GEN x, GEN y);
 extern GEN vecdiv(GEN x, GEN y);
 extern GEN vecmul(GEN x, GEN y);
-extern GEN mul_real(GEN x, GEN y);
 
 #define SFB_MAX 2
 #define SFB_STEP 2
@@ -103,7 +104,8 @@ desallocate(GEN **M)
 static GEN
 subFBgen(long N,long m,long minsFB,GEN vperm, long *ptss)
 {
-  long av = avma,i,j, lv=lg(vectbase),s=0,s1=0,n=0,ss=0,z=0;
+  long i, j, lv=lg(vectbase), s=0, s1=0, n=0, ss=0, z=0;
+  gpmem_t av = avma;
   GEN y1,y2,subFB,perm,perm1,P,Q;
   double prod;
 
@@ -174,7 +176,7 @@ subFBgen(long N,long m,long minsFB,GEN vperm, long *ptss)
 static GEN
 mulred(GEN nf,GEN x, GEN I, long prec)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN y = cgetg(3,t_VEC);
 
   y[1] = (long)idealmulh(nf,I,(GEN)x[1]);
@@ -228,8 +230,7 @@ FBgen(GEN nf,long n2,long n)
 {
   byteptr delta=diffptr;
   long KC2,i,j,k,p,lon,ip,nor, N = degpol(nf[1]);
-  GEN p2,p1,NormP,Res;
-  long prim[] = { evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3),0 };
+  GEN prim,p2,p1,NormP,Res;
 
   numFB    = cgetg(n2+1,t_VECSMALL);
   FB       = cgetg(n2+1,t_VECSMALL);
@@ -237,10 +238,10 @@ FBgen(GEN nf,long n2,long n)
   idealbase= (GEN*)cgetg(n2+1,t_VEC);
 
   Res=realun(DEFAULTPREC);
-  p=*delta++; i=0; ip=0; KC=0;
+  prim = icopy(gun); p=*delta++; i=0; ip=0; KC=0;
   while (p<=n2)
   {
-    long av = avma, av1;
+    gpmem_t av = avma, av1;
     if (DEBUGLEVEL>=2) { fprintferr(" %ld",p); flusherr(); }
     prim[2] = p; p1 = primedec(nf,prim); lon=lg(p1);
     av1 = avma;
@@ -368,7 +369,7 @@ factorgen(GEN nf,GEN I,GEN m,long kcz,long limp)
 
 /* can we factor x ? Nx = norm(x) */
 static long
-factorelt(GEN nf,GEN cbase,GEN x,GEN Nx,long kcz,long limp)
+factorelt(GEN nf,GEN x,GEN Nx,long kcz,long limp)
 {
   long i,j,n1,ip,v,p,k,lo,ifinal;
   GEN q,r,P,p1,listexpo;
@@ -385,7 +386,6 @@ factorelt(GEN nf,GEN cbase,GEN x,GEN Nx,long kcz,long limp)
   }
   if (cmpis(Nx,limp) > 0) return 0;
 
-  if (cbase) x = gmul(cbase,x);
   ifinal=i; lo = 0;
   for (i=1; i<=ifinal; i++)
   {
@@ -397,7 +397,7 @@ factorelt(GEN nf,GEN cbase,GEN x,GEN Nx,long kcz,long limp)
       for (j=1; j<n1; j++)
       {
         P = (GEN)p1[j];
-	v = int_elt_val(nf,x,(GEN)P[1],(GEN)P[5], NULL, k);
+	v = int_elt_val(nf,x,(GEN)P[1],(GEN)P[5], NULL);
 	if (v)
 	{
 	  primfact[++lo]=ip+j; expoprimfact[lo]=v;
@@ -415,7 +415,7 @@ factorelt(GEN nf,GEN cbase,GEN x,GEN Nx,long kcz,long limp)
   for (k=1,j=1; j<n1; j++)
   {
     P = (GEN)p1[j];
-    v = int_elt_val(nf,x,(GEN)P[1],(GEN)P[5], NULL, k);
+    v = int_elt_val(nf,x,(GEN)P[1],(GEN)P[5], NULL);
     if (v)
     {
       primfact[++lo]=ip+j; expoprimfact[lo]=v;
@@ -429,7 +429,8 @@ factorelt(GEN nf,GEN cbase,GEN x,GEN Nx,long kcz,long limp)
 static GEN
 cleancol(GEN x,long N,long PRECREG)
 {
-  long i,j,av,tetpil,tx=typ(x),R1,RU;
+  long i, j, tx=typ(x), R1, RU;
+  gpmem_t av, tetpil;
   GEN s,s2,re,pi4,im,y;
 
   if (tx==t_MAT)
@@ -459,7 +460,7 @@ cleancol(GEN x,long N,long PRECREG)
 #define LARGE 1
 #define PRECI 2
 static GEN
-not_given(long av, long flun, long reason)
+not_given(gpmem_t av, long flun, long reason)
 {
   if (labs(flun)==2)
   {
@@ -529,7 +530,8 @@ gauss_realimag(GEN x, GEN y)
 GEN
 getfu(GEN nf,GEN *ptxarch,GEN reg,long flun,long *pte,long prec)
 {
-  long av = avma,e,i,j,R1,RU,N=degpol(nf[1]);
+  long e, i, j, R1, RU, N=degpol(nf[1]);
+  gpmem_t av = avma;
   GEN p1,p2,u,y,matep,s,xarch,vec;
   GEN *gptr[2];
 
@@ -550,7 +552,7 @@ getfu(GEN nf,GEN *ptxarch,GEN reg,long flun,long *pte,long prec)
     for (   ; i<=RU; i++) p1[i] = ladd(s, gmul2n(gcoeff(xarch,i,j),-1));
   }
   if (prec <= 0) prec = gprecision(xarch);
-  u = lllintern(greal(matep),1,prec);
+  u = lllintern(greal(matep),100,1,prec);
   if (!u) return not_given(av,flun,PRECI);
 
   p1 = gmul(matep,u);
@@ -595,7 +597,8 @@ getfu(GEN nf,GEN *ptxarch,GEN reg,long flun,long *pte,long prec)
 GEN
 buchfu(GEN bnf)
 {
-  long av = avma, c;
+  long c;
+  gpmem_t av = avma;
   GEN nf,xarch,reg,res, y = cgetg(3,t_VEC);
 
   bnf = checkbnf(bnf); xarch = (GEN)bnf[3]; nf = (GEN)bnf[7];
@@ -665,7 +668,8 @@ get_norm_fact(GEN gen, GEN ex, GEN *pd)
 static long
 factorgensimple(GEN nf,GEN ideal)
 {
-  long N,i,v,av1 = avma,lo, L = lg(vectbase);
+  long N, i, v, lo, L = lg(vectbase);
+  gpmem_t av1 = avma;
   GEN x;
   if (typ(ideal) != t_MAT) ideal = (GEN)ideal[1]; /* idele */
   x = dethnf_i(ideal);
@@ -698,6 +702,7 @@ factorgensimple(GEN nf,GEN ideal)
     { /* divisible by P | p not in FB */
       long k,l,n;
       k = N - sum_ef;
+      if (k == 0) err(talker,"not an ideal in factorgensimple: %Z", ideal);
       if (vx % k) break;
       k = vx / k; /* ideal / p^k may factor on FB */
       for (l = lo0+1; l <= lo; l++)
@@ -775,14 +780,14 @@ split_ideal(GEN nf, GEN x0, long prec, GEN vperm)
 
   z = init_idele(nf); ru = lg(z[2]);
   z[2] = lgetg(1, t_MAT);
-  vdir = cgetg(ru,t_VEC);
-  for (i=2; i<ru; i++) vdir[i]=zero;
+  vdir = cgetg(ru,t_VECSMALL);
+  for (i=2; i<ru; i++) vdir[i]=0;
   for (i=1; i<ru; i++)
   {
-    vdir[i]=lstoi(10);
+    vdir[i] = 10;
     y = ideallllred(nf,x0,vdir,prec);
     if (factorgensimple(nf,y)) return y;
-    vdir[i]=zero;
+    vdir[i] = 0;
   }
   nbtest = 0; nbtest_lim = (ru-1) << 2; lgsub = 3;
   init_sub(lgsub, vperm, &v, &ex);
@@ -804,13 +809,13 @@ split_ideal(GEN nf, GEN x0, long prec, GEN vperm)
     }
     if (id == x0) continue;
 
-    for (i=1; i<ru; i++) vdir[i] = lstoi(mymyrand() >> randshift);
+    for (i=1; i<ru; i++) vdir[i] = mymyrand() >> randshift;
     for (bou=1; bou<ru; bou++)
     {
       if (bou>1)
       {
-        for (i=1; i<ru; i++) vdir[i]=zero;
-        vdir[bou]=lstoi(10);
+        for (i=1; i<ru; i++) vdir[i] = 0;
+        vdir[bou] = 10;
       }
       nbtest++;
       y = ideallllred(nf,id,vdir,prec);
@@ -892,7 +897,7 @@ red_mod_units(GEN col, GEN z, long prec)
   RU = lg(mat); x = cgetg(RU+1,t_COL);
   for (i=1; i<RU; i++) x[i]=lreal((GEN)col[i]);
   x[RU] = (long)N2;
-  x = lllintern(concatsp(mat,x), 1,prec);
+  x = lllintern(concatsp(mat,x),100, 1,prec);
   if (!x) return NULL;
   x = (GEN)x[RU];
   if (signe(x[RU]) < 0) x = gneg_i(x);
@@ -996,7 +1001,7 @@ isprincipalarch(GEN bnf, GEN col, GEN kNx, GEN e, GEN dx, long *pe)
 static int
 fact_ok(GEN nf, GEN y, GEN C, GEN g, GEN e)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long i, c = lg(e);
   GEN z = C? C: gun;
   for (i=1; i<c; i++)
@@ -1023,6 +1028,8 @@ isprincipalall0(GEN bnf, GEN x, long *ptprec, long flag)
   U = (GEN)clg2[1];
   cyc = gmael3(bnf,8,1,2); c = lg(cyc)-1;
   gen = gmael3(bnf,8,1,3);
+  ex = cgetg(c+1,t_COL);
+  if (c == 0 && !(flag & nf_GEN)) return ex;
 
   vectbase = (GEN)bnf[5]; /* needed by factorgensimple */
 
@@ -1042,7 +1049,6 @@ isprincipalall0(GEN bnf, GEN x, long *ptprec, long flag)
   A = gsub(Wex, gmul(B,Bex));
   if (old_format) U = ginv(U);
   Q = gmul(U, A);
-  ex = cgetg(c+1,t_COL);
   for (i=1; i<=c; i++)
     Q[i] = (long)truedvmdii((GEN)Q[i],(GEN)cyc[i],(GEN*)(ex+i));
   if (!(flag & nf_GEN)) return gcopy(ex);
@@ -1105,7 +1111,8 @@ triv_gen(GEN nf, GEN x, long c, long flag)
 GEN
 isprincipalall(GEN bnf,GEN x,long flag)
 {
-  long av = avma,c,pr, tx = typ(x);
+  long c, pr, tx = typ(x);
+  gpmem_t av = avma;
   GEN nf;
 
   bnf = checkbnf(bnf); nf = (GEN)bnf[7];
@@ -1129,12 +1136,12 @@ isprincipalall(GEN bnf,GEN x,long flag)
   c = getrand();
   for (;;)
   {
-    long av1 = avma;
+    gpmem_t av1 = avma;
     GEN y = isprincipalall0(bnf,x,&pr,flag);
     if (y) return gerepileupto(av,y);
 
     if (DEBUGLEVEL) err(warnprec,"isprincipalall0",pr);
-    avma = av1; bnf = bnfnewprec(bnf,pr); setrand(c);
+    avma = av1; bnf = bnfnewprec(bnf,pr); (void)setrand(c);
   }
 }
 
@@ -1142,7 +1149,8 @@ isprincipalall(GEN bnf,GEN x,long flag)
 GEN
 isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag)
 {
-  long av = avma, l = lg(e), i,prec,c;
+  long l = lg(e), i, prec, c;
+  gpmem_t av = avma;
   GEN id,id2, nf = checknf(bnf), z = NULL; /* gcc -Wall */
   int gen = flag & (nf_GEN | nf_GENMAT);
 
@@ -1160,15 +1168,16 @@ isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag)
       id2 = idealpowred(bnf,z, (GEN)e[i],prec);
       id = id? idealmulred(nf,id,id2,prec): id2;
     }
-  if (id == C)
+  if (id == C) /* e = 0 */
   {
-    if (!C) id = gun;
-    return isprincipalall(bnf, id, flag);
+    if (!C) return isprincipalall(bnf, gun, flag);
+    C = idealhermite(nf,C); id = z;
+    if (gen) id[1] = (long)C; else id = C;
   }
   c = getrand();
   for (;;)
   {
-    long av1 = avma;
+    gpmem_t av1 = avma;
     GEN y = isprincipalall0(bnf, gen? (GEN)id[1]: id,&prec,flag);
     if (y)
     {
@@ -1191,14 +1200,14 @@ isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag)
       avma = av; return stoi(prec);
     }
     if (DEBUGLEVEL) err(warnprec,"isprincipalall0",prec);
-    avma = av1; bnf = bnfnewprec(bnf,prec); setrand(c);
+    avma = av1; bnf = bnfnewprec(bnf,prec); (void)setrand(c);
   }
 }
 
 GEN
 isprincipal(GEN bnf,GEN x)
 {
-  return isprincipalall(bnf,x,nf_REGULAR);
+  return isprincipalall(bnf,x,0);
 }
 
 GEN
@@ -1222,7 +1231,8 @@ isprincipalgenforce(GEN bnf,GEN x)
 GEN
 isunit(GEN bnf,GEN x)
 {
-  long av=avma,tetpil,tx = typ(x),i,R1,RU,n;
+  long tx = typ(x), i, R1, RU, n;
+  gpmem_t av=avma, tetpil;
   GEN p1,logunit,y,ex,nf,z,pi2_sur_w,gn,emb;
 
   bnf = checkbnf(bnf); nf=(GEN)bnf[7];
@@ -1303,7 +1313,8 @@ isunit(GEN bnf,GEN x)
 GEN
 signunits(GEN bnf)
 {
-  long av,i,j,R1,RU,mun;
+  long i, j, R1, RU, mun;
+  gpmem_t av;
   GEN matunit,y,p1,p2,nf,pi;
 
   bnf = checkbnf(bnf); nf = (GEN)bnf[7];
@@ -1323,46 +1334,29 @@ signunits(GEN bnf)
   return y;
 }
 
-/* LLL-reduce ideal and return T2 | ideal */
+/* LLL-reduce ideal and return Cholesky for T2 | ideal */
 static GEN
-red_ideal(GEN *ideal,GEN T2vec,GEN prvec)
+red_ideal(GEN *ideal,GEN Gvec,GEN prvec)
 {
   long i;
-  for (i=1; i<lg(T2vec); i++)
+  for (i=1; i<lg(Gvec); i++)
   {
-    GEN p1,base, T2 = (GEN)T2vec[i];
+    GEN p1,u, G = (GEN)Gvec[i];
     long prec = prvec[i];
 
-    p1 = qf_base_change(T2, *ideal, 1);
-    base = lllgramintern(p1,100,1,prec);
-    if (base)
+    p1 = gmul(G, *ideal);
+    u = lllintern(p1,100,1,prec);
+    if (u)
     {
-      p1 = sqred1intern(qf_base_change(p1,base,1),1);
-      if (p1) { *ideal = gmul(*ideal,base); return p1; }
+      p1 = gmul(p1, u);
+      p1 = sqred1_from_QR(p1, prec);
+      if (p1) { *ideal = gmul(*ideal,u); return p1; }
     }
     if (DEBUGLEVEL) err(warner, "prec too low in red_ideal[%ld]: %ld",i,prec);
   }
   return NULL;
 }
 
-static double
-get_minkovski(long N, long R1, GEN D, GEN gborne)
-{
-  const long prec = DEFAULTPREC;
-  GEN p1,p2, pi = mppi(prec);
-  double bound;
-
-  p1 = gsqrt(gmulsg(N,gmul2n(pi,1)),prec);
-  p1 = gdiv(p1, gexp(stoi(N),prec));
-  p1 = gmulsg(N, gpui(p1, dbltor(2./(double)N),prec));
-  p2 = gpui(gdivsg(4,pi), gdivgs(stoi(N-R1),N),prec);
-  p1 = gmul(p1,p2);
-  bound = gtodouble(gmul(p1, gpui(absi(D), dbltor(1./(double)N),prec)));
-  bound *= gtodouble(gborne);
-  if (DEBUGLEVEL) { fprintferr("Bound for norms = %.0f\n",bound); flusherr(); }
-  return bound;
-}
-
 static void
 wr_rel(long *col)
 {
@@ -1411,54 +1405,44 @@ small_to_mat_i(GEN z, long d)
   return x;
 }
 
-#define MAXTRY 20
-
 /* return -1 in case of precision problems. t = current number of relations */
 static long
 small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
-                       GEN nf,GEN gborne,long nbrelpid,GEN invp,GEN L,GEN LLLnf)
+                       GEN nf,GEN gborne,long nbrelpid,GEN invp,GEN L)
 {
+  const int maxtry_DEP  = 20;
+  const int maxtry_FACT = 500;
   const double eps = 0.000001;
-  double *y,*z,**q,*v, MINKOVSKI_BOUND,BOUND;
-  ulong av = avma, av1,av2,limpile;
+  double *y,*z,**q,*v, BOUND;
+  gpmem_t av = avma, av1, av2, limpile;
   long j,k,noideal, nbrel = lg(mat)-1;
-  long alldep = 0, nbsmallnorm,nbfact,R1, N = degpol(nf[1]);
-  GEN x,xembed,M,T2,r,cbase,invcbase,T2vec,prvec;
+  long nbsmallnorm,nbfact,R1, N = degpol(nf[1]);
+  GEN x,xembed,M,G,r,Gvec,prvec;
 
   if (gsigne(gborne)<=0) return cglob;
   if (DEBUGLEVEL)
     fprintferr("\n#### Looking for %ld relations (small norms)\n",nbrel);
   xembed = NULL; /* gcc -Wall */
   nbsmallnorm = nbfact = 0;
-  R1 = itos(gmael(nf,2,1));
-  T2 = (GEN)LLLnf[1];
-  M  = (GEN)LLLnf[2];
-  cbase   =(GEN)LLLnf[3];
-  invcbase=(GEN)LLLnf[4];
+  R1 = nf_get_r1(nf);
+  M = gmael(nf,5,1);
+  G = gmael(nf,5,2);
 
-  prvec = cgetg(3,t_VECSMALL);
-  prvec[1] = MEDDEFAULTPREC;
-  prvec[2] = PRECREG;
-  T2vec = cgetg(3,t_VEC);
-  T2vec[1] = (long)gprec_w(T2,prvec[1]);
-  T2vec[2] = (long)T2;
+  prvec = cgetg(3,t_VECSMALL); Gvec = cgetg(3,t_VEC);
+  prvec[1] = MEDDEFAULTPREC;   Gvec[1] = (long)gprec_w(G,prvec[1]);
+  prvec[2] = PRECREG;          Gvec[2] = (long)G;
   minim_alloc(N+1, &q, &x, &y, &z, &v);
   av1 = avma;
-  MINKOVSKI_BOUND = get_minkovski(N,R1,(GEN)nf[3],gborne);
   for (noideal=KC; noideal; noideal--)
   {
-    ulong av0 = avma;
-    long nbrelideal=0, dependent = 0, oldcglob = cglob;
+    gpmem_t av0 = avma;
+    long nbrelideal=0, dependent = 0, try_factor = 0, oldcglob = cglob;
     GEN IDEAL, ideal = (GEN)vectbase[noideal];
-    GEN normideal = idealnorm(nf,ideal);
-
-    if (alldep > 2*N) break;
 
     if (DEBUGLEVEL>1) fprintferr("\n*** Ideal no %ld: %Z\n", noideal, ideal);
     ideal = prime_to_ideal(nf,ideal);
-    IDEAL = invcbase? gmul(invcbase, ideal): ideal;
-    IDEAL = gmul(IDEAL, lllint(IDEAL)); /* should be almost T2-reduced */
-    r = red_ideal(&IDEAL,T2vec,prvec);
+    IDEAL = lllint_ip(ideal,4); /* should be almost T2-reduced */
+    r = red_ideal(&IDEAL,Gvec,prvec);
     if (!r) return -1; /* precision problem */
 
     for (k=1; k<=N; k++)
@@ -1468,17 +1452,19 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
       if (DEBUGLEVEL>3) fprintferr("v[%ld]=%.4g ",k,v[k]);
     }
 
-    BOUND = MINKOVSKI_BOUND * pow(gtodouble(normideal),2./(double)N);
+    BOUND = v[2] + v[1] * q[1][2] * q[1][2];
+    if (BOUND < v[1]) BOUND = v[1];
+    BOUND *= 2; /* at most twice larger than smallest known vector */
     if (DEBUGLEVEL>1)
     {
       if (DEBUGLEVEL>3) fprintferr("\n");
       fprintferr("BOUND = %.4g\n",BOUND); flusherr();
     }
-    BOUND += eps; av2=avma; limpile = stack_lim(av2,1);
+    BOUND *= 1 + eps; av2=avma; limpile = stack_lim(av2,1);
     k = N; y[N]=z[N]=0; x[N]= (long) sqrt(BOUND/v[N]);
     for(;; x[1]--)
     {
-      ulong av3 = avma;
+      gpmem_t av3 = avma;
       double p;
       GEN col;
 
@@ -1506,12 +1492,13 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
 	  if (ccontent(x)==1) /* primitive */
 	  {
             GEN Nx, gx = gmul_mati_smallvec(IDEAL,x);
-            long av4;
+            gpmem_t av4;
             if (!isnfscalar(gx))
             {
               xembed = gmul(M,gx); av4 = avma; nbsmallnorm++;
+              if (++try_factor > maxtry_FACT) goto ENDIDEAL;
               Nx = ground(norm_by_embed(R1,xembed)); setsigne(Nx, 1);
-              if (factorelt(nf,cbase,gx,Nx,KCZ,LIMC)) { avma = av4; break; }
+              if (factorelt(nf,gx,Nx,KCZ,LIMC)) { avma = av4; break; }
               if (DEBUGLEVEL > 1) { fprintferr("."); flusherr(); }
             }
 	    avma = av3;
@@ -1526,12 +1513,12 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
       { /* Q-dependent from previous ones: forget it */
         cglob--; unset_fact(col);
         if (DEBUGLEVEL>1) { fprintferr("*"); flusherr(); }
-        if (++dependent > MAXTRY) { alldep++; break; }
+        if (++dependent > maxtry_DEP) break;
         avma = av3; continue;
       }
       /* record archimedean part */
       set_log_embed((GEN)matarch[cglob], xembed, R1, PRECREG);
-      alldep = dependent = 0;
+      dependent = 0;
 
       if (DEBUGLEVEL)
       {
@@ -1564,14 +1551,13 @@ END:
   }
   avma = av; return cglob;
 }
-#undef MAXTRY
 
-/* I assumed to be integral HNF, T2 a weighted T2 matrix. Return an
- * irrational m in I with T2(m) small */
+/* I assumed to be integral HNF, G the Cholesky form of a weighted T2 matrix.
+ * Return an irrational m in I with T2(m) small */
 static GEN
-pseudomin(GEN I, GEN T2)
+pseudomin(GEN I, GEN G)
 {
-  GEN m, y = lllgramintern(qf_base_change(T2,I,1),100,1, 0);
+  GEN m, y = lllintern(gmul(G, I), 100,1, 0);
   if (!y) return NULL;
 
   m = gmul(I,(GEN)y[1]);
@@ -1598,7 +1584,7 @@ dbg_newrel(long jideal, long jdir, long phase, long cglob, long *col,
 static void
 dbg_cancelrel(long jideal,long jdir,long phase, long *col)
 {
-  fprintferr("rel. cancelled. phase %ld: %ld ",phase);
+  fprintferr("rel. cancelled. phase %ld: ",phase);
   if (DEBUGLEVEL>3) fprintferr("(jideal=%ld,jdir=%ld)",jideal,jdir);
   wr_rel(col); flusherr();
 }
@@ -1611,7 +1597,7 @@ dbg_outrel(long phase,long cglob, GEN vperm,GEN *mat,GEN maarch)
 
   if (phase == 0)
   {
-    ulong av = avma; p2=cgetg(cglob+1,t_MAT);
+    gpmem_t av = avma; p2=cgetg(cglob+1, t_MAT);
     for (j=1; j<=cglob; j++)
     {
       p1=cgetg(KC+1,t_COL); p2[j]=(long)p1;
@@ -1669,38 +1655,45 @@ static GEN
 remove_content(GEN I)
 {
   long N = lg(I)-1;
-  if (!gcmp1(gcoeff(I,N,N))) I = primpart(I);
+  if (!gcmp1(gcoeff(I,N,N))) I = Q_primpart(I);
   return I;
 }
 
 /* if phase != 1 re-initialize static variables. If <0 return immediately */
 static long
 random_relation(long phase,long cglob,long LIMC,long PRECREG,
-                GEN nf,GEN subFB,GEN vecT2,GEN *mat,GEN matarch,GEN list_jideal)
+                GEN nf,GEN subFB,GEN vecG,GEN *mat,GEN matarch,GEN list_jideal)
 {
   static long jideal, jdir;
-  long lim,i,av,av1,cptzer,nbT2,lgsub,r1, jlist = 1;
+  long i, maxcglob, cptlist, cptzer, nbG, lgsub, r1, jlist = 1;
+  gpmem_t av, av1;
   GEN arch,col,colarch,ideal,m,P,ex;
 
   if (phase != 1) { jideal=jdir=1; if (phase<0) return 0; }
 
   r1 = nf_get_r1(nf);
-  lim = lg(mat)-1; /* requested number of relations */
-  nbT2 = lg(vecT2)-1;
+  maxcglob = lg(mat)-1; /* requested number of relations */
+  nbG = lg(vecG)-1;
   lgsub = lg(subFB); ex = cgetg(lgsub, t_VECSMALL);
-  cptzer = 0;
+  cptzer = cptlist = 0;
   if (DEBUGLEVEL && list_jideal)
     fprintferr("looking hard for %Z\n",list_jideal);
   P = NULL; /* gcc -Wall */
   for (av = avma;;)
   {
-    if (list_jideal && jlist < lg(list_jideal) && jdir <= nbT2)
-      jideal = list_jideal[jlist++];
-    if (!list_jideal || jdir <= nbT2)
+    if (list_jideal && jlist < lg(list_jideal) && jdir <= nbG)
+    {
+      jideal = list_jideal[jlist++]; cptlist = 0;
+    }
+    if (!list_jideal || jdir <= nbG)
     {
       avma = av;
       P = prime_to_ideal(nf, (GEN)vectbase[jideal]);
     }
+    else
+    {
+      if (++cptlist > 300) return -1;
+    }
     ideal = P;
     do {
       for (i=1; i<lgsub; i++)
@@ -1713,12 +1706,12 @@ random_relation(long phase,long cglob,long LIMC,long PRECREG,
     ideal = remove_content(ideal);
 
     if (phase != 1) jdir = 1; else phase = 2;
-    for (av1 = avma; jdir <= nbT2; jdir++, avma = av1)
+    for (av1 = avma; jdir <= nbG; jdir++, avma = av1)
     { /* reduce along various directions */
       if (DEBUGLEVEL>2)
         fprintferr("phase=%ld,jideal=%ld,jdir=%ld,rand=%ld\n",
                    phase,jideal,jdir,getrand());
-      m = pseudomin(ideal,(GEN)vecT2[jdir]);
+      m = pseudomin(ideal,(GEN)vecG[jdir]);
       if (!m) return -2;
       if (!factorgen(nf,ideal,m,KCZ,LIMC))
       {
@@ -1755,10 +1748,10 @@ random_relation(long phase,long cglob,long LIMC,long PRECREG,
       /* arch = archimedean component (MULTIPLICATIVE form) of ideal */
       arch = vecdiv(arch, gmul(gmael(nf,5,1), m));
       set_log_embed(colarch, arch, r1, PRECREG);
-      if (DEBUGLEVEL) dbg_newrel(jideal,jdir,phase,cglob,col,colarch,lim);
+      if (DEBUGLEVEL) dbg_newrel(jideal,jdir,phase,cglob,col,colarch,maxcglob);
 
       /* Need more, try next P */
-      if (cglob < lim) break;
+      if (cglob < maxcglob) break;
 
       /* We have found enough. Return */
       if (phase)
@@ -1780,9 +1773,9 @@ random_relation(long phase,long cglob,long LIMC,long PRECREG,
 static long
 be_honest(GEN nf,GEN subFB,long PRECLLL)
 {
-  ulong av;
+  gpmem_t av;
   long ex,i,j,J,k,iz,nbtest,ru, lgsub = lg(subFB);
-  GEN MCtw,MC,M,P,ideal,m,exu, D = (GEN)nf[3];
+  GEN G,M,P,ideal,m,exu, D = (GEN)nf[3];
 
   if (KCZ2 <= KCZ) return 1;
   if (DEBUGLEVEL)
@@ -1792,10 +1785,9 @@ be_honest(GEN nf,GEN subFB,long PRECLLL)
   }
   if (!powsubFB) powsubFBgen(nf,subFB,CBUCHG+1,0);
   M = gprec_w(gmael(nf,5,1), PRECLLL);
-  MC= gprec_w(gmael(nf,5,2), PRECLLL);
-  ru = lg(MC);
+  G = gprec_w(gmael(nf,5,2), PRECLLL);
+  ru = lg(G);
   exu = cgetg(ru, t_VECSMALL);
-  MCtw= cgetg(ru, t_MAT);
   av = avma;
   for (iz=KCZ+1; iz<=KCZ2; iz++, avma = av)
   {
@@ -1806,7 +1798,7 @@ be_honest(GEN nf,GEN subFB,long PRECLLL)
     for (j=1; j<J; j++)
     {
       GEN ideal0 = prime_to_ideal(nf,(GEN)P[j]);
-      ulong av2 = avma;
+      gpmem_t av2 = avma;
       for(nbtest=0;;)
       {
 	ideal = ideal0;
@@ -1825,9 +1817,7 @@ be_honest(GEN nf,GEN subFB,long PRECLLL)
 	    for (i=1; i<ru; i++) exu[i] = 0;
             exu[k] = 10;
 	  }
-          for (i=1; i<ru; i++)
-            MCtw[i] = exu[i]? lmul2n((GEN)MC[i],exu[i]<<1): MC[i];
-          m = pseudomin(ideal, mulmat_real(MCtw,M));
+          m = pseudomin(ideal, computeGtwist(nf, exu));
           if (m && factorgen(nf,ideal,m,iz-1,FB[iz-1])) break;
 	  if (++nbtest==200) return 0;
 	}
@@ -1855,7 +1845,8 @@ static GEN
 compute_multiple_of_R(GEN xarch,long RU,long N,GEN *ptlambda)
 {
   GEN T,v,mdet,mdet_t,Im_mdet,kR,xreal,lambda, *gptr[2];
-  long av = avma, i, zc = lg(xarch)-1, R1 = 2*RU - N;
+  long i, zc = lg(xarch)-1, R1 = 2*RU - N;
+  gpmem_t av = avma;
 
   if (DEBUGLEVEL) fprintferr("\n#### Computing regulator multiple\n");
   xreal = greal(xarch); /* = (log |sigma_i(u_j)|) */
@@ -1913,7 +1904,7 @@ bestappr_noer(GEN x, GEN k)
 static int
 compute_R(GEN lambda, GEN z, GEN *ptL, GEN *ptkR)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long r;
   GEN L,H,gc,den,R;
   double c;
@@ -1933,7 +1924,7 @@ compute_R(GEN lambda, GEN z, GEN *ptL, GEN *ptkR)
     return PRECI;
   }
   L = gmul(lambda,den);
-  H = hnf(L); r = lg(H)-1;
+  H = hnfall_i(L, NULL, 1); r = lg(H)-1;
   R = gdiv(dethnf_i(H), gpowgs(den, r));
   R = mpabs(gmul(*ptkR,R)); /* tentative regulator */
   c = gtodouble(gmul(R,z)); /* should be 2n (= 2 if we are done) */
@@ -1997,10 +1988,10 @@ class_group_gen(GEN nf,GEN W,GEN C,GEN Vbase,long prec, GEN nf0,
                 GEN *ptclg1,GEN *ptclg2)
 {
   GEN z,G,Ga,ga,GD,cyc,X,Y,D,U,V,Ur,Ui,Uir,I,J;
-  long i,j,s,lo,lo0;
+  long i,j,lo,lo0;
 
   if (DEBUGLEVEL)
-    { fprintferr("\n#### Computing class group generators\n"); timer2(); }
+    { fprintferr("\n#### Computing class group generators\n"); (void)timer2(); }
   D = smithall(W,&U,&V); /* UWV = D, D diagonal, G = g Ui (G=new gens, g=old) */
   Ui = ginv(U);
   lo0 = lo = lg(D);
@@ -2025,14 +2016,12 @@ class_group_gen(GEN nf,GEN W,GEN C,GEN Vbase,long prec, GEN nf0,
   {
     GEN p1 = gcoeff(Uir,1,j);
     z[1]=Vbase[1]; I = idealpowred(nf0,z,p1,prec);
-    if (signe(p1)<0) I[1] = lmul((GEN)I[1],denom((GEN)I[1]));
     for (i=2; i<lo0; i++)
     {
-      p1 = gcoeff(Uir,i,j); s = signe(p1);
-      if (s)
+      p1 = gcoeff(Uir,i,j); 
+      if (signe(p1))
       {
 	z[1]=Vbase[i]; J = idealpowred(nf0,z,p1,prec);
-        if (s<0) J[1] = lmul((GEN)J[1],denom((GEN)J[1]));
 	I = idealmulh(nf0,I,J);
 	I = ideallllred(nf0,I,NULL,prec);
       }
@@ -2081,42 +2070,50 @@ class_group_gen(GEN nf,GEN W,GEN C,GEN Vbase,long prec, GEN nf0,
   if (DEBUGLEVEL) msgtimer("classgroup generators");
 }
 
-/* real(MC * M), where columns a and b of MC have been multiplied by 2^20+1 */
-static GEN
-shift_t2(GEN T2, GEN M, GEN MC, long a, long b)
+static void
+shift_embed(GEN G, GEN Gtw, long a, long r1, long r2)
 {
-  long i,j,N = lg(T2);
-  GEN z, t2 = cgetg(N,t_MAT);
-  for (j=1; j<N; j++)
+  long j, k, l = lg(G);
+  if (a <= r1)
+    for (j=1; j<l; j++) coeff(G,a,j) = coeff(Gtw,a,j);
+  else
   {
-    t2[j] = lgetg(N,t_COL);
-    for (i=1; i<=j; i++)
+    k = (a<<1) - r1;
+    for (j=1; j<l; j++)
     {
-      z = mul_real(gcoeff(MC,i,a), gcoeff(M,a,j));
-      if (a!=b) z = gadd(z, mul_real(gcoeff(MC,i,b), gcoeff(M,b,j)));
-      coeff(t2,j,i) = coeff(t2,i,j) = ladd(gcoeff(T2,i,j), gmul2n(z,20));
+      coeff(G,k-1,j) = coeff(Gtw,k-1,j);
+      coeff(G,k  ,j) = coeff(Gtw,k,  j);
     }
   }
-  return t2;
+}
+
+/* G where embeddings a and b are multiplied by 2^10 */
+static GEN
+shift_G(GEN G, GEN Gtw, long a, long b, long r1, long r2)
+{
+  GEN g = dummycopy(G);
+  shift_embed(g,Gtw,a,r1,r2);
+  shift_embed(g,Gtw,b,r1,r2); return g;
 }
 
 static GEN
-compute_vecT2(GEN nf,long prec)
+compute_vecG(GEN nf,long prec)
 {
-  GEN vecT2, M = gmael(nf,5,1), MC = gmael(nf,5,2), T2 = gmael(nf,5,3);
-  long i,j,ind, n = min(lg(M[1])-1, 9);
+  GEN vecG, Gtw, M = gmael(nf,5,1), G = gmael(nf,5,2);
+  long r1,r2,i,j,ind, n = min(lg(M[1])-1, 9);
 
-  vecT2=cgetg(1 + n*(n+1)/2,t_VEC);
+  nf_get_sign(nf,&r1,&r2);
+  vecG=cgetg(1 + n*(n+1)/2,t_VEC);
   if (nfgetprec(nf) > prec)
   {
     M = gprec_w(M,prec);
-    MC= gprec_w(MC,prec);
-    T2= gprec_w(T2,prec);
+    G = gprec_w(G,prec);
   }
+  Gtw = gmul2n(G, 10);
   for (ind=j=1; j<=n; j++)
-    for (i=1; i<=j; i++) vecT2[ind++] = (long)shift_t2(T2,M,MC,i,j);
-  if (DEBUGLEVEL) msgtimer("weighted T2 matrices");
-  return vecT2;
+    for (i=1; i<=j; i++) vecG[ind++] = (long)shift_G(G,Gtw,i,j,r1,r2);
+  if (DEBUGLEVEL) msgtimer("weighted G matrices");
+  return vecG;
 }
 
 /* cf. relationrank()
@@ -2165,7 +2162,8 @@ addcolumntomatrix(GEN V, GEN invp, GEN L)
 static GEN
 relationrank(GEN *A, long r, GEN L)
 {
-  long i, n = lg(L)-1, av = avma, lim = stack_lim(av,1);
+  long i, n = lg(L)-1;
+  gpmem_t av = avma, lim = stack_lim(av, 1);
   GEN invp = idmat(n);
 
   if (!r) return invp;
@@ -2186,7 +2184,8 @@ relationrank(GEN *A, long r, GEN L)
 static GEN
 codeprime(GEN bnf, GEN pr)
 {
-  long j,av=avma,tetpil;
+  long j;
+  gpmem_t av=avma, tetpil;
   GEN p,al,fa,p1;
 
   p=(GEN)pr[1]; al=(GEN)pr[2]; fa=primedec(bnf,p);
@@ -2203,7 +2202,8 @@ codeprime(GEN bnf, GEN pr)
 static GEN
 decodeprime(GEN nf, GEN co)
 {
-  long n,indi,av=avma;
+  long n, indi;
+  gpmem_t av=avma;
   GEN p,rem,p1;
 
   n=lg(nf[7])-1; p=dvmdis(co,n,&rem); indi=itos(rem)+1;
@@ -2258,14 +2258,8 @@ makecycgen(GEN bnf)
       if (y) { h[i] = (long)to_famat_all(y,gun); continue; }
     }
     y = isprincipalfact(bnf, gen, (GEN)D[i], NULL,
-                        nf_GEN|nf_GENMAT|nf_FORCE|nf_GIVEPREC);
-    if (typ(y) != t_INT)
-      h[i] = y[2];
-    else
-    {
-      y = idealpow(nf, (GEN)gen[i], (GEN)cyc[i]);
-      h[i] = isprincipalgenforce(bnf,y)[2];
-    }
+                        nf_GEN|nf_GENMAT|nf_FORCE);
+    h[i] = y[2];
   }
   return h;
 }
@@ -2315,7 +2309,7 @@ makematal(GEN bnf)
     prec = itos(y); j--;
     if (DEBUGLEVEL) err(warnprec,"makematal",prec);
     nf = nfnewprec(nf,prec);
-    bnf = bnfinit0(nf,1,NULL,prec); setrand(c);
+    bnf = bnfinit0(nf,1,NULL,prec); (void)setrand(c);
   }
   if (DEBUGLEVEL>1) fprintferr("\n");
   return ma;
@@ -2347,7 +2341,7 @@ check_and_build_cycgen(GEN bnf)
   GEN cycgen = get_cycgen((GEN)bnf[10]);
   if (!cycgen)
   {
-    long av = avma;
+    gpmem_t av = avma;
     if (DEBUGLEVEL) err(warner,"completing bnf (building cycgen)");
     bnfinsert(bnf, makecycgen(bnf), 2); avma = av;
     cycgen = get_cycgen((GEN)bnf[10]);
@@ -2361,7 +2355,7 @@ check_and_build_matal(GEN bnf)
   GEN matal = get_matal((GEN)bnf[10]);
   if (!matal)
   {
-    long av = avma;
+    gpmem_t av = avma;
     if (DEBUGLEVEL) err(warner,"completing bnf (building matal)");
     bnfinsert(bnf, makematal(bnf), 1); avma = av;
     matal = get_matal((GEN)bnf[10]);
@@ -2372,7 +2366,8 @@ check_and_build_matal(GEN bnf)
 GEN
 smallbuchinit(GEN pol,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,long minsFB,long prec)
 {
-  long av=avma,av1,k;
+  long k;
+  gpmem_t av=avma, av1;
   GEN y,bnf,pFB,vp,nf,mas,res,uni,v1,v2,v3;
 
   if (typ(pol)==t_VEC) bnf = checkbnf(pol);
@@ -2406,7 +2401,7 @@ smallbuchinit(GEN pol,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpi
 static GEN
 get_regulator(GEN mun,long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN p1;
 
   if (lg(mun)==1) return gun;
@@ -2424,7 +2419,7 @@ log_poleval(GEN P, GEN *ro, long i, GEN nf, long prec0)
   {
     prec = (prec-2)<<1;
     if (DEBUGLEVEL) err(warnprec,"log_poleval",prec);
-    *ro = get_roots((GEN)nf[1], itos(gmael(nf,2,1)),lg(*ro)-1,prec);
+    *ro = get_roots((GEN)nf[1], nf_get_r1(nf), prec);
     x = poleval(P, (GEN)(*ro)[i]);
   }
   if (k > prec0) x = gprec_w(x,prec0);
@@ -2447,7 +2442,7 @@ get_arch2_i(GEN nf, GEN a, long prec, int units)
     GEN pol = (GEN)nf[1];
     N = cgetg(la,t_VEC);
     for (k=1; k<la; k++) N[k] = (long)gabs(subres(pol,(GEN)a[k]),0);
-    N = gdivgs(glog(N,prec), - (degpol(pol))); /* - log(|norm|) / [K:Q] */
+    N = gdivgs(glog(N,prec), -degpol(pol)); /* - log(|norm|) / [K:Q] */
   }
   for (k=1; k<la; k++)
   {
@@ -2465,19 +2460,19 @@ get_arch2_i(GEN nf, GEN a, long prec, int units)
 static GEN
 get_arch2(GEN nf, GEN a, long prec, int units)
 {
-  long av = avma;
+  gpmem_t av = avma;
   return gerepilecopy(av, get_arch2_i(nf,a,prec,units));
 }
 
 static void
 my_class_group_gen(GEN bnf, long prec, GEN nf0, GEN *ptcl, GEN *ptcl2)
 {
-  ulong av = avma;
-  GEN W=(GEN)bnf[1], C=(GEN)bnf[4], nf=(GEN)bnf[7], *gptr[2];
+  gpmem_t av = avma;
+  GEN W=(GEN)bnf[1], C=(GEN)bnf[4], nf=(GEN)bnf[7];
   GEN Vbase = get_Vbase(W, (GEN)bnf[5], (GEN)bnf[6]);
 
   class_group_gen(nf,W,C,Vbase,prec,nf0, ptcl,ptcl2);
-  gptr[0]=ptcl; gptr[1]=ptcl2; gerepilemany(av,gptr,2);
+  gerepileall(av, 2, ptcl, ptcl2);
 }
 
 GEN
@@ -2535,56 +2530,62 @@ bnrnewprec(GEN bnr, long prec)
   return y;
 }
 
+static void
+nfbasic_from_sbnf(GEN sbnf, nfbasic_t *T)
+{
+  T->x    = (GEN)sbnf[1];
+  T->dK   = (GEN)sbnf[3];
+  T->bas  = (GEN)sbnf[4];
+  T->index= get_nfindex(T->bas);
+  T->r1   = itos((GEN)sbnf[2]);
+  T->dx   = NULL;
+  T->lead = NULL;
+  T->basden = NULL;
+}
+
 GEN
 bnfmake(GEN sbnf, long prec)
 {
-  long av = avma, j,k,n,r1,r2,ru,lpf;
-  GEN p1,x,bas,ro,nf,mun,funits,index;
+  long j, k, l;
+  gpmem_t av = avma;
+  GEN p1,bas,ro,nf,mun,funits;
   GEN pfc,vp,mc,clgp,clgp2,res,y,W,racu,reg,matal,vectbase,Vbase;
+  nfbasic_t T;
 
-  if (typ(sbnf)!=t_VEC || lg(sbnf)!=13)
-    err(talker,"incorrect sbnf in bnfmake");
-  x=(GEN)sbnf[1]; bas=(GEN)sbnf[4]; n=lg(bas)-1;
-  r1=itos((GEN)sbnf[2]); r2=(n-r1)>>1; ru=r1+r2;
-  ro=(GEN)sbnf[5];
-  if (prec > gprecision(ro)) ro=get_roots(x,r1,ru,prec);
-  index = gun;
-  for (j=2; j<=n; j++) index = mulii(index, denom(leading_term(bas[j])));
+  if (typ(sbnf) != t_VEC || lg(sbnf) != 13) err(typeer,"bnfmake");
+  if (prec < DEFAULTPREC) prec = DEFAULTPREC;
 
-  nf=cgetg(10,t_VEC);
-  nf[1]=sbnf[1]; p1=cgetg(3,t_VEC); p1[1]=lstoi(r1); p1[2]=lstoi(r2);
-  nf[2]=(long)p1;
-  nf[3]=sbnf[3];
-  nf[4]=(long)index;
-  nf[6]=(long)ro;
-  nf[7]=(long)bas;
-  get_nf_matrices(nf, 0);
+  nfbasic_from_sbnf(sbnf, &T);
+  ro = (GEN)sbnf[5];
+  if (prec > gprecision(ro)) ro = get_roots(T.x,T.r1,prec);
+  nf = nfbasic_to_nf(&T, ro, prec);
+  bas = (GEN)nf[7];
 
-  funits=cgetg(ru,t_VEC); p1 = (GEN)sbnf[11];
-  for (k=1; k < lg(p1); k++)
-    funits[k] = lmul(bas,(GEN)p1[k]);
+  p1 = (GEN)sbnf[11]; l = lg(p1); funits = cgetg(l, t_VEC);
+  for (k=1; k < l; k++) funits[k] = lmul(bas, (GEN)p1[k]);
   mun = get_arch2_i(nf,funits,prec,1);
 
-  prec=gprecision(ro); if (prec<DEFAULTPREC) prec=DEFAULTPREC;
+  prec = gprecision(ro);
   matal = get_matal((GEN)sbnf[12]);
   if (!matal) matal = (GEN)sbnf[12];
   mc = get_arch2_i(nf,matal,prec,0);
 
-  pfc=(GEN)sbnf[9]; lpf=lg(pfc);
-  vectbase=cgetg(lpf,t_COL); vp=cgetg(lpf,t_COL);
-  for (j=1; j<lpf; j++)
+  pfc = (GEN)sbnf[9]; l = lg(pfc);
+  vectbase = cgetg(l,t_COL);
+  vp       = cgetg(l,t_COL);
+  for (j=1; j<l; j++)
   {
-    vp[j]=lstoi(j);
-    vectbase[j]=(long)decodeprime(nf,(GEN)pfc[j]);
+    vectbase[j] = (long)decodeprime(nf,(GEN)pfc[j]);
+    vp[j]       = lstoi(j);
   }
   W = (GEN)sbnf[7];
   Vbase = get_Vbase(W,vectbase,vp);
   class_group_gen(nf,W,mc,Vbase,prec,NULL, &clgp,&clgp2);
 
   reg = get_regulator(mun,prec);
-  p1=cgetg(3,t_VEC); racu=(GEN)sbnf[10];
-  p1[1]=racu[1]; p1[2]=lmul(bas,(GEN)racu[2]);
-  racu=p1;
+  p1 = cgetg(3,t_VEC); racu = (GEN)sbnf[10];
+  p1[1] = racu[1];
+  p1[2] = lmul(bas,(GEN)racu[2]); racu = p1;
 
   res=cgetg(7,t_VEC);
   res[1]=(long)clgp; res[2]=(long)reg;    res[3]=(long)dbltor(1.0);
@@ -2601,7 +2602,8 @@ static GEN
 classgroupall(GEN P, GEN data, long flag, long prec)
 {
   long court[3],doubl[4];
-  long av=avma,flun,lx, minsFB=3,nbrelpid=4;
+  long flun, lx, minsFB=3, nbrelpid=4;
+  gpmem_t av=avma;
   GEN bach=doubl,bach2=doubl,RELSUP=court,borne=gun;
 
   if (!data) lx=1;
@@ -2664,7 +2666,7 @@ bnfinit0(GEN P, long flag, GEN data, long prec)
 GEN
 classgrouponly(GEN P, GEN data, long prec)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN z;
 
   if (typ(P)==t_INT)
@@ -2679,7 +2681,7 @@ classgrouponly(GEN P, GEN data, long prec)
 GEN
 regulator(GEN P, GEN data, long prec)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN z;
 
   if (typ(P)==t_INT)
@@ -2760,7 +2762,8 @@ buchall_end(GEN nf,GEN CHANGE,long fl,long k, GEN fu, GEN clg1, GEN clg2,
 static GEN
 buchall_for_degree_one_pol(GEN nf, GEN CHANGE, long flun)
 {
-  long av = avma, k = EXP220;
+  long k = EXP220;
+  gpmem_t av = avma;
   GEN W,B,xarch,matarch,vectbase,vperm;
   GEN fu=cgetg(1,t_VEC), reg=gun, c_1=gun, zu=cgetg(3,t_VEC);
   GEN clg1=cgetg(4,t_VEC), clg2=cgetg(4,t_VEC);
@@ -2774,36 +2777,21 @@ buchall_for_degree_one_pol(GEN nf, GEN CHANGE, long flun)
   return gerepilecopy(av, buchall_end(nf,CHANGE,flun,k,fu,clg1,clg2,reg,c_1,zu,W,B,xarch,matarch,vectbase,vperm));
 }
 
-static GEN
-get_LLLnf(GEN nf, long prec)
-{
-  GEN M  = gmael(nf,5,1);
-  GEN T2 = gmael(nf,5,3);
-  GEN cbase = lllgramintern(T2,100,1,prec);
-  GEN v = cgetg(5,t_VEC);
-  if (!cbase) return NULL;
-  if (gegal(cbase, idmat(lg(T2)-1))) cbase = NULL;
-  v[1] = (long) (cbase? qf_base_change(T2, cbase, 1): T2);
-  v[2] = (long) (cbase? gmul(M, cbase): M);
-  v[3] = (long) cbase;
-  v[4] = (long) (cbase? ZM_inv(cbase,gun): NULL); return v;
-}
-
 GEN
 buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
         long minsFB,long flun,long prec)
 {
-  ulong av = avma,av0,av1,limpile;
+  gpmem_t av = avma, av0, av1, limpile;
   long N,R1,R2,RU,PRECREG,PRECLLL,PRECLLLadd,KCCO,RELSUP,LIMC,LIMC2,lim;
   long nlze,zc,nrelsup,nreldep,phase,matmax,i,j,k,ss,cglob;
   long sfb_increase=0, sfb_trials=0, precdouble=0, precadd=0;
   double cbach,cbach2,drc,LOGD2;
-  GEN p1,vecT2,fu,zu,nf,LLLnf,D,xarch,W,R,Res,z,h,vperm,subFB;
+  GEN p1,vecG,fu,zu,nf,D,xarch,W,R,Res,z,h,vperm,subFB;
   GEN L,resc,B,C,c1,lambda,pdep,liste,invp,clg1,clg2,Vbase, *mat;
   GEN CHANGE=NULL, extramat=NULL, extraC=NULL, list_jideal=NULL;
   char *precpb = NULL;
 
-  if (DEBUGLEVEL) timer2();
+  if (DEBUGLEVEL) (void)timer2();
 
   P = get_nfpol(P, &nf);
   if (typ(gRELSUP) != t_INT) gRELSUP = gtrunc(gRELSUP);
@@ -2817,7 +2805,7 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
   PRECLLLadd = DEFAULTPREC;
   if (!nf)
   {
-    nf = initalgall0(P, nf_REGULAR, PRECREG);
+    nf = initalg(P, PRECREG);
     /* P was non-monic and nfinit changed it ? */
     if (lg(nf)==3) { CHANGE = (GEN)nf[2]; nf = (GEN)nf[1]; }
   }
@@ -2832,6 +2820,9 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
   lim = (long) (exp(-(double)N) * sqrt(2*PI*N*drc) * pow(4/PI,(double)R2));
   if (lim < 3) lim = 3;
   cbach = min(12., gtodouble(gcbach)); cbach /= 2;
+  if (cbach == 0.) err(talker,"Bach constant = 0 in bnfxxx");
+  if (nbrelpid < 0) err(talker,"negative #relation / ideal in bnfxxx");
+   
   cbach2 = gtodouble(gcbach2);
   /* resc = sqrt(D) w / 2^r1 (2pi)^r2 ~ hR / Res(zeta_K,1) */
   resc = gmul(gmul2n(gpuigs(shiftr(mppi(DEFAULTPREC),1),-R2),-R1),
@@ -2866,9 +2857,6 @@ START:
   precadd = 0;
 
   /* T2-LLL-reduce nf.zk */
-  LLLnf = get_LLLnf(nf, PRECREG);
-  if (!LLLnf) { precpb = "LLLnf"; goto START; }
-
   LIMC = (long)(cbach*LOGD2); if (LIMC < 20) LIMC = 20;
   LIMC2= max(3 * N, (long)(cbach2*LOGD2));
   if (LIMC2 < LIMC) LIMC2 = LIMC;
@@ -2930,13 +2918,13 @@ START:
 
   /* relations through elements of small norm */
   cglob = small_norm_for_buchall(cglob,mat,C,(long)LIMC,PRECREG,
-                                 nf,gborne,nbrelpid,invp,liste,LLLnf);
+                                 nf,gborne,nbrelpid,invp,liste);
   if (cglob < 0) { precpb = "small_norm"; goto START; }
   avma = av1; limpile = stack_lim(av1,1);
 
   phase = 0;
   nlze = matmax = 0; /* for lint */
-  vecT2 = NULL;
+  vecG = NULL;
   list_jideal = NULL;
 
   /* random relations */
@@ -2977,9 +2965,9 @@ MORE:
       for (j=cglob+1; j<=slim; j++) mat[j] = col_0(KC);
       matarch = extraC - cglob; /* start at 0, the others at cglob */
     }
-    if (!vecT2)
+    if (!vecG)
     {
-      vecT2 = compute_vecT2(nf,PRECLLL);
+      vecG = compute_vecG(nf,PRECLLL);
       av1 = avma;
     }
     if (!powsubFB)
@@ -2988,7 +2976,7 @@ MORE:
       av1 = avma;
     }
     ss = random_relation(phase,cglob,(long)LIMC,PRECREG,
-                         nf,subFB,vecT2,mat,matarch,list_jideal);
+                         nf,subFB,vecG,mat,matarch,list_jideal);
     if (ss < 0)
     { /* could not find relations */
       if (ss != -1)
