@@ -1,4 +1,4 @@
-/* $Id: buch4.c,v 1.19 2001/11/23 22:36:01 karim Exp $
+/* $Id: buch4.c,v 1.27 2002/05/12 16:58:50 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -44,7 +44,8 @@ psquare(GEN a,GEN p)
 static long
 lemma6(GEN pol,GEN p,long nu,GEN x)
 {
-  long i,lambda,mu,ltop=avma;
+  long i, lambda, mu;
+  gpmem_t ltop=avma;
   GEN gx,gpx;
 
   for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
@@ -65,7 +66,9 @@ lemma6(GEN pol,GEN p,long nu,GEN x)
 
 static long
 lemma7(GEN pol,long nu,GEN x)
-{ long i,odd4,lambda,mu,mnl,ltop=avma;
+{
+  long i,odd4,lambda,mu,mnl;
+  gpmem_t ltop=avma;
   GEN gx,gpx,oddgx;
 
   for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
@@ -96,7 +99,8 @@ lemma7(GEN pol,long nu,GEN x)
 static long
 zpsol(GEN pol,GEN p,long nu,GEN pnu,GEN x0)
 {
-  long i,result,ltop=avma;
+  long i, result;
+  gpmem_t ltop=avma;
   GEN x,pnup;
 
   result = (cmpis(p,2)) ? lemma6(pol,p,nu,x0) : lemma7(pol,nu,x0);
@@ -137,7 +141,7 @@ qpsoluble(GEN pol,GEN p)
 static long
 psquarenf(GEN nf,GEN a,GEN pr)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long v;
   GEN norm;
 
@@ -169,7 +173,8 @@ check2(GEN nf, GEN a, GEN zinit)
 static long
 psquare2nf(GEN nf,GEN a,GEN pr,GEN zinit)
 {
-  long v, ltop = avma;
+  long v;
+  gpmem_t ltop = avma;
 
   if (gcmp0(a)) return 1;
   v = idealval(nf,a,pr); if (v&1) return 0;
@@ -182,7 +187,8 @@ psquare2nf(GEN nf,GEN a,GEN pr,GEN zinit)
 static long
 psquare2qnf(GEN nf,GEN a,GEN p,long q)
 {
-  long v, ltop=avma;
+  long v;
+  gpmem_t ltop=avma;
   GEN zinit = zidealstarinit(nf,idealpows(nf,p,q));
 
   v = check2(nf,a,zinit); avma = ltop; return v;
@@ -191,7 +197,8 @@ psquare2qnf(GEN nf,GEN a,GEN p,long q)
 static long
 lemma6nf(GEN nf,GEN pol,GEN p,long nu,GEN x)
 {
-  long i,lambda,mu,ltop=avma;
+  long i, lambda, mu;
+  gpmem_t ltop=avma;
   GEN gx,gpx;
 
   for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
@@ -212,7 +219,8 @@ lemma6nf(GEN nf,GEN pol,GEN p,long nu,GEN x)
 static long
 lemma7nf(GEN nf,GEN pol,GEN p,long nu,GEN x,GEN zinit)
 {
-  long res,i,lambda,mu,q,ltop=avma;
+  long res, i, lambda, mu, q;
+  gpmem_t ltop=avma;
   GEN gx,gpx,p1;
 
   for (i=lgef(pol)-2, gx=(GEN) pol[i+1]; i>1; i--)
@@ -245,7 +253,8 @@ lemma7nf(GEN nf,GEN pol,GEN p,long nu,GEN x,GEN zinit)
 static long
 zpsolnf(GEN nf,GEN pol,GEN p,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit)
 {
-  long i,result,ltop=avma;
+  long i, result;
+  gpmem_t ltop=avma;
   GEN pnup;
 
   nf=checknf(nf);
@@ -297,7 +306,7 @@ long
 qpsolublenf(GEN nf,GEN pol,GEN pr)
 {
   GEN repr,zinit,p1;
-  long ltop=avma;
+  gpmem_t ltop=avma;
 
   if (gcmp0(pol)) return 1;
   if (typ(pol)!=t_POL) err(notpoler,"qpsolublenf");
@@ -333,7 +342,7 @@ long
 zpsolublenf(GEN nf,GEN pol,GEN p)
 {
   GEN repr,zinit;
-  long ltop=avma;
+  gpmem_t ltop=avma;
 
   if (gcmp0(pol)) return 1;
   if (typ(pol)!=t_POL) err(notpoler,"zpsolublenf");
@@ -359,7 +368,7 @@ zpsolublenf(GEN nf,GEN pol,GEN p)
 static long
 hilb2nf(GEN nf,GEN a,GEN b,GEN p)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long rep;
   GEN pol;
 
@@ -377,9 +386,9 @@ hilb2nf(GEN nf,GEN a,GEN b,GEN p)
 long
 nfhilbertp(GEN nf,GEN a,GEN b,GEN pr)
 {
-  GEN ord, ordp, p, prhall,t;
+  GEN ord, ordp, T,p, modpr,t;
   long va, vb, rep;
-  ulong av = avma;
+  gpmem_t av = avma;
 
   if (gcmp0(a) || gcmp0(b)) err (talker,"0 argument in nfhilbertp");
   checkprimeid(pr); nf = checknf(nf);
@@ -398,9 +407,14 @@ nfhilbertp(GEN nf,GEN a,GEN b,GEN pr)
   /* quad. symbol is image of t by the quadratic character  */
   ord = subis( idealnorm(nf,pr), 1 ); /* |(O_K / pr)^*| */
   ordp= subis( p, 1);                 /* |F_p^*|        */
-  prhall = nfmodprinit(nf, pr);
-  t = element_powmodpr(nf, t, divii(ord, ordp), prhall); /* in F_p^* */
-  t = lift_intern((GEN)t[1]);
+  modpr = nf_to_ff_init(nf, &pr,&T,&p);
+  t = nf_to_ff(nf,t,modpr);
+  t = FpXQ_pow(t, diviiexact(ord, ordp), T,p); /* in F_p^* */
+  if (typ(t) == t_POL)
+  {
+    if (degpol(t)) err(bugparier,"nfhilbertp");
+    t = constant_term(t);
+  }
   rep = kronecker(t, p);
   avma = av; return rep;
 }
@@ -413,7 +427,7 @@ nfhilbertp(GEN nf,GEN a,GEN b,GEN pr)
 long
 nfhilbert(GEN nf,GEN a,GEN b)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long r1, i;
   GEN S, al, bl, ro;
 
@@ -462,6 +476,7 @@ extern GEN isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag);
 extern GEN vconcat(GEN Q1, GEN Q2);
 extern GEN mathnfspec(GEN x, GEN *ptperm, GEN *ptdep, GEN *ptB, GEN *ptC);
 extern GEN factorback_i(GEN fa, GEN e, GEN nf, int red);
+extern GEN detcyc(GEN cyc);
 /* S a list of prime ideal in primedec format. Return res:
  * res[1] = generators of (S-units / units), as polynomials
  * res[2] = [perm, HB, den], for bnfissunit
@@ -473,7 +488,7 @@ extern GEN factorback_i(GEN fa, GEN e, GEN nf, int red);
 GEN
 bnfsunit(GEN bnf,GEN S,long prec)
 {
-  ulong ltop = avma;
+  gpmem_t ltop = avma;
   long i,j,ls;
   GEN p1,nf,classgp,gen,M,U,H;
   GEN sunit,card,sreg,res,pow;
@@ -511,12 +526,12 @@ bnfsunit(GEN bnf,GEN S,long prec)
     GEN D,U, ClS = cgetg(4,t_VEC);
 
     D = smithall(H, &U, NULL);
-    card = dethnf_i(D);
-    ClS[1] = (long)card; /* h */
     for(i=1; i<lg(D); i++)
       if (gcmp1((GEN)D[i])) break;
-    setlg(D,i);
-    ClS[2]=(long)D; /* cyc */
+    setlg(D,i); D = mattodiagonal_i(D); /* cf smithrel */
+    card = detcyc(D);
+    ClS[1] = (long)card; /* h */
+    ClS[2] = (long)D; /* cyc */
 
     p1=cgetg(i,t_VEC); pow=ZM_inv(U,gun);
     for(i--; i; i--)
@@ -584,7 +599,8 @@ bnfsunit(GEN bnf,GEN S,long prec)
 GEN
 bnfissunit(GEN bnf,GEN suni,GEN x)
 {
-  long lB,cH,i,k,ls,tetpil, av = avma;
+  long lB, cH, i, k, ls;
+  gpmem_t tetpil, av = avma;
   GEN den,gen,S,v,p1,xp,xm,xb,N,HB,perm;
 
   bnf = checkbnf(bnf);
@@ -607,7 +623,7 @@ bnfissunit(GEN bnf,GEN suni,GEN x)
   HB = (GEN)p1[2]; den = (GEN)p1[3];
   cH = lg(HB[1]) - 1;
   lB = lg(HB) - cH;
-  xb = algtobasis(bnf,x); p1 = denom(content(xb));
+  xb = algtobasis(bnf,x); p1 = Q_denom(xb);
   N = mulii(gnorm(gmul(x,p1)), p1); /* relevant primes divide N */
   v = cgetg(ls, t_VECSMALL);
   for (i=1; i<ls; i++)
@@ -679,8 +695,8 @@ GEN
 rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)
 {
   long lgsunitrelnf,i;
-  ulong ltop = avma;
-  GEN relnf,aux,vec,tors,xnf,H,Y,M,A,suni,sunitrelnf,sunitnormnf,prod;
+  gpmem_t ltop = avma;
+  GEN listp,relnf,aux,vec,tors,xnf,H,Y,M,A,suni,sunitrelnf,sunitnormnf,prod;
   GEN res = cgetg(3,t_VEC), S1,S2;
 
   if (typ(ext)!=t_VEC || lg(ext)!=4) err (typeer,"bnfisnorm");
@@ -688,7 +704,7 @@ rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)
   bnf = checkbnf(bnf); relnf = (GEN)ext[3];
   if (gcmp0(x) || gcmp1(x) || (gcmp_1(x) && (degpol(ext[1])&1)))
   {
-    avma = (long)res; res[1]=lcopy(x); res[2]=un; return res;
+    avma = (gpmem_t)res; res[1]=lcopy(x); res[2]=un; return res;
   }
 
 /* construction de l'ensemble S des ideaux
@@ -705,18 +721,24 @@ rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)
     vecconcat(bnf,relnf,vec,&prod,&S1,&S2);
   }
 
-  if (flag>1)
+  if (flag > 1)
   {
-    for (i=2; i<=flag; i++)
-      if (isprime(stoi(i)) && signe(resis(prod,i)))
-      {
-	prod=mulis(prod,i);
-	S1=concatsp(S1,primedec(bnf,stoi(i)));
-	S2=concatsp(S2,primedec(relnf,stoi(i)));
-      }
+    i = cgcd(flag, smodis(prod,flag));
+    if (i > 1) flag /= i; /* don't include same primes twice */
+    listp = (GEN)factor(stoi(flag))[1];
+    for (i=1; i<lg(listp); i++)
+    {
+      GEN p = (GEN)listp[i];
+      prod = mulii(prod,p);
+      S1 = concatsp(S1,primedec(bnf,p));
+      S2 = concatsp(S2,primedec(relnf,p));
+    }
+  }
+  else if (flag < 0)
+  {
+    listp = (GEN)factor(stoi(-flag))[1];
+    vecconcat(bnf,relnf,listp,&prod,&S1,&S2);
   }
-  else if (flag<0)
-    vecconcat(bnf,relnf,(GEN)factor(stoi(-flag))[1],&prod,&S1,&S2);
 
   if (flag)
   {
@@ -725,7 +747,7 @@ rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)
     vecconcat(bnf,relnf,(GEN) factor(absi(normdiscrel))[1],
 	      &prod,&S1,&S2);
   }
-  vec=(GEN) idealfactor(bnf,x)[1]; aux=cgetg(2,t_VEC);
+  vec=(GEN)idealfactor(bnf,x)[1]; aux=cgetg(2,t_VEC);
   for (i=1; i<lg(vec); i++)
     if (signe(resii(prod,gmael(vec,i,1))))
     {
@@ -795,7 +817,7 @@ rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)
 GEN
 bnfisnorm(GEN bnf,GEN x,long flag,long PREC)
 {
-  long ltop = avma, lbot;
+  gpmem_t lbot, ltop = avma;
   GEN ext = cgetg(4,t_VEC);
 
   bnf = checkbnf(bnf);
