@@ -1,4 +1,4 @@
-/* $Id: gen1.c,v 1.41 2001/11/27 22:06:01 karim Exp $
+/* $Id: gen1.c,v 1.50 2002/05/10 15:14:28 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -30,13 +30,13 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
 /* assume z[1] was created last */
 #define fix_frac_if_int(z) if (is_pm1(z[2]))\
-  z = gerepileupto((long)(z+3), (GEN)z[1]);
+  z = gerepileupto((gpmem_t)(z+3), (GEN)z[1]);
 
 /* assume z[1] was created last */
 #define fix_frac_if_int_GC(z,tetpil) { if (is_pm1(z[2]))\
-  z = gerepileupto((long)(z+3), (GEN)z[1]);\
+  z = gerepileupto((gpmem_t)(z+3), (GEN)z[1]);\
 else\
-  gerepilemanyvec((long)z, tetpil, z+1, 2); }
+  gerepilemanyvec((gpmem_t)z, tetpil, z+1, 2); }
 
 GEN quickmul(GEN a, GEN b, long na, long nb);
 
@@ -46,7 +46,7 @@ static GEN
 op_polmod(GEN f(GEN,GEN), GEN x, GEN y, long tx)
 {
   GEN mod,k,l, z=cgetg(3,t_POLMOD);
-  long av,tetpil;
+  gpmem_t av, tetpil;
 
   l=(GEN)y[1];
   if (tx==t_POLMOD)
@@ -105,7 +105,7 @@ gred_rfrac_simple(GEN x1, GEN x2)
 
   c = denom(x1);
   y = cgetg(3,t_RFRAC);
-  y[1] = (long)numer(x1);
+  y[1] = lmul(x1,c);
   y[2] = lmul(x2,c); return y;
 }
 
@@ -165,7 +165,7 @@ gred_rfrac_i(GEN x)
 GEN 
 gred_rfrac2(GEN x1, GEN x2)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   return gerepileupto(av, gred_rfrac2_i(x1, x2));
 }
 
@@ -180,7 +180,7 @@ GEN
 gred_frac2(GEN x1, GEN x2)
 {
   GEN p1, y = dvmdii(x1,x2,&p1);
-  ulong av;
+  gpmem_t av;
 
   if (p1 == gzero) return y; /* gzero intended */
   av = avma;
@@ -221,7 +221,7 @@ gred(GEN x)
 GEN
 gsub(GEN x, GEN y)
 {
-  long tetpil, av = avma;
+  gpmem_t tetpil, av = avma;
   y=gneg_i(y); tetpil=avma;
   return gerepile(av,tetpil,gadd(x,y));
 }
@@ -235,7 +235,7 @@ gsub(GEN x, GEN y)
 static GEN
 addpadic(GEN x, GEN y)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long c,d,e,r,rx,ry;
   GEN u,z,p,mod;
 
@@ -277,8 +277,8 @@ addpadic(GEN x, GEN y)
 static GEN
 gaddpex(GEN x, GEN y)
 {
-  ulong av;
-  long tx,d,r,e;
+  gpmem_t av;  
+  long tx,vy,py,d,r,e;
   GEN z,q,p,p1,p2,mod,u;
 
   if (gcmp0(x)) return gcopy(y);
@@ -287,8 +287,8 @@ gaddpex(GEN x, GEN y)
   e = (tx == t_INT)? pvaluation(x,p,&p1)
                     : pvaluation((GEN)x[1],p,&p1) -
                       pvaluation((GEN)x[2],p,&p2);
-  d = valp(y)-e; r = d+precp(y);
-  if (r<=0) { avma = av; return gcopy(y); }
+  vy = valp(y); d = vy - e; py = precp(y); r = d + py;
+  if (r <= 0) { avma = av; return gcopy(y); }
   mod = (GEN)y[3];
   u   = (GEN)y[4];
   (void)new_chunk(5 + lgefint(mod) + lgefint(p)*labs(d));
@@ -304,15 +304,15 @@ gaddpex(GEN x, GEN y)
   else if (d < 0)
   {
     q = gpowgs(p,-d);
-    mod = divii(mod, q);
-    u   = modii(u, mod);
     if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, mpinvmod(p2,mod));
     p1 = mulii(p1, q);
     u = addii(u, p1);
+    r = py; e = vy;
   }
   else
   {
     long c;
+    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, mpinvmod(p2,mod));
     u = addii(u, p1);
     if (!signe(u) || (c = pvaluation(u,p,&u)) >= r)
     {
@@ -335,7 +335,8 @@ gaddpex(GEN x, GEN y)
 static long
 kro_quad(GEN x, GEN y)
 {
-  long k, av=avma;
+  long k;
+  gpmem_t av=avma;
 
   x = subii(sqri((GEN)x[3]), shifti((GEN)x[2],2));
   k = kronecker(x,y); avma=av; return k;
@@ -354,20 +355,20 @@ addfrac(GEN x, GEN y)
   if (is_pm1(delta))
   {
     p1 = mulii(x1,y2);
-    p2 = mulii(y1,x2); avma = (long)z;
+    p2 = mulii(y1,x2); avma = (gpmem_t)z;
     z[1] = laddii(p1,p2);
     z[2] = lmulii(x2,y2); return z;
   }
   x2 = divii(x2,delta);
   y2 = divii(y2,delta);
   n = addii(mulii(x1,y2), mulii(y1,x2));
-  if (!signe(n)) { avma = (long)(z+3); return gzero; }
+  if (!signe(n)) { avma = (gpmem_t)(z+3); return gzero; }
   d = mulii(x2, y2);
   p1 = dvmdii(n, delta, &p2);
   if (p2 == gzero)
   {
-    if (is_pm1(d)) { avma = (long)(z+3); return icopy(p1); }
-    avma = (long)z;
+    if (is_pm1(d)) { avma = (gpmem_t)(z+3); return icopy(p1); }
+    avma = (gpmem_t)z;
     z[1] = licopy(p1);
     z[2] = licopy(d); return z;
   }
@@ -378,7 +379,7 @@ addfrac(GEN x, GEN y)
     n = divii(n, p1);
   }
   d = mulii(d,delta);
-  avma = (long)z;
+  avma = (gpmem_t)z;
   z[1] = licopy(n);
   z[2] = licopy(d); return z;
 }
@@ -389,7 +390,7 @@ addrfrac(GEN x, GEN y)
   GEN z = cgetg(3,t_RFRAC);
   GEN x1 = (GEN)x[1], x2 = (GEN)x[2];
   GEN y1 = (GEN)y[1], y2 = (GEN)y[2], p1,p2,n,d,delta;
-  long tetpil;
+  gpmem_t tetpil;
 
   delta = ggcd(x2,y2);
   if (gcmp1(delta))
@@ -397,13 +398,13 @@ addrfrac(GEN x, GEN y)
     p1 = gmul(x1,y2);
     p2 = gmul(y1,x2);
     tetpil = avma; /* numerator is non-zero */
-    z[1] = lpile((long)z,tetpil, gadd(p1,p2));
+    z[1] = lpile((gpmem_t)z,tetpil, gadd(p1,p2));
     z[2] = lmul(x2, y2); return z;
   }
   x2 = gdeuc(x2,delta);
   y2 = gdeuc(y2,delta);
   n = gadd(gmul(x1,y2), gmul(y1,x2));
-  if (gcmp0(n)) return gerepileupto((long)(z+3), n);
+  if (gcmp0(n)) return gerepileupto((gpmem_t)(z+3), n);
   tetpil = avma; d = gmul(x2, y2);
   p1 = poldivres(n, delta, &p2); /* we want gcd(n,delta) */
   if (gcmp0(p2))
@@ -413,10 +414,10 @@ addrfrac(GEN x, GEN y)
       d = (GEN)d[2];
            if (gcmp_1(d)) p1 = gneg(p1);
       else if (!gcmp1(d)) p1 = gdiv(p1, d);
-      return gerepileupto((long)(z+3), p1);
+      return gerepileupto((gpmem_t)(z+3), p1);
     }
     z[1]=(long)p1; z[2]=(long)d;
-    gerepilemanyvec((long)z,tetpil,z+1,2); return z;
+    gerepilemanyvec((gpmem_t)z,tetpil,z+1,2); return z;
   }
   p1 = ggcd(delta, p2);
   if (gcmp1(p1))
@@ -431,14 +432,14 @@ addrfrac(GEN x, GEN y)
     z[1] = ldeuc(n,p1);
   }
   z[2] = lmul(d,delta);
-  gerepilemanyvec((long)z,tetpil,z+1,2); return z;
+  gerepilemanyvec((gpmem_t)z,tetpil,z+1,2); return z;
 }
 
 static GEN
 addscalrfrac(GEN x, GEN y)
 {
   GEN p1,num, z = cgetg(3,t_RFRAC);
-  long tetpil, av;
+  gpmem_t tetpil, av;
 
   p1 = gmul(x,(GEN)y[2]); tetpil = avma;
   num = gadd(p1,(GEN)y[1]);
@@ -452,11 +453,11 @@ addscalrfrac(GEN x, GEN y)
       tetpil = avma;
       z[1] = ldiv(num, p1);
       z[2] = ldiv((GEN)y[2], p1);
-      gerepilemanyvec((long)z,tetpil,z+1,2); return z;
+      gerepilemanyvec((gpmem_t)z,tetpil,z+1,2); return z;
     }
   }
   avma = av;
-  z[1]=lpile((long)z,tetpil, num);
+  z[1]=lpile((gpmem_t)z,tetpil, num);
   z[2]=lcopy((GEN)y[2]); return z;
 }
 
@@ -477,7 +478,8 @@ to_polmod(GEN x, GEN mod)
 GEN
 gadd(GEN x, GEN y)
 {
-  long tx = typ(x), ty = typ(y), vx,vy,lx,ly,i,j,k,l,av,tetpil;
+  long tx = typ(x), ty = typ(y), vx, vy, lx, ly, i, j, k, l;
+  gpmem_t av, tetpil;
   GEN z,p1,p2;
 
   if (is_const_t(tx) && is_const_t(ty))
@@ -493,13 +495,13 @@ gadd(GEN x, GEN y)
 	
 	  case t_INTMOD: z=cgetg(3,t_INTMOD); p2=(GEN)y[1];
 	    (void)new_chunk(lgefint(p2)+1); /* HACK */
-            p1 = addii(modii(x,p2),(GEN)y[2]); avma = (long)z;
+            p1 = addii(modii(x,p2),(GEN)y[2]); avma = (gpmem_t)z;
             z[2] = (cmpii(p1,p2) >=0)? lsubii(p1,p2): licopy(p1);
 	    icopyifstack(p2,z[1]); return z;
 
 	  case t_FRAC: case t_FRACN: z=cgetg(3,ty);
             (void)new_chunk(lgefint(x)+lgefint(y[1])+lgefint(y[2])+1); /*HACK*/
-            p1 = mulii((GEN)y[2],x); avma = (long)z;
+            p1 = mulii((GEN)y[2],x); avma = (gpmem_t)z;
 	    z[1] = laddii((GEN)y[1], p1);
 	    z[2] = licopy((GEN)y[2]); return z;
 
@@ -571,7 +573,7 @@ gadd(GEN x, GEN y)
 	  case t_FRAC: case t_FRACN: z=cgetg(3,t_INTMOD); p2=(GEN)x[1];
 	    (void)new_chunk(lgefint(p2)<<2); /* HACK */
             p1 = mulii((GEN)y[1], mpinvmod((GEN)y[2],p2));
-            p1 = addii(modii(p1,p2), (GEN)x[2]); avma=(long)z;
+            p1 = addii(modii(p1,p2), (GEN)x[2]); avma=(gpmem_t)z;
             z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 	
 	  case t_COMPLEX: z=cgetg(3,t_COMPLEX);
@@ -579,10 +581,10 @@ gadd(GEN x, GEN y)
             z[2]=lcopy((GEN)y[2]); return z;
 	
 	  case t_PADIC:
-	    l=avma; p1=cgetg(3,t_INTMOD);
+	    av=avma; p1=cgetg(3,t_INTMOD);
 	    p1[1]=x[1]; p1[2]=lgeti(lgefint(x[1]));
 	    gaffect(y,p1); tetpil=avma;
-	    return gerepile(l,tetpil,gadd(p1,x));
+	    return gerepile(av,tetpil,gadd(p1,x));
 	
 	  case t_QUAD: z=cgetg(4,t_QUAD);
             copyifstack(y[1], z[1]);
@@ -594,10 +596,10 @@ gadd(GEN x, GEN y)
         switch (ty)
 	{
 	  case t_FRAC: return addfrac(x,y);
-	  case t_FRACN: z=cgetg(3,t_FRACN); l=avma;
+	  case t_FRACN: z=cgetg(3,t_FRACN); av=avma;
 	    p1=mulii((GEN)x[1],(GEN)y[2]);
 	    p2=mulii((GEN)x[2],(GEN)y[1]);
-	    tetpil=avma; z[1]=lpile(l,tetpil,addii(p1,p2));
+	    tetpil=avma; z[1]=lpile(av,tetpil,addii(p1,p2));
 	    z[2]=lmulii((GEN)x[2],(GEN)y[2]);
 	    return z;
 
@@ -768,7 +770,7 @@ gadd(GEN x, GEN y)
           for (i=2; i<ly; i++) z[i]=ladd((GEN)x[i],(GEN)y[i]);
           for (   ; i<lx; i++) z[i]=lcopy((GEN)x[i]);
           (void)normalizepol_i(z, lx);
-          if (lgef(z) == 2) { avma = (long)(z + lx); z = zeropol(vx); }
+          if (lgef(z) == 2) { avma = (gpmem_t)(z + lx); z = zeropol(vx); }
           return z;
 
 	case t_SER:
@@ -912,7 +914,8 @@ GEN
 mulscalrfrac(GEN x, GEN y)
 {
   GEN p1,z,y1,y2,cx,cy1,cy2;
-  long tetpil,tx;
+  long tx;
+  gpmem_t tetpil;
 
   if (gcmp0(x)) return gcopy(x);
 
@@ -943,8 +946,8 @@ mulscalrfrac(GEN x, GEN y)
   z[2] = lmul(y2, cy2);
   z[1] = lmul(y1, cy1);
   p1 = fix_rfrac_if_pol((GEN)z[1],(GEN)z[2]);
-  if (p1) return gerepileupto((long)(z+3), p1);
-  gerepilemanyvec((long)z,tetpil,z+1,2); return z;
+  if (p1) return gerepileupto((gpmem_t)(z+3), p1);
+  gerepilemanyvec((gpmem_t)z,tetpil,z+1,2); return z;
 }
 
 static GEN
@@ -953,7 +956,7 @@ mulrfrac(GEN x, GEN y)
   GEN z = cgetg(3,t_RFRAC), p1;
   GEN x1 = (GEN)x[1], x2 = (GEN)x[2];
   GEN y1 = (GEN)y[1], y2 = (GEN)y[2];
-  long tetpil;
+  gpmem_t tetpil;
 
   p1 = ggcd(x1, y2); if (!gcmp1(p1)) { x1 = gdiv(x1,p1); y2 = gdiv(y2,p1); }
   p1 = ggcd(x2, y1); if (!gcmp1(p1)) { x2 = gdiv(x2,p1); y1 = gdiv(y1,p1); }
@@ -961,8 +964,8 @@ mulrfrac(GEN x, GEN y)
   z[2] = lmul(x2,y2);
   z[1] = lmul(x1,y1);
   p1 = fix_rfrac_if_pol((GEN)z[1],(GEN)z[2]);
-  if (p1) return gerepileupto((long)(z+3), p1);
-  gerepilemanyvec((long)z,tetpil,z+1,2); return z;
+  if (p1) return gerepileupto((gpmem_t)(z+3), p1);
+  gerepilemanyvec((gpmem_t)z,tetpil,z+1,2); return z;
 }
 
 GEN
@@ -1073,7 +1076,8 @@ gmul_err(GEN x, GEN y, long tx, long ty)
 GEN
 gmul(GEN x, GEN y)
 {
-  long tx,ty,lx,ly,vx,vy,i,j,k,l,av,tetpil;
+  long tx, ty, lx, ly, vx, vy, i, j, k, l;
+  gpmem_t av, tetpil;
   GEN z,p1,p2,p3,p4;
 
   if (x == y) return gsqr(x);
@@ -1093,7 +1097,7 @@ gmul(GEN x, GEN y)
 
 	  case t_INTMOD: z=cgetg(3,t_INTMOD); p2=(GEN)y[1];
 	    (void)new_chunk(lgefint(p2)<<2); /* HACK */
-            p1=mulii(modii(x,p2),(GEN)y[2]); avma=(long)z;
+            p1=mulii(modii(x,p2),(GEN)y[2]); avma=(gpmem_t)z;
 	    z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
 	  case t_FRAC:
@@ -1102,7 +1106,7 @@ gmul(GEN x, GEN y)
             p1 = mppgcd(x,(GEN)y[2]);
             if (is_pm1(p1))
             {
-              avma = (long)z;
+              avma = (gpmem_t)z;
               z[2] = licopy((GEN)y[2]);
               z[1] = lmulii((GEN)y[1], x);
             }
@@ -1125,8 +1129,8 @@ gmul(GEN x, GEN y)
 	
 	  case t_PADIC:
 	    if (!signe(x)) return gzero;
-	    l=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
-	    return gerepile(l,tetpil,gmul(p1,y));
+	    av=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
+	    return gerepile(av,tetpil,gmul(p1,y));
 	
 	  case t_QUAD: z=cgetg(4,t_QUAD);
 	    copyifstack(y[1], z[1]);
@@ -1140,16 +1144,16 @@ gmul(GEN x, GEN y)
 	  case t_REAL: return mulrr(x,y);
 
 	  case t_FRAC: case t_FRACN:
-	    l=avma; p1=mulri(x,(GEN)y[1]); tetpil=avma;
-            return gerepile(l, tetpil, divri(p1, (GEN)y[2]));
+	    av=avma; p1=mulri(x,(GEN)y[1]); tetpil=avma;
+            return gerepile(av, tetpil, divri(p1, (GEN)y[2]));
 	
 	  case t_COMPLEX: z=cgetg(3,t_COMPLEX);
 	    z[1]=lmul(x,(GEN)y[1]);
 	    z[2]=lmul(x,(GEN)y[2]); return z;
 	
 	  case t_QUAD:
-	    l=avma; p1=co8(y,lg(x)); tetpil=avma;
-	    return gerepile(l,tetpil,gmul(p1,x));
+	    av=avma; p1=co8(y,lg(x)); tetpil=avma;
+	    return gerepile(av,tetpil,gmul(p1,x));
 	
 	  default: err(operf,"*",x,y);
 	}
@@ -1177,7 +1181,7 @@ gmul(GEN x, GEN y)
 	  case t_FRAC: case t_FRACN: z=cgetg(3,t_INTMOD); p2=(GEN)x[1];
             (void)new_chunk(lgefint(p2)<<2); /* HACK */
             p1 = mulii((GEN)y[1], mpinvmod((GEN)y[2],p2));
-            p1 = mulii(modii(p1,p2),(GEN)x[2]); avma=(long)z;
+            p1 = mulii(modii(p1,p2),(GEN)x[2]); avma=(gpmem_t)z;
             z[2] = lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 	
 	  case t_COMPLEX: z=cgetg(3,t_COMPLEX);
@@ -1185,10 +1189,10 @@ gmul(GEN x, GEN y)
 	    z[2]=lmul(x,(GEN)y[2]); return z;
 	
 	  case t_PADIC:
-	    l=avma; p1=cgetg(3,t_INTMOD);
+	    av=avma; p1=cgetg(3,t_INTMOD);
 	    p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
 	    gaffect(y,p1); tetpil=avma;
-	    return gerepile(l,tetpil,gmul(x,p1));
+	    return gerepile(av,tetpil,gmul(x,p1));
 
 	  case t_QUAD: z=cgetg(4,t_QUAD);
             copyifstack(y[1], z[1]);
@@ -1223,8 +1227,8 @@ gmul(GEN x, GEN y)
 	
 	  case t_PADIC:
 	    if (!signe(x[1])) return gzero;
-	    l=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
-            return gerepile(l,tetpil,gmul(p1,y));
+	    av=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
+            return gerepile(av,tetpil,gmul(p1,y));
 	
 	  case t_QUAD: z=cgetg(4,t_QUAD);
 	    copyifstack(y[1], z[1]);
@@ -1235,14 +1239,14 @@ gmul(GEN x, GEN y)
       case t_COMPLEX:
         switch(ty)
 	{
-	  case t_COMPLEX: z=cgetg(3,t_COMPLEX); l=avma;
+	  case t_COMPLEX: z=cgetg(3,t_COMPLEX); av=avma;
             p1=gmul((GEN)x[1],(GEN)y[1]);
             p2=gmul((GEN)x[2],(GEN)y[2]);
 	    x=gadd((GEN)x[1],(GEN)x[2]);
             y=gadd((GEN)y[1],(GEN)y[2]);
 	    y=gmul(x,y); x=gadd(p1,p2);
 	    tetpil=avma; z[1]=lsub(p1,p2); z[2]=lsub(y,x);
-	    gerepilemanyvec(l,tetpil,z+1,2); return z;
+	    gerepilemanyvec(av,tetpil,z+1,2); return z;
 	
 	  case t_PADIC:
 	    if (krosg(-1,(GEN)y[2]))
@@ -1262,8 +1266,8 @@ gmul(GEN x, GEN y)
 	
 	  case t_QUAD:
 	    lx=precision(x); if (!lx) err(operi,"*",x,y);
-	    l=avma; p1=co8(y,lx); tetpil=avma;
-	    return gerepile(l,tetpil,gmul(p1,x));
+	    av=avma; p1=co8(y,lx); tetpil=avma;
+	    return gerepile(av,tetpil,gmul(p1,x));
 	}
 	
       case t_PADIC:
@@ -1297,7 +1301,7 @@ gmul(GEN x, GEN y)
         p1=(GEN)x[1]; p2=(GEN)y[1];
         if (!gegal(p1,p2)) err(operi,"*",x,y);
 
-        copyifstack(p2, z[1]); l=avma;
+        copyifstack(p2, z[1]); av=avma;
         p2=gmul((GEN)x[2],(GEN)y[2]);
         p3=gmul((GEN)x[3],(GEN)y[3]);
         p4=gmul(gneg_i((GEN)p1[2]),p3);
@@ -1305,17 +1309,17 @@ gmul(GEN x, GEN y)
         if (gcmp0((GEN)p1[3]))
         {
           tetpil=avma;
-          z[2]=lpile(l,tetpil,gadd(p4,p2)); l=avma;
+          z[2]=lpile(av,tetpil,gadd(p4,p2)); av=avma;
           p2=gmul((GEN)x[2],(GEN)y[3]);
           p3=gmul((GEN)x[3],(GEN)y[2]); tetpil=avma;
-          z[3]=lpile(l,tetpil,gadd(p2,p3)); return z;
+          z[3]=lpile(av,tetpil,gadd(p2,p3)); return z;
         }
 
         p1 = gadd(gmul((GEN)x[2],(GEN)y[3]), gmul((GEN)x[3],(GEN)y[2]));
         tetpil=avma;
         z[2]=ladd(p2,p4);
         z[3]=ladd(p1,p3);
-        gerepilemanyvec(l,tetpil,z+2,2); return z;
+        gerepilemanyvec(av,tetpil,z+2,2); return z;
     }
     err(bugparier,"multiplication");
   }
@@ -1350,13 +1354,13 @@ gmul(GEN x, GEN y)
         {
           case t_COL:
             if (lx!=ly) err(operi,"*",x,y);
-            z=gzero; l=avma;
+            z=gzero; av=avma;
             for (i=1; i<lx; i++)
             {
               p1=gmul((GEN)x[i],(GEN)y[i]);
               z=gadd(z,p1);
             }
-            return gerepileupto(l,z);
+            return gerepileupto(av,z);
 
           case t_MAT:
             if (ly==1) return cgetg(1,t_VEC);
@@ -1417,13 +1421,13 @@ gmul(GEN x, GEN y)
             lx=lg(x[1]); z=cgetg(lx,t_COL);
             for (i=1; i<lx; i++)
             {
-              p1=gzero; l=avma;
+              p1=gzero; av=avma;
               for (j=1; j<ly; j++)
               {
                 p2=gmul(gcoeff(x,i,j),(GEN)y[j]);
                 p1=gadd(p1,p2);
               }
-              z[i]=lpileupto(l,p1);
+              z[i]=lpileupto(av,p1);
             }
             return z;
 
@@ -1512,7 +1516,7 @@ gmul(GEN x, GEN y)
  */
           GEN a = x,b = y
           GEN p = NULL, pol = NULL;
-          long av = avma;
+          gpmem_t av = avma;
           if (ff_poltype(&x,&p,&pol) && ff_poltype(&y,&p,&pol))
           {
             /* fprintferr("HUM"); */
@@ -1551,6 +1555,8 @@ gmul(GEN x, GEN y)
       switch (ty)
       {
 	case t_SER:
+        {
+          long mix, miy;
 	  if (gcmp0(x) || gcmp0(y)) return zeroser(vx, valp(x)+valp(y));
           lx=lg(x); ly=lg(y);
 	  if (lx>ly) { k=ly; ly=lx; lx=k; p1=y; y=x; x=p1; }
@@ -1558,23 +1564,27 @@ gmul(GEN x, GEN y)
 	  z[1] = evalvalp(valp(x)+valp(y)) | evalvarn(vx) | evalsigne(1);
           x += 2; y += 2; z += 2; lx -= 3;
           p2 = (GEN)gpmalloc((lx+1)*sizeof(long));
+          p3 = (GEN)gpmalloc((ly+1)*sizeof(long));
+          mix = miy = 0;
 	  for (i=0; i<=lx; i++)
           {
-	    p2[i] = !isexactzero((GEN)y[i]);
+	    p2[i] = !isexactzero((GEN)y[i]); if (p2[i]) miy = i;
+	    p3[i] = !isexactzero((GEN)x[i]); if (p3[i]) mix = i;
             p1 = gzero; av = avma;
-            for (j=0; j<=i; j++)
-              if (p2[j])
-                p1 = gadd(p1, gmul((GEN)y[j],(GEN)x[i-j]));
+            for (j=i-mix; j<=min(i,miy); j++)
+              if (p2[j]) p1 = gadd(p1, gmul((GEN)y[j],(GEN)x[i-j]));
             z[i] = lpileupto(av,p1);
           }
           z -= 2; /* back to normalcy */
-          free(p2); return normalize(z);
+          free(p2);
+          free(p3); return normalize(z);
+        }
 
 	case t_RFRAC: case t_RFRACN:
 	  if (gcmp0(y)) return zeropol(vx);
 	  if (gcmp0(x)) return zeroser(vx, valp(x)+gval(y,vx));
-	  l=avma; p1=gmul((GEN)y[1],x); tetpil=avma;
-          return gerepile(l,tetpil,gdiv(p1,(GEN)y[2]));
+	  av=avma; p1=gmul((GEN)y[1],x); tetpil=avma;
+          return gerepile(av,tetpil,gdiv(p1,(GEN)y[2]));
 	
 	default: err(operf,"*",x,y);
       }
@@ -1593,7 +1603,8 @@ gmul(GEN x, GEN y)
 GEN
 gsqr(GEN x)
 {
-  long tx=typ(x),lx,i,j,k,l,av,tetpil;
+  long tx=typ(x), lx, i, j, k, l;
+  gpmem_t av, tetpil;
   GEN z,p1,p2,p3,p4;
 
   if (is_scalar_t(tx))
@@ -1607,7 +1618,7 @@ gsqr(GEN x)
 	
       case t_INTMOD: z=cgetg(3,t_INTMOD); p2=(GEN)x[1];
         (void)new_chunk(lgefint(p2)<<2); /* HACK */
-        p1=sqri((GEN)x[2]); avma=(long)z;
+        p1=sqri((GEN)x[2]); avma=(gpmem_t)z;
         z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
       case t_FRAC: case t_FRACN:
@@ -1617,13 +1628,13 @@ gsqr(GEN x)
 	return z;
 
       case t_COMPLEX:
-	z=cgetg(3,t_COMPLEX); l=avma;
+	z=cgetg(3,t_COMPLEX); av=avma;
 	p1=gadd((GEN)x[1],(GEN)x[2]);
 	p2=gadd((GEN)x[1],gneg_i((GEN)x[2]));
 	p3=gmul((GEN)x[1],(GEN)x[2]);
 	tetpil=avma;
 	z[1]=lmul(p1,p2); z[2]=lshift(p3,1);
-	gerepilemanyvec(l,tetpil,z+1,2);
+	gerepilemanyvec(av,tetpil,z+1,2);
 	return z;
 	
       case t_PADIC:
@@ -1633,33 +1644,33 @@ gsqr(GEN x)
         z[1] = evalprecp(precp(x)+i) | evalvalp(2*valp(x));
 	icopyifstack(x[2], z[2]);
         z[3] = lshifti((GEN)x[3], i); av = avma;
-	z[4] = (long)gerepileuptoint(av, modii(sqri((GEN)x[4]), (GEN)z[3]));
+	z[4] = lpileuptoint(av, modii(sqri((GEN)x[4]), (GEN)z[3]));
 	return z;
 	
       case t_QUAD:
-	p1=(GEN)x[1]; z=cgetg(4,t_QUAD); l=avma;
+	p1=(GEN)x[1]; z=cgetg(4,t_QUAD); av=avma;
 	p2=gsqr((GEN)x[2]); p3=gsqr((GEN)x[3]);
 	p4=gmul(gneg_i((GEN)p1[2]),p3);
 
 	if (gcmp0((GEN)p1[3]))
 	{
 	  tetpil=avma;
-	  z[2]=lpile(l,tetpil,gadd(p4,p2));
-	  l=avma; p2=gmul((GEN)x[2],(GEN)x[3]); tetpil=avma;
-	  z[3]=lpile(l,tetpil,gmul2n(p2,1));
+	  z[2]=lpile(av,tetpil,gadd(p4,p2));
+	  av=avma; p2=gmul((GEN)x[2],(GEN)x[3]); tetpil=avma;
+	  z[3]=lpile(av,tetpil,gmul2n(p2,1));
 	  copyifstack(p1,z[1]); return z;
 	}
 
 	p1=gmul((GEN)x[2],(GEN)x[3]);
 	p1=gmul2n(p1,1); tetpil=avma;
 	z[2]=ladd(p2,p4); z[3]=ladd(p1,p3);
-	gerepilemanyvec(l,tetpil,z+2,2);
+	gerepilemanyvec(av,tetpil,z+2,2);
 	copyifstack(x[1],z[1]); return z;
 
       case t_POLMOD:
         z=cgetg(3,t_POLMOD); copyifstack(x[1],z[1]);
-	l=avma; p1=gsqr((GEN)x[2]); tetpil=avma;
-        z[2]=lpile(l,tetpil, gres(p1,(GEN)z[1]));
+	av=avma; p1=gsqr((GEN)x[2]); tetpil=avma;
+        z[2]=lpile(av,tetpil, gres(p1,(GEN)z[1]));
 	return z;
     }
 
@@ -1686,25 +1697,27 @@ gsqr(GEN x)
     }
 
     case t_SER:
+    {
+      long mi;
       if (gcmp0(x)) return zeroser(varn(x), 2*valp(x));
       lx = lg(x); z = cgetg(lx,tx);
       z[1] = evalsigne(1) | evalvalp(2*valp(x)) | evalvarn(varn(x));
       x += 2; z += 2; lx -= 3;
       p2 = (GEN)gpmalloc((lx+1)*sizeof(long));
+      mi = 0;
       for (i=0; i<=lx; i++)
       {
-	p2[i] = !isexactzero((GEN)x[i]);
-        p1=gzero; av=avma; l=(i+1)>>1;
-        for (j=0; j<l; j++)
-          if (p2[j] && p2[i-j])
-            p1 = gadd(p1, gmul((GEN)x[j],(GEN)x[i-j]));
+	p2[i] = !isexactzero((GEN)x[i]); if (p2[i]) mi = i;
+        p1=gzero; av=avma; l=((i+1)>>1) - 1;
+        for (j=i-mi; j<=min(l,mi); j++)
+          if (p2[j] && p2[i-j]) p1 = gadd(p1, gmul((GEN)x[j],(GEN)x[i-j]));
         p1 = gshift(p1,1);
         if ((i&1) == 0 && p2[i>>1])
           p1 = gadd(p1, gsqr((GEN)x[i>>1]));
         z[i] = lpileupto(av,p1);
       }
       z -= 2; free(p2); return normalize(z);
-
+    }
     case t_RFRAC: case t_RFRACN:
       z=cgetg(3,tx);
       z[1]=lsqr((GEN)x[1]);
@@ -1720,10 +1733,10 @@ gsqr(GEN x)
         z[j]=lgetg(lx,t_COL);
         for (i=1; i<lx; i++)
         {
-          p1=gzero; l=avma;
+          p1=gzero; av=avma;
           for (k=1; k<lx; k++)
             p1 = gadd(p1, gmul(gcoeff(x,i,k),gcoeff(x,k,j)));
-          coeff(z,i,j)=lpileupto(l,p1);
+          coeff(z,i,j)=lpileupto(av,p1);
         }
       }
       return z;
@@ -1764,7 +1777,8 @@ GEN divrfrac(GEN x, GEN y)
 GEN
 gdiv(GEN x, GEN y)
 {
-  long tx = typ(x), ty = typ(y), lx,ly,vx,vy,i,j,k,l,av,tetpil;
+  long tx = typ(x), ty = typ(y), lx, ly, vx, vy, i, j, k, l;
+  gpmem_t av, tetpil;
   GEN z,p1,p2,p3;
 
   if (y == gun) return gcopy(x);
@@ -1793,7 +1807,7 @@ gdiv(GEN x, GEN y)
       case t_INTMOD:
         z=cgetg(3,t_INTMOD); p2=(GEN)y[1];
         (void)new_chunk(lgefint(p2)<<2); /* HACK */
-        p1=mulii(modii(x,p2), mpinvmod((GEN)y[2],p2)); avma=(long)z;
+        p1=mulii(modii(x,p2), mpinvmod((GEN)y[2],p2)); avma=(gpmem_t)z;
         z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
       case t_FRAC:
@@ -1801,7 +1815,7 @@ gdiv(GEN x, GEN y)
         p1 = mppgcd(x,(GEN)y[1]);
         if (is_pm1(p1))
         {
-          avma = (long)z; tetpil = 0;
+          avma = (gpmem_t)z; tetpil = 0;
           z[2] = licopy((GEN)y[1]);
         }
         else
@@ -1824,12 +1838,12 @@ gdiv(GEN x, GEN y)
         fix_frac(z); return z;
 
       case t_PADIC:
-        l=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
-        return gerepile(l,tetpil,gdiv(p1,y));
+        av=avma; p1=cgetp(y); gaffect(x,p1); tetpil=avma;
+        return gerepile(av,tetpil,gdiv(p1,y));
             
       case t_COMPLEX: case t_QUAD:
-        l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
-        return gerepile(l,tetpil,gdiv(p2,p1));
+        av=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
+        return gerepile(av,tetpil,gdiv(p2,p1));
     }
   }
   if (gcmp0(y) && ty != t_MAT) err(gdiver2);
@@ -1848,22 +1862,22 @@ gdiv(GEN x, GEN y)
 	    return divrr(x,y);
 
 	  case t_FRAC: case t_FRACN:
-	    l=avma; p1=mulri(x,(GEN)y[2]); tetpil=avma;
-	    return gerepile(l, tetpil, divri(p1,(GEN)y[1]));
+	    av=avma; p1=mulri(x,(GEN)y[2]); tetpil=avma;
+	    return gerepile(av, tetpil, divri(p1,(GEN)y[1]));
 
 	  case t_COMPLEX: z=cgetg(3,t_COMPLEX);
-            l=avma; p1=gnorm(y);
+            av=avma; p1=gnorm(y);
 	    p2=gmul(x,(GEN)y[1]);
             p3=gmul(x,(GEN)y[2]);
 	    if (!gcmp0(p3)) p3 = gneg_i(p3);
             tetpil=avma;
 	    z[1]=ldiv(p2,p1);
             z[2]=ldiv(p3,p1);
-	    gerepilemanyvec(l,tetpil,z+1,2); return z;
+	    gerepilemanyvec(av,tetpil,z+1,2); return z;
 
 	  case t_QUAD:
-	    l=avma; p1=co8(y,lg(x)); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(x,p1));
+	    av=avma; p1=co8(y,lg(x)); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(x,p1));
 
 	  case t_INTMOD: case t_PADIC: err(operf,"/",x,y);
 	}
@@ -1873,7 +1887,7 @@ gdiv(GEN x, GEN y)
 	{
 	  case t_INT: z=cgetg(3,t_INTMOD); p2=(GEN)x[1];
 	    (void)new_chunk(lgefint(p2)<<2); /* HACK */
-	    p1=mulii((GEN)x[2], mpinvmod(y,p2)); avma=(long)z;
+	    p1=mulii((GEN)x[2], mpinvmod(y,p2)); avma=(gpmem_t)z;
             z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
 	  case t_INTMOD: z=cgetg(3,t_INTMOD); p2=(GEN)x[1]; p1=(GEN)y[1];
@@ -1888,20 +1902,20 @@ gdiv(GEN x, GEN y)
 	  case t_FRAC: z=cgetg(3,t_INTMOD); p2=(GEN)x[1];
             (void)new_chunk(lgefint(p2)<<2); /* HACK */
             p1=mulii((GEN)y[2], mpinvmod((GEN)y[1],p2));
-            p1=mulii(modii(p1,p2),(GEN)x[2]); avma=(long)z;
+            p1=mulii(modii(p1,p2),(GEN)x[2]); avma=(gpmem_t)z;
             z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
 	  case t_FRACN:
-	    l=avma; p1=gred(y); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(x,p1));
+	    av=avma; p1=gred(y); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(x,p1));
 
 	  case t_COMPLEX: case t_QUAD:
-	    l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p2,p1));
+	    av=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p2,p1));
 
 	  case t_PADIC:
-	    l=avma; p1=cgetg(3,t_INTMOD); p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
-	    gaffect(y,p1); tetpil=avma; return gerepile(l,tetpil,gdiv(x,p1));
+	    av=avma; p1=cgetg(3,t_INTMOD); p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
+	    gaffect(y,p1); tetpil=avma; return gerepile(av,tetpil,gdiv(x,p1));
 
 	  case t_REAL: err(operf,"/",x,y);
 	}
@@ -1916,7 +1930,7 @@ gdiv(GEN x, GEN y)
             p1 = mppgcd(y,(GEN)x[1]);
             if (is_pm1(p1))
             {
-              avma = (long)z; tetpil = 0;
+              avma = (gpmem_t)z; tetpil = 0;
               z[1] = licopy((GEN)x[1]);
             }
             else
@@ -1936,13 +1950,13 @@ gdiv(GEN x, GEN y)
           return z;
 
 	  case t_REAL:
-	    l=avma; p1=mulri(y,(GEN)x[2]); tetpil=avma;
-            return gerepile(l, tetpil, divir((GEN)x[1], p1));
+	    av=avma; p1=mulri(y,(GEN)x[2]); tetpil=avma;
+            return gerepile(av, tetpil, divir((GEN)x[1], p1));
 
 	  case t_INTMOD: z=cgetg(3,t_INTMOD); p2=(GEN)y[1];
             (void)new_chunk(lgefint(p2)<<2); /* HACK */
 	    p1=mulii((GEN)y[2],(GEN)x[2]);
-	    p1=mulii(mpinvmod(p1,p2), modii((GEN)x[1],p2)); avma=(long)z;
+	    p1=mulii(mpinvmod(p1,p2), modii((GEN)x[1],p2)); avma=(gpmem_t)z;
 	    z[2]=lmodii(p1,p2); icopyifstack(p2,z[1]); return z;
 
 	  case t_FRAC: if (tx == t_FRACN) ty=t_FRACN;
@@ -1971,23 +1985,23 @@ gdiv(GEN x, GEN y)
             return z;
 
 	  case t_COMPLEX: z=cgetg(3,t_COMPLEX);
-            l=avma; p1=gnorm(y);
+            av=avma; p1=gnorm(y);
 	    p2=gmul(x,(GEN)y[1]);
 	    p3=gmul(x,(GEN)y[2]);
             if(!gcmp0(p3)) p3 = gneg_i(p3);
 	    tetpil=avma;
 	    z[1]=ldiv(p2,p1); z[2]=ldiv(p3,p1);
-	    gerepilemanyvec(l,tetpil,z+1,2); return z;
+	    gerepilemanyvec(av,tetpil,z+1,2); return z;
 
 	  case t_PADIC:
 	    if (!signe(x[1])) return gzero;
 
-	    l=avma; p1=cgetp(y); gaffect(x,p1);
-	    tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));
+	    av=avma; p1=cgetp(y); gaffect(x,p1);
+	    tetpil=avma; return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_QUAD:
-	    l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p2,p1));
+	    av=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p2,p1));
 	}
 
       case t_COMPLEX:
@@ -1999,8 +2013,8 @@ gdiv(GEN x, GEN y)
 	    z[2]=ldiv((GEN)x[2],y); return z;
 
 	  case t_COMPLEX:
-	    l=avma; p1=gnorm(y); p2=gconj(y); p2=gmul(x,p2); tetpil=avma;
-	    return gerepile(l,tetpil, gdiv(p2,p1));
+	    av=avma; p1=gnorm(y); p2=gconj(y); p2=gmul(x,p2); tetpil=avma;
+	    return gerepile(av,tetpil, gdiv(p2,p1));
 
 	  case t_PADIC:
 	    if (krosg(-1,(GEN)y[2])== -1)
@@ -2014,24 +2028,24 @@ gdiv(GEN x, GEN y)
 
 	  case t_QUAD:
 	    lx=precision(x); if (!lx) err(operi,"/",x,y);
-	    l=avma; p1=co8(y,lx); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(x,p1));
+	    av=avma; p1=co8(y,lx); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(x,p1));
 	}
 
       case t_PADIC:
 	switch(ty)
 	{
 	  case t_INT: case t_FRAC: case t_FRACN:
-	    l=avma;
+	    av=avma;
 	    if (signe(x[4])) { p1=cgetp(x); gaffect(y,p1); }
 	    else p1=cvtop(y,(GEN)x[2],(valp(x)>0)?valp(x):1);
-	    tetpil=avma; return gerepile(l,tetpil,gdiv(x,p1));
+	    tetpil=avma; return gerepile(av,tetpil,gdiv(x,p1));
 
 	  case t_INTMOD:
-	    l=avma; p1=cgetg(3,t_INTMOD);
+	    av=avma; p1=cgetg(3,t_INTMOD);
 	    p1[1]=y[1]; p1[2]=lgeti(lg(y[1]));
 	    gaffect(x,p1); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p1,y));
+	    return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_PADIC:
 	    if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"/",x,y);
@@ -2042,15 +2056,15 @@ gdiv(GEN x, GEN y)
 	    }
 
 	    p1=(precp(x)>precp(y)) ? y : x;
-	    z=cgetp(p1); l=avma;
+	    z=cgetp(p1); av=avma;
 	    setvalp(z,valp(x)-valp(y));
 	    p2=mpinvmod((GEN)y[4],(GEN)p1[3]);
 	    modiiz(mulii((GEN)x[4],p2),(GEN)p1[3],(GEN)z[4]);
-	    avma=l; return z;
+	    avma=av; return z;
 
 	  case t_COMPLEX: case t_QUAD:
-	    l=avma; p1=gmul(x,gconj(y)); p2=gnorm(y); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p1,p2));
+	    av=avma; p1=gmul(x,gconj(y)); p2=gnorm(y); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p1,p2));
 
 	  case t_REAL:
 	    err(operf,"/",x,y);
@@ -2066,23 +2080,23 @@ gdiv(GEN x, GEN y)
 	    return z;
 
 	  case t_REAL:
-	    l=avma; p1=co8(x,lg(y)); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p1,y));
+	    av=avma; p1=co8(x,lg(y)); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_PADIC:
-	    l=avma; p1=cvtop(x,(GEN)y[2],precp(y));
-	    tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));
+	    av=avma; p1=cvtop(x,(GEN)y[2],precp(y));
+	    tetpil=avma; return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_COMPLEX:
 	    ly=precision(y); if (!ly) err(operi,"/",x,y);
-	    l=avma; p1=co8(x,ly); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p1,y));
+	    av=avma; p1=co8(x,ly); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_QUAD:
 	    k=x[1]; l=y[1];
 	    if (!gegal((GEN)k,(GEN)l)) err(operi,"/",x,y);
-	    l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
-	    return gerepile(l,tetpil,gdiv(p2,p1));
+	    av=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
+	    return gerepile(av,tetpil,gdiv(p2,p1));
 	}
     }
     err(bugparier,"division");
@@ -2145,8 +2159,8 @@ gdiv(GEN x, GEN y)
       case t_SER:
 	if (gcmp0(x))
 	{
-          l=avma; p1=ginv(y); tetpil=avma; /* a ameliorer !!!! */
-	  return gerepile(l,tetpil,gmul(x,p1));
+          av=avma; p1=ginv(y); tetpil=avma; /* a ameliorer !!!! */
+	  return gerepile(av,tetpil,gmul(x,p1));
 	}
         p1 = (GEN)gpmalloc(ly*sizeof(long));
         p1[0] = evaltyp(t_SER) | evallg(ly);
@@ -2160,8 +2174,8 @@ gdiv(GEN x, GEN y)
         z[2]=lcopy((GEN)y[1]); return z;
 
       case t_MAT:
-	l=avma; p1=invmat(y); tetpil=avma;
-	return gerepile(l,tetpil,gmul(x,p1));
+	av=avma; p1=invmat(y); tetpil=avma;
+	return gerepile(av,tetpil,gmul(x,p1));
 
       case t_VEC: case t_COL: err(operf,"/",x,y);
     }
@@ -2249,8 +2263,8 @@ gdiv(GEN x, GEN y)
         }
 
 	case t_RFRAC: case t_RFRACN:
-	  l=avma; p2=gmul(x,(GEN)y[2]); tetpil=avma;
-	  return gerepile(l,tetpil,gdiv(p2,(GEN)y[1]));
+	  av=avma; p2=gmul(x,(GEN)y[2]); tetpil=avma;
+	  return gerepile(av,tetpil,gdiv(p2,(GEN)y[1]));
 
 	default: err(operf,"/",x,y);
       }
@@ -2265,8 +2279,8 @@ gdiv(GEN x, GEN y)
 	  z[1]=lcopy((GEN)x[1]); return z;
 
 	case t_SER:
-	  l=avma; p2=gmul((GEN)x[2],y); tetpil=avma;
-	  return gerepile(l,tetpil, gdiv((GEN)x[1],p2));
+	  av=avma; p2=gmul((GEN)x[2],y); tetpil=avma;
+	  return gerepile(av,tetpil, gdiv((GEN)x[1],p2));
 
 	case t_RFRAC: case t_RFRACN:
 	  if (tx == t_RFRACN) ty=t_RFRACN;
@@ -2286,8 +2300,8 @@ gdiv(GEN x, GEN y)
 	return z;
       }
       if (ty!=t_MAT || ly==1 || lg(y[1])!=ly) err(operi,"/",x,y);
-      l=avma; p1=invmat(y); tetpil=avma;
-      return gerepile(l,tetpil,gmul(x,p1));
+      av=avma; p1=invmat(y); tetpil=avma;
+      return gerepile(av,tetpil,gmul(x,p1));
      case t_QFI:case t_QFR:
        break;
      default: err(operf,"/",x,y);
