@@ -1,4 +1,4 @@
-/* $Id: gen2.c,v 1.41 2002/01/10 19:41:57 karim Exp $
+/* $Id: gen2.c,v 1.51 2002/06/03 09:14:46 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -32,31 +32,31 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 void
 gop1z(GEN (*f)(GEN), GEN x, GEN y)
 {
-  long av=avma; gaffect(f(x),y); avma=av;
+  gpmem_t av=avma; gaffect(f(x), y); avma=av;
 }
 
 void
 gop2z(GEN (*f)(GEN, GEN), GEN x, GEN y, GEN z)
 {
-  long av=avma; gaffect(f(x,y),z); avma=av;
+  gpmem_t av=avma; gaffect(f(x, y), z); avma=av;
 }
 
 void
 gops2gsz(GEN (*f)(GEN, long), GEN x, long s, GEN z)
 {
-  long av=avma; gaffect(f(x,s),z); avma=av;
+  gpmem_t av=avma; gaffect(f(x, s), z); avma=av;
 }
 
 void
 gops2sgz(GEN (*f)(long, GEN), long s, GEN y, GEN z)
 {
-  long av=avma; gaffect(f(s,y),z); avma=av;
+  gpmem_t av=avma; gaffect(f(s, y), z); avma=av;
 }
 
 void
 gops2ssz(GEN (*f)(long, long), long s, long y, GEN z)
 {
-  long av=avma; gaffect(f(s,y),z); avma=av;
+  gpmem_t av=avma; gaffect(f(s, y), z); avma=av;
 }
 
 /*******************************************************************/
@@ -66,7 +66,7 @@ gops2ssz(GEN (*f)(long, long), long s, long y, GEN z)
 /*******************************************************************/
 
 /* small int prototype */
-static long court_p[] = { evaltyp(t_INT) | m_evallg(3),0,0 };
+static long court_p[] = { evaltyp(t_INT) | _evallg(3),0,0 };
 
 GEN
 gopsg2(GEN (*f)(GEN, GEN), long s, GEN y)
@@ -89,14 +89,14 @@ opgs2(int (*f)(GEN, GEN), GEN y, long s)
 void
 gopsg2z(GEN (*f)(GEN, GEN), long s, GEN y, GEN z)
 {
-  long av=avma;
+  gpmem_t av=avma;
   affsi(s,court_p); gaffect(f(court_p,y),z); avma=av;
 }
 
 void
 gopgs2z(GEN (*f)(GEN, GEN), GEN y, long s, GEN z)
 {
-  long av=avma;
+  gpmem_t av=avma;
   affsi(s,court_p); gaffect(f(y,court_p),z); avma=av;
 }
 
@@ -214,7 +214,8 @@ greffe(GEN x, long l, long use_stack)
 long
 gtolong(GEN x)
 {
-  long y,tx=typ(x),av=avma;
+  long y, tx=typ(x);
+  gpmem_t av=avma;
 
   switch(tx)
   {
@@ -336,6 +337,7 @@ gcmp1(GEN x)
 int
 gcmp_1(GEN x)
 {
+  gpmem_t av;
   long l,y;
   GEN p1;
 
@@ -355,7 +357,7 @@ gcmp_1(GEN x)
       return 0;
 
     case t_INTMOD:
-      l=avma; y=egalii(addsi(1,(GEN)x[2]), (GEN)x[1]); avma=l; return y;
+      av=avma; y=egalii(addsi(1,(GEN)x[2]), (GEN)x[1]); avma=av; return y;
 
     case t_FRAC: case t_FRACN:
       l = signe(x[1]);
@@ -368,11 +370,11 @@ gcmp_1(GEN x)
       return gcmp_1((GEN)x[2]) && gcmp0((GEN)x[3]);
 
     case t_PADIC:
-      l=avma; y=gegal(addsi(1,(GEN)x[4]), (GEN)x[3]); avma=l; return y;
+      av=avma; y=gegal(addsi(1,(GEN)x[4]), (GEN)x[3]); avma=av; return y;
 
     case t_POLMOD:
-      l=avma; p1=gadd(gun,(GEN)x[2]);
-      y = signe(p1) && !gegal(p1,(GEN)x[1]); avma=l; return !y;
+      av=avma; p1=gadd(gun,(GEN)x[2]);
+      y = signe(p1) && !gegal(p1,(GEN)x[1]); avma=av; return !y;
 
     case t_POL:
       return lgef(x)==3 && gcmp_1((GEN)x[2]);
@@ -384,7 +386,8 @@ gcmp_1(GEN x)
 int
 gcmp(GEN x, GEN y)
 {
-  long tx,ty,f,av;
+  long tx, ty, f;
+  gpmem_t av;
 
   tx=typ(x); ty=typ(y);
   if (is_intreal_t(tx))
@@ -394,7 +397,9 @@ gcmp(GEN x, GEN y)
     if (tx==t_STR)
     {
       if (ty != t_STR) return 1;
-      return strcmp(GSTR(x),GSTR(y));
+      f = strcmp(GSTR(x),GSTR(y));
+      return f > 0? 1
+                  : f? -1: 0;
     }
     if (!is_frac_t(tx)) err(typeer,"comparison");
   }
@@ -499,7 +504,8 @@ vecegal(GEN x, GEN y)
 int
 gegal(GEN x, GEN y)
 {
-  ulong av,tx;
+  ulong tx;
+  gpmem_t av;
   long i;
   
   if (x == y) return 1;
@@ -547,8 +553,8 @@ gegal(GEN x, GEN y)
         return 1;
     }
   {
+    VOLATILE gpmem_t av = avma;
     jmp_buf env;
-    VOLATILE long av = avma;
     void *c;
     if (setjmp(env)) i = 0;
     else
@@ -611,7 +617,8 @@ polvaluation(GEN x, GEN *Z)
 long
 ggval(GEN x, GEN p)
 {
-  long tx=typ(x), tp=typ(p), av, limit,vx,v,i,val;
+  long tx=typ(x), tp=typ(p), vx, v, i, val;
+  gpmem_t av, limit;
   GEN p1,p2;
 
   if (isexactzero(x)) return VERYBIGINT;
@@ -689,7 +696,7 @@ ggval(GEN x, GEN p)
 
 /* x is non-zero */
 long
-svaluation(ulong x, ulong p, long *py)
+svaluation(ulong x, ulong p, ulong *py)
 {
   ulong v = 0;
   for(;;)
@@ -703,7 +710,8 @@ svaluation(ulong x, ulong p, long *py)
 long
 pvaluation(GEN x, GEN p, GEN *py)
 {
-  long av,v;
+  long v;
+  gpmem_t av, av2;
   GEN p1,p2;
 
   if (egalii(p,gdeux))
@@ -718,14 +726,17 @@ pvaluation(GEN x, GEN p, GEN *py)
     if (py) { *py = v? negi(x): icopy(x); }
     return v;
   }
-  if (!is_bigint(x))
+  if (lgefint(x) == 3)
   {
-    long y;
-    if (!is_bigint(p))
+    if (lgefint(p) == 3)
     {
-      v = svaluation(x[2],p[2], &y);
-      if (signe(x) < 0) y = -y;
-      if (py) *py = stoi(y);
+      ulong y;
+      v = svaluation((ulong)x[2],(ulong)p[2], &y);
+      if (py)
+      {
+        *py = utoi(y);
+        if (signe(x) < 0) setsigne(*py, -1);
+      }
     }
     else
     {
@@ -735,11 +746,13 @@ pvaluation(GEN x, GEN p, GEN *py)
     return v;
   }
   av = avma; v = 0; (void)new_chunk(lgefint(x));
+  av2= avma;
   for(;;)
   {
     p1 = dvmdii(x,p,&p2);
     if (p2 != gzero) { avma=av; if (py) *py = icopy(x); return v; }
     v++; x = p1;
+    if ((v & 0xff) == 0) p1 = gerepileuptoint(av2, p1);
   }
 }
 
@@ -869,7 +882,8 @@ gneg_i(GEN x)
 GEN
 gabs(GEN x, long prec)
 {
-  long tx=typ(x),lx,i,l,tetpil;
+  long tx=typ(x), lx, i;
+  gpmem_t av, tetpil;
   GEN y,p1;
 
   switch(tx)
@@ -882,27 +896,27 @@ gabs(GEN x, long prec)
       y[2]=labsi((GEN)x[2]); return y;
 
     case t_COMPLEX:
-      l=avma; p1=gnorm(x);
+      av=avma; p1=gnorm(x);
       switch(typ(p1))
       {
         case t_INT:
           if (!carrecomplet(p1, &y)) break;
-          return gerepileupto(l, y);
+          return gerepileupto(av, y);
         case t_FRAC:
         case t_FRACN:
         {
           GEN a,b;
           if (!carrecomplet((GEN)p1[1], &a)) break;
           if (!carrecomplet((GEN)p1[2], &b)) break;
-          return gerepileupto(l, gdiv(a,b));
+          return gerepileupto(av, gdiv(a,b));
         }
       }
       tetpil=avma;
-      return gerepile(l,tetpil,gsqrt(p1,prec));
+      return gerepile(av,tetpil,gsqrt(p1,prec));
 
     case t_QUAD:
-      l=avma; p1=gmul(x, realun(prec)); tetpil=avma;
-      return gerepile(l,tetpil,gabs(p1,prec));
+      av=avma; p1=gmul(x, realun(prec)); tetpil=avma;
+      return gerepile(av,tetpil,gabs(p1,prec));
 
     case t_POL:
       lx=lgef(x); if (lx<=2) return gcopy(x);
@@ -1023,7 +1037,6 @@ void
 gaffsg(long s, GEN x)
 {
   long l,i,v;
-  GEN p;
 
   switch(typ(x))
   {
@@ -1043,11 +1056,13 @@ gaffsg(long s, GEN x)
       gaffsg(s,(GEN)x[1]); gaffsg(0,(GEN)x[2]); break;
 
     case t_PADIC:
+    {
+      GEN y;
       if (!s) { padicaff0(x); break; }
-      p = (GEN)x[2];
-      v = (is_bigint(p))? 0: svaluation(s,p[2], &i);
-      setvalp(x,v); modsiz(i,(GEN)x[3],(GEN)x[4]);
+      v = pvaluation(stoi(s), (GEN)x[2], &y);
+      setvalp(x,v); modiiz(y,(GEN)x[3],(GEN)x[4]);
       break;
+    }
 
     case t_QUAD:
       gaffsg(s,(GEN)x[2]); gaffsg(0,(GEN)x[3]); break;
@@ -1094,7 +1109,8 @@ gaffsg(long s, GEN x)
 void
 gaffect(GEN x, GEN y)
 {
-  long i,j,k,l,v,vy,lx,ly,tx,ty,av;
+  long i, j, k, l, v, vy, lx, ly, tx, ty;
+  gpmem_t av;
   GEN p1,num,den;
 
 
@@ -1138,10 +1154,10 @@ gaffect(GEN x, GEN y)
 
 	    case t_PADIC:
               if (!signe(x)) { padicaff0(y); break; }
-	      l=avma;
+	      av=avma;
 	      setvalp(y, pvaluation(x,(GEN)y[2],&p1));
 	      modiiz(p1,(GEN)y[3],(GEN)y[4]);
-	      avma=l; break;
+	      avma=av; break;
 
 	    case t_QUAD:
 	      gaffect(x,(GEN)y[2]); gaffsg(0,(GEN)y[3]); break;
@@ -1382,28 +1398,25 @@ gaffect(GEN x, GEN y)
   }
 
   if (is_const_t(ty)) {
-#define initial_value(ep) ((ep)+1)
-      if (tx == t_POL) {
-	  entree *var = varentries[ordvar[varn(x)]];
-	  /* Is a bound expression, thus should not loop: */
-	  if (var && var->value != (void*)initial_value(var)) {
-	      gaffect(geval(x),y);
-	      return;
-	  }	  
-      } else if (is_rfrac_t(tx)) {
-	  GEN num = (GEN)x[1], den = (GEN)x[2];
-	  entree *varnum = varentries[ordvar[varn(num)]];
-	  entree *varden = varentries[ordvar[varn(den)]];
-
-	  /* Are bound expressions, thus should not loop: */
-	  if (varnum && varnum->value != (void*)initial_value(varnum)
-	   && varden && varden->value != (void*)initial_value(varden)) {
-	      gaffect(geval(x),y);
-	      return;
-	  }
+    entree *varnum, *varden;
+    long vnum, vden;
+    GEN num, den;
+    if (tx == t_POL) {
+      vnum = varn(x); varnum = varentries[ordvar[vnum]];
+      if (varnum) {
+        x = geval(x); tx = typ(x);
+        if (tx != t_POL || varn(x) != vnum) { gaffect(x, y); return; }
+      }	  
+    } else if (is_rfrac_t(tx)) {
+      num = (GEN)x[1]; vnum = gvar(num); varnum = varentries[ordvar[vnum]];
+      den = (GEN)x[2]; vden = gvar(den); varden = varentries[ordvar[vden]];
+      if (varnum && varden) {
+        vnum = min(vnum, vden);
+        x = geval(x); tx = typ(x);
+        if (!is_rfrac_t(tx) || gvar(x) != vnum) { gaffect(x, y); return; }
       }
-#undef initial_value
-      err(operf,"",x,y);
+    }
+    err(operf,"",x,y);
   }
   
   switch(tx)
@@ -1528,7 +1541,7 @@ gaffect(GEN x, GEN y)
 GEN
 co8(GEN x, long prec)
 {
-  long av=avma,tetpil;
+  gpmem_t av=avma, tetpil;
   GEN p1, p = (GEN) x[1];
 
   p1 = subii(sqri((GEN)p[3]), shifti((GEN)p[2],2));
@@ -1551,7 +1564,8 @@ GEN
 cvtop(GEN x, GEN p, long l)
 {
   GEN p1,p2,p3;
-  long av,tetpil,n;
+  long n;
+  gpmem_t av, tetpil;
 
   if (typ(p)!=t_INT)
     err(talker,"not an integer modulus in cvtop or gcvtop");
@@ -1639,7 +1653,8 @@ gcvtop(GEN x, GEN p, long r)
 long
 gexpo(GEN x)
 {
-  long tx=typ(x),lx,e,i,y,av;
+  long tx=typ(x), lx, e, i, y;
+  gpmem_t av;
 
   switch(tx)
   {
@@ -1687,19 +1702,19 @@ normalize(GEN x)
   long i,j, lx = lg(x);
 
   if (typ(x)!=t_SER) err(typeer,"normalize");
-  if (lx==2) { setsigne(x,0); avma = (long) x; return x; }
+  if (lx==2) { setsigne(x,0); avma = (gpmem_t) x; return x; }
   if (! isexactzero((GEN)x[2])) { setsigne(x,1); return x; }
 
   for (i=3; i<lx; i++)
     if (! isexactzero((GEN)x[i]))
     {
-      long tetpil = avma;
+      gpmem_t tetpil = avma;
       GEN p1 = cgetg(lx-i+2,t_SER);
       p1[1] = evalsigne(1) | evalvalp(valp(x)+i-2) | evalvarn(varn(x));
       j=i; i=2; while (j<lx) p1[i++] = lcopy((GEN)x[j++]);
-      return gerepile((long) (x+lx),tetpil,p1);
+      return gerepile((gpmem_t) (x+lx),tetpil,p1);
     }
-  avma = (long) (x+lx); return zeroser(varn(x),lx-2);
+  avma = (gpmem_t) (x+lx); return zeroser(varn(x),lx-2);
 }
 #else
 
@@ -1872,7 +1887,8 @@ listconcat(GEN list1, GEN list2)
 GEN
 listsort(GEN list, long flag)
 {
-  long i,av=avma, c=list[1], lx = lgef(list)-1;
+  long i, c=list[1], lx = lgef(list)-1;
+  gpmem_t av=avma;
   GEN perm,vec,l;
 
   if (typ(list) != t_LIST) err(typeer,"listsort");
