@@ -1,4 +1,4 @@
-/* $Id: rootpol.c,v 1.44 2002/01/10 21:32:59 karim Exp $
+/* $Id: rootpol.c,v 1.58 2002/06/08 14:02:29 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -63,7 +63,7 @@ quickmulcc(GEN x, GEN y)
     }
     if (ty==t_COMPLEX)
     {
-      long av,tetpil;
+      gpmem_t av, tetpil;
       GEN p1,p2;
 
       z=cgetg(3,t_COMPLEX); av=avma;
@@ -95,7 +95,8 @@ static GEN
 mysquare(GEN p)
 {
   GEN s,aux1,aux2;
-  long i,j,n=degpol(p),nn,ltop,lbot;
+  long i, j, n=degpol(p), nn;
+  gpmem_t ltop, lbot;
 
   if (n==-1) return gcopy(p);
   nn=n<<1; s=cgetg(nn+3,t_POL);
@@ -140,7 +141,7 @@ karasquare(GEN p)
 {
   GEN p1,s0,s1,s2,aux;
   long n=degpol(p),n0,n1,i,var,nn0;
-  ulong ltop;
+  gpmem_t ltop;
 
   if (n<=KARASQUARE_LIMIT) return mysquare(p);
   ltop=avma;
@@ -172,7 +173,7 @@ cook_square(GEN p)
 {
   GEN p0,p1,p2,p3,q,aux0,aux1,r,aux,plus,moins;
   long n=degpol(p),n0,n3,i,j,var;
-  ulong ltop = avma;
+  gpmem_t ltop = avma;
 
   if (n<=COOK_SQUARE_LIMIT) return karasquare(p);
 
@@ -345,7 +346,7 @@ log2ir(GEN x)
   return 1.+ (double) expo(x)+log2( (double)(ulong) x[2]) - (double) BITS_IN_LONG;
 }
 
-static double
+double
 mylog2(GEN z)
 {
   double x,y;
@@ -572,7 +573,8 @@ at a distance eps */
 static double
 lower_bound(GEN p, long *k, double eps)
 {
-  long n=degpol(p),i,j,ltop=avma;
+  long n=degpol(p), i, j;
+  gpmem_t ltop=avma;
   GEN a,s,S,icd;
   double r,*rho;
 
@@ -604,11 +606,11 @@ lower_bound(GEN p, long *k, double eps)
 }
 
 /* returns the maximum of the modulus of p with a relative error tau */
-static GEN
+GEN
 max_modulus(GEN p, double tau)
 {
   GEN r,q,aux,gunr;
-  ulong av, ltop = avma;
+  gpmem_t av, ltop = avma;
   long i,k,n=degpol(p),nn,bitprec,M,e;
   double rho,eps, tau2 = (tau > 3.0)? 0.5: tau/6.;
 
@@ -640,12 +642,13 @@ max_modulus(GEN p, double tau)
     nn -= polvaluation(q, &q);
     set_karasquare_limit(gexpo(q));
     q = gerepileupto(av, graeffe(q));
-    tau2 *= 1.5; eps = -1/log(tau2);
+    tau2 *= 1.5; if (tau2 > 0.9) tau2 = 0.5;
+    eps = -1/log(tau2); /* > 0 */
     e = findpower(q);
   }
   if (!signe(r)) { avma = ltop; return realun(DEFAULTPREC); } 
-  aux = cgetr(DEFAULTPREC);
-  affir(r,aux); r = aux; setexpo(r, expo(r) - M);
+  r = itor(r, DEFAULTPREC);
+  setexpo(r, expo(r) - M);
   rho = rtodbl(r);
   /* rho = sum e_i 2^-i */
   avma = ltop;
@@ -657,7 +660,8 @@ static GEN
 modulus(GEN p, long k, double tau)
 {
   GEN q,gunr;
-  long i,kk=k,imax,n=degpol(p),nn,av,ltop=avma,bitprec,decprec,e;
+  long i, kk=k, imax, n=degpol(p), nn, bitprec, decprec, e;
+  gpmem_t av, ltop=avma;
   double tau2,r;
 
   tau2 = tau/6;
@@ -697,7 +701,8 @@ static GEN
 pre_modulus(GEN p, long k, double tau, GEN rmin, GEN rmax)
 {
   GEN R, q, aux;
-  long n=degpol(p),i,imax,imax2,bitprec,ltop=avma, av;
+  long n=degpol(p), i, imax, imax2, bitprec;
+  gpmem_t ltop=avma, av;
   double tau2, aux2;
 
   tau2=tau/6.;
@@ -732,7 +737,7 @@ pre_modulus(GEN p, long k, double tau, GEN rmin, GEN rmax)
 static GEN
 min_modulus(GEN p, double tau)
 {
-  long av=avma;
+  gpmem_t av=avma;
   GEN r;
 
   if (isexactzero((GEN)p[2])) return realzero(3);
@@ -746,7 +751,8 @@ static long
 dual_modulus(GEN p, GEN R, double tau, long l)
 {
   GEN q;
-  long i,imax,k,delta_k=0,n=degpol(p),nn,nnn,valuat,ltop=avma,bitprec,ll=l;
+  long i, imax, k, delta_k=0, n=degpol(p), nn, nnn, valuat, bitprec, ll=l;
+  gpmem_t ltop=avma;
   double logmax,aux,tau2;
 
   tau2=7.*tau/8.;
@@ -807,7 +813,7 @@ gmulbyi(GEN z)
 static void
 fft(GEN Omega, GEN p, GEN f, long Step, long l)
 {
-  ulong ltop;
+  gpmem_t ltop;
   long i,l1,l2,l3,rap=Lmax/l,rapi,Step4;
   GEN f1,f2,f3,f02,f13,g02,g13,ff;
 
@@ -861,20 +867,6 @@ fft(GEN Omega, GEN p, GEN f, long Step, long l)
   for (i=0; i<l; i++) f[i]=ff[i+1];
 }
 
-extern void mpsincos(GEN x, GEN *s, GEN *c);
-
-/* return exp(ix), x a t_REAL */
-static GEN
-exp_i(GEN x)
-{
-  GEN v;
-
-  if (!signe(x)) return realun(lg(x)); /* should not happen */
-  v = cgetg(3,t_COMPLEX);
-  mpsincos(x, (GEN*)(v+2), (GEN*)(v+1));
-  return v;
-}
-
 /* e(1/N) */
 static GEN
 RUgen(long N, long bitprec)
@@ -883,7 +875,7 @@ RUgen(long N, long bitprec)
   if (N == 2) return mpneg(realun(bitprec));
   if (N == 4) return gi;
   pi2 = gmul2n(mppi(bitprec/BITS_IN_LONG+3), 1);
-  return exp_i(gdivgs(pi2,N));
+  return exp_Ir(gdivgs(pi2,N));
 }
 
 /* N=2^k. returns a vector RU which contains exp(2*i*k*Pi/N), k=0..N-1 */
@@ -923,7 +915,7 @@ initRUgen(long N, long bitprec)
 }
 
 /* returns 1 if p has only real coefficients, 0 else */
-static long
+static int
 isreal(GEN p)
 {
   long n=degpol(p),i=0;
@@ -934,10 +926,11 @@ isreal(GEN p)
 
 static void
 parameters(GEN p, double *mu, double *gamma,
-           long polreal, double param, double param2)
+           int polreal, double param, double param2)
 {
   GEN q,pc,Omega,coef,RU,prim,aux,aux0,ggamma,gx,mygpi;
-  long ltop=avma,limite=stack_lim(ltop,1),n=degpol(p),bitprec,NN,K,i,j,ltop2;
+  long n=degpol(p), bitprec, NN, K, i, j;
+  gpmem_t ltop2, ltop=avma, limite=stack_lim(ltop, 1);
   double lx;
 
   bitprec=gexpo(p)+(long)param2+8;
@@ -945,7 +938,7 @@ parameters(GEN p, double *mu, double *gamma,
   K=NN/Lmax; if (K%2==1) K++; NN=Lmax*K;
   mygpi=mppi(bitprec/BITS_IN_LONG+3);
   aux = gdivgs(mygpi,NN/2); /* 2 Pi/NN */
-  prim = exp_i(aux);
+  prim = exp_Ir(aux);
   aux = gmulbyi(aux);
   RU = myrealun(bitprec);
 
@@ -1004,11 +997,11 @@ dft(GEN p, long k, long NN, long bitprec, GEN F, GEN H, long polreal)
 {
   GEN Omega,q,qd,pc,pdc,alpha,beta,gamma,RU,aux,U,W,mygpi,prim,prim2,*gptr[3];
   long n=degpol(p),i,j,K;
-  ulong ltop;
+  gpmem_t ltop;
 
   mygpi=mppi(bitprec/BITS_IN_LONG+3);
   aux = gdivgs(mygpi,NN/2); /* 2 Pi/NN */
-  prim = exp_i(aux);
+  prim = exp_Ir(aux);
   aux = gmulbyi(aux);
   prim2 = myrealun(bitprec);
   RU=cgetg(n+2,t_VEC); RU++;
@@ -1101,7 +1094,7 @@ static GEN
 refine_H(GEN F, GEN G, GEN HH, long bitprec, long shiftbitprec)
 {
   GEN H=HH,D,aux;
-  ulong ltop=avma, limite=stack_lim(ltop,1);
+  gpmem_t ltop=avma, limite=stack_lim(ltop, 1);
   long error=0,i,bitprec1,bitprec2;
 
   D=gsub(gun,gres(gmul(HH,G),F)); error=gexpo(D);
@@ -1135,8 +1128,9 @@ static long
 refine_F(GEN p, GEN *F, GEN *G, GEN H, long bitprec, double gamma)
 {
   GEN pp,FF,GG,r,HH,f0;
-  long error,i,bitprec1=0,bitprec2,ltop=avma,shiftbitprec;
-  long shiftbitprec2,n=degpol(p),enh,normF,normG,limite=stack_lim(ltop,1);
+  long error, i, bitprec1=0, bitprec2, shiftbitprec;
+  long shiftbitprec2, n=degpol(p), enh, normF, normG;
+  gpmem_t ltop=avma, limite=stack_lim(ltop, 1);
 
   FF=*F; HH=H;
   GG=poldivres(p,*F,&r);
@@ -1192,8 +1186,9 @@ split_fromU(GEN p, long k, double delta, long bitprec,
             GEN *F, GEN *G, double param, double param2)
 {
   GEN pp,FF,GG,H;
-  long n=degpol(p),NN,bitprec2,
-  ltop=avma,polreal=isreal(p);
+  long n=degpol(p),NN,bitprec2;
+  int polreal=isreal(p);
+  gpmem_t ltop;
   double mu,gamma;
 
   pp=gdiv(p,(GEN)p[2+n]);
@@ -1268,7 +1263,8 @@ extern GEN addshiftpol(GEN x, GEN y, long d);
 static GEN
 shiftpol(GEN p, GEN b)
 {
-  long av = avma,i, limit = stack_lim(av,1);
+  long i;
+  gpmem_t av = avma, limit = stack_lim(av, 1);
   GEN q = gzero;
 
   if (gcmp0(b)) return p;
@@ -1293,7 +1289,7 @@ conformal_pol(GEN p, GEN a, long bitprec)
 {
   GEN r,pui,num,aux, unr = myrealun(bitprec);
   long n=degpol(p), i;
-  ulong av, limit;
+  gpmem_t av, limit;
 
   aux = pui = cgetg(4,t_POL);
   pui[1] = evalsigne(1) | evalvarn(varn(p)) | evallgef(4);
@@ -1384,7 +1380,8 @@ static void
 conformal_mapping(GEN *radii, GEN ctr, GEN p, long k, long bitprec,
                   double aux, GEN *F,GEN *G)
 {
-  long bitprec2,n=degpol(p),decprec,i,ltop = avma, av;
+  long bitprec2, n=degpol(p), decprec, i;
+  gpmem_t ltop = avma, av;
   GEN q,FF,GG,a,R, *gptr[2];
   GEN rho,invrho;
   double delta,param,param2;
@@ -1399,7 +1396,7 @@ conformal_mapping(GEN *radii, GEN ctr, GEN p, long k, long bitprec,
   for (i=1; i<=n; i++)
     if (signe(radii[i])) /* updating array radii */
     {
-      long a = avma;
+      gpmem_t a = avma;
       GEN p1 = gsqr(radii[i]);
       /* 2(r^2 - 1) / (r^2 - 3(r-1)) */
       p1 = divrr(gmul2n((subrs(p1,1)),1),
@@ -1758,7 +1755,7 @@ a_posteriori_errors(GEN p, GEN roots_pol, long err)
   setexpo(sigma, err + (long)log2((double)n) + 1);
   overn=dbltor(1./n);
   shatzle=gdiv(gpui(sigma,overn,0),
-	       gsub(gpui(gsub(gun,sigma),overn,0),
+	       gsub(gpui(gsub(realun(DEFAULTPREC),sigma),overn,0),
 		    gpui(sigma,overn,0)));
   shatzle=gmul2n(shatzle,1); e_max=-pariINFINITY;
   for (i=1; i<=n; i++)
@@ -1787,13 +1784,14 @@ append_root(GEN roots_pol, GEN a)
 static GEN
 split_complete(GEN p, long bitprec, GEN roots_pol)
 {
-  long n=degpol(p),decprec,ltop;
+  long n=degpol(p), decprec;
+  gpmem_t ltop;
   GEN p1,F,G,a,b,m1,m2,m;
 
   if (n==1)
   {
     a=gneg_i(gdiv((GEN)p[2],(GEN)p[3]));
-    append_root(roots_pol,a); return p;
+    (void)append_root(roots_pol,a); return p;
   }
   ltop = avma;
   if (n==2)
@@ -1820,18 +1818,21 @@ split_complete(GEN p, long bitprec, GEN roots_pol)
 GEN
 cauchy_bound(GEN p)
 {
-  long i,n=degpol(p);
-  GEN x=gzero,y,lc;
+  long i, n = degpol(p), prec = DEFAULTPREC;
+  GEN lc, y, q = gmul(p, realun(prec)), x = gzero;
 
-  lc=gabs((GEN)p[n+2],DEFAULTPREC); /* leading coefficient */
-  lc=gdiv(dbltor(1.),lc);
+  if (n <= 0) err(constpoler,"cauchy_bound");
+
+  lc = gabs((GEN)q[n+2],prec); /* leading coefficient */
+  lc = ginv(lc);
   for (i=0; i<n; i++)
   {
-    y=gmul(gabs((GEN) p[i+2],DEFAULTPREC),lc);
-    y=gpui(y,dbltor(1./(n-i)),DEFAULTPREC);
-    if (gcmp(y,x) > 0) x=y;
+    y = (GEN)q[i+2]; if (gcmp0(y)) continue;
+    y = gmul(gabs(y,prec), lc);
+    y = divrs(mplog(y), n-i);
+    if (gcmp(y,x) > 0) x = y;
   }
-  return x;
+  return gexp(x, prec);
 }
 
 static GEN
@@ -1904,7 +1905,7 @@ all_roots(GEN p, long bitprec)
 {
   GEN lc,pd,q,roots_pol,m;
   long bitprec0, bitprec2,n=degpol(p),i,e,h;
-  ulong av;
+  gpmem_t av;
 
   pd = poldeflate(p, &h); lc = leading_term(pd);
   e = 2*gexpo(cauchy_bound(pd)); if (e<0) e=0;
@@ -2054,7 +2055,7 @@ isrealappr(GEN x, long e)
 static int
 isconj(GEN x, GEN y, long e)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long i= (gexpo( gsub((GEN)x[1],(GEN)y[1]) ) < e
         && gexpo( gadd((GEN)x[2],(GEN)y[2]) ) < e);
   avma = av; return i;
@@ -2066,7 +2067,7 @@ isconj(GEN x, GEN y, long e)
 GEN
 roots(GEN p, long l)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   long n,i,k,s,t,e;
   GEN c,L,p1,res,rea,com;
 
