@@ -1,4 +1,4 @@
-/* $Id: trans1.c,v 1.46 2001/12/15 14:14:59 karim Exp $
+/* $Id: trans1.c,v 1.68 2002/06/08 14:48:35 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -41,7 +41,8 @@ void
 constpi(long prec)
 {
   GEN p1,p2,p3,tmppi;
-  long l,n,n1,av1,av2;
+  long l, n, n1;
+  gpmem_t av1, av2;
   double alpha;
 
 #define k1  545140134
@@ -56,10 +57,10 @@ constpi(long prec)
 
   prec++;
 
-  n=(long)(1+(prec-2)/alpha2);
-  n1=6*n-1; p1=cgetr(prec);
-  p2=addsi(k2,mulss(n,k1));
-  affir(p2,p1);
+  n = (long)(1 + (prec-2)/alpha2);
+  n1 = 6*n - 1;
+  p2 = addsi(k2, mulss(n,k1));
+  p1 = itor(p2, prec);
 
   /* initialisation longueur mantisse */
   if (prec>=4) l=4; else l=prec;
@@ -96,6 +97,27 @@ mppi(long prec)
   constpi(prec); affrr(gpi,x); return x;
 }
 
+/* Pi * 2^n */
+GEN
+Pi2n(long n, long prec)
+{
+  GEN x = mppi(prec); setexpo(x, 1+n);
+  return x;
+}
+
+/* I * Pi * 2^n */
+GEN
+PiI2n(long n, long prec)
+{
+  GEN z = cgetg(3,t_COMPLEX);
+  z[1] = zero;
+  z[2] = (long)Pi2n(n, prec); return z;
+}
+
+/* 2I * Pi */
+GEN
+PiI2(long prec) { return PiI2n(1, prec); }
+
 /********************************************************************/
 /**                                                                **/
 /**                      FONCTION EULER                            **/
@@ -106,7 +128,8 @@ void
 consteuler(long prec)
 {
   GEN u,v,a,b,tmpeuler;
-  long l,n,k,x,av1,av2;
+  long l, n, k, x;
+  gpmem_t av1, av2;
 
   if (geuler && lg(geuler) >= prec) return;
 
@@ -116,9 +139,9 @@ consteuler(long prec)
   prec++;
 
   l = prec+1; x = (long) (1 + (bit_accuracy(l) >> 2) * LOG2);
-  a=cgetr(l); affsr(x,a); u=mplog(a); setsigne(u,-1); affrr(u,a);
-  b=realun(l);
-  v=realun(l);
+  a = stor(x,l); u=mplog(a); setsigne(u,-1); affrr(u,a);
+  b = realun(l);
+  v = realun(l);
   n=(long)(1+3.591*x); /* z=3.591: z*[ ln(z)-1 ]=1 */
   av2 = avma;
   if (x < SQRTVERYBIGINT)
@@ -162,7 +185,8 @@ GEN
 transc(GEN (*f)(GEN,long), GEN x, long prec)
 {
   GEN p1,p2,y;
-  long lx,i,av,tetpil;
+  long lx, i;
+  gpmem_t av, tetpil;
 
   av=avma;
   switch(typ(x))
@@ -286,7 +310,7 @@ _muli(void *data /* ignored */, GEN x, GEN y) {
 static GEN
 puissii(GEN a, GEN n, long s)
 {
-  ulong av;
+  gpmem_t av;
   GEN y;
 
   if (!signe(a)) return gzero; /* a==0 */
@@ -305,29 +329,28 @@ puissii(GEN a, GEN n, long s)
   setsigne(y,s); return gerepileuptoint(av,y);
 }
 
-struct muldata {
+typedef struct {
   long prec, a;
   GEN (*sqr)(GEN);
   GEN (*mulsg)(long,GEN);
-  GEN unr;
-};
+} sr_muldata;
 
 static GEN
-_rpowsi_mul(void *data, GEN x, GEN y/* base; ignored */)
+_rpowsi_mul(void *data, GEN x, GEN y/*unused*/)
 {
-  struct muldata *D = (struct muldata *)data;
+  sr_muldata *D = (sr_muldata *)data;
   return D->mulsg(D->a, x);
 }
 
 static GEN
 _rpowsi_sqr(void *data, GEN x)
 {
-  struct muldata *D = (struct muldata *)data;
-  if (lgefint(x) >= D->prec && typ(x) == t_INT)
+  sr_muldata *D = (sr_muldata *)data;
+  if (typ(x) == t_INT && lgefint(x) >= D->prec)
   { /* switch to t_REAL */
     D->sqr   = &gsqr;
     D->mulsg = &mulsr;
-    affir(x, D->unr); x = D->unr;
+    x = itor(x, D->prec);
   }
   return D->sqr(x);
 }
@@ -337,28 +360,29 @@ _rpowsi_sqr(void *data, GEN x)
 GEN
 rpowsi(ulong a, GEN n, long prec)
 {
-  ulong av = avma;
-  GEN y, unr = realun(prec);
-  struct muldata D;
+  gpmem_t av;
+  GEN y;
+  sr_muldata D;
 
-  if (a == 1) return unr;
-  if (a == 2) { setexpo(unr, itos(n)); return unr; }
-  if (is_pm1(n)) { affsr(a, unr); return unr; }
+  if (a == 1) return realun(prec);
+  if (a == 2) return real2n(itos(n), prec);
+  if (is_pm1(n)) return stor(a, prec);
+  av = avma;
   D.sqr   = &sqri; 
   D.mulsg = &mulsi;
   D.prec = prec;
-  D.unr = unr;
   D.a = a;
   y = leftright_pow(stoi(a), n, (void*)&D, &_rpowsi_sqr, &_rpowsi_mul);
-  if (typ(y) == t_INT) { affir(y, unr); y = unr ; }
-  return gerepileuptoleaf(av,y);
+  if (typ(y) == t_INT) y = itor(y, prec);
+  return gerepileuptoleaf(av, y);
 }
 
 GEN
 gpowgs(GEN x, long n)
 {
-  long lim,av,m,tx;
-  static long gn[3] = {evaltyp(t_INT)|m_evallg(3), 0, 0};
+  long m, tx;
+  gpmem_t lim, av;
+  static long gn[3] = {evaltyp(t_INT)|_evallg(3), 0, 0};
   GEN y;
 
   if (n == 0) return puiss0(x);
@@ -442,7 +466,8 @@ gpowgs(GEN x, long n)
 GEN
 pow_monome(GEN x, GEN nn)
 {
-  long n,m,i,j,dd,tetpil, av = avma;
+  long n, m, i, j, dd;
+  gpmem_t tetpil, av = avma;
   GEN p1,y;
   if (is_bigint(nn)) err(talker,"power overflow in pow_monome");
   n = itos(nn); m = labs(n);
@@ -526,7 +551,7 @@ powgi(GEN x, GEN n)
       else
       {
         mod = mulii(mod, gpowgs(p,v));
-        mod = gerepileuptoint((long)y, mod);
+        mod = gerepileuptoint((gpmem_t)y, mod);
       }
       y[1] = evalprecp(precp(x)+v) | evalvalp(e);
       icopyifstack(p, y[2]);
@@ -540,7 +565,7 @@ powgi(GEN x, GEN n)
       if (ismonome(x)) return pow_monome(x,n);
     default:
     {
-      ulong av = avma;
+      gpmem_t av = avma;
       y = leftright_pow(x, n, NULL, &_sqr, &_mul);
       if (sn < 0) y = ginv(y);
       return av==avma? gcopy(y): gerepileupto(av,y);
@@ -552,42 +577,47 @@ powgi(GEN x, GEN n)
 static GEN
 ser_pui(GEN x, GEN n, long prec)
 {
-  long av,tetpil,lx,i,j;
-  GEN p1,p2,y,alp;
+  gpmem_t av, tetpil;
+  GEN y, p1, p2, lead;
 
-  if (gvar9(n) > varn(x))
+  if (gvar9(n) <= varn(x)) return gexp(gmul(n, glog(x,prec)), prec);
+  lead = (GEN)x[2];
+  if (gcmp1(lead))
   {
-    GEN lead = (GEN)x[2];
-    if (gcmp1(lead))
+    GEN X, Y, alp;
+    long lx, mi, i, j, d;
+
+    av = avma; alp = gclone(gadd(n,gun)); avma = av;
+    lx = lg(x);
+    y = cgetg(lx,t_SER);
+    y[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
+    X = x+2;
+    Y = y+2;
+
+    d = mi = lx-3; while (mi>=1 && gcmp0((GEN)X[mi])) mi--;
+    Y[0] = un;
+    for (i=1; i<=d; i++)
     {
-      av=avma; alp = gclone(gadd(n,gun)); avma=av;
-      y=cgetg(lx=lg(x),t_SER);
-      y[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
-      y[2] = un;
-      for (i=3; i<lx; i++)
+      av = avma; p1 = gzero;
+      for (j=1; j<=min(i,mi); j++)
       {
-	av=avma; p1=gzero;
-	for (j=1; j<i-1; j++)
-	{
-	  p2 = gsubgs(gmulgs(alp,j),i-2);
-	  p1 = gadd(p1,gmul(gmul(p2,(GEN)x[j+2]),(GEN)y[i-j]));
-	}
-	tetpil=avma; y[i]=lpile(av,tetpil,gdivgs(p1,i-2));
+        p2 = gsubgs(gmulgs(alp,j),i);
+        p1 = gadd(p1, gmul(gmul(p2,(GEN)X[j]),(GEN)Y[i-j]));
       }
-      gunclone(alp); return y;
+      tetpil = avma; Y[i] = lpile(av,tetpil, gdivgs(p1,i));
     }
-    av=avma; p1=gdiv(x,lead); p1[2] = un; /* in case it's inexact */
-    p1=gpow(p1,n,prec); p2=gpow(lead,n,prec);
-    tetpil=avma; return gerepile(av,tetpil,gmul(p1,p2));
+    gunclone(alp); return y;
   }
-  av=avma; y=gmul(n,glog(x,prec)); tetpil=avma;
-  return gerepile(av,tetpil,gexp(y,prec));
+  p1 = gdiv(x,lead); p1[2] = un; /* in case it's inexact */
+  p1 = gpow(p1,  n, prec);
+  p2 = gpow(lead,n, prec); return gmul(p1, p2);
 }
 
 GEN
 gpow(GEN x, GEN n, long prec)
 {
-  long av,tetpil,i,lx,tx;
+  long i, lx, tx;
+  gpmem_t av, tetpil;
   GEN y;
 
   if (typ(n)==t_INT) return powgi(x,n);
@@ -598,12 +628,18 @@ gpow(GEN x, GEN n, long prec)
     for (i=1; i<lx; i++) y[i]=lpui((GEN)x[i],n,prec);
     return y;
   }
+  if (tx==t_POL)
+  {
+    av=avma; y=tayl(x,gvar(x),precdl); tetpil=avma;
+    return gerepile(av,tetpil,gpow(y,n,prec));
+  }
   if (tx==t_SER)
   {
     if (valp(x))
       err(talker,"not an integer exponent for non invertible series in gpow");
     if (lg(x) == 2) return gcopy(x); /* O(1) */
-    return ser_pui(x,n,prec);
+    av = avma;
+    return gerepileupto(av, ser_pui(x, n, prec));
   }
   av=avma;
   if (gcmp0(x))
@@ -624,7 +660,7 @@ gpow(GEN x, GEN n, long prec)
   if (tx==t_INTMOD && typ(n)==t_FRAC)
   {
     GEN p1;
-    if (!isprime((GEN)x[1])) err(talker,"modulus must be prime in gpow");
+    if (!BSW_psp((GEN)x[1])) err(talker,"modulus must be prime in gpow");
     y=cgetg(3,tx); copyifstack(x[1],y[1]);
     av=avma;
     p1=mpsqrtnmod((GEN)x[2],(GEN)n[2],(GEN)x[1],NULL);
@@ -647,7 +683,7 @@ gpow(GEN x, GEN n, long prec)
 GEN
 mpsqrt(GEN x)
 {
-  ulong av, av0;
+  gpmem_t av, av0;
   long l,l0,l1,l2,s,eps,n,i,ex;
   double beta;
   GEN y,p1,p2,p3;
@@ -708,7 +744,7 @@ sqrt_2adic(GEN x, long pp)
 {
   GEN z = mod16(x)==1? gun: stoi(3);
   long zp;
-  ulong av, lim;
+  gpmem_t av, lim;
 
   if (pp == 4) return z;
   zp = 3; /* number of correct bits in z (compared to sqrt(x)) */
@@ -740,7 +776,7 @@ sqrt_padic(GEN x, GEN modx, long pp, GEN p)
 {
   GEN mod, z = mpsqrtmod(x, p);
   long zp = 1;
-  ulong av, lim;
+  gpmem_t av, lim;
 
   if (!z) err(sqrter5);
   if (pp <= zp) return z; 
@@ -772,7 +808,7 @@ padic_sqrt(GEN x)
 {
   long pp, e = valp(x);
   GEN z,y,mod, p = (GEN)x[2];
-  ulong av;
+  gpmem_t av;
 
   if (gcmp0(x)) return padiczero(p, (e+1) >> 1);
   if (e & 1) err(sqrter6);
@@ -820,7 +856,8 @@ padic_sqrt(GEN x)
 GEN
 gsqrt(GEN x, long prec)
 {
-  long av,tetpil,e;
+  long e;
+  gpmem_t av, tetpil;
   GEN y,p1,p2;
 
   switch(typ(x))
@@ -883,17 +920,17 @@ gsqrt(GEN x, long prec)
       return padic_sqrt(x);
 
     case t_SER:
-      e=valp(x);
+      e = valp(x);
       if (gcmp0(x)) return zeroser(varn(x), (e+1)>>1);
       if (e & 1) err(sqrter6);
       av = avma;
-      /* trick: ser_pui assumes valp(x) = 0 */
-      y = ser_pui(x,ghalf,prec);
+      y = dummycopy(x); setvalp(y, 0);
+      y = ser_pui(y, ghalf, prec);
       if (typ(y) == t_SER) /* generic case */
         y[1] = evalsigne(1) | evalvalp(e>>1) | evalvarn(varn(x));
       else /* e.g coeffs are POLMODs */
-        y = gerepileupto(av, gmul(y, gpowgs(polx[varn(x)], e>>1)));
-      return y;
+        y = gmul(y, gpowgs(polx[varn(x)], e>>1));
+      return gerepileupto(av, y);
   }
   return transc(gsqrt,x,prec);
 }
@@ -901,7 +938,8 @@ gsqrt(GEN x, long prec)
 void
 gsqrtz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"gsqrtz");
   gaffect(gsqrt(x,prec),y); avma=av;
@@ -914,26 +952,19 @@ gsqrtz(GEN x, GEN y)
 GEN
 rootsof1complex(GEN n, long prec)
 {
-  ulong ltop=avma;
-  GEN z,a;
+  gpmem_t ltop = avma;
+  GEN z;
   if (is_pm1(n)) return realun(prec);
-  if (lgefint(n)==3 && n[2]==2)
-  {
-    a=realun(prec);
-    setsigne(a,-1);
-    return a;
-  }
-  a=mppi(prec); setexpo(a,2); /* a=2*pi */
-  a=divri(a,n);
-  z = cgetg(3,t_COMPLEX);
-  gsincos(a,(GEN*)(z+2),(GEN*)(z+1),prec);
+  if (lgefint(n)==3 && n[2]==2) return stor(-1, prec);
+
+  z = exp_Ir( divri(Pi2n(1, prec), n) ); /* exp(2Ipi/n) */
   return gerepileupto(ltop,z);  
 }
 /*Only the O() of y is used*/
 GEN
 rootsof1padic(GEN n, GEN y)
 {
-  ulong ltop=avma;
+  gpmem_t ltop=avma;
   GEN z,r;
   (void)mpsqrtnmod(gun,n,(GEN)y[2],&z);
   if (z==gzero){avma=ltop;return gzero;}/*should not happen*/
@@ -946,9 +977,10 @@ rootsof1padic(GEN n, GEN y)
 }
 static GEN paexp(GEN x);
 /*compute the p^e th root of x p-adic*/ 
-GEN padic_sqrtn_ram(GEN x, long e)
+GEN
+padic_sqrtn_ram(GEN x, long e)
 {
-  ulong ltop=avma;
+  gpmem_t ltop=avma;
   GEN n,a;
   GEN p=(GEN)x[2];
   long v=0;
@@ -975,9 +1007,10 @@ GEN padic_sqrtn_ram(GEN x, long e)
   return a;
 }
 /*compute the nth root of x p-adic p prime with n*/ 
-GEN padic_sqrtn_unram(GEN x, GEN n, GEN *zetan)
+GEN
+padic_sqrtn_unram(GEN x, GEN n, GEN *zetan)
 {
-  ulong ltop=avma,tetpil;
+  gpmem_t ltop=avma, tetpil;
   GEN a,r;
   GEN p=(GEN)x[2];
   long v=0;
@@ -1015,9 +1048,11 @@ GEN padic_sqrtn_unram(GEN x, GEN n, GEN *zetan)
   else
     return gerepile(ltop,tetpil,r);
 }
-GEN padic_sqrtn(GEN x, GEN n, GEN *zetan)
+
+GEN
+padic_sqrtn(GEN x, GEN n, GEN *zetan)
 {
-  ulong ltop=avma,tetpil;
+  gpmem_t ltop=avma, tetpil;
   GEN p=(GEN)x[2];
   GEN q;
   long e;
@@ -1073,7 +1108,8 @@ GEN padic_sqrtn(GEN x, GEN n, GEN *zetan)
 GEN
 gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
 {
-  long av,tetpil,i,lx,tx;
+  long i, lx, tx;
+  gpmem_t av, tetpil;
   long e,m;
   GEN y,z;
   if (zetan) *zetan=gzero;
@@ -1102,19 +1138,19 @@ gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
   case t_SER:   
     e=valp(x);m=itos(n);
     if (gcmp0(x)) return zeroser(varn(x), (e+m-1)/m);
-    if (e % m) err(talker,"incorrect valuation in gsqrt");
+    if (e % m) err(talker,"incorrect valuation in gsqrtn");
     av = avma;
-    /* trick: ser_pui assumes valp(x) = 0 */
-    y = ser_pui(x,ginv(n),prec);
+    y = dummycopy(x); setvalp(y, 0);
+    y = ser_pui(y, ginv(n), prec);
     if (typ(y) == t_SER) /* generic case */
       y[1] = evalsigne(1) | evalvalp(e/m) | evalvarn(varn(x));
     else /* e.g coeffs are POLMODs */
-      y = gerepileupto(av, gmul(y, gpowgs(polx[varn(x)], e/m)));
-    return y;
+      y = gmul(y, gpowgs(polx[varn(x)], e/m));
+    return gerepileupto(av, y);
   case t_INTMOD:
     z=gzero;
     /*This is not great, but else it will generate too much trouble*/
-    if (!isprime((GEN)x[1])) err(talker,"modulus must be prime in gsqrtn");
+    if (!BSW_psp((GEN)x[1])) err(talker,"modulus must be prime in gsqrtn");
     if (zetan) 
     {
       z=cgetg(3,tx); copyifstack(x[1],z[1]);
@@ -1167,7 +1203,8 @@ gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
 GEN
 mpexp1(GEN x)
 {
-  long l,l1,l2,i,n,m,ex,s,av,av2, sx = signe(x);
+  long l, l1, l2, i, n, m, ex, s, sx = signe(x);
+  gpmem_t av, av2;
   double a,b,alpha,beta, gama = 2.0 /* optimized for SUN3 */;
                                     /* KB: 3.0 is better for UltraSparc */
   GEN y,p1,p2,p3,p4,unr;
@@ -1234,20 +1271,22 @@ mpexp1(GEN x)
 GEN
 mpexp(GEN x)
 {
-  long av, sx = signe(x);
+  long sx = signe(x);
+  gpmem_t av;
   GEN y;
 
   if (!sx) return addsr(1,x);
   if (sx<0) setsigne(x,1);
   av = avma; y = addsr(1, mpexp1(x));
-  if (sx<0) { y = divsr(1,y); setsigne(x,-1); }
+  if (sx<0) { y = ginv(y); setsigne(x,-1); }
   return gerepileupto(av,y);
 }
 
 static GEN
 paexp(GEN x)
 {
-  long k,av, e = valp(x), pp = precp(x), n = e + pp;
+  long k, e = valp(x), pp = precp(x), n = e + pp;
+  gpmem_t av;
   GEN y,r,p1;
 
   if (gcmp0(x)) return gaddgs(x,1);
@@ -1273,7 +1312,7 @@ static GEN
 cxexp(GEN x, long prec)
 {
   GEN r,p1,p2, y = cgetg(3,t_COMPLEX);
-  ulong av = avma, tetpil;
+  gpmem_t av = avma, tetpil;
   r=gexp((GEN)x[1],prec);
   gsincos((GEN)x[2],&p2,&p1,prec);
   tetpil = avma;
@@ -1286,16 +1325,19 @@ cxexp(GEN x, long prec)
 static GEN
 serexp(GEN x, long prec)
 {
-  ulong av;
-  long i,j,lx,ly,ex;
+  gpmem_t av;
+  long i,j,lx,ly,ex,mi;
   GEN p1,y,xd,yd;
  
-  if (gcmp0(x)) return gaddsg(1,x);
-  lx = lg(x); ex = valp(x);
+  ex = valp(x);
   if (ex < 0) err(negexper,"gexp");
+  if (gcmp0(x)) return gaddsg(1,x);
+  lx = lg(x);
   if (ex)
   {
     ly = lx+ex; y = cgetg(ly,t_SER);
+    mi = lx-1; while (mi>=3 && gcmp0((GEN)x[mi])) mi--;
+    mi += ex-2;
     y[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
     /* zd[i] = coeff of X^i in z */
     xd = x+2-ex; yd = y+2; ly -= 2;
@@ -1304,7 +1346,7 @@ serexp(GEN x, long prec)
     for (   ; i<ly; i++)
     {
       av = avma; p1 = gzero;
-      for (j=ex; j<=i; j++)
+      for (j=ex; j<=min(i,mi); j++)
         p1 = gadd(p1, gmulgs(gmul((GEN)xd[j],(GEN)yd[i-j]),j));
       yd[i] = lpileupto(av, gdivgs(p1,i));
     }
@@ -1335,7 +1377,8 @@ gexp(GEN x, long prec)
 void
 gexpz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"gexpz");
   gaffect(gexp(x,prec),y); avma=av;
@@ -1346,63 +1389,112 @@ gexpz(GEN x, GEN y)
 /**                      FONCTION LOGARITHME                       **/
 /**                                                                **/
 /********************************************************************/
+/* 2 * atanh(1/3) */
+GEN
+constlog2(long prec)
+{
+  static GEN glog2 = NULL;
+  const long _3 = 3, _9 = _3*_3;
+  gpmem_t av0, av;
+  long k, l, G;
+  GEN s, u, S, U, tmplog2;
+
+  if (glog2 && lg(glog2) >= prec) return glog2;
+
+  tmplog2 = newbloc(prec);
+  *tmplog2 = evaltyp(t_REAL) | evallg(prec);
+  av0 = avma;
+  l = prec+1; G = bit_accuracy(l+1);
+
+  s = S = divrs(realun(l), _3);
+  u = U = mpcopy(s);
+  av = avma;
+  for (k = 3; ; k += 2)
+  {
+    u = divrs(u, _9);
+    if (bit_accuracy(l) - expo(u) > G) {
+      l--; if (l <= 2) break;
+      setlg(U,l);
+      affrr(s,S); s = S;
+      affrr(u,U); u = U; avma = av;
+    }
+    s = addrr(s, divrs(u,k));
+  }
+  setexpo(s, -1); affrr(s, tmplog2);
+  gunclone(glog2); glog2 = tmplog2;
+  avma = av0; return glog2;
+} 
+
+GEN
+mplog2(long prec)
+{
+  GEN x = cgetr(prec);
+  affrr(constlog2(prec), x); return x;
+} 
 
 GEN
 mplog(GEN x)
 {
-  ulong av, ltop;
-  long l,l1,l2,m,m1,n,i,ex,s,sgn;
+  gpmem_t ltop, av;
+  long EX,l,l1,l2,m,n,k,ex,s;
   double alpha,a,b;
-  GEN y,p1,p2,p3,p4,p5,unr;
+  GEN z,p1,y,y2,p4,p5,unr;
 
   if (typ(x)!=t_REAL) err(typeer,"mplog");
   if (signe(x)<=0) err(talker,"non positive argument in mplog");
-  l=lg(x); sgn = cmpsr(1,x); if (!sgn) return realzero(l);
-  y=cgetr(l); ltop=avma;
 
-  l2 = l+1; p2=p1=cgetr(l2); affrr(x,p1);
-  av=avma;
-  if (sgn > 0) p2 = divsr(1,p1); /* x<1 change x to 1/x */
-  for (m1=1; expo(p2)>0; m1++) p2 = mpsqrt(p2);
-  if (m1 > 1 || sgn > 0) { affrr(p2,p1); avma=av; }
+  av = avma;
+  l = lg(x); EX = expo(x);
+  z = cgetr(l); ltop = avma;
 
+  l2 = l+1; y=p1=cgetr(l2); affrr(x,p1);
+  setexpo(p1, 0);
+  if (gcmp1(p1)) {
+    if (!EX) { avma = av; return realzero(l); }
+    affrr(mulsr(EX, mplog2(l)), z);
+    avma = ltop; return z;
+  }
+  /* 1 < p1 < 2 */
+  av = avma; l -= 2;
   alpha = 1+p1[2]/C31; if (!alpha) alpha = 0.00000001;
-  l -= 2; alpha = -log(alpha);
-  a = alpha/LOG2;
-  b = sqrt((BITS_IN_LONG/2)*l/3.0);
-  if (a<=b)
+  a = -log(alpha)/LOG2;
+  b = sqrt(BITS_IN_HALFULONG*l/3.0);
+  if (a <= b)
   {
-    n = (long)(1+sqrt((BITS_IN_LONG/2)*3.0*l));
-    m = (long)(1+b-a);
+    n = 1 + (long)sqrt(BITS_IN_HALFULONG*3.0*l);
+    m = 1 + (long)(b-a);
     l2 += m>>TWOPOTBITS_IN_LONG;
     p4 = cgetr(l2); affrr(p1,p4);
     p1 = p4; av = avma;
-    for (i=1; i<=m; i++) p1 = mpsqrt(p1);
+    for (k=1; k<=m; k++) p1 = mpsqrt(p1);
     affrr(p1,p4); avma = av;
   }
   else
   {
-    double beta = (BITS_IN_LONG/2)*l*LOG2;
-    n = (long)(1+beta/alpha);
+    n = 1 + (long)(BITS_IN_HALFULONG*l / a);
     m = 0; p4 = p1;
   }
   unr = realun(l2);
-  p2 = cgetr(l2);
-  p3 = cgetr(l2); av = avma;
+  y  = cgetr(l2);
+  y2 = cgetr(l2); av = avma;
+
+  /* want to compute log(X), X ~ 1  (X = p4) */
+  /* y = (X-1)/(X+1).  log(X) = log(1+y) - log(1-y) = 2 \sum_{k odd} y^k / * k */
 
   /* affrr needed here instead of setlg since prec may DECREASE */
   p1 = cgetr(l2); affrr(subrr(p4,unr), p1);
 
   p5 = addrr(p4,unr); setlg(p5,l2);
-  affrr(divrr(p1,p5), p2);
-  affrr(mulrr(p2,p2), p3);
-  affrr(divrs(unr,2*n+1), p4); setlg(p4,4); avma = av;
+  affrr(divrr(p1,p5), y); /* = (X-1) / (X+1) ~ 0 */
+  affrr(gsqr(y), y2); /* = y^2 */
+  k = 2*n + 1;
+  affrr(divrs(unr,k), p4); setlg(p4,4); avma = av;
 
-  s = 0; ex = expo(p3); l1 = 4;
-  for (i=n; i>=1; i--)
-  {
-    setlg(p3, l1); p5 = mulrr(p4,p3);
-    setlg(unr,l1); p1 = divrs(unr,2*i-1);
+  s = 0; ex = expo(y2); l1 = 4;
+  for (k -= 2; k > 0; k -= 2)
+  { /* compute sum_i=0^n  y^2i / (2i + 1), k = 2i+1 */
+    setlg(y2, l1); p5 = mulrr(p4,y2);
+    setlg(unr,l1); p1 = divrs(unr, k);
     s -= ex;
     l1 += s>>TWOPOTBITS_IN_LONG; if (l1>l2) l1=l2;
     s &= (BITS_IN_LONG-1);
@@ -1410,17 +1502,19 @@ mplog(GEN x)
     setlg(p4, l1);
     setlg(p5, l1); affrr(addrr(p1,p5), p4); avma=av;
   }
-  setlg(p4, l2); affrr(mulrr(p2,p4), y);
-  setexpo(y, expo(y)+m+m1);
-  if (sgn > 0) setsigne(y, -1);
-  avma = ltop; return y;
+  setlg(p4, l2);
+  y = mulrr(y,p4); /* = log(X)/2 */
+  setexpo(y, expo(y)+m+1);
+  if (EX) y = addrr(y, mulsr(EX, mplog2(l2)));
+  affrr(y, z); avma = ltop; return z;
 }
 
 GEN
 teich(GEN x)
 {
   GEN p,q,y,z,aux,p1;
-  long av,n,k;
+  long n, k;
+  gpmem_t av;
 
   if (typ(x)!=t_PADIC) err(talker,"not a p-adic argument in teichmuller");
   if (!signe(x[4])) return gcopy(x);
@@ -1477,7 +1571,7 @@ palogaux(GEN x)
 GEN
 palog(GEN x)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN y, p = (GEN)x[2];
 
   if (!signe(x[4])) err(talker,"zero argument in palog");
@@ -1512,7 +1606,7 @@ log0(GEN x, long flag,long prec)
 GEN
 glog(GEN x, long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN y,p1,p2;
 
   switch(typ(x))
@@ -1533,7 +1627,8 @@ glog(GEN x, long prec)
       return palog(x);
 
     case t_SER:
-      av=avma; if (valp(x)) err(negexper,"glog");
+      av=avma;
+      if (valp(x) || gcmp0(x)) err(talker,"log is not analytic at 0");
       p1=gdiv(derivser(x),x);
       tetpil=avma; p1=integ(p1,varn(x));
       if (gcmp1((GEN)x[2])) return gerepile(av,tetpil,p1);
@@ -1550,7 +1645,8 @@ glog(GEN x, long prec)
 void
 glogz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"glogz");
   gaffect(glog(x,prec),y); avma=av;
@@ -1558,131 +1654,128 @@ glogz(GEN x, GEN y)
 
 /********************************************************************/
 /**                                                                **/
-/**                      FONCTION SINCOS-1                         **/
+/**                        SINE, COSINE                            **/
 /**                                                                **/
 /********************************************************************/
 
+/* Reduce x0 mod Pi/2 to x in [-Pi/4, Pi/4]. Return cos(x)-1 */
 static GEN
-mpsc1(GEN x, long *ptmod8)
+mpsc1(GEN x0, long *ptmod8)
 {
-  const long mmax = 23169;
+  const long mmax = 23169; /* largest m such that (2m+2)(2m+1) < 2^31 */
  /* on a 64-bit machine with true 128 bit/64 bit division, one could
-  * take mmax=1518500248; on the alpha it does not seem worthwhile
-  */
-  long l,l0,l1,l2,l4,ee,i,n,m,s,t,av;
+  * take mmax=1518500248; on the alpha it does not seem worthwhile */
+  long l, l0, l1, l2, l4, i, n, m, s, t;
+  gpmem_t av;
   double alpha,beta,a,b,c,d;
-  GEN y,p1,p2,p3,p4,pitemp;
+  GEN y,p1,p2,p3,x2,pitemp, x = x0;
 
-  if (typ(x)!=t_REAL) err(typeer,"mpsc1");
-  if (!signe(x)) { *ptmod8=0; return realzero_bit((expo(x)<<1) - 1); }
-  l=lg(x); y=cgetr(l); av=avma;
+  if (typ(x) != t_REAL) err(typeer,"mpsc1");
+  l = lg(x); y = cgetr(l); av = avma;
 
-  l++; pitemp = mppi(l+1); setexpo(pitemp,-1);
-  p1 = addrr(x,pitemp); setexpo(pitemp,0);
-  if (expo(p1) >= bit_accuracy(min(l,lg(p1))) + 3)
-    err(precer,"mpsc1");
-  p3 = divrr(p1,pitemp); p2 = mpent(p3);
-  if (signe(p2)) x = subrr(x, mulir(p2,pitemp));
-  p1 = cgetr(l+1); affrr(x, p1);
+  l++; pitemp = mppi(l);
+  setexpo(pitemp,-1);
+  p1 = addrr(x,pitemp); /* = x + Pi/4 */
+  l0 = min(l, lg(p1));
+  if (expo(p1) >= bit_accuracy(l0) + 3) err(precer,"mpsc1");
 
-  *ptmod8 = (signe(p1) < 0)? 4: 0;
-  if (signe(p2))
+  setexpo(pitemp, 0);
+  p1 = mpent( divrr(p1,pitemp) ); /* round ( x / (Pi/2) ) */
+  if (signe(p1)) x = subrr(x, mulir(p1,pitemp)); /* x mod Pi/2  */
+  p2 = cgetr(l); affrr(x, p2); x = p2;
+
+  *ptmod8 = (signe(x) < 0)? 4: 0;
+  if (signe(p1))
   {
-    long mod4 = mod4(p2);
-    if (signe(p2) < 0 && mod4) mod4 = 4-mod4;
-    *ptmod8 += mod4;
+    long k = mod4(p1);
+    if (signe(p1) < 0 && k) k = 4-k;
+    /* x = x0 - k*Pi/2  (mod 2Pi) */
+    *ptmod8 += k;
   }
-  if (gcmp0(p1)) alpha=1000000.0;
+
+  if (gcmp0(x)) alpha = 1000000.0;
   else
   {
-    m=expo(p1); alpha=(m<-1022)? -1-m*LOG2: -1-log(fabs(rtodbl(p1)));
+    long e = expo(x);
+    alpha = -1 - ((e<-1022)? e*LOG2: log(fabs(rtodbl(x))));
   }
   beta = 5 + bit_accuracy(l)*LOG2;
-  a=0.5/LOG2; b=0.5*a;
-  c = a+sqrt((beta+b)/LOG2);
-  d = ((beta/c)-alpha-log(c))/LOG2;
-  if (d>=0)
+  a = 0.5 / LOG2;
+  b = 0.5 * a;
+  c = a + sqrt((beta+b) / LOG2);
+  d = ((beta/c) - alpha - log(c)) / LOG2;
+  if (d >= 0)
   {
-    m=(long)(1+d);
-    n=(long)((1+c)/2.0);
-    setexpo(p1,expo(p1)-m);
+    m = (long)(1+d);
+    n = (long)((1+c) / 2.0);
+    setexpo(x, expo(x)-m);
   }
-  else { m=0; n=(long)((1+beta/alpha)/2.0); }
+  else { m = 0; n = (long)((2+beta/alpha) / 2.0); }
   l2=l+1+(m>>TWOPOTBITS_IN_LONG);
-  p2=realun(l2); setlg(p2,4);
-  p4=cgetr(l2); av = avma;
-  affrr(gsqr(p1),p4); setlg(p4,4);
-
-  if (n>mmax)
-    p3 = divrs(divrs(p4,2*n+2),2*n+1);
+  p2 = realun(l2);
+  x2 = cgetr(l2); av = avma;
+  affrr(gsqr(x), x2);
+  
+  setlg(x2, 3);
+  if (n > mmax)
+    p3 = divrs(divrs(x2, 2*n+2), 2*n+1);
   else
-    p3 = divrs(p4, (2*n+2)*(2*n+1));
-  ee = -expo(p3); s=0;
-  l4 = l1 = 3 + (ee>>TWOPOTBITS_IN_LONG);
-  if (l4<=l2) { setlg(p2,l4); setlg(p4,l4); }
-  for (i=n; i>mmax; i--)
+    p3 = divrs(x2, (2*n+2)*(2*n+1));
+  s = -expo(p3);
+  l4 = l1 = 3 + (s>>TWOPOTBITS_IN_LONG);
+  if (l4<=l2) { setlg(p2,l4); setlg(x2,l4); }
+  s = 0;
+  for (i=n; i>=2; i--)
   {
-    p3 = divrs(divrs(p4,2*i),2*i-1); s -= expo(p3);
-    t=s&(BITS_IN_LONG-1); l0=(s>>TWOPOTBITS_IN_LONG);
+    if (i > mmax)
+      p3 = divrs(divrs(x2, 2*i), 2*i-1);
+    else
+      p3 = divrs(x2, 2*i*(2*i-1));
+    s -= expo(p3);
+    t = s & (BITS_IN_LONG-1); l0 = (s>>TWOPOTBITS_IN_LONG);
     if (t) l0++;
     l1 += l0; if (l1>l2) { l0 += l2-l1; l1=l2; }
     l4 += l0;
     p3 = mulrr(p3,p2);
-    if (l4<=l2) { setlg(p2,l4); setlg(p4,l4); }
-    subsrz(1,p3,p2); avma=av;
+    if (l4<=l2) { setlg(p2,l4); setlg(x2,l4); }
+    subsrz(1,p3, p2); avma = av;
   }
-  for (  ; i>=2; i--)
-  {
-    p3 = divrs(p4, 2*i*(2*i-1)); s -= expo(p3);
-    t=s&(BITS_IN_LONG-1); l0=(s>>TWOPOTBITS_IN_LONG);
-    if (t) l0++;
-    l1 += l0; if (l1>l2) { l0 += l2-l1; l1=l2; }
-    l4 += l0;
-    p3 = mulrr(p3,p2);
-    if (l4<=l2) { setlg(p2,l4); setlg(p4,l4); }
-    subsrz(1,p3,p2); avma=av;
-  }
-  if (l4<=l2) { setlg(p2,l4); setlg(p4,l4); }
-  setexpo(p4,expo(p4)-1); setsigne(p4, -signe(p4));
-  p2 = mulrr(p4,p2);
+  setlg(p2,l2); setlg(x2,l2);
+  setexpo(p2, expo(p2)-1); setsigne(p2, -signe(p2));
+  p2 = mulrr(x2,p2);
+  /* Now p2 = sum {1<= i <=n} (-1)^i x^(2i) / (2i)! ~ cos(x) - 1 */
   for (i=1; i<=m; i++)
-  {
-    p2 = mulrr(p2,addsr(2,p2)); setexpo(p2,expo(p2)+1);
+  { /* p2 = cos(x)-1 --> cos(2x)-1 */
+    p2 = mulrr(p2, addsr(2,p2)); setexpo(p2, expo(p2)+1);
   }
   affrr(p2,y); avma=av; return y;
 }
 
-/********************************************************************/
-/**                                                                **/
-/**                      FONCTION sqrt(-x*(x+2))                   **/
-/**                                                                **/
-/********************************************************************/
-
+/* sqrt (1 - (1+x)^2) = sqrt(-x*(x+2)). Sends cos(x)-1 to |sin(x)| */
 static GEN
 mpaut(GEN x)
 {
-  long av = avma;
-  GEN p1 = mulrr(x,addsr(2,x));
-  setsigne(p1,-signe(p1));
+  gpmem_t av = avma;
+  GEN p1 = mulrr(x, addsr(2,x));
+  setsigne(p1, -signe(p1));
   return gerepileuptoleaf(av, mpsqrt(p1));
 }
 
 /********************************************************************/
-/**                                                                **/
-/**                       FONCTION COSINUS                         **/
-/**                                                                **/
+/**                            COSINE                              **/
 /********************************************************************/
 
 static GEN
 mpcos(GEN x)
 {
-  long mod8,av,tetpil;
+  long mod8;
+  gpmem_t av;
   GEN y,p1;
 
   if (typ(x)!=t_REAL) err(typeer,"mpcos");
   if (!signe(x)) return addsr(1,x);
 
-  av=avma; p1=mpsc1(x,&mod8); tetpil=avma;
+  av = avma; p1 = mpsc1(x,&mod8);
   switch(mod8)
   {
     case 0: case 4:
@@ -1694,13 +1787,13 @@ mpcos(GEN x)
     default: /* case 3: case 5: */
       y=mpaut(p1); break;
   }
-  return gerepile(av,tetpil,y);
+  return gerepileuptoleaf(av, y);
 }
 
 GEN
 gcos(GEN x, long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN r,u,v,y,p1,p2;
 
   switch(typ(x))
@@ -1735,28 +1828,28 @@ gcos(GEN x, long prec)
 void
 gcosz(GEN x, GEN y)
 {
-  long av = avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av = avma;
 
   if (!prec) err(infprecer,"gcosz");
   gaffect(gcos(x,prec),y); avma=av;
 }
 
 /********************************************************************/
-/**                                                                **/
-/**                       FONCTION SINUS                           **/
-/**                                                                **/
+/**                             SINE                               **/
 /********************************************************************/
 
 GEN
 mpsin(GEN x)
 {
-  long mod8,av,tetpil;
+  long mod8;
+  gpmem_t av;
   GEN y,p1;
 
   if (typ(x)!=t_REAL) err(typeer,"mpsin");
   if (!signe(x)) return realzero_bit(expo(x));
 
-  av=avma; p1=mpsc1(x,&mod8); tetpil=avma;
+  av = avma; p1 = mpsc1(x,&mod8);
   switch(mod8)
   {
     case 0: case 6:
@@ -1768,13 +1861,13 @@ mpsin(GEN x)
     default: /* case 3: case 7: */
       y=subsr(-1,p1); break;
   }
-  return gerepile(av,tetpil,y);
+  return gerepileuptoleaf(av, y);
 }
 
 GEN
 gsin(GEN x, long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN r,u,v,y,p1,p2;
 
   switch(typ(x))
@@ -1789,7 +1882,8 @@ gsin(GEN x, long prec)
       p1=gsub(p2,p1);
       gsincos((GEN)x[1],&u,&v,prec);
       tetpil=avma;
-      y[1]=lmul(p2,u); y[2]=lmul(p1,v);
+      y[1]=lmul(p2,u);
+      y[2]=lmul(p1,v);
       gerepilemanyvec(av,tetpil,y+1,2);
       return y;
 
@@ -1809,29 +1903,29 @@ gsin(GEN x, long prec)
 void
 gsinz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"gsinz");
   gaffect(gsin(x,prec),y); avma=av;
 }
 
 /********************************************************************/
-/**                                                                **/
-/**                    PROCEDURE SINUS,COSINUS                     **/
-/**                                                                **/
+/**                       SINE, COSINE together                    **/
 /********************************************************************/
 
 void
 mpsincos(GEN x, GEN *s, GEN *c)
 {
-  long av,tetpil,mod8;
+  long mod8;
+  gpmem_t av, tetpil;
   GEN p1, *gptr[2];
 
   if (typ(x)!=t_REAL) err(typeer,"mpsincos");
   if (!signe(x))
   {
-    *s=realzero_bit(expo(x));
-    *c=addsr(1,x); return;
+    *s = realzero_bit(expo(x));
+    *c = addsr(1,x); return;
   }
 
   av=avma; p1=mpsc1(x,&mod8); tetpil=avma;
@@ -1850,10 +1944,22 @@ mpsincos(GEN x, GEN *s, GEN *c)
   gerepilemanysp(av,tetpil,gptr,2);
 }
 
+/* return exp(ix), x a t_REAL */
+GEN
+exp_Ir(GEN x)
+{
+  gpmem_t av = avma;
+  GEN v = cgetg(3,t_COMPLEX);
+  mpsincos(x, (GEN*)(v+2), (GEN*)(v+1));
+  if (!signe(x)) return gerepilecopy(av, (GEN)v[1]);
+  return v;
+}
+
 void
 gsincos(GEN x, GEN *s, GEN *c, long prec)
 {
-  long av,tetpil,ii,i,j,ex,ex2,lx,ly;
+  long ii, i, j, ex, ex2, lx, ly, mi;
+  gpmem_t av, tetpil;
   GEN r,u,v,u1,v1,p1,p2,p3,p4,ps,pc;
   GEN *gptr[4];
 
@@ -1880,7 +1986,8 @@ gsincos(GEN x, GEN *s, GEN *c, long prec)
       p3=gmul(v1,v); p4=gmul(r,u);
       gptr[0]=&p1; gptr[1]=&p2; gptr[2]=&p3; gptr[3]=&p4;
       gerepilemanysp(av,tetpil,gptr,4);
-      ps[1]=(long)p1; ps[2]=(long)p2; pc[1]=(long)p3; pc[2]=(long)p4;
+      ps[1]=(long)p1; pc[1]=(long)p3;
+      ps[2]=(long)p2; pc[2]=(long)p4;
       return;
 
     case t_SER:
@@ -1908,6 +2015,8 @@ gsincos(GEN x, GEN *s, GEN *c, long prec)
       }
 
       ly=lx+2*ex;
+      mi = lx-1; while (mi>=3 && gcmp0((GEN)x[mi])) mi--;
+      mi += ex-2;
       pc=cgetg(ly,t_SER); *c=pc;
       ps=cgetg(lx,t_SER); *s=ps;
       pc[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
@@ -1917,14 +2026,14 @@ gsincos(GEN x, GEN *s, GEN *c, long prec)
       for (i=ex2; i<ly; i++)
       {
 	ii=i-ex; av=avma; p1=gzero;
-	for (j=ex; j<ii-1; j++)
+	for (j=ex; j<=min(ii-2,mi); j++)
 	  p1=gadd(p1,gmulgs(gmul((GEN)x[j-ex+2],(GEN)ps[ii-j]),j));
 	tetpil=avma;
 	pc[i]=lpile(av,tetpil,gdivgs(p1,2-i));
 	if (ii<lx)
 	{
 	  av=avma; p1=gzero;
-	  for (j=ex; j<=i-ex2; j++)
+	  for (j=ex; j<=min(i-ex2,mi); j++)
 	    p1=gadd(p1,gmulgs(gmul((GEN)x[j-ex+2],(GEN)pc[i-j]),j));
 	  p1=gdivgs(p1,i-2); tetpil=avma;
 	  ps[i-ex]=lpile(av,tetpil,gadd(p1,(GEN)x[i-ex]));
@@ -1958,7 +2067,7 @@ gsincos(GEN x, GEN *s, GEN *c, long prec)
 static GEN
 mptan(GEN x)
 {
-  long av=avma,tetpil;
+  gpmem_t av=avma, tetpil;
   GEN s,c;
 
   mpsincos(x,&s,&c); tetpil=avma;
@@ -1968,7 +2077,7 @@ mptan(GEN x)
 GEN
 gtan(GEN x, long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN s,c;
 
   switch(typ(x))
@@ -1992,7 +2101,8 @@ gtan(GEN x, long prec)
 void
 gtanz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"gtanz");
   gaffect(gtan(x,prec),y); avma=av;
@@ -2001,7 +2111,7 @@ gtanz(GEN x, GEN y)
 static GEN
 mpcotan(GEN x)
 {
-  long av=avma,tetpil;
+  gpmem_t av=avma, tetpil;
   GEN s,c;
 
   mpsincos(x,&s,&c); tetpil=avma;
@@ -2011,7 +2121,7 @@ mpcotan(GEN x)
 GEN
 gcotan(GEN x, long prec)
 {
-  long av,tetpil;
+  gpmem_t av, tetpil;
   GEN s,c;
 
   switch(typ(x))
@@ -2020,8 +2130,8 @@ gcotan(GEN x, long prec)
       return mpcotan(x);
 
     case t_SER: 
-      if (gcmp0(x)) err(diver9);
-      if (valp(x)<0) err(negexper,"gcotan"); /* fall through */
+      if (gcmp0(x)) err(talker,"0 argument in cotan");
+      if (valp(x) < 0) err(negexper,"cotan"); /* fall through */
     case t_COMPLEX:
       av=avma; gsincos(x,&s,&c,prec); tetpil=avma;
       return gerepile(av,tetpil,gdiv(c,s));
@@ -2035,7 +2145,8 @@ gcotan(GEN x, long prec)
 void
 gcotanz(GEN x, GEN y)
 {
-  long av=avma, prec = precision(y);
+  long prec = precision(y);
+  gpmem_t av=avma;
 
   if (!prec) err(infprecer,"gcotanz");
   gaffect(gcotan(x,prec),y); avma=av;
