@@ -1,4 +1,4 @@
-/* $Id: gp.c,v 1.102 2002/01/10 23:59:19 karim Exp $
+/* $Id: gp.c,v 1.121 2002/05/02 18:30:46 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -37,20 +37,16 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
   int readline_init = 1;
 BEGINEXTERN
 #  if defined(__cplusplus) && defined(__SUNPRO_CC)
-  /* readline.h gives a bad definition of readline() */
-  extern char*readline(char*);
+  extern char*readline(char*); /* bad prototype for readline() in readline.h */
 #  else
 #   ifdef READLINE_LIBRARY
 #     include <readline.h>
-#     include <history.h>
 #   else
 #     include <readline/readline.h>
-#     include <readline/history.h>
 #   endif
 #  endif
   extern int isatty(int);
 ENDEXTERN
-static void gp_add_history(char *s);
 #endif
 
 char*  _analyseur(void);
@@ -80,6 +76,7 @@ int    whatnow(char *s, int flag);
 
 #define MAX_PROMPT_LEN 128
 #define DFT_PROMPT "? "
+#define BREAK_LOOP_PROMPT "> "
 #define COMMENTPROMPT "comment> "
 #define CONTPROMPT ""
 #define DFT_INPROMPT ""
@@ -91,28 +88,18 @@ static char thestring[256];
 static char *prettyprinter;
 static char *prettyprinter_dft = "tex2mail -TeX -noindent -ragged -by_par";
 static pariFILE *prettyprinter_file;
-static long prettyp, test_mode, quiet_mode, gpsilent, simplifyflag;
-static long chrono, pariecho, primelimit, strictmatch;
-static long tglobal, histsize, paribufsize, lim_lines;
+static ulong test_mode, quiet_mode, gpsilent, simplifyflag;
+static ulong chrono, pariecho, primelimit, strictmatch;
+static ulong tglobal, histsize, paribufsize, lim_lines;
 static int tm_is_waiting = 0, handle_C_C = 0;
-static gp_format fmt;
-
-typedef struct Buffer {
-  char *buf;
-  long len;
-  jmp_buf env;
-  int flenv;
-} Buffer;
+static pariout_t *fmt;
 
 #define current_buffer (bufstack?((Buffer*)(bufstack->value)):NULL)
 static stack *bufstack = NULL;
 
-#define LBRACE '{'
-#define RBRACE '}'
 #define pariputs_opt(s) if (!quiet_mode) pariputs(s)
 #define skip_space(s) while (isspace((int)*s)) s++
 #define skip_alpha(s) while (isalpha((int)*s)) s++
-#define ask_filtre(t) filtre("",NULL,t)
 
 static void
 usage(char *s)
@@ -138,8 +125,8 @@ gp_preinit(void)
   long i;
 
   primelimit = 500000; 
-  bot = 0;
-  top = 1000000*sizeof(long);
+  bot = (gpmem_t)0;
+  top = (gpmem_t)(1000000*sizeof(long));
   strcpy(prompt, DFT_PROMPT);
   strcpy(prompt_cont, CONTPROMPT);
 
@@ -158,18 +145,18 @@ gp_preinit(void)
 # endif
   if (help_prg) help_prg = pari_strdup(help_prg);
 
-  prettyp = f_PRETTYMAT;
   strictmatch = simplifyflag = 1;
   tglobal = 0;
   bufstack = NULL;
   secure = test_mode = under_emacs = under_texmacs = chrono = pariecho = 0;
   prettyprinter = prettyprinter_dft;
   prettyprinter_file = NULL;
-  fmt.format = 'g'; fmt.field = 0;
+  fmt = &DFLT_OUTPUT;
+  fmt->prettyp= f_PRETTYMAT;
 #ifdef LONG_IS_64BIT
-  fmt.nb = 38;
+  fmt->sigd     = 38;
 #else
-  fmt.nb = 28;
+  fmt->sigd     = 28;
 #endif
   lim_lines = 0;
   histsize = 5000; paribufsize = 1024;
@@ -278,41 +265,23 @@ tm_end_output(void)
   tm_is_waiting = 0;
 }
 
-static void
-gp_output(GEN x)
-{
-  long tx=typ(x);
-
-  if (fmt.nb >= 0 && is_intreal_t(tx))
-    ecrire(x, fmt.format, fmt.nb, fmt.field);
-  else
-    switch(prettyp)
-    {
-      case f_PRETTYMAT: matbrute(x, fmt.format, fmt.nb); break;
-      case f_PRETTY:
-      case f_PRETTYOLD: sor(x, fmt.format, fmt.nb, fmt.field); break;
-      case f_RAW      : brute(x, fmt.format, fmt.nb); break;
-      case f_TEX      : texe(x, fmt.format, fmt.nb); break;
-    }
-}
-
 /* print a sequence of (NULL terminated) GEN */
 void
 print0(GEN *g, long flag)
 {
-  int old=prettyp;
+  pariout_t T = *fmt;
 
   added_newline = (flag & f_NOEOL) == 0;
-  prettyp = flag & ~f_NOEOL;
+  T.prettyp = flag & ~f_NOEOL;
 
   for( ; *g; g++)
     if (typ(*g)==t_STR)
       pariputs(GSTR(*g)); /* otherwise it's surrounded by "" */
     else
-      gp_output(*g);
+      gen_output(*g, &T);
 
   if (added_newline) pariputc('\n');
-  prettyp=old; pariflush();
+  pariflush();
 }
 
 /* write a sequence of (NULL terminated) GEN, to file s */
@@ -421,14 +390,14 @@ do_strftime(char *s, char *buf)
 }
 
 static GEN
-sd_numeric(char *v, int flag, char *s, long *ptn, ulong Min, ulong Max,
+sd_ulong(char *v, int flag, char *s, ulong *ptn, ulong Min, ulong Max,
            char **msg)
 {
-  long n;
+  ulong n;
   if (*v == 0) n = *ptn;
   else
   {
-    n = get_int(v,0);
+    n = (ulong)get_int(v,0);
     if (*ptn == n) return gnil;
     if (n > Max || n < Min)
     {
@@ -440,7 +409,7 @@ sd_numeric(char *v, int flag, char *s, long *ptn, ulong Min, ulong Max,
   }
   switch(flag)
   {
-    case d_RETURN: return stoi(n);
+    case d_RETURN: return utoi(n);
     case d_ACKNOWLEDGE:
       if (msg)
       {
@@ -467,19 +436,19 @@ sd_realprecision(char *v, int flag)
 {
   if (*v)
   {
-    long newnb = get_int(v, fmt.nb);
+    long newnb = get_int(v, fmt->sigd);
     long newprec = (long) (newnb*pariK1 + 3);
 
-    if (fmt.nb == newnb && prec == newprec) return gnil;
+    if (fmt->sigd == newnb && prec == newprec) return gnil;
     if (newnb < 0) err(talker,"default: negative real precision");
-    fmt.nb = newnb; prec = newprec;
+    fmt->sigd = newnb; prec = newprec;
   }
-  if (flag == d_RETURN) return stoi(fmt.nb);
+  if (flag == d_RETURN) return stoi(fmt->sigd);
   if (flag == d_ACKNOWLEDGE)
   {
     long n = PRECDIGIT;
     pariputsf("   realprecision = %ld significant digits", n);
-    if (n != fmt.nb) pariputsf(" (%ld digits displayed)", fmt.nb);
+    if (n != fmt->sigd) pariputsf(" (%ld digits displayed)", fmt->sigd);
     pariputc('\n');
   }
   return gnil;
@@ -490,7 +459,7 @@ static GEN
 sd_seriesprecision(char *v, int flag)
 {
   char *msg[] = {NULL, "significant terms"};
-  return sd_numeric(v,flag,"seriesprecision",&precdl, 0,LGBITS,msg);
+  return sd_ulong(v,flag,"seriesprecision",(ulong*)&precdl, 0,LGBITS,msg);
 }
 
 static GEN
@@ -501,24 +470,24 @@ sd_format(char *v, int flag)
     char c = *v;
     if (c!='e' && c!='f' && c!='g')
       err(talker2,"default: inexistent format",v,v);
-    fmt.format = c; v++;
+    fmt->format = c; v++;
 
     if (isdigit((int)*v))
-      { fmt.field=atol(v); while (isdigit((int)*v)) v++; }
+      { fmt->fieldw=atol(v); while (isdigit((int)*v)) v++; }
     if (*v++ == '.')
     {
-      if (*v == '-') fmt.nb = -1;
+      if (*v == '-') fmt->sigd = -1;
       else
-	if (isdigit((int)*v)) fmt.nb=atol(v);
+	if (isdigit((int)*v)) fmt->sigd=atol(v);
     }
   }
   if (flag == d_RETURN)
   {
-    sprintf(thestring, "%c%ld.%ld", fmt.format, fmt.field, fmt.nb);
+    sprintf(thestring, "%c%ld.%ld", fmt->format, fmt->fieldw, fmt->sigd);
     return strtoGENstr(thestring,0);
   }
   if (flag == d_ACKNOWLEDGE)
-    pariputsf("   format = %c%ld.%ld\n", fmt.format, fmt.field, fmt.nb);
+    pariputsf("   format = %c%ld.%ld\n", fmt->format, fmt->fieldw, fmt->sigd);
   return gnil;
 }
 
@@ -558,6 +527,7 @@ sd_colors(char *v, int flag)
   long c,l;
   if (*v && !under_emacs && !under_texmacs)
   {
+    char *v0;
     disable_color=1;
     l = strlen(v);
     if (l <= 2 && strncmp(v, "no", l) == 0)
@@ -568,9 +538,10 @@ sd_colors(char *v, int flag)
       v = "1, 6, 3, 4, 5, 2, 3";	/* Assume recent ReadLine. */
     if (l <= 6 && strncmp(v, "boldfg", l) == 0)	/* Good for darkbg consoles */
       v = "[1,,1], [5,,1], [3,,1], [7,,1], [6,,1], [2,,1], [3,,1]";
-    v = filtre(v,NULL, f_INIT|f_REG);
+    v0 = v = filtre(v, 0);
     for (c=c_ERR; c < c_LAST; c++)
       gp_colors[c] = gp_get_color(&v);
+    free(v0);
   }
   if (flag == d_ACKNOWLEDGE || flag == d_RETURN)
   {
@@ -612,8 +583,8 @@ sd_compatible(char *v, int flag)
     "(use old functions, don't ignore case)",
     "(use old functions, ignore case)", NULL
   };
-  long old = compatible;
-  GEN r = sd_numeric(v,flag,"compatible",&compatible, 0,3,msg);
+  ulong old = compatible;
+  GEN r = sd_ulong(v,flag,"compatible",&compatible, 0,3,msg);
 
   if (old != compatible && flag != d_INITRC)
   {
@@ -632,16 +603,16 @@ sd_secure(char *v, int flag)
     fprintferr("[secure mode]: Do you want to modify the 'secure' flag? (^C if not)\n");
     hit_return();
   }
-  return sd_numeric(v,flag,"secure",&secure, 0,1,NULL);
+  return sd_ulong(v,flag,"secure",&secure, 0,1,NULL);
 }
 
 static GEN
 sd_buffersize(char *v, int flag)
-{ return sd_numeric(v,flag,"buffersize",&paribufsize, 1,
+{ return sd_ulong(v,flag,"buffersize",&paribufsize, 1,
                     (VERYBIGINT / sizeof(long)) - 1,NULL); }
 static GEN
 sd_debug(char *v, int flag)
-{ return sd_numeric(v,flag,"debug",&DEBUGLEVEL, 0,20,NULL); }
+{ return sd_ulong(v,flag,"debug",&DEBUGLEVEL, 0,20,NULL); }
 
 static GEN
 sd_rl(char *v, int flag)
@@ -651,34 +622,34 @@ sd_rl(char *v, int flag)
       init_readline();
       readline_init = 1;
     }
-    return sd_numeric(v,flag,"readline",&use_readline, 0,20,NULL);
+    return sd_ulong(v,flag,"readline",(ulong*)&use_readline, 0,20,NULL);
 #else	/* !( defined READLINE ) */
-    long dummy;
-    return sd_numeric(v,flag,"readline",&dummy, 0,20,NULL);
+    ulong dummy;
+    return sd_ulong(v,flag,"readline",&dummy, 0,20,NULL);
 #endif
 }
 
 static GEN
 sd_debugfiles(char *v, int flag)
-{ return sd_numeric(v,flag,"debugfiles",&DEBUGFILES, 0,20,NULL); }
+{ return sd_ulong(v,flag,"debugfiles",&DEBUGFILES, 0,20,NULL); }
 
 static GEN
 sd_debugmem(char *v, int flag)
-{ return sd_numeric(v,flag,"debugmem",&DEBUGMEM, 0,20,NULL); }
+{ return sd_ulong(v,flag,"debugmem",&DEBUGMEM, 0,20,NULL); }
 
 static GEN
 sd_echo(char *v, int flag)
-{ return sd_numeric(v,flag,"echo",&pariecho, 0,1,NULL); }
+{ return sd_ulong(v,flag,"echo",&pariecho, 0,1,NULL); }
 
 static GEN
 sd_lines(char *v, int flag)
-{ return sd_numeric(v,flag,"lines",&lim_lines, 0,VERYBIGINT,NULL); }
+{ return sd_ulong(v,flag,"lines",&lim_lines, 0,VERYBIGINT,NULL); }
 
 static GEN
 sd_histsize(char *v, int flag)
 {
-  long n = histsize;
-  GEN r = sd_numeric(v,flag,"histsize",&n, 1,
+  ulong n = histsize;
+  GEN r = sd_ulong(v,flag,"histsize",&n, 1,
                      (VERYBIGINT / sizeof(long)) - 1,NULL);
   if (n != histsize)
   {
@@ -714,8 +685,8 @@ sd_histsize(char *v, int flag)
 static GEN
 sd_log(char *v, int flag)
 {
-  long vlog = logfile? 1: 0, old = vlog;
-  GEN r = sd_numeric(v,flag,"log",&vlog, 0,1,NULL);
+  ulong vlog = logfile? 1: 0, old = vlog;
+  GEN r = sd_ulong(v,flag,"log",&vlog, 0,1,NULL);
   if (vlog != old)
   {
     if (vlog)
@@ -740,13 +711,13 @@ static GEN
 sd_output(char *v, int flag)
 {
   char *msg[] = {"(raw)", "(prettymatrix)", "(prettyprint)", "(external prettyprint)", NULL};
-  return sd_numeric(v,flag,"output",&prettyp, 0,3,msg);
+  return sd_ulong(v,flag,"output",(ulong*) &(fmt->prettyp), 0,3,msg);
 }
 
 extern void err_clean(void);
 
 void
-allocatemem0(unsigned long newsize)
+allocatemem0(size_t newsize)
 {
   (void)allocatemoremem(newsize);
   err_clean();
@@ -756,11 +727,11 @@ allocatemem0(unsigned long newsize)
 static GEN
 sd_parisize(char *v, int flag)
 {
-  long n = top-bot;
-  GEN r = sd_numeric(v,flag,"parisize",&n, 10000,VERYBIGINT,NULL);
-  if (n != top-bot)
+  ulong n = top-bot;
+  GEN r = sd_ulong(v,flag,"parisize",&n, 10000,VERYBIGINT,NULL);
+  if (n != (ulong)top-bot)
   {
-    if (!bot) top = n; /* no stack allocated yet */
+    if (!bot) top = (gpmem_t)n; /* no stack allocated yet */
     if (flag != d_INITRC) allocatemem0(n);
   }
   return r;
@@ -769,8 +740,8 @@ sd_parisize(char *v, int flag)
 static GEN
 sd_primelimit(char *v, int flag)
 {
-  long n = primelimit;
-  GEN r = sd_numeric(v,flag,"primelimit",&n, 0,VERYBIGINT,NULL);
+  ulong n = primelimit;
+  GEN r = sd_ulong(v,flag,"primelimit",&n, 0,VERYBIGINT,NULL);
   if (n != primelimit)
   {
     if (flag != d_INITRC)
@@ -785,15 +756,15 @@ sd_primelimit(char *v, int flag)
 
 static GEN
 sd_simplify(char *v, int flag)
-{ return sd_numeric(v,flag,"simplify",&simplifyflag, 0,1,NULL); }
+{ return sd_ulong(v,flag,"simplify",&simplifyflag, 0,1,NULL); }
 
 static GEN
 sd_strictmatch(char *v, int flag)
-{ return sd_numeric(v,flag,"strictmatch",&strictmatch, 0,1,NULL); }
+{ return sd_ulong(v,flag,"strictmatch",&strictmatch, 0,1,NULL); }
 
 static GEN
 sd_timer(char *v, int flag)
-{ return sd_numeric(v,flag,"timer",&chrono, 0,1,NULL); }
+{ return sd_ulong(v,flag,"timer",&chrono, 0,1,NULL); }
 
 static GEN
 sd_filename(char *v, int flag, char *s, char **f)
@@ -1446,9 +1417,9 @@ aide0(char *s, int flag)
   s = get_sep(s);
   if (isdigit((int)*s))
   {
-    n=atoi(s);
+    n = atoi(s);
     if (n == 12) { community(); return; }
-    if (n<0 || n > 12)
+    if (n < 0 || n > 12)
       err(talker2,"no such section in help: ?",s,s);
     if (long_help) external_help(s,3); else commands(n);
     return;
@@ -1640,7 +1611,7 @@ Type ?12 for how to get moral (and possibly technical) support.\n\n");
   pariputsf("\nparisize = %lu, primelimit = %lu\n", top-bot, primelimit);
 }
 
-static void
+void
 fix_buffer(Buffer *b, long newlbuf)
 {
   b->len = paribufsize = newlbuf;
@@ -1675,7 +1646,7 @@ gp_history(long p, long flag, char *old, char *entrypoint)
     p = tglobal - p;
     if (p <= 0) er1 = 1;
   }
-  else if (p > tglobal)
+  else if ((ulong)p > tglobal)
     err(talker2,"I can't see into the future",old,entrypoint);
   if (!p) p = tglobal;
   if (tglobal - p >= histsize) er1 = 1;
@@ -1685,12 +1656,15 @@ gp_history(long p, long flag, char *old, char *entrypoint)
   return hist[p];
 }
 
-extern char *GENtostr0(GEN x, void(*do_out)(GEN));
-
 static void
 texmacs_output(GEN z, long n)
 {
-  char *sz = GENtostr0(z, &outtex);
+  pariout_t T = *fmt;
+  char *sz;
+
+  T.prettyp = f_TEX;
+  T.fieldw = 0;
+  sz = GENtostr0(z, &T, &gen_output);
   printf("%clatex:", DATA_BEGIN);
   printf("\\magenta\\%%%ld = $\\blue ", n);
   printf("%s$%c", sz,DATA_END); free(sz);
@@ -1729,17 +1703,16 @@ static int
 tex2mail_output(GEN z, long n)
 {
   FILE *o_out;
-  int o_prettyp;
+  pariout_t T = *fmt;
   
   if (!(prettyprinter && prettyp_init())) return 0;
   o_out = pari_outfile; /* save state */
-  o_prettyp = prettyp;
 
   /* Emit first: there may be lines before the prompt */
   if (n) term_color(c_OUTPUT);
   pariflush();
   pari_outfile = prettyprinter_file->file;
-  prettyp = f_TEX;
+  T.prettyp = f_TEX;
 
   /* history number */
   if (n)
@@ -1748,7 +1721,7 @@ tex2mail_output(GEN z, long n)
     {
       char col1[80];
       strcpy(col1, term_get_color(c_HIST));
-      sprintf(thestring, "\\LITERALnoLENGTH{%s}\\%%%ld = \\LITERALnoLENGTH{%s}",
+      sprintf(thestring, "\\LITERALnoLENGTH{%s}\\%%%ld =\\LITERALnoLENGTH{%s} ",
               col1, n, term_get_color(c_OUTPUT));
     }
     else
@@ -1756,11 +1729,10 @@ tex2mail_output(GEN z, long n)
     pariputs_opt(thestring);
   }
   /* output */
-  gp_output(z);
+  gen_output(z, &T);
 
   /* flush and restore */
   prettyp_wait();
-  prettyp = o_prettyp;
   pari_outfile = o_out;
   if (n) term_color(c_NONE);
   return 1;
@@ -1776,7 +1748,7 @@ normal_output(GEN z, long n)
   /* output */
   term_color(c_OUTPUT);
   init_lim_lines(thestring,lim_lines);
-  gp_output(z);
+  gen_output(z, fmt);
   init_lim_lines(NULL,lim_lines);
   term_color(c_NONE); pariputc('\n');
 }
@@ -1835,10 +1807,10 @@ escape0(char *tch)
       x = gp_history(d, 0, tch+1,tch-1);
       switch (c)
       {
-	case 'a': brute   (x, fmt.format, -1); break;
-	case 'm': matbrute(x, fmt.format, -1); break;
+	case 'a': brute   (x, fmt->format, -1); break;
+	case 'm': matbrute(x, fmt->format, -1); break;
 	case 'B': if (tex2mail_output(x,0)) return;  /* fall through */
-	case 'b': sor     (x, fmt.format, -1, fmt.field); break;
+	case 'b': sor     (x, fmt->format, -1, fmt->fieldw); break;
 	case 'x': voir(x, get_int(s, -1)); 
         case 'w':
 	{
@@ -2018,9 +1990,18 @@ gprc_get(void)
   return f;
 }
 
-static int get_line_from_file(char *prompt, Buffer *b, FILE *file);
+static int get_line_from_file(char *prompt, filtre_t *F, FILE *file);
 #define err_gprc(s,t,u) { fprintferr("\n"); err(talker2,s,t,u); }
 
+static void 
+init_filtre(filtre_t *F, void *data)
+{
+  F->data = data;
+  F->in_string  = 0;
+  F->in_comment = 0;
+  F->downcase = 0;
+}
+
 static char **
 gp_initrc(void)
 {
@@ -2028,13 +2009,15 @@ gp_initrc(void)
   FILE *file = gprc_get();
   long fnum = 4, find = 0;
   Buffer *b;
+  filtre_t F;
 
   if (!file) return NULL;
   flist = (char **) gpmalloc(fnum * sizeof(char*));
   b = new_buffer();
+  init_filtre(&F, (void*)b);
   for(;;)
   {
-    if (!get_line_from_file(NULL,b,file))
+    if (!get_line_from_file(NULL,&F,file))
     {
       del_buffer(b);
       if (!quiet_mode) fprintferr("Done.\n\n");
@@ -2212,7 +2195,7 @@ brace_color(char *s, int c, int force)
 }
 
 static char *
-do_prompt(char *p)
+do_prompt(int in_comment, char *p)
 {
   static char buf[MAX_PROMPT_LEN + 24]; /* + room for color codes */
   char *s;
@@ -2222,7 +2205,7 @@ do_prompt(char *p)
   /* escape sequences bug readline, so use special bracing (if available) */
   brace_color(s, c_PROMPT, 0);
   s += strlen(s);
-  if (filtre(s,NULL, f_COMMENT)) 
+  if (in_comment) 
     strcpy(s, COMMENTPROMPT);
   else
     do_strftime(p,s);
@@ -2230,19 +2213,16 @@ do_prompt(char *p)
   brace_color(s, c_INPUT, 1); return buf;
 }
 
-static void
-unblock_SIGINT(void)
+static char *
+fgets_texmacs(char *s, int n, FILE *f)
 {
-#ifdef USE_SIGRELSE
-  sigrelse(SIGINT);
-#elif USE_SIGSETMASK
-  sigsetmask(0);
-#endif
+  tm_start_output(); tm_end_output(); /* tell TeXmacs we need input */
+  return fgets(s,n,f);
 }
 
 /* Read from file (up to '\n' or EOF) and copy at s0 (points in b->buf) */
 static char *
-file_input(Buffer *b, char **s0, FILE *file, int TeXmacs)
+file_input(Buffer *b, char **s0, input_method *IM)
 {
   int first = 1;
   char *s = *s0;
@@ -2251,9 +2231,7 @@ file_input(Buffer *b, char **s0, FILE *file, int TeXmacs)
   used0 = used;
   for(;;)
   {
-    long left = b->len - used, ls;
-    /* if from TeXmacs, tell him we need input */
-    if (TeXmacs) { tm_start_output(); tm_end_output(); }
+    long left = b->len - used, l;
 
     if (left < 512)
     {
@@ -2262,118 +2240,74 @@ file_input(Buffer *b, char **s0, FILE *file, int TeXmacs)
       *s0 = b->buf + used0;
     }
     s = b->buf + used;
-    if (! fgets(s, left, file)) return first? NULL: *s0; /* EOF */
-    ls = strlen(s); first = 0;
-    if (ls+1 < left || s[ls-1] == '\n') return *s0; /* \n */
-    used += ls;
+    if (! IM->fgets(s, left, IM->file))
+      return first? NULL: *s0; /* EOF */
+     
+    l = strlen(s); first = 0;
+    if (l+1 < left || s[l-1] == '\n') return *s0; /* \n */
+    used += l;
   }
 }
 
-#ifdef READLINE
-/* Read line; returns a malloc()ed string of the user input or NULL on EOF.
-   Increments the buffer size appropriately if needed; fix *endp if so.
- */
-static char *
-gprl_input(Buffer *b, char **endp, char *prompt)
+/* Read a "complete line" and filter it. Return: 0 if EOF, 1 otherwise */
+int
+input_loop(filtre_t *F, input_method *IM)
 {
-  long used = *endp - b->buf;
-  long left = b->len - used, l;
-  char *s;
-  
-  if (! (s = readline(prompt)) ) return NULL; /* EOF */
-  l = strlen(s);
-  if ((ulong)left < l)
-  {
-    long incr = b->len;
-
-    if (incr < l) incr = l;
-    fix_buffer(b, b->len + incr);
-    *endp = b->buf + used;
+  Buffer *b = (Buffer*)F->data;
+  char *to_read, *s = b->buf;
+ 
+  /* read first line */
+  handle_C_C = 0;
+  while (! (to_read = IM->getline(b,&s,IM)) )
+  { /* EOF */
+    if (!handle_C_C) { check_filtre(F); return 0; }
+    /* received ^C in getline and coming back from break_loop();
+     * retry (as if "\n" were input) */
+    handle_C_C = 0;
   }
-  return s;
-}
-#endif
-
-#define ask_filtre(t) filtre("",NULL,t)
-
-/* True if more than one line read */
-static int
-input_loop(Buffer *b, char *already_read, FILE *file, char *prompt)
-{
-  const int TeXmacs = (under_texmacs && file == stdin);
-  const int f_flag = prompt? f_REG: f_REG | f_KEEPCASE;
-  char *end, *s = b->buf;
-  int wait_for_brace = 0;
-  int wait_for_input = 0;
-  int read = 0;
 
   /* buffer is not empty, init filter */
-  (void)ask_filtre(f_INIT);
+  F->in_string = 0;
+  F->more_input= 0;
+  F->wait_for_brace = 0;
   for(;;)
   {
-    char *t = already_read;
-    if (!ask_filtre(f_COMMENT))
-    { /* not in comment */
-      while (isspace((int)*t)) t++; /* Skip space */
-      if (*t == LBRACE) { t++; wait_for_input = wait_for_brace = 1; }
-    }
-    end = filtre(t,s, f_flag);
+    F->s = to_read;
+    F->t = s;
+    (void)filtre0(F);
+    if (IM->free) free(to_read);
+   
+    if (! F->more_input) break;
 
-    if (!*s) { if (!wait_for_input) break; }
-    else
-    {
-      if (*(b->buf) == '?') break;
-
-      s = end-1; /* *s = last input char */
-      if (*s == '\\')
-      {
-      }
-      else if (*s == '=')
-      {
-        wait_for_input = 1; s++;
-      }
-      else
-      {
-	if (!wait_for_brace) break;
-	if (*s == RBRACE) { *s=0; break; }
-	s++;
-      }
-    }
     /* read continuation line */
-#ifdef READLINE
-    if (!file) {
-      free(already_read);
-      already_read = gprl_input(b, &s, do_prompt(prompt_cont));
-      gp_add_history(already_read);	/* Makes a copy */
-    } else
-#endif
-      already_read = file_input(b,&s,file,TeXmacs);
-    if (!already_read) break;
-    read++;
+    s = F->end;
+    if (IM->prompt) IM->prompt = do_prompt(F->in_comment, prompt_cont);
+    to_read = IM->getline(b,&s, IM);
+    if (!to_read) break;
   }
-  if (!file && already_read) free(already_read);
-  return read;
+  return 1;
 }
 
 /* prompt = NULL --> from gprc. Return 1 if new input, and 0 if EOF */
 static int
-get_line_from_file(char *prompt, Buffer *b, FILE *file)
+get_line_from_file(char *prompt, filtre_t *F, FILE *file)
 {
   const int TeXmacs = (under_texmacs && file == stdin);
-  char *buf, *s =  b->buf;
-
-  handle_C_C = 0;
-  while (! (buf = file_input(b,&s,file,TeXmacs)) )
-  { /* EOF */
-    if (!handle_C_C)
-    {
-      if (TeXmacs) tm_start_output();
-      return 0;
-    }
-    /* received ^C  in fgets, retry (as is "\n" were input) */
+  char *s;
+  input_method IM;
+ 
+  IM.file = file;
+  IM.fgets= TeXmacs? &fgets_texmacs: &fgets;
+  IM.prompt = NULL;
+  IM.getline= &file_input;
+  IM.free = 0;
+  if (! input_loop(F,&IM))
+  {
+    if (TeXmacs) tm_start_output();
+    return 0;
   }
-  input_loop(b,buf,file,prompt);
-
+ 
+  s = ((Buffer*)F->data)->buf;
   if (*s && prompt) /* don't echo if from gprc */
   {
     if (pariecho)
@@ -2386,73 +2320,12 @@ get_line_from_file(char *prompt, Buffer *b, FILE *file)
   return 1;
 }
 
-/* request one line interactively.
- * Return 0: EOF
- *        1: got one line from readline or infile */
-#ifndef READLINE
 static int
-get_line_from_user(char *prompt, Buffer *b)
+get_line_from_user(char *prompt, filtre_t *F)
 {
   pariputs(prompt);
-  return get_line_from_file(prompt,b,infile);
+  return get_line_from_file(prompt,F,infile);
 }
-#else	/* READLINE */
-static int
-history_is_new(char *s)
-{
-  return (*s && (!history_length ||
-                  strcmp(s, history_get(history_length)->line)));
-}
-
-static void
-gp_add_history(char *s)
-{
-  if (history_is_new(s)) add_history(s);
-}
-
-static int
-get_line_from_user(char *prompt, Buffer *b)
-{
-  if (use_readline)
-  {
-    char *buf, *s = b->buf;
-    int index, added;
-
-    if (! (buf = gprl_input(b,&s, prompt)) )
-    { /* EOF */
-      pariputs("\n"); return 0;
-    }
-    /* Put the original read line into history */
-    index = history_length;
-    gp_add_history(buf);	/* Copies the entry */
-
-    added = input_loop(b,buf,NULL,prompt); /* free()s buf */
-    unblock_SIGINT(); /* bug in readline 2.0: need to unblock ^C */
-
-    if (*s)
-    {				/* XXXX Better use b->buf ?! */
-      if (added)
-      { /* Remove incomplete lines */
-        int i = history_length;
-        while (i > index) {
-          HIST_ENTRY *e = remove_history(--i);
-          free(e->line); free(e);
-        }
-        gp_add_history(s);
-      }
-    
-      /* update logfile */
-      if (logfile) fprintf(logfile, "%s%s\n",prompt,s);
-    }
-    return 1;
-  }
-  else
-  {
-    pariputs(prompt);
-    return get_line_from_file(prompt,b,infile);
-  }
-}
-#endif
 
 static int
 is_interactive(void)
@@ -2465,14 +2338,28 @@ is_interactive(void)
 #endif
 }
 
+extern int get_line_from_readline(char *prompt, filtre_t *F);
+
 /* return 0 if no line could be read (EOF) */
 static int
-read_line(char *promptbuf, Buffer *b)
+read_line(filtre_t *F, char *PROMPT)
 {
+  int res;
+  if (compatible == OLDALL) F->downcase = 1;
   if (is_interactive())
-    return get_line_from_user(promptbuf, b);
+  {
+    if (!PROMPT) PROMPT = do_prompt(F->in_comment, prompt);
+#ifdef READLINE
+    if (use_readline)
+      res = get_line_from_readline(PROMPT, F);
+    else
+#endif
+    res = get_line_from_user(PROMPT, F);
+    if (!disable_color) term_color(c_NONE);
+  }
   else
-    return get_line_from_file(DFT_PROMPT,b,infile);
+    res = get_line_from_file(DFT_PROMPT,F,infile);
+  return res;
 }
 
 static int
@@ -2488,6 +2375,8 @@ chron(char *s)
   return 1;
 }
 
+/* return 0: can't interpret *buf as a metacommand
+ *        1: did interpret *buf as a metacommand or empty command */
 static int
 check_meta(char *buf)
 {
@@ -2496,7 +2385,7 @@ check_meta(char *buf)
     case '?': aide(buf, h_REGULAR); break;
     case '#': return chron(buf);
     case '\\': escape(buf); break;
-    case '\0': return 2;
+    case '\0': break;
     default: return 0;
   }
   return 1;
@@ -2507,20 +2396,25 @@ check_meta(char *buf)
 static GEN
 gp_main_loop(int ismain)
 {
-  long av, i,j;
+  long i, j;
+  gpmem_t av;
   VOLATILE GEN z = gnil;
   Buffer *b = new_buffer();
+  filtre_t F;
+
   if (!setjmp(b->env))
   {
     b->flenv = 1;
     push_stack(&bufstack, (void*)b);
   }
-  for(; ; setjmp(b->env))
+  init_filtre(&F, (void*)b);
+
+  for (;; setjmp(b->env))
   {
     if (ismain)
     {
       static long tloc, outtyp;
-      tloc = tglobal; outtyp = prettyp; recover(0);
+      tloc = tglobal; outtyp = fmt->prettyp; recover(0);
       if (setjmp(environnement))
       {
         char *s = (char*)global_err_data;
@@ -2533,7 +2427,7 @@ gp_main_loop(int ismain)
 	  if (!i) i = histsize;
 	  i--; j--;
 	}
-        tglobal = tloc; prettyp = outtyp;
+        tglobal = tloc; fmt->prettyp = outtyp;
         kill_all_buffers(b);
       }
     }
@@ -2542,10 +2436,7 @@ gp_main_loop(int ismain)
 
     for(;;)
     {
-      int r;
-      r = read_line(do_prompt(prompt), b);
-      if (!disable_color) term_color(c_NONE);
-      if (!r)
+      if (! read_line(&F, NULL))
       {
 #ifdef _WIN32
 	Sleep(10); if (win32ctrlc) dowin32ctrlc();
@@ -2573,12 +2464,12 @@ gp_main_loop(int ismain)
     avma = av;
     if (gpsilent) continue;
 
-    if (test_mode) { init80(0); gp_output(z); pariputc('\n'); }
+    if (test_mode) { init80(0); gen_output(z, fmt); pariputc('\n'); }
     else
     {
       if (under_texmacs)
         texmacs_output(z,tglobal);
-      else if (prettyp != f_PRETTY || !tex2mail_output(z,tglobal))
+      else if (fmt->prettyp != f_PRETTY || !tex2mail_output(z,tglobal))
         normal_output(z,tglobal);
     }
     pariflush();
@@ -2628,10 +2519,12 @@ GEN
 input0(void)
 {
   Buffer *b = new_buffer();
+  filtre_t F;
   GEN x;
 
+  init_filtre(&F, (void*)b);
   push_stack(&bufstack, (void*)b);
-  while (! get_line_from_file(DFT_INPROMPT,b,infile))
+  while (! get_line_from_file(DFT_INPROMPT,&F,infile))
     if (popinfile()) { fprintferr("no input ???"); gp_quit(); }
   x = lisseq(b->buf);
   pop_buffer(); return x;
@@ -2668,6 +2561,7 @@ break_loop(long numerr)
   static Buffer *b = NULL;
   VOLATILE int go_on = 0;
   char *s, *t, *msg;
+  filtre_t F;
 
   if (b) jump_to_given_buffer(b);
   push_stack(&bufstack, (void*)new_buffer());
@@ -2687,6 +2581,7 @@ break_loop(long numerr)
     if (!s || !s[-1] || s < t || s >= t + oldb->len) s = NULL;
     b->flenv = 1; oldinfile = infile;
   }
+  init_filtre(&F, (void*)b);
 
   term_color(c_ERR); pariputc('\n');
   errcontext(msg, s, t); if (s) pariputc('\n');
@@ -2695,9 +2590,8 @@ break_loop(long numerr)
   infile = stdin;
   for(;;)
   {
-    int flag;
-    if (! read_line("> ", b)) break;
-    if (!(flag = check_meta(b->buf)))
+    if (! read_line(&F, BREAK_LOOP_PROMPT)) break;
+    if (! check_meta(b->buf))
     {
       GEN x = lisseq(b->buf);
       if (did_break())
@@ -2711,10 +2605,11 @@ break_loop(long numerr)
       }
       if (x == gnil) continue;
 
-      term_color(c_OUTPUT); gp_output(x);
+      term_color(c_OUTPUT); gen_output(x, fmt);
       term_color(c_NONE); pariputc('\n');
     }
-    if (numerr == siginter && flag == 2) { handle_C_C = go_on = 1; break; }
+    /* break loop initiated by ^C. Empty input --> continue computation */
+    if (numerr == siginter && *(b->buf) == 0) { handle_C_C = go_on = 1; break; }
   }
   if (old && !s) _set_analyseur(old);
   b = NULL; infile = oldinfile;
@@ -2737,9 +2632,9 @@ gp_exception_handler(long numerr)
 long
 setprecr(long n)
 {
-  long m = fmt.nb;
+  long m = fmt->sigd;
 
-  if (n>0) {fmt.nb = n; prec = (long)(n*pariK1 + 3);}
+  if (n>0) {fmt->sigd = n; prec = (long)(n*pariK1 + 3);}
   return m;
 }
 
@@ -2765,7 +2660,6 @@ read_opt(long argc, char **argv)
 {
   char *b=NULL, *p=NULL, *s=NULL, **pre;
   int i=1, initrc=1;
-  long TOP;
 
   pari_outfile=stderr;
   while (i<argc)
@@ -2801,10 +2695,9 @@ read_opt(long argc, char **argv)
   pre = initrc? gp_initrc(): NULL;
 
   /* override the values from gprc */
-  testint(b, &paribufsize); if (paribufsize < 10) paribufsize = 10;
-  testint(p, &primelimit);
-  TOP = (long)top;
-  testint(s, &TOP); top = (ulong)TOP;
+  testint(b, (long*)&paribufsize); if (paribufsize < 10) paribufsize = 10;
+  testint(p, (long*)&primelimit);
+  testint(s, (long*)&top);
   if (under_emacs || under_texmacs) disable_color=1;
   pari_outfile=stdout; return pre;
 }
@@ -2849,7 +2742,6 @@ main(int argc, char **argv)
 #endif
   gp_history_fun = gp_history;
   whatnow_fun = whatnow;
-  output_fun = gp_output;
   default_exception_handler = gp_exception_handler;
   gp_expand_path(path);
 
