@@ -1,4 +1,4 @@
-/* $Id: highlvl.c,v 1.10 2001/12/06 17:53:14 karim Exp $
+/* $Id: highlvl.c,v 1.12 2002/04/19 19:49:46 bill Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -28,7 +28,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #endif
 
 void kill0(entree *ep);
-long secure;
+ulong secure;
 
 #ifdef HAS_DLOPEN
 #include <dlfcn.h>
@@ -49,8 +49,14 @@ install0(char *name, char *code, char *gpname, char *lib)
 #endif
   if (! *gpname) gpname=name;
   if (lib) lib = expand_tilde(lib);
-  
+
+/* OSF1 has dlopen but not RTLD_GLOBAL*/
+#ifndef RTLD_GLOBAL
+#define RTLD_GLOBAL 0
+#endif
+
   handle = dlopen(lib,RTLD_LAZY|RTLD_GLOBAL);
+
   if (!handle)
   {
     const char *s = dlerror(); if (s) fprintferr("%s\n\n",s);
