@@ -1,4 +1,4 @@
-/* $Id: anal.c,v 1.79 2002/01/09 15:30:02 karim Exp $
+/* $Id: anal.c,v 1.96 2002/06/08 14:37:45 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -50,7 +50,6 @@ static entree *installep(void *f,char *name,int l,int v,int add,entree **table);
 static entree *skipentry(void);
 
 extern void killbloc0(GEN x, int inspect);
-extern char *GENtostr0(GEN x, void(*do_out)(GEN));
 
 /* last time we began parsing an object of specified type */
 static struct
@@ -113,7 +112,8 @@ static GEN br_res = NULL;
  *  or    .entry
  *
  *  truc:
- *      ! truc
+ *      ! facteur
+ *  or  # facteur
  *  or  ' entry
  *  or  identifier
  *  or  constante
@@ -183,7 +183,7 @@ _set_analyseur(char *s)
 static GEN
 lisseq0(char *t, GEN (*f)(void))
 {
-  const ulong av = avma;
+  const gpmem_t av = avma;
   char *olds = analyseur, *olde = mark.start;
   GEN res;
 
@@ -212,7 +212,7 @@ lisseq0(char *t, GEN (*f)(void))
 static GEN
 flisseq0(char *s, GEN (*f)(void))
 {
-  char *t = filtre(s,NULL, f_INIT | f_REG);
+  char *t = filtre(s, (compatible == OLDALL));
   GEN x = lisseq0(t, f);
   free(t); return x;
 }
@@ -441,7 +441,7 @@ kill0(entree *ep)
 static GEN
 seq(void)
 {
-  const ulong av = avma, lim = stack_lim(av,1);
+  const gpmem_t av = avma, lim = stack_lim(av,1);
   GEN res = gnil;
 
   for(;;)
@@ -471,7 +471,7 @@ gshift_r(GEN x, GEN n) { return gshift(x,-itos(n)); }
 static GEN
 expr(void)
 {
-  ulong av = avma, lim = stack_lim(av,2);
+  gpmem_t av = avma, lim = stack_lim(av, 2);
   GEN aux,e,e1,e2,e3;
   F2GEN F1,F2,F3;
   int F0 = 0;
@@ -613,7 +613,7 @@ err_match(char *s, char c)
 static long
 readlong()
 {
-  const ulong av = avma;
+  const gpmem_t av = avma;
   const char *old = analyseur;
   long m;
   GEN arg = expr();
@@ -695,10 +695,16 @@ matcell(GEN p, matcomp *C)
         c = check_array_index(lgef(p)-1) + 1;
         pt = (GEN*)(p + c); match(']'); break;
 
-      case t_VEC: case t_COL: case t_VECSMALL:
+      case t_VEC: case t_COL:
         c = check_array_index(lg(p));
         pt = (GEN*)(p + c); match(']'); break;
 
+      case t_VECSMALL:
+        c = check_array_index(lg(p));
+        pt = (GEN*)(p + c); match(']');
+        if (*analyseur == '[') err(caracer1,analyseur,mark.start);
+        break;
+
       case t_MAT:
         if (lg(p)==1) err(talker2,"a 0x0 matrix has no elements",
                                   analyseur,mark.start);
@@ -764,7 +770,7 @@ facteur(void)
       case '^':
 	analyseur++; p1 = facteur();
         if (br_status) err(breaker,"here (after ^)");
-        x = gpui(x,p1,prec); break;
+        x = gpow(x,p1,prec); break;
       case '\'':
 	analyseur++; x = deriv(x,gvar9(x)); break;
       case '~':
@@ -813,7 +819,7 @@ truc(void)
 
   if (*analyseur == '!') /* NOT */
   {
-    analyseur++; p1 = truc();
+    analyseur++; p1 = facteur();
     if (br_status) err(breaker,"here (after !)");
     return gcmp0(p1)? gun: gzero;
   }
@@ -832,7 +838,7 @@ truc(void)
   }
   if (*analyseur == '#') /* CARD */
   {
-    analyseur++; p1 = truc();
+    analyseur++; p1 = facteur();
     if (br_status) err(breaker,"here (after #)");
     return stoi(glength(p1));
   }
@@ -896,7 +902,7 @@ truc(void)
 static GEN
 double_op()
 {
-  static long mun[] = { evaltyp(t_INT) | m_evallg(3), 
+  static long mun[] = { evaltyp(t_INT) | _evallg(3), 
                         evalsigne(-1)|evallgefint(3), 1 };
   char c = *analyseur;
   if (c == analyseur[1])
@@ -1060,10 +1066,10 @@ expand_string(char *bp, char **ptbuf, char **ptlimit)
   }
   else
   {
-    long av = avma;
+    gpmem_t av = avma;
     GEN p1 = expr();
     if (br_status) err(breaker,"here (expanding string)");
-    tmp = GENtostr0(p1, output_fun);
+    tmp = GENtostr0(p1, &DFLT_OUTPUT, &gen_output);
     len = strlen(tmp); avma = av;
     alloc = 1;
   }
@@ -1165,7 +1171,7 @@ strtoGENstr_t()
   old++; /* skip '"' */
   n = (n+BYTES_IN_LONG) >> TWOPOTBYTES_IN_LONG;
   x = cgetg(n+1, t_STR);
-  translate(&old, GSTR(x), NULL,NULL);
+  (void)translate(&old, GSTR(x), NULL,NULL);
   return x;
 }
 
@@ -1263,7 +1269,7 @@ check_pointers(unsigned int ptrs, matcomp *init[])
         *pt = x;
       }
       else
-        change_compo(*c, x);
+        (void)change_compo(*c, x);
       free((void*)c);
     }
 }
@@ -1307,7 +1313,7 @@ check_args()
     if (*analyseur == '=')
     {
       char *old = ++analyseur;
-      ulong av = avma;
+      gpmem_t av = avma;
       skipexpr();
       cell[1] = lclone(_strtoGENstr(old, analyseur-old));
       avma = av;
@@ -1352,7 +1358,8 @@ static GEN
 num_deriv(void *call, GEN argvec[])
 {
   GEN eps,a,b, y, x = argvec[0];
-  long fpr,pr,l,e,ex, av = avma;
+  long fpr, pr, l, e, ex;
+  gpmem_t av = avma;
   if (!is_const_t(typ(x)))
   {
     a = do_call(call, x, argvec);
@@ -1366,7 +1373,7 @@ num_deriv(void *call, GEN argvec[])
   l = 2+pr;
   e = fpr * BITS_IN_HALFULONG; /* 1/2 required prec (in sig. bits) */
 
-  eps = realun(l); setexpo(eps, -e);
+  eps = real2n(-e, l);
   y = fix(gsub(x, eps), l); a = do_call(call, y, argvec);
   y = fix(gadd(x, eps), l); b = do_call(call, y, argvec);
   setexpo(eps, e-1);
@@ -1378,7 +1385,8 @@ static GEN
 num_derivU(GEN p, GEN *arg, GEN *loc, int narg, int nloc)
 {
   GEN eps,a,b, x = *arg;
-  long fpr,pr,l,e,ex, av = avma;
+  long fpr, pr, l, e, ex;
+  gpmem_t av = avma;
 
   if (!is_const_t(typ(x)))
   {
@@ -1393,7 +1401,7 @@ num_derivU(GEN p, GEN *arg, GEN *loc, int narg, int nloc)
   l = 2+pr;
   e = fpr * BITS_IN_HALFULONG; /* 1/2 required prec (in sig. bits) */
 
-  eps = realun(l); setexpo(eps, -e);
+  eps = real2n(-e, l);
   *arg = fix(gsub(x, eps), l); a = call_fun(p,arg,loc,narg,nloc);
   *arg = fix(gadd(x, eps), l); b = call_fun(p,arg,loc,narg,nloc);
   setexpo(eps, e-1);
@@ -1408,7 +1416,8 @@ num_derivU(GEN p, GEN *arg, GEN *loc, int narg, int nloc)
 static GEN
 identifier(void)
 {
-  long m,i,av,matchcomma, deriv;
+  long m, i, matchcomma, deriv;
+  gpmem_t av;
   char *ch1;
   entree *ep;
   GEN res, newfun, ptr;
@@ -1757,7 +1766,7 @@ identifier(void)
           analyseur=ch1; ep = entry();
           if (*analyseur == '=')
           {
-            long av=avma; analyseur++;
+            gpmem_t av=avma; analyseur++;
             res = expr();
             if (br_status) err(breaker,"here (defining global var)");
             changevalue(ep, res); avma=av;
@@ -1849,7 +1858,12 @@ identifier(void)
       /* checking arguments */
       match('('); ch1 = analyseur;
       narg = check_args(); nloc = 0;
-      match(')'); match('=');
+      match(')'); 
+      /* Dirty, but don't want to define a local() function */
+      if (*analyseur != '=' && strcmp(ep->name, "local") == 0)
+        err(talker2, "local() bloc must appear before any other expression",
+                     mark.identifier,mark.start);
+      match('=');
       while (strncmp(analyseur,"local(",6) == 0)
       {
         analyseur += 6;
@@ -1900,7 +1914,7 @@ identifier(void)
       ep->value = (void *)ptr;
       ep->valence = EpUSER;
       check_new_fun=NULL;
-      avma = (long)tmpargs; return gnil;
+      avma = (gpmem_t)tmpargs; return gnil;
     }
   }
   err(valencer1); return NULL; /* not reached */
@@ -1922,7 +1936,8 @@ constante()
 {
   static long pw10[] = { 1, 10, 100, 1000, 10000, 100000, 1000000,
                         10000000, 100000000, 1000000000 };
-  long i,l,m,n = 0,nb, av = avma;
+  long i, l, m, n = 0, nb;
+  gpmem_t av = avma;
   GEN z,y;
 
   y = stoi(number(&nb)); i = 0;
@@ -1971,13 +1986,13 @@ constante()
   l=lgefint(y); if (l<prec) l=prec;
   if (n)
   {
-    new_chunk(l); /* hack: mulrr and divrr need exactly l words */
-    z=cgetr(l); affir(y,z);
-    y=cgetr(l); affsr(10,y); y = gpuigs(y, labs(n));
+    (void)new_chunk(l); /* HACK: mulrr and divrr need exactly l words */
+    z = itor(y, l);
+    y = gpowgs(stor(10,l), labs(n));
     avma = av; /* hidden gerepile */
     return n > 0 ?  mulrr(z,y) : divrr(z,y);
   }
-  z=cgetr(l); affir(y,z); return z;
+  return itor(y, l);
 }
 
 /********************************************************************/
@@ -2138,14 +2153,14 @@ fetch_named_var(char *s, int doerr)
   }
   ep = installep(NULL,s,strlen(s),EpVAR, 7*sizeof(long),
                  functions_hash + hashvalue(s));
-  manage_var(0,ep); return ep;
+  (void)manage_var(0,ep); return ep;
 }
 
 long
 fetch_user_var(char *s)
 {
   entree *ep = is_entry(s);
-  long av;
+  gpmem_t av;
   GEN p1;
 
   if (ep)
@@ -2164,7 +2179,7 @@ fetch_user_var(char *s)
 void
 delete_named_var(entree *ep)
 {
-  manage_var(5, (entree*)varn(initial_value(ep)));
+  (void)manage_var(5, (entree*)varn(initial_value(ep)));
   kill0(ep);
 }
 
@@ -2216,7 +2231,7 @@ entry(void)
     { n=7*sizeof(long); val=EpVAR; }
   ep = installep(NULL,old,len,val,n, functions_hash + hash);
 
-  if (n) manage_var(0,ep); /* Variable */
+  if (n) (void)manage_var(0,ep); /* Variable */
   return ep;
 }
 
@@ -2225,6 +2240,7 @@ entry(void)
 /**                          SKIP FUNCTIONS                        **/
 /**                                                                **/
 /********************************************************************/
+extern int term_width(void);
 
 /* as skipseq without modifying analyseur && al */
 static void
@@ -2413,10 +2429,10 @@ skiptruc(void)
   switch(*analyseur)
   {
     case '"': skipstring(); return;
-    case '!': case '#': analyseur++; skiptruc(); return;
+    case '!': case '#': analyseur++; skipfacteur(); return;
     case '&': case '\'':
       analyseur++; check_var_name();
-      skipentry(); return;
+      (void)skipentry(); return;
   }
   if (isalpha((int)*analyseur)) { skipidentifier(); return; }
   if (isdigit((int)*analyseur) || *analyseur== '.') { skipconstante(); return; }
@@ -2451,7 +2467,7 @@ skiptruc(void)
       }
     case '%':
       if (*analyseur == '`') { while (*++analyseur == '`') /*empty*/; return; }
-      number(&n); return;
+      (void)number(&n); return;
   }
   err(caracer1,analyseur-1,mark.start);
 }
@@ -2550,7 +2566,7 @@ skipidentifier(void)
         break;
 
       case 'S': match_comma();
-        check_var_name(); skipentry(); break;
+        check_var_name(); (void)skipentry(); break;
       case '&': match_comma(); match('&'); check_matcell(); break;
       case 'V': match_comma(); check_var(); break;
 
@@ -2731,7 +2747,7 @@ f(GEN x)
 static GEN
 p(GEN x)
 {
-  int t; get_nf(x,&t);
+  int t; (void)get_nf(x,&t);
   if (t == typ_GAL)
     return gmael(x,2,1);
   x = get_primeid(x);
@@ -2815,7 +2831,7 @@ pol(GEN x) /* polynomial */
 static GEN
 mod(GEN x) /* modulus */
 {
-  int t; get_nf(x,&t);
+  int t; (void)get_nf(x,&t);
   if (t == typ_GAL)
     return gmael(x,2,3);
   switch(typ(x))
@@ -2854,7 +2870,7 @@ t2(GEN x) /* T2 matrix */
 {
   int t; x = nfmats(get_nf(x,&t));
   if (!x) err(member,"t2",mark.member,mark.start);
-  return (GEN)x[3];
+  return gram_matrix((GEN)x[2]);
 }
 
 static GEN
@@ -3066,7 +3082,7 @@ gen(GEN x)
     x[1] = y[1];
     x[2] = y[2]; return x;
   }
-  get_nf(x,&t);
+  (void)get_nf(x,&t);
   if (t == typ_GAL)
     return (GEN)x[7];
   x = clgp(x);
@@ -3078,8 +3094,7 @@ gen(GEN x)
 static GEN
 group(GEN x)
 {
-  int t;
-  get_nf(x,&t);
+  int t; (void)get_nf(x,&t);
   if (t == typ_GAL)
     return (GEN)x[6];
   err(member,"group",mark.member,mark.start);
@@ -3088,8 +3103,7 @@ group(GEN x)
 static GEN
 orders(GEN x)
 {
-  int t;
-  get_nf(x,&t);
+  int t; (void)get_nf(x,&t);
   if (t == typ_GAL)
     return (GEN)x[8];
   err(member,"orders",mark.member,mark.start);
@@ -3400,7 +3414,7 @@ alias0(char *s, char *old)
   ep = do_alias(ep); x = newbloc(2);
   x[0] = evaltyp(t_STR)|evallg(2); /* for getheap */
   x[1] = (long)ep;
-  installep(x, s, strlen(s), EpALIAS, 0, functions_hash + hash);
+  (void)installep(x, s, strlen(s), EpALIAS, 0, functions_hash + hash);
 }
 
 extern void err_leave_default(long n);
@@ -3409,7 +3423,8 @@ extern void err_leave_default(long n);
 GEN
 trap0(char *e, char *r, char *f)
 {
-  VOLATILE long av = avma, numerr = -1;
+  VOLATILE gpmem_t av = avma;
+  VOLATILE long numerr = -1;
   VOLATILE GEN x = gnil;
   char *F;
        if (!strcmp(e,"errpile")) numerr = errpile;
