@@ -1,4 +1,4 @@
-/* $Id: anal.h,v 1.18 2001/12/04 21:29:45 karim Exp $
+/* $Id: anal.h,v 1.24 2002/04/27 18:42:13 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -73,13 +73,12 @@ void hit_return(void);
 void push_val(entree *ep, GEN a);
 void pop_val(entree *ep);
 
-extern long prec, secure;
+extern ulong prec;
 extern GEN gnil;
 
 extern char *current_function;
 extern GEN  (*gp_history_fun)(long, long, char *, char *);
 extern int  (*whatnow_fun)(char *, int);
-extern void (*output_fun)(GEN);
 extern void *foreignHandler;
 extern GEN  (*foreignExprHandler)(char*);
 extern char foreignExprSwitch;
@@ -111,7 +110,7 @@ extern char   *oldhelpmessage[], *helpmessages_oldgp[];
 extern entree  oldfonctions[], functions_oldgp[];
 
 /* backward compatibility */
-extern long compatible;
+extern ulong compatible;
 enum { NONE, WARN, OLDFUN, OLDALL };
 #define new_fun_set (compatible == NONE || compatible == WARN)
 
@@ -195,8 +194,51 @@ enum { c_ERR, c_HIST, c_PROMPT, c_INPUT, c_OUTPUT, c_HELP, c_TIME, c_LAST };
 #define mf_TEST 32
 
 /* for filtre */
-#define f_COMMENT  0
-#define f_INIT     1
-#define f_KEEPCASE 2
-#define f_REG      4
-#define f_ENDFILE  8
+typedef struct {
+  char *s, *t, *end; /* source, target, last char read */
+  int in_string, in_comment, more_input, wait_for_brace, downcase;
+  void *data;
+} filtre_t;
+
+#define LBRACE '{'
+#define RBRACE '}'
+
+extern char *filtre0(filtre_t *F);
+extern char *filtre(char *s, int flag);
+extern void check_filtre(filtre_t *F);
+
+typedef struct Buffer {
+  char *buf;
+  ulong len;
+  jmp_buf env;
+  int flenv;
+} Buffer;
+
+typedef struct input_method {
+  int free;
+  char *prompt;
+  FILE *file;
+  char * (*fgets)(char *,int,FILE*);
+  char * (*getline)(Buffer*, char**, struct input_method*);
+} input_method;
+
+/* output format */
+enum { f_RAW, f_PRETTYMAT, f_PRETTYOLD, f_PRETTY, f_TEX, f_NOEOL = 16 };
+
+/* for output */
+typedef struct {
+  char format; /* e,f,g */
+  long fieldw; /* 0 (ignored) or field width */
+  long sigd; /* -1 (all) or number of sign. digits printed */
+  int sp;
+  int initial;
+  int prettyp;
+} pariout_t;
+
+extern void gen_output(GEN x, pariout_t *T);
+extern char *GENtostr0(GEN x, pariout_t *T, void(*do_out)(GEN, pariout_t *));
+extern void bruti(GEN g, pariout_t *T, int nosign);
+extern void matbruti(GEN g, pariout_t *T);
+extern void sori(GEN g, pariout_t *T);
+extern void texi(GEN g, pariout_t *T, int nosign);
+extern pariout_t DFLT_OUTPUT;
