@@ -1,4 +1,4 @@
-/* $Id: compat.c,v 1.17 2001/12/05 19:37:14 karim Exp $
+/* $Id: compat.c,v 1.26 2002/05/12 16:58:52 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -38,12 +38,13 @@ subres0(GEN x, GEN y) { return subres(x,y); }
 static long
 rounderror(GEN x)
 {
-  long e, av=avma;
-  grndtoi(x,&e); avma=av; return (long)(e*L2SL10);
+  gpmem_t av = avma;
+  long e; (void)grndtoi(x,&e);
+  avma = av; return (long)(e*L2SL10);
 }
 
 static GEN
-_factpol(GEN x, long unused, long hint) { return factpol(x,hint); }
+_factpol(GEN x, long t/*unused*/, long hint) { return factpol(x,hint); }
 
 static void
 suppressed(void) {err(talker,"this function has been suppressed");}
@@ -179,8 +180,8 @@ entree oldfonctions[]={
 {"factor",1,(void *)factor,4,"G"},
 {"factoredbasis",28,(void *)factoredbase,6,"GGf"},
 {"factoreddiscf",2,(void *)factoreddiscf,6,"GG"},
-{"factoredpolred",2,(void *)factoredpolred,6,"GGp"},
-{"factoredpolred2",2,(void *)factoredpolred2,6,"GGp"},
+{"factoredpolred",2,(void *)factoredpolred,6,"GG"},
+{"factoredpolred2",2,(void *)factoredpolred2,6,"GG"},
 {"factornf",2,(void *)polfnf,6,"GG"},
 {"factorpadic",32,(void *)factorpadic4,7,"GGL"},
 {"factorpadic2",32,(void *)factorpadic2,7,"GGL"},
@@ -318,7 +319,7 @@ entree oldfonctions[]={
 {"keri",1,(void *)keri,8,"G"},
 {"kerint",1,(void *)kerint,8,"G"},
 {"kerint1",1,(void *)kerint1,8,"G"},
-{"kerint2",1,(void *)kerint2,8,"G"},
+{"kerint2",1,(void *)suppressed,8,"G"},
 {"kro",2,(void *)gkronecker,4,"GG"},
 {"label",0,(void*)suppressed,11,"s*"},
 {"lambdak",2,(void *)glambdak,6,"GGp"},
@@ -332,10 +333,10 @@ entree oldfonctions[]={
 {"lindep",1,(void *)lindep,8,"Gp"},
 {"lindep2",23,(void *)lindep2,8,"GLp"},
 {"lll",1,(void *)lll,8,"Gp"},
-{"lll1",1,(void *)lll1,8,"Gp"},
+{"lll1",1,(void *)suppressed,8,"Gp"},
 {"lllgen",1,(void *)lllgen,8,"Gp"},
 {"lllgram",1,(void *)lllgram,8,"Gp"},
-{"lllgram1",1,(void *)lllgram1,8,"Gp"},
+{"lllgram1",1,(void *)suppressed,8,"Gp"},
 {"lllgramgen",1,(void *)lllgramgen,8,"G"},
 {"lllgramint",1,(void *)lllgramint,8,"G"},
 {"lllgramkerim",1,(void *)lllgramkerim,8,"G"},
@@ -344,7 +345,7 @@ entree oldfonctions[]={
 {"lllintpartial",1,(void *)lllintpartial,8,"G"},
 {"lllkerim",1,(void *)lllkerim,8,"G"},
 {"lllkerimgen",1,(void *)lllkerimgen,8,"G"},
-{"lllrat",1,(void *)lllrat,8,"G"},
+{"lllrat",1,(void *)suppressed,8,"G"},
 {"ln",1,(void *)glog,3,"Gp"},
 {"lngamma",1,(void *)glngamma,3,"Gp"},
 {"localred",2,(void *)localreduction,5,"GG"},
@@ -407,13 +408,13 @@ entree oldfonctions[]={
 {"pnqn",1,(void *)pnqn,4,"G"},
 {"pointell",2,(void *)pointell,5,"GGp"},
 {"polint",31,(void *)polint,7,"GGGD&"},
-{"polred",1,(void *)polred,6,"Gp"},
-{"polred2",1,(void *)polred2,6,"Gp"},
-{"polredabs",1,(void *)polredabs,6,"Gp"},
-{"polredabs2",1,(void *)polredabs2,6,"Gp"},
-{"polredabsall",1,(void *)polredabsall,6,"Gp"},
+{"polred",1,(void *)polred,6,"G"},
+{"polred2",1,(void *)polred2,6,"G"},
+{"polredabs",1,(void *)polredabs,6,"G"},
+{"polredabs2",1,(void *)polredabs2,6,"G"},
+{"polredabsall",1,(void *)polredabsall,6,"G"},
 {"polredabsfast",0,(void *)suppressed,6,""},
-{"polredabsnored",1,(void *)polredabsnored,6,"Gp"},
+{"polredabsnored",1,(void *)polredabs,6,"G"},
 {"polsym",21,(void *)polsym,7,"GL"},
 {"polvar",1,(void *)gpolvar,2,"G"},
 {"poly",14,(void *)gtopoly,2,"Gn"},
@@ -503,8 +504,8 @@ entree oldfonctions[]={
 {"smalldiscf",1,(void *)smalldiscf,6,"G"},
 {"smallfact",1,(void *)smallfact,4,"G"},
 {"smallinitell",1,(void *)smallinitell,5,"Gp"},
-{"smallpolred",1,(void *)smallpolred,6,"Gp"},
-{"smallpolred2",1,(void *)smallpolred2,6,"Gp"},
+{"smallpolred",1,(void *)smallpolred,6,"G"},
+{"smallpolred2",1,(void *)smallpolred2,6,"G"},
 {"smith",1,(void *)smith,8,"G"},
 {"smith2",1,(void *)smith2,8,"G"},
 {"smithclean",1,(void *)smithclean,8,"G"},
@@ -517,7 +518,7 @@ entree oldfonctions[]={
 {"srgcd",2,(void *)srgcd,7,"GG"},
 {"sturm",10,(void *)sturm0,7,"lG"},
 {"sturmpart",30,(void *)sturmpart,7,"lGGG"},
-{"subcyclo",2,(void *)subcyclo,6,"GGDn"},
+{"subcyclo",2,(void *)subcyclo,6,"LLDn"},
 {"subell",3,(void *)subell,5,"GGGp"},
 {"subst",26,(void *)gsubst,7,"GnG"},
 {"sum",48,(void *)sum0,9,"GV=GGI"},
