@@ -1,4 +1,4 @@
-/* $Id: init.c,v 1.119 2002/01/09 12:04:52 karim Exp $
+/* $Id: init.c,v 1.149 2002/06/09 22:45:47 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -45,10 +45,11 @@ entree  **varentries;
 void    *global_err_data;
 jmp_buf environnement;
 long    *ordvar;
-long    DEBUGFILES,DEBUGLEVEL,DEBUGMEM,compatible;
-long    prec,precdl;
+ulong   DEBUGFILES,DEBUGLEVEL,DEBUGMEM,compatible;
+ulong   prec,precdl;
 ulong   init_opts = INIT_JMPm | INIT_SIGm;
-ulong   bot = 0, top = 0, avma, memused;
+gpmem_t bot = 0, top = 0, avma;
+size_t memused;
 
 void *foreignHandler; 	              /* Handler for foreign commands.   */
 char foreignExprSwitch = 3; 	      /* Just some unprobable char.      */
@@ -59,11 +60,20 @@ void (*foreignFuncFree)(entree *);    /* How to free external entree.    */
 int  (*default_exception_handler)(long);
 GEN  (*gp_history_fun)(long, long, char *, char *);
 int  (*whatnow_fun)(char *, int);
-void (*output_fun)(GEN);
+pariout_t DFLT_OUTPUT = { 'g', 0, -1, 1, 0, f_RAW };
 
 extern void  initout(int initerr);
 extern int   term_width(void);
 
+#ifdef BOTH_GNUPLOT_AND_X11
+/* Satisfy DLL dependencies: dummy only */
+#define EXTERM_DLL_DPES *PL_markstack_ptr, PL_stack_max, *PL_Sv, *PL_stack_sp, \
+  *PL_tmps_floor, *PL_tmps_ix, *PL_markstack_max, *PL_stack_base, *PL_na, \
+  *PL_sv_yes, *PL_sv_no, *PL_curpad, *PL_op
+extern int EXTERM_DLL_DPES;
+int EXTERM_DLL_DPES;
+#endif	/* defined BOTH_GNUPLOT_AND_X11 */ 
+
 typedef struct cell {
   void *env;
   void *data;
@@ -183,7 +193,7 @@ static void
 pari_sighandler(int sig)
 {
   char *msg;
-  os_signal(sig,pari_sighandler);
+  (void)os_signal(sig,pari_sighandler);
   switch(sig)
   {
 #ifdef SIGBREAK
@@ -447,43 +457,43 @@ init_universal_constants(void)
 }
 
 static long
-fix_size(long a)
+fix_size(size_t a)
 {
   /* BYTES_IN_LONG*ceil(a/BYTES_IN_LONG) */
-  ulong b = a+BYTES_IN_LONG - (((a-1) & (BYTES_IN_LONG-1)) + 1);
+  size_t b = a+BYTES_IN_LONG - (((a-1) & (BYTES_IN_LONG-1)) + 1);
   if (b > VERYBIGINT) err(talker,"stack too large");
   if (b < 1024) b = 1024;
   return b;
 }
 
 static long
-init_stack(long size)
+init_stack(size_t size)
 {
-  long s = fix_size(size), old = 0;
+  size_t s = fix_size(size), old = 0;
   if (bot)
   {
     old = top - bot;
     free((void*)bot);
   }
   /* NOT gpmalloc, memer would be deadly */
-  bot = (ulong)__gpmalloc(s);
+  bot = (gpmem_t)__gpmalloc(s);
   if (!bot)
-    for(s = old;; s>>=1)
+    for (s = old;; s>>=1)
     {
       if (!s) err(memer); /* no way out. Die */
       err(warner,"not enough memory, new stack %lu",s);
-      bot = (ulong)__gpmalloc(s);
+      bot = (gpmem_t)__gpmalloc(s);
       if (bot) break;
     }
-  memused = avma = top = bot+s;
-  return s;
+  avma = top = bot+s;
+  memused = 0; return s;
 }
 
 /* initialise les donnees de la bibliotheque PARI. Peut tre prcde d'un
  * appel  pari_addfunctions si on ajoute d'autres fonctions au pool de base.
  */
 void
-pari_init(long parisize, long maxprime)
+pari_init(size_t parisize, long maxprime)
 {
   long i;
 
@@ -509,7 +519,7 @@ pari_init(long parisize, long maxprime)
   polvar[0] = evaltyp(t_VEC) | evallg(1);
   for (i=0; i <= MAXVARN; i++) { ordvar[i] = i; varentries[i] = NULL; }
 
-  fetch_var(); /* create polx/polun[MAXVARN] */
+  (void)fetch_var(); /* create polx/polun[MAXVARN] */
   primetab = (GEN) gpmalloc(1 * sizeof(long));
   primetab[0] = evaltyp(t_VEC) | evallg(1);
 
@@ -534,14 +544,12 @@ pari_init(long parisize, long maxprime)
 
   gp_history_fun = NULL;
   whatnow_fun = NULL;
-  output_fun = &outbrute;
   err_catch_array = (long *) gpmalloc((noer + 1) *sizeof(long));
   reset_traps(0);
   default_exception_handler = NULL;
 
   (void)manage_var(2,NULL); /* init nvar */
-  (void)get_timer(-1); /* init timers */
-  var_not_changed = 1; fetch_named_var("x", 0);
+  var_not_changed = 1; (void)fetch_named_var("x", 0);
   try_to_recover=1;
 }
 
@@ -577,7 +585,7 @@ freeall(void)
   free(current_psfile);
 
   if (gp_history_fun)
-    gp_history_fun(0,-1,NULL,NULL);
+    (void)gp_history_fun(0,-1,NULL,NULL);
 }
 
 GEN
@@ -698,7 +706,8 @@ gunclone(GEN x) { killbloc0(x,0); }
 GEN
 changevar(GEN x, GEN y)
 {
-  long tx,ty,lx,vx,vy,i,av,tetpil;
+  long tx, ty, lx, vx, vy, i;
+  gpmem_t av, tetpil;
   GEN  p1,p2,z;
 
   if (var_not_changed && y==polvar) return x;
@@ -863,7 +872,7 @@ recover(int flag)
   }
 #endif
   try_to_recover=1;
-  os_signal(SIGINT, sigfun);
+  (void)os_signal(SIGINT, sigfun);
 }
 
 void
@@ -1031,7 +1040,6 @@ err_recover(long numerr)
 {
   initout(0);
   disable_dbg(-1);
-  get_timer(-1);
   killallfiles(0);
 
   if (pariErr->die) pariErr->die();    /* Caller wants to catch exceptions? */
@@ -1284,7 +1292,7 @@ gcopy_av(GEN x, GEN *AVMA)
   long i,lx,tx=typ(x);
   GEN y;
 
-  if (tx == t_SMALL) return (*--AVMA = x);
+  if (tx == t_SMALL) return x;
   lx = lg(x); *AVMA = y = *AVMA - lx;
   if (! is_recursive_t(tx))
   {
@@ -1309,7 +1317,7 @@ gcopy_av0(GEN x, GEN *AVMA)
   if (! is_recursive_t(tx))
   {
     if (tx == t_INT && !signe(x)) return NULL; /* special marker */
-    if (tx == t_SMALL) return (*--AVMA = x);
+    if (tx == t_SMALL) return x;
     lx = lg(x); *AVMA = y = *AVMA - lx;
     for (i=0; i<lx; i++) y[i] = x[i];
   }
@@ -1417,7 +1425,7 @@ bin_copy(GENbin *p)
   GEN x,y,base;
   long dx,len;
 
-  x   = p->x; if (!x) return gzero;
+  x   = p->x; if (!x) { free(p); return gzero; }
   len = p->len;
   base= p->base; dx = x - base;
   y = (GEN)memcpy((void*)new_chunk(len), (void*)GENbase(p), len*sizeof(long));
@@ -1442,7 +1450,7 @@ stackdummy(GEN z, long l) { z[0] = evaltyp(t_VECSMALL) | evallg(l); }
 
 /* gerepileupto(av, forcecopy(x)) */
 GEN
-gerepilecopy(long av, GEN x)
+gerepilecopy(gpmem_t av, GEN x)
 {
   GENbin *p = copy_bin(x);
   avma = av; return bin_copy(p);
@@ -1452,29 +1460,43 @@ gerepilecopy(long av, GEN x)
  * objects to contiguous locations and cleans up the stack between
  * av and avma. */
 void
-gerepilemany(long av, GEN* gptr[], long n)
+gerepilemany(gpmem_t av, GEN* gptr[], int n)
 {
   GENbin **l = (GENbin**)gpmalloc(n*sizeof(GENbin*));
-  long i;
+  int i;
   for (i=0; i<n; i++) l[i] = copy_bin(*(gptr[i]));
   avma = av;
   for (i=0; i<n; i++) *(gptr[i]) = bin_copy(l[i]);
   free(l);
 }
 
-void
-gerepilemanycoeffs(long av, GEN x, long n)
+void 
+gerepileall(gpmem_t av, int n, ...)
 {
-  long i;
+  GENbin **l = (GENbin**)gpmalloc(n*sizeof(GENbin*));
+  GEN **gptr  = (GEN**)  gpmalloc(n*sizeof(GEN*));
+  int i;
+  va_list a; va_start(a, n);
+
+  for (i=0; i<n; i++) { gptr[i] = va_arg(a,GEN*); l[i] = copy_bin(*(gptr[i])); }
+  avma = av;
+  for (--i; i>=0; i--) *(gptr[i]) = bin_copy(l[i]);
+  free(l); free(gptr);
+}
+
+void
+gerepilemanycoeffs(gpmem_t av, GEN x, int n)
+{
+  int i;
   for (i=0; i<n; i++) x[i] = (long)copy_bin((GEN)x[i]);
   avma = av;
   for (i=0; i<n; i++) x[i] = (long)bin_copy((GENbin*)x[i]);
 }
 
 void
-gerepilemanycoeffs2(long av, GEN x, long n, GEN y, long o)
+gerepilemanycoeffs2(gpmem_t av, GEN x, int n, GEN y, int o)
 {
-  long i;
+  int i;
   for (i=0; i<n; i++) x[i] = (long)copy_bin((GEN)x[i]);
   for (i=0; i<o; i++) y[i] = (long)copy_bin((GEN)y[i]);
   avma = av;
@@ -1485,19 +1507,20 @@ gerepilemanycoeffs2(long av, GEN x, long n, GEN y, long o)
 /* Takes an array of pointers to GENs, of length n.
  * Cleans up the stack between av and tetpil, updating those GENs. */
 void
-gerepilemanysp(long av, long tetpil, GEN* gptr[], long n)
+gerepilemanysp(gpmem_t av, gpmem_t tetpil, GEN* gptr[], int n)
 {
-  const long av2 = avma, dec = av-tetpil;
-  long i;
+  const gpmem_t av2 = avma;
+  const size_t dec = av-tetpil;
+  int i;
 
   (void)gerepile(av,tetpil,NULL);
   for (i=0; i<n; i++)
   {
-    ulong *g1 = (ulong*) gptr[i];
-    if (*g1 < (ulong)tetpil)
+    gpmem_t *g1 = (gpmem_t*) gptr[i];
+    if (*g1 < tetpil)
     {
-      if (*g1 >= (ulong)av2) *g1 += dec; /* Update address if in stack */
-      else if (*g1 >=(ulong)av) err(gerper);
+      if (*g1 >= av2) *g1 += dec; /* Update address if in stack */
+      else if (*g1 >= av) err(gerper);
     }
   }
 }
@@ -1505,26 +1528,27 @@ gerepilemanysp(long av, long tetpil, GEN* gptr[], long n)
 /* Takes an array of GENs (cast to longs), of length n.
  * Cleans up the stack between av and tetpil, updating those GENs. */
 void
-gerepilemanyvec(long av, long tetpil, long *g, long n)
+gerepilemanyvec(gpmem_t av, gpmem_t tetpil, long *g, int n)
 {
-  const long av2 = avma, dec = av-tetpil;
-  long i;
+  const gpmem_t av2 = avma;
+  const size_t dec = av-tetpil;
+  int i;
 
   (void)gerepile(av,tetpil,NULL);
   for (i=0; i<n; i++,g++)
-    if ((ulong)*g < (ulong)tetpil)
+    if ((gpmem_t)*g < tetpil)
     {
-      if ((ulong)*g >= (ulong)av2) *g += dec;/* Update addresses if in stack */
-      else if ((ulong)*g >= (ulong)av) err(gerper);
+      if ((gpmem_t)*g >= av2) *g += dec;/* Update addresses if in stack */
+      else if ((gpmem_t)*g >= av) err(gerper);
     }
 }
 
 GEN
-gerepileupto(long av, GEN q)
+gerepileupto(gpmem_t av, GEN q)
 {
-  if (!isonstack(q)) { avma=av; return q; } /* universal object */
+  if (!isonstack(q)) { avma = av; return q; } /* universal object */
   /* empty garbage */
-  if ((ulong)av <= (ulong)q) return q;
+  if (av <= (gpmem_t)q) return q;
   /* The garbage is only empty when av==q. It's probably a mistake if
    * av < q. But "temporary variables" from sumiter are a problem since
    * ep->values are returned as-is by identifier() and they can be in the
@@ -1533,27 +1557,27 @@ gerepileupto(long av, GEN q)
    */
 
   /* Beware: (long)(q+i) --> ((long)q)+i*sizeof(long) */
-  return gerepile(av, (long) (q+lg(q)), q);
+  return gerepile(av, (gpmem_t) (q+lg(q)), q);
 }
 
 /* internal */
 GEN
-gerepileuptoleaf(long av, GEN q)
+gerepileuptoleaf(gpmem_t av, GEN q)
 {
   long i;
   GEN q0;
 
-  if (!isonstack(q) || av==(long)q) { avma=av; return q; }
-  i=lg(q); avma = (long)(((GEN)av) -  i);
+  if (!isonstack(q) || av==(gpmem_t)q) { avma = av; return q; }
+  i=lg(q); avma = (gpmem_t)(((GEN)av) -  i);
   q0 = (GEN)avma; while (--i >= 0) q0[i]=q[i];
   return q0;
 }
 /* internal */
 GEN
-gerepileuptoint(long av, GEN q)
+gerepileuptoint(gpmem_t av, GEN q)
 {
-  if (!isonstack(q) || av==(long)q) { avma=av; return q; }
-  avma = (long)icopy_av(q, (GEN)av);
+  if (!isonstack(q) || (GEN)av==q) { avma = av; return q; }
+  avma = (gpmem_t)icopy_av(q, (GEN)av);
   return (GEN)avma;
 }
 
@@ -1585,19 +1609,20 @@ int
 ok_gerepileupto(GEN x) { return _ok_gerepileupto(x, x); }
 
 GEN
-gerepile(long av, long tetpil, GEN q)
+gerepile(gpmem_t av, gpmem_t tetpil, GEN q)
 {
-  long avmb, dec = av - tetpil;
+  gpmem_t avmb;
+  size_t dec = av - tetpil;
   GEN ll,a,b;
 
   if (dec==0) return q;
-  if (dec<0) err(talker,"lbot>ltop in gerepile");
+  if ((long)dec<0) err(talker,"lbot>ltop in gerepile");
 
-  if ((ulong)q>=(ulong)avma && (ulong)q<(ulong)tetpil)
-    q = (GEN) (((long)q) + dec);
+  if ((gpmem_t)q >= avma && (gpmem_t)q < tetpil)
+    q = (GEN) (((gpmem_t)q) + dec);
 
   for (ll=(GEN)av, a=(GEN)tetpil; a > (GEN)avma; ) *--ll= *--a;
-  avmb = (long)ll;
+  avmb = (gpmem_t)ll;
   while (ll < (GEN)av)
   {
     const long tl=typ(ll);
@@ -1606,16 +1631,16 @@ gerepile(long av, long tetpil, GEN q)
     a = ll+lontyp[tl];
     if (tl==t_POL) { b=ll+lgef(ll); ll+=lg(ll); } else { ll+=lg(ll); b=ll; }
     for (  ; a<b; a++)
-      if ((ulong)*a < (ulong)av && (ulong)*a >= (ulong)avma)
+      if ((gpmem_t)*a < av && (gpmem_t)*a >= avma)
       {
-	if ((ulong)*a < (ulong)tetpil) *a += dec; else err(gerper);
+	if ((gpmem_t)*a < tetpil) *a += dec; else err(gerper);
       }
   }
   avma = avmb; return q;
 }
 
 long
-allocatemoremem(ulong newsize)
+allocatemoremem(size_t newsize)
 {
   if (!newsize)
   {
@@ -1632,9 +1657,9 @@ switch_stack(stackzone *z, long n)
 {
   if (!z)
   { /* create parallel stack */
-    long size = n*sizeof(long) + sizeof(stackzone);
+    size_t size = n*sizeof(long) + sizeof(stackzone);
     z = (stackzone*) gpmalloc(size);
-    z->zonetop = ((long)z) + size;
+    z->zonetop = ((gpmem_t)z) + size;
     return z;
   }
 
@@ -1644,10 +1669,10 @@ switch_stack(stackzone *z, long n)
     z->top     = top;
     z->avma    = avma;
     z->memused = memused;
-    bot     = (long) (z+1);
+    bot     = (gpmem_t) (z+1);
     top     = z->zonetop;
     avma    = top;
-    memused = (ulong)-1;
+    memused = (size_t)-1;
   }
   else
   { /* back to normalcy */
@@ -1663,10 +1688,10 @@ switch_stack(stackzone *z, long n)
 void
 checkmemory(GEN z)
 {
-  if (DEBUGMEM && memused != (ulong)-1 &&
+  if (DEBUGMEM && memused != (size_t)-1 &&
        ((GEN)memused > z + MEMSTEP || z > (GEN)memused + MEMSTEP))
   {
-    memused=(ulong)z;
+    memused=(size_t)z;
 #if MEMSTEP >= 1048576
     fprintferr("...%4.0lf Mbytes used\n",(top-memused)/1048576.);
 #else
@@ -1688,149 +1713,99 @@ fill_stack(void)
 /*                               TIMER                             */
 /*                                                                 */
 /*******************************************************************/
-#define MAX_TIMER 32
-#define MIN_TIMER 3
+long
+_get_time(pari_timer *T, long Ticks, long TickPerSecond)
+{
+  long s  = Ticks / TickPerSecond;
+  long us = (long) (Ticks % TickPerSecond) * (1000000. / TickPerSecond);
+  long delay = 1000 * (s - T->s) + (us - T->us) / 1000;
+  T->us = us;
+  T->s  = s; return delay;
+}
 
 #ifdef WINCE
-  static long
-  timer_proto(int i)
-  {
-    static DWORD oldticks[MAX_TIMER];
-    DWORD ticks = GetTickCount();
-    DWORD delay = ticks - oldticks[i];
-    oldticks[i] = ticks;
-    return delay;
-  }
+long
+TIMER(pari_timer *T)
+{
+  return _get_time(T, GetTickCount(), 1000);
+}
 #elif defined(macintosh)
 # include <Events.h>
-  static long
-  timer_proto(int i)
-  {
-    static long oldticks[MAX_TIMER];
-    long ticks = TickCount(), delay = ticks - oldticks[i];
-
-    oldticks[i] = ticks;
-    return 50 * delay / 3;
-  }
+long
+TIMER(pari_timer *T)
+{
+  return _get_time(T, TickCount(), 60);
+}
 #elif USE_TIMES
 
 # include <sys/times.h>
 # include <sys/time.h>
 # include <time.h>
-  static long
-  timer_proto(int i)
-  {
-    static clock_t oldticks[MAX_TIMER];
-    struct tms t;
-    long delay;
-
-    times(&t);
-    delay = (long)((t.tms_utime - oldticks[i]) * (1000. / CLK_TCK));
-    oldticks[i] = t.tms_utime;
-    return delay;
-  }
+long
+TIMER(pari_timer *T)
+{
+  struct tms t; times(&t);
+  return _get_time(T, t.tms_utime, CLK_TCK);
+}
 #elif USE_GETRUSAGE
 
 # include <sys/time.h>
 # include <sys/resource.h>
-  static long
-  timer_proto(int i)
-  {
-    static long oldmusec[MAX_TIMER],oldsec[MAX_TIMER];
-    struct rusage r;
-    struct timeval t;
-    long delay;
+long
+TIMER(pari_timer *T)
+{
+  struct rusage r;
+  struct timeval t;
+  long delay;
 
-    getrusage(0,&r); t=r.ru_utime;
-    delay = 1000 * (t.tv_sec - oldsec[i]) + (t.tv_usec - oldmusec[i]) / 1000;
-    oldmusec[i] = t.tv_usec; oldsec[i] = t.tv_sec;
-    return delay;
-  }
+  getrusage(0,&r); t = r.ru_utime;
+  delay = 1000 * (t.tv_sec - T->s) + (t.tv_usec - T->us) / 1000;
+  T->us = t.tv_usec;
+  T->s  = t.tv_sec; return delay;
+}
 #elif USE_FTIME
 
 # include <sys/timeb.h>
-  static long
-  timer_proto(int i)
-  {
-    static long oldmsec[MAX_TIMER],oldsec[MAX_TIMER];
-    struct timeb t;
-    long delay;
+long
+TIMER(pari_timer *T)
+{
+  struct timeb t;
+  long delay;
 
-    ftime(&t);
-    delay = 1000 * (t.time - oldsec[i]) + (t.millitm - oldmsec[i]);
-    oldmsec[i] = t.millitm; oldsec[i] = t.time;
-    return delay;
-  }
+  ftime(&t);
+  delay = 1000 * (t.time - T->s) + (t.millitm - T->us / 1000);
+  T->us = t.millitm * 1000;
+  T->s  = t.time; return delay;
+}
 #else
 
 # include <time.h>
 # ifndef CLOCKS_PER_SEC
 #   define CLOCKS_PER_SEC 1000000 /* may be false on YOUR system */
 # endif
-  static long
-  timer_proto(int i)
-  {
-    static clock_t oldclocks[MAX_TIMER];
-    clock_t t = clock();
-    long delay = (long)((t-oldclocks[i]) * 1000. / CLOCKS_PER_SEC);
-
-    oldclocks[i] = t;
-    return delay;
-  }
+long
+TIMER(pari_timer *T)
+{
+  return _get_time(T, clock(), CLOCKS_PER_SEC);
+}
 #endif
 
-#define is_valid_timer(t) ((t) < MAX_TIMER || (t) >= MIN_TIMER)
 long
-gptimer(void) {return timer_proto(0);}
+gptimer(void) { static pari_timer T; return TIMER(&T);}
 long
-timer(void)   {return timer_proto(1);}
+timer(void)   { static pari_timer T; return TIMER(&T);}
 long
-timer2(void)  {return timer_proto(2);}
-long
-gentimer(long t)
-{
-  if (!is_valid_timer(t))
-    err(talker,"not an available timer (%ld)",t);
-  return timer_proto(t);
-}
-
-/* internal */
-
-long
-get_timer(long t)
-{
-  static int used[MAX_TIMER];
-  int i;
-  if (!t)
-  { /* get new timer */
-    for (i=MIN_TIMER; i < MAX_TIMER; i++)
-      if (!used[i]) { used[i] = 1; t = i; break; }
-    if (!t) { t = 2; err(warner, "no timers left! Using timer2()"); }
-    timer_proto(t); /* init timer */
-  }
-  else if (t < 0)
-  { /* initialize */
-    for (i=MIN_TIMER; i < MAX_TIMER; i++) used[i] = 0;
-  }
-  else
-  { /* delete */
-    if (!is_valid_timer(t) || !used[t])
-      err(warner, "timer %ld wasn't in use", t);
-    else
-      used[t] = 0;
-  }
-  return t;
-}
+timer2(void)  { static pari_timer T; return TIMER(&T);}
 
 void
-genmsgtimer(long t, char *format, ...)
+msgTIMER(pari_timer *T, char *format, ...)
 {
   va_list args;
   PariOUT *out = pariOut; pariOut = pariErr;
 
   pariputs("Time "); va_start(args, format);
   vpariputs(format,args); va_end(args);
-  pariputsf(": %ld\n", timer_proto(t)); pariflush();
+  pariputsf(": %ld\n", TIMER(T)); pariflush();
   pariOut = out;
 }
 
@@ -1842,7 +1817,7 @@ msgtimer(char *format, ...)
 
   pariputs("Time "); va_start(args, format);
   vpariputs(format,args); va_end(args);
-  pariputsf(": %ld\n", timer_proto(2)); pariflush();
+  pariputsf(": %ld\n", timer2()); pariflush();
   pariOut = out;
 }
 
@@ -1976,8 +1951,13 @@ entree functions_basic[]={
 {"bernfrac",11,(void*)bernfrac,3,"L"},
 {"bernreal",99,(void*)bernreal,3,"Lp"},
 {"bernvec",11,(void*)bernvec,3,"L"},
+{"besselh1",29,(void*)hbessel1,3,"GGp"},
+{"besselh2",29,(void*)hbessel2,3,"GGp"},
+{"besseli",29,(void*)ibessel,3,"GGp"},
+{"besselj",29,(void*)jbessel,3,"GGp"},
 {"besseljh",29,(void*)jbesselh,3,"GGp"},
 {"besselk",99,(void*)kbessel0,3,"GGD0,L,p"},
+{"besseln",29,(void*)nbessel,3,"GGp"},
 {"bestappr",99,(void*)bestappr0,4,"GGDG"},
 {"bezout",2,(void*)vecbezout,4,"GG"},
 {"bezoutres",2,(void*)vecbezoutres,4,"GG"},
@@ -2006,7 +1986,7 @@ entree functions_basic[]={
 {"bnfsignunit",18,(void*)signunits,6,"G"},
 {"bnfsunit",99,(void*)bnfsunit,6,"GGp"},
 {"bnfunit",18,(void*)buchfu,6,"G"},
-{"bnrL1",99,(void*)bnrL1,6,"GGD0,L,p"},
+{"bnrL1",99,(void*)bnrL1,6,"GDGD0,L,p"},
 {"bnrclass",99,(void*)bnrclass0,6,"GGD0,L,"},
 {"bnrclassno",2,(void*)rayclassno,6,"GG"},
 {"bnrclassnolist",2,(void*)rayclassnolist,6,"GG"},
@@ -2018,7 +1998,7 @@ entree functions_basic[]={
 {"bnrisconductor",99,(void*)bnrisconductor,6,"lGDGDG"},
 {"bnrisprincipal",99,(void*)isprincipalrayall,6,"GGD1,L,"},
 {"bnrrootnumber",99,(void*)bnrrootnumber,6,"GGD0,L,p"},
-{"bnrstark",99,(void*)bnrstark,6,"GGD0,L,p"},
+{"bnrstark",99,(void*)bnrstark,6,"GDGD0,L,p"},
 {"break",0,(void*)break0,11,"D1,L,"},
 {"ceil",18,(void*)gceil,2,"G"},
 {"centerlift",99,(void*)centerlift0,2,"GDn"},
@@ -2064,6 +2044,7 @@ entree functions_basic[]={
 {"ellj",1,(void*)jell,5,"Gp"},
 {"elllocalred",2,(void*)localreduction,5,"GG"},
 {"elllseries",99,(void*)lseriesell,5,"GGDGp"},
+{"ellminimalmodel",99,(void*)ellminimalmodel,5,"GD&"},
 {"ellorder",2,(void*)orderell,5,"GG"},
 {"ellordinate",29,(void*)ordell,5,"GGp"},
 {"ellpointtoz",29,(void*)zell,5,"GGp"},
@@ -2104,7 +2085,9 @@ entree functions_basic[]={
 {"galoisinit",99,(void*)galoisinit,6,"GDGD0,L,"},
 {"galoisisabelian",99,(void*)galoisisabelian,6,"GD0,L,"},
 {"galoispermtopol",2,(void*)galoispermtopol,6,"GG"},
-{"galoissubcyclo",99,(void*)galoissubcyclo,6,"LGDGDnD0,L,"},
+{"galoissubcyclo",99,(void*)galoissubcyclo,6,"GDGD0,L,Dn"},
+{"galoissubfields",99,(void*)galoissubfields,6,"GD0,L,Dn"},
+{"galoissubgroups",18,(void*)galoissubgroups,6,"G"},
 {"gamma",1,(void*)ggamma,3,"Gp"},
 {"gammah",1,(void*)ggamd,3,"Gp"},
 {"gcd",99,(void*)gcd0,4,"GDGD0,L,"},
@@ -2146,7 +2129,7 @@ entree functions_basic[]={
 {"intnum",99,(void*)intnum0,9,"V=GGED0,L,p"},
 {"isfundamental",18,(void*)gisfundamental,4,"G"},
 {"isprime",99,(void*)gisprime,4,"GD0,L,"},
-{"ispseudoprime",18,(void*)gispsp,4,"G"},
+{"ispseudoprime",18,(void*)gispseudoprime,4,"GD0,L,"},
 {"issquare",99,(void*)gcarrecomplet,4,"GD&"},
 {"issquarefree",18,(void*)gissquarefree,4,"G"},
 {"kronecker",2,(void*)gkronecker,4,"GG"},
@@ -2218,7 +2201,7 @@ entree functions_basic[]={
 {"nfeltpow",3,(void*)element_pow,6,"GGG"},
 {"nfeltpowmodpr",4,(void*)element_powmodpr,6,"GGGG"},
 {"nfeltreduce",3,(void*)element_reduce,6,"GGG"},
-{"nfeltreducemodpr",3,(void*)nfreducemodpr2,6,"GGG"},
+{"nfeltreducemodpr",3,(void*)nfreducemodpr,6,"GGG"},
 {"nfeltval",30,(void*)element_val,6,"lGGG"},
 {"nffactor",99,(void*)nffactor,6,"GG"},
 {"nffactormod",99,(void*)nffactormod,6,"GGG"},
@@ -2261,15 +2244,15 @@ entree functions_basic[]={
 {"pollead",99,(void*)pollead,7,"GDn"},
 {"pollegendre",99,(void*)legendre,7,"LDn"},
 {"polrecip",18,(void*)polrecip,7,"G"},
-{"polred",99,(void*)polred0,6,"GD0,L,DGp"},
-{"polredabs",99,(void*)polredabs0,6,"GD0,L,p"},
-{"polredord",1,(void*)ordred,6,"Gp"},
+{"polred",99,(void*)polred0,6,"GD0,L,DG"},
+{"polredabs",99,(void*)polredabs0,6,"GD0,L,"},
+{"polredord",1,(void*)ordred,6,"G"},
 {"polresultant",99,(void*)polresultant0,7,"GGDnD0,L,"},
 {"polroots",99,(void*)roots0,7,"GD0,L,p"},
 {"polrootsmod",25,(void*)rootmod0,7,"GGD0,L,"},
 {"polrootspadic",32,(void*)rootpadic,7,"GGL"},
 {"polsturm",99,(void*)sturmpart,7,"lGDGDG"},
-{"polsubcyclo",99,(void*)subcyclo,7,"GGDn"},
+{"polsubcyclo",99,(void*)polsubcyclo,7,"LLDn"},
 {"polsylvestermatrix",2,(void*)sylvestermatrix,7,"GG"},
 {"polsym",21,(void*)polsym,7,"GL"},
 {"poltchebi",99,(void*)tchebi,7,"LDn"},
@@ -2338,11 +2321,11 @@ entree functions_basic[]={
 {"rnfinit",29,(void*)rnfinitalg,6,"GGp"},
 {"rnfisfree",20,(void*)rnfisfree,6,"lGG"},
 {"rnfisnorm",99,(void*)rnfisnorm,6,"GGGD1,L,p"},
-{"rnfkummer",99,(void*)rnfkummer,6,"GGD0,L,p"},
+{"rnfkummer",99,(void*)rnfkummer,6,"GDGD0,L,p"},
 {"rnflllgram",99,(void*)rnflllgram,6,"GGGp"},
 {"rnfnormgroup",2,(void*)rnfnormgroup,6,"GG"},
 {"rnfpolred",29,(void*)rnfpolred,6,"GGp"},
-{"rnfpolredabs",99,(void*)rnfpolredabs,6,"GGD0,L,p"},
+{"rnfpolredabs",99,(void*)rnfpolredabs,6,"GGD0,L,"},
 {"rnfpseudobasis",2,(void*)rnfpseudobasis,6,"GG"},
 {"rnfsteinitz",2,(void*)rnfsteinitz,6,"GG"},
 {"round",99,(void*)round0,2,"GD&"},
