@@ -1,4 +1,4 @@
-/* $Id: sumiter.c,v 1.27 2001/11/16 14:25:14 karim Exp $
+/* $Id: sumiter.c,v 1.33 2002/06/08 14:48:36 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -31,7 +31,7 @@ extern void changevalue_p(entree *ep, GEN x);
 void
 forpari(entree *ep, GEN a, GEN b, char *ch)
 {
-  ulong av,av0 = avma, lim;
+  gpmem_t av, av0 = avma, lim;
 
   b = gcopy(b); av=avma; lim = stack_lim(av,1);
  /* gcopy nedeed in case b gets overwritten in ch, as in
@@ -40,7 +40,7 @@ forpari(entree *ep, GEN a, GEN b, char *ch)
   push_val(ep, a);
   while (gcmp(a,b) <= 0)
   {
-    ulong av1=avma; (void)lisseq(ch); avma=av1;
+    gpmem_t av1=avma; (void)lisseq(ch); avma=av1;
     if (loop_break()) break;
     a = (GEN) ep->value; a = gadd(a,gun);
     if (low_stack(lim, stack_lim(av,1)))
@@ -58,7 +58,8 @@ static int negcmp(GEN x, GEN y) { return gcmp(y,x); }
 void
 forstep(entree *ep, GEN a, GEN b, GEN s, char *ch)
 {
-  long ss, av,av0 = avma, lim, i;
+  long ss, i;
+  gpmem_t av, av0 = avma, lim;
   GEN v = NULL;
   int (*cmp)(GEN,GEN);
 
@@ -75,7 +76,7 @@ forstep(entree *ep, GEN a, GEN b, GEN s, char *ch)
   i = 0;
   while (cmp(a,b) <= 0)
   {
-    long av1=avma; (void)lisseq(ch); avma=av1;
+    gpmem_t av1=avma; (void)lisseq(ch); avma=av1;
     if (loop_break()) break;
     if (v)
     {
@@ -153,8 +154,9 @@ prime_loop_init(GEN ga, GEN gb, long *a, long *b, long prime[])
 void
 forprime(entree *ep, GEN ga, GEN gb, char *ch)
 {
-  long prime[] = {evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
-  long av = avma, a,b;
+  long prime[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
+  long a, b;
+  gpmem_t av = avma;
   byteptr p;
 
   p = prime_loop_init(ga,gb, &a,&b, prime);
@@ -179,7 +181,8 @@ forprime(entree *ep, GEN ga, GEN gb, char *ch)
 void
 fordiv(GEN a, entree *ep, char *ch)
 {
-  long i,av2,l, av = avma;
+  long i, l;
+  gpmem_t av2, av = avma;
   GEN t = divisors(a);
 
   push_val(ep, NULL); l=lg(t); av2 = avma;
@@ -198,111 +201,107 @@ fordiv(GEN a, entree *ep, char *ch)
  *   fl = 1: impose a1 <= ... <= an
  *   fl = 2:        a1 <  ... <  an
  */
-static GEN *fv_a, *fv_m, *fv_M;
-static long fv_n, fv_fl;
-static char *fv_ch;
+typedef struct {
+  GEN *a, *m, *M;
+  long n,fl;
+  char *ch;
+} fvdat;
 
 /* general case */
 static void
-fvloop(long i)
+fvloop(long i, fvdat *d)
 {
-  fv_a[i] = fv_m[i];
-  if (fv_fl && i > 1)
+  d->a[i] = d->m[i];
+  if (d->fl && i > 1)
   {
-    GEN p1 = gsub(fv_a[i], fv_a[i-1]);
+    GEN p1 = gsub(d->a[i], d->a[i-1]);
     if (gsigne(p1) < 0)
-      fv_a[i] = gadd(fv_a[i], gceil(gneg_i(p1)));
-    if (fv_fl == 2 && gegal(fv_a[i], fv_a[i-1]))
-      fv_a[i] = gadd(fv_a[i], gun);
+      d->a[i] = gadd(d->a[i], gceil(gneg_i(p1)));
+    if (d->fl == 2 && gegal(d->a[i], d->a[i-1]))
+      d->a[i] = gadd(d->a[i], gun);
   }
-  if (i+1 == fv_n)
-    while (gcmp(fv_a[i], fv_M[i]) <= 0)
+  if (i+1 == d->n)
+    while (gcmp(d->a[i], d->M[i]) <= 0)
     {
-      long av = avma; (void)lisseq(fv_ch); avma = av;
-      if (loop_break()) { fv_n = 0; return; }
-      fv_a[i] = gadd(fv_a[i], gun);
+      gpmem_t av = avma; (void)lisseq(d->ch); avma = av;
+      if (loop_break()) { d->n = 0; return; }
+      d->a[i] = gadd(d->a[i], gun);
     }
   else
-    while (gcmp(fv_a[i], fv_M[i]) <= 0)
+    while (gcmp(d->a[i], d->M[i]) <= 0)
     {
-      long av = avma; fvloop(i+1); avma = av;
-      if (!fv_n) return;
-      fv_a[i] = gadd(fv_a[i], gun);
+      gpmem_t av = avma; fvloop(i+1, d); avma = av;
+      if (!d->n) return;
+      d->a[i] = gadd(d->a[i], gun);
     }
 }
 
 /* we run through integers */
 static void
-fvloop_i(long i)
+fvloop_i(long i, fvdat *d)
 {
-  fv_a[i] = setloop(fv_m[i]);
-  if (fv_fl && i > 1)
+  d->a[i] = setloop(d->m[i]);
+  if (d->fl && i > 1)
   {
-    int c = cmpii(fv_a[i], fv_a[i-1]);
+    int c = cmpii(d->a[i], d->a[i-1]);
     if (c < 0)
     {
-      fv_a[i] = setloop(fv_a[i-1]);
+      d->a[i] = setloop(d->a[i-1]);
       c = 0;
     }
-    if (c == 0 && fv_fl == 2)
-      fv_a[i] = incloop(fv_a[i]);
+    if (c == 0 && d->fl == 2)
+      d->a[i] = incloop(d->a[i]);
   }
-  if (i+1 == fv_n)
-    while (gcmp(fv_a[i], fv_M[i]) <= 0)
+  if (i+1 == d->n)
+    while (gcmp(d->a[i], d->M[i]) <= 0)
     {
-      long av = avma; (void)lisseq(fv_ch); avma = av;
-      if (loop_break()) { fv_n = 0; return; }
-      fv_a[i] = incloop(fv_a[i]);
+      gpmem_t av = avma; (void)lisseq(d->ch); avma = av;
+      if (loop_break()) { d->n = 0; return; }
+      d->a[i] = incloop(d->a[i]);
     }
   else
-    while (gcmp(fv_a[i], fv_M[i]) <= 0)
+    while (gcmp(d->a[i], d->M[i]) <= 0)
     {
-      long av = avma; fvloop_i(i+1); avma = av;
-      if (!fv_n) return;
-      fv_a[i] = incloop(fv_a[i]);
+      gpmem_t av = avma; fvloop_i(i+1, d); avma = av;
+      if (!d->n) return;
+      d->a[i] = incloop(d->a[i]);
     }
 }
 
 void
 forvec(entree *ep, GEN x, char *c, long flag)
 {
-  long i, av = avma, tx = typ(x), n = fv_n, fl = fv_fl;
-  GEN *a = fv_a, *M = fv_M, *m = fv_m;
-  char *ch = fv_ch;
-  void (*fv_fun)(long) = fvloop_i;
+  gpmem_t av = avma;
+  long tx = typ(x);
+  fvdat D, *d = &D;
 
   if (!is_vec_t(tx)) err(talker,"not a vector in forvec");
-  if (fl<0 || fl>2) err(flagerr);
-  fv_n = lg(x);
-  fv_ch = c;
-  fv_fl = flag;
-  fv_a = (GEN*)cgetg(fv_n,t_VEC); push_val(ep, (GEN)fv_a);
-  fv_m = (GEN*)cgetg(fv_n,t_VEC);
-  fv_M = (GEN*)cgetg(fv_n,t_VEC);
-  if (fv_n == 1) lisseq(fv_ch);
+  if (flag<0 || flag>2) err(flagerr);
+  d->n = lg(x);
+  d->ch = c;
+  d->a = (GEN*)cgetg(d->n,t_VEC); push_val(ep, (GEN)d->a);
+  if (d->n == 1) (void)lisseq(d->ch);
   else
   {
-    for (i=1; i<fv_n; i++)
+    long i, t = t_INT;
+    d->fl = flag;
+    d->m = (GEN*)cgetg(d->n,t_VEC);
+    d->M = (GEN*)cgetg(d->n,t_VEC);
+    for (i=1; i<d->n; i++)
     {
-      GEN *c = (GEN*) x[i];
-      tx = typ(c);
-      if (! is_vec_t(tx) || lg(c)!=3)
+      GEN *e = (GEN*) x[i];
+      tx = typ(e);
+      if (! is_vec_t(tx) || lg(e)!=3)
 	err(talker,"not a vector of two-component vectors in forvec");
-      if (gcmp(c[1],c[2]) > 0) fv_n = 0;
-      /* in case x is an ep->value and c destroys it, we have to copy */
-      if (typ(c[1]) != t_INT) fv_fun = fvloop;
-      fv_m[i] = gcopy(c[1]);
-      fv_M[i] = gcopy(c[2]);
+      if (gcmp(e[1],e[2]) > 0) d->n = 0;
+      if (typ(e[1]) != t_INT) t = t_REAL;
+      /* in case x is an ep->value and lisexpr(d->ch) kills it, we have to copy */
+      d->m[i] = gcopy(e[1]);
+      d->M[i] = gcopy(e[2]);
     }
-    fv_fun(1);
+    if (t == t_INT) fvloop_i(1, d); else fvloop(1, d);
   }
-  pop_val(ep);
-  fv_n = n;
-  fv_ch = ch;
-  fv_fl = fl;
-  fv_a = a;
-  fv_m = m;
-  fv_M = M; avma = av;
+  pop_val(ep); avma = av;
 }
 
 /********************************************************************/
@@ -314,7 +313,7 @@ forvec(entree *ep, GEN x, char *c, long flag)
 GEN
 somme(entree *ep, GEN a, GEN b, char *ch, GEN x)
 {
-  long av,av0 = avma, lim;
+  gpmem_t av, av0 = avma, lim;
   GEN p1;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sum");
@@ -343,7 +342,8 @@ somme(entree *ep, GEN a, GEN b, char *ch, GEN x)
 GEN
 suminf(entree *ep, GEN a, char *ch, long prec)
 {
-  long fl,G,tetpil, av0 = avma, av,lim;
+  long fl, G;
+  gpmem_t tetpil, av0 = avma, av, lim;
   GEN p1,x = realun(prec);
 
   if (typ(a) != t_INT) err(talker,"non integral index in suminf");
@@ -373,7 +373,7 @@ suminf(entree *ep, GEN a, char *ch, long prec)
 GEN
 divsum(GEN num, entree *ep, char *ch)
 {
-  ulong av = avma;
+  gpmem_t av = avma;
   GEN z, y = gzero, t = divisors(num);
   long i, l = lg(t);
 
@@ -397,7 +397,7 @@ divsum(GEN num, entree *ep, char *ch)
 GEN
 produit(entree *ep, GEN a, GEN b, char *ch, GEN x)
 {
-  long av,av0 = avma, lim;
+  gpmem_t av, av0 = avma, lim;
   GEN p1;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sum");
@@ -438,7 +438,7 @@ prodinf0(entree *ep, GEN a, char *ch, long flag, long prec)
 GEN
 prodinf(entree *ep, GEN a, char *ch, long prec)
 {
-  ulong av0 = avma, av,lim;
+  gpmem_t av0 = avma, av, lim;
   long fl,G;
   GEN p1,x = realun(prec);
 
@@ -465,7 +465,7 @@ prodinf(entree *ep, GEN a, char *ch, long prec)
 GEN
 prodinf1(entree *ep, GEN a, char *ch, long prec)
 {
-  ulong av0 = avma, av,lim;
+  gpmem_t av0 = avma, av, lim;
   long fl,G;
   GEN p1,p2,x = realun(prec);
 
@@ -492,9 +492,9 @@ prodinf1(entree *ep, GEN a, char *ch, long prec)
 GEN
 prodeuler(entree *ep, GEN ga, GEN gb, char *ch, long prec)
 {
-  long prime[] = {evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
+  long prime[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
   long a,b;
-  ulong av,av0 = avma, lim;
+  gpmem_t av, av0 = avma, lim;
   GEN p1,x = realun(prec);
   byteptr p;
   
@@ -530,8 +530,8 @@ prodeuler(entree *ep, GEN ga, GEN gb, char *ch, long prec)
 GEN
 direulerall(entree *ep, GEN ga, GEN gb, char *ch, GEN c)
 {
-  long prime[] = {evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
-  ulong av0 = avma,av, lim = (av0+bot)>>1;
+  long prime[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
+  gpmem_t av0 = avma, av, lim = stack_lim(av0, 1);
   long p,n,i,j,k,tx,lx,a,b;
   GEN x,y,s,polnum,polden;
   byteptr d;
@@ -636,7 +636,7 @@ vecteur(GEN nmax, entree *ep, char *ch)
 {
   GEN y,p1;
   long i,m;
-  long c[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
+  long c[]={evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
 
   if (typ(nmax)!=t_INT || signe(nmax) < 0)
     err(talker,"bad number of components in vector");
@@ -661,7 +661,7 @@ vecteursmall(GEN nmax, entree *ep, char *ch)
 {
   GEN y,p1;
   long i,m;
-  long c[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
+  long c[]={evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 0};
 
   if (typ(nmax)!=t_INT || signe(nmax) < 0)
     err(talker,"bad number of components in vector");
@@ -694,8 +694,8 @@ matrice(GEN nlig, GEN ncol,entree *ep1, entree *ep2, char *ch)
 {
   GEN y,z,p1;
   long i,j,m,n,s;
-  long c1[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 1};
-  long c2[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 1};
+  long c1[]={evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 1};
+  long c2[]={evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3), 1};
 
   s=signe(ncol);
   if (ep1 == ep2 && ep1) err(talker, "identical index variables in matrix");
@@ -769,7 +769,8 @@ sumpos0(entree *ep, GEN a, char *ch, long flag, long prec)
 GEN
 sumalt(entree *ep, GEN a, char *ch, long prec)
 {
-  long av = avma, tetpil,k,N;
+  long k, N;
+  gpmem_t av = avma, tetpil;
   GEN s,az,c,x,e1,d;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sumalt");
@@ -777,7 +778,7 @@ sumalt(entree *ep, GEN a, char *ch, long prec)
   push_val(ep, a);
   e1=addsr(3,gsqrt(stoi(8),prec));
   N=(long)(0.4*(bit_accuracy(prec) + 7));
-  d=gpuigs(e1,N); d=shiftr(addrr(d,divsr(1,d)),-1);
+  d=gpuigs(e1,N); d=shiftr(addrr(d, ginv(d)),-1);
   az=negi(gun); c=d; s=gzero;
   for (k=0; ; k++)
   {
@@ -794,7 +795,8 @@ sumalt(entree *ep, GEN a, char *ch, long prec)
 GEN
 sumalt2(entree *ep, GEN a, char *ch, long prec)
 {
-  long av = avma, tetpil,k,N;
+  long k, N;
+  gpmem_t av = avma, tetpil;
   GEN x,s,dn,pol;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sumalt");
@@ -818,7 +820,8 @@ sumalt2(entree *ep, GEN a, char *ch, long prec)
 GEN
 sumpos(entree *ep, GEN a, char *ch, long prec)
 {
-  long av = avma,tetpil,k,kk,N,G;
+  long k, kk, N, G;
+  gpmem_t av = avma, tetpil;
   GEN p1,r,q1,reel,s,az,c,x,e1,d, *stock;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sumpos");
@@ -827,7 +830,7 @@ sumpos(entree *ep, GEN a, char *ch, long prec)
   a=subis(a,1); reel=cgetr(prec);
   e1=addsr(3,gsqrt(stoi(8),prec));
   N=(long)(0.4*(bit_accuracy(prec) + 7));
-  d=gpuigs(e1,N); d=shiftr(addrr(d,divsr(1,d)),-1);
+  d=gpuigs(e1,N); d=shiftr(addrr(d, ginv(d)),-1);
   az=negi(gun); c=d; s=gzero;
   G = -bit_accuracy(prec) - 5;
   stock=(GEN*)new_chunk(N+1); for (k=1; k<=N; k++) stock[k]=NULL;
@@ -862,7 +865,8 @@ sumpos(entree *ep, GEN a, char *ch, long prec)
 GEN
 sumpos2(entree *ep, GEN a, char *ch, long prec)
 {
-  long av = avma,tetpil,k,kk,N,G;
+  long k, kk, N, G;
+  gpmem_t av = avma, tetpil;
   GEN p1,r,q1,reel,s,pol,dn,x, *stock;
 
   if (typ(a) != t_INT) err(talker,"non integral index in sumpos2");
@@ -941,7 +945,8 @@ GEN
 qromb(entree *ep, GEN a, GEN b, char *ch, long prec)
 {
   GEN ss,dss,s,h,p1,p2,qlint,del,x,sum;
-  long av = avma, av1, tetpil,j,j1,j2,lim,it,sig;
+  long j, j1, j2, it, sig, lim;
+  gpmem_t av = avma, av1, tetpil;
 
   a = fix(a,prec);
   b = fix(b,prec);
@@ -989,7 +994,8 @@ GEN
 qromo(entree *ep, GEN a, GEN b, char *ch, long prec)
 {
   GEN ss,dss,s,h,p1,qlint,del,ddel,x,sum;
-  long av = avma,av1,tetpil,j,j1,j2,lim,it,sig;
+  long j, j1, j2, it, sig, lim;
+  gpmem_t av = avma, av1, tetpil;
 
   a = fix(a, prec);
   b = fix(b, prec);
@@ -1041,7 +1047,8 @@ GEN
 qromi(entree *ep, GEN a, GEN b, char *ch, long prec)
 {
   GEN ss,dss,s,h,q1,p1,qlint,del,ddel,x,sum;
-  long av = avma, av1,tetpil,j,j1,j2,lim,it,sig;
+  long j, j1, j2, it, sig, lim;
+  gpmem_t av = avma, av1, tetpil;
 
   a = fix(ginv(a), prec);
   b = fix(ginv(b), prec);
@@ -1093,7 +1100,8 @@ GEN
 rombint(entree *ep, GEN a, GEN b, char *ch, long prec)
 {
   GEN aa,bb,mun,p1,p2,p3;
-  long l,av,tetpil;
+  long l;
+  gpmem_t av, tetpil;
 
   l=gcmp(b,a); if (!l) return gzero;
   if (l<0) { bb=a; aa=b; } else { bb=b; aa=a; }
@@ -1128,7 +1136,8 @@ rombint(entree *ep, GEN a, GEN b, char *ch, long prec)
 GEN
 polzag(long n, long m)
 {
-  long d1,d,r,k,av=avma,tetpil;
+  long d1, d, r, k;
+  gpmem_t av=avma, tetpil;
   GEN p1,p2,pol1,g,s;
 
   if (m>=n || m<0)
@@ -1158,7 +1167,8 @@ polzag(long n, long m)
 GEN
 polzagreel(long n, long m, long prec)
 {
-  long d1,d,r,j,k,k2,av=avma,tetpil;
+  long d1, d, r, j, k, k2;
+  gpmem_t av=avma, tetpil;
   GEN p2,pol1,g,h,v,b,gend,s;
 
   if (m>=n || m<0)
@@ -1209,7 +1219,8 @@ polzagreel(long n, long m, long prec)
 GEN
 zbrent(entree *ep, GEN a, GEN b, char *ch, long prec)
 {
-  long av = avma,sig,iter,itmax;
+  long sig, iter, itmax;
+  gpmem_t av = avma;
   GEN c,d,e,tol,tol1,min1,min2,fa,fb,fc,p,q,r,s,xm;
 
   a = fix(a,prec);
