@@ -1,4 +1,4 @@
-/* $Id: kummer.c,v 1.26 2002/01/09 14:05:07 karim Exp $
+/* $Id: kummer.c,v 1.38 2002/06/01 14:27:42 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -19,6 +19,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 /*                                                                 */
 /*******************************************************************/
 #include "pari.h"
+#include "parinf.h"
 extern GEN check_and_build_cycgen(GEN bnf);
 extern GEN get_arch_real(GEN nf,GEN x,GEN *emb,long prec);
 extern GEN vecmul(GEN x, GEN y);
@@ -27,11 +28,7 @@ extern GEN T2_from_embed_norm(GEN x, long r1);
 extern GEN pol_to_vec(GEN x, long N);
 extern GEN famat_to_nf(GEN nf, GEN f);
 extern GEN vconcat(GEN A, GEN B);
-extern long int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *newx, long v);
-    
-
-static long rc,ell,degK,degKz,m,d,vnf,dv;
-static GEN invexpoteta1,nfz,gell;
+extern long int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *newx);
 
 /* row vector B x matrix T : c_j=prod_i (b_i^T_ij) */
 static GEN
@@ -50,7 +47,7 @@ groupproduct(GEN B, GEN T)
 }
 
 static int
-ok_x(GEN X, GEN arch, GEN vecmunit2, GEN msign)
+ok_x(GEN X, GEN arch, GEN vecmunit2, GEN msign, GEN gell)
 {
   long i, l = lg(vecmunit2);
   GEN p1;
@@ -64,10 +61,10 @@ ok_x(GEN X, GEN arch, GEN vecmunit2, GEN msign)
 }
 
 static GEN
-get_listx(GEN arch,GEN msign,GEN munit,GEN vecmunit2,long ell,long lSp,long nbvunit)
+get_listx(GEN arch,GEN msign,GEN munit,GEN vecmunit2,GEN gell,long lSp,long nbvunit)
 {
   GEN kermat,p2,X,y, listx = cgetg(1,t_VEC);
-  long i,j,cmpt,lker;
+  long i,j,cmpt,lker, ell = itos(gell);
 
   kermat = FpM_ker(munit,gell); lker=lg(kermat)-1;
   if (!lker) return listx;
@@ -82,7 +79,7 @@ get_listx(GEN arch,GEN msign,GEN munit,GEN vecmunit2,long ell,long lSp,long nbvu
     cmpt = 0; for (j=1; j<=lSp; j++) if (gcmp0((GEN)X[j+nbvunit])) cmpt++;
     if (!cmpt)
     {
-      if (ok_x(X, arch, vecmunit2, msign))
+      if (ok_x(X, arch, vecmunit2, msign, gell))
         { p2[1] = (long)X; listx = concatsp(listx,p2); }
     }
     if (!lSp)
@@ -91,7 +88,7 @@ get_listx(GEN arch,GEN msign,GEN munit,GEN vecmunit2,long ell,long lSp,long nbvu
       if (j<lker && y[j] == 1)
       {
         X = gsub(X,(GEN)kermat[lker]);
-        if (ok_x(X, arch, vecmunit2, msign))
+        if (ok_x(X, arch, vecmunit2, msign, gell))
           { p2[1] = (long)X; listx = concatsp(listx,p2); }
       }
     }
@@ -135,6 +132,95 @@ prank(GEN cyc, long ell)
   return i-1;
 }
 
+static GEN
+tauofelt(GEN nfz, GEN nf, GEN x, GEN U)
+{
+  return algtobasis(nfz, poleval(gmul((GEN)nf[7],x), U));
+}
+static GEN
+tauofalg(GEN x, GEN U)
+{
+  return gsubst(lift(x),varn(U[1]),U);
+}
+
+static GEN
+lifttoKz(GEN nfz, GEN nf, GEN id, GEN U)
+{
+  GEN I = ideal_two_elt(nf,id);
+  I[2] = (long)tauofelt(nfz,nf,(GEN)I[2],U);
+  return prime_to_ideal(nfz,I);
+}
+
+static GEN
+tauofideal(GEN nfz, GEN id, GEN U)
+{
+  GEN I = ideal_two_elt(nfz,id);
+  I[2] = (long)tauofelt(nfz,nfz,(GEN)I[2],U);
+  return prime_to_ideal(nfz,I);
+}
+
+static int
+isprimeidealconj(GEN nfz, GEN pr1, GEN pr2, GEN U)
+{
+  GEN p = (GEN)pr1[1];
+  GEN x = (GEN)pr1[2];
+  GEN b1= (GEN)pr1[5];
+  GEN b2= (GEN)pr2[5];
+  if (!egalii(p, (GEN)pr2[1])
+   || !egalii((GEN)pr1[3], (GEN)pr2[3])
+   || !egalii((GEN)pr1[4], (GEN)pr2[4])) return 0;
+  if (gegal(x,(GEN)pr2[2])) return 1;
+  for(;;)
+  {
+    if (int_elt_val(nfz,x,p,b2,NULL)) return 1;
+    x = FpV_red(tauofelt(nfz,nfz,x,U), p);
+    if (int_elt_val(nfz,x,p,b1,NULL)) return 0;
+  }
+}
+
+static int
+isconjinprimelist(GEN nfz, GEN S, GEN pr, GEN U)
+{
+  long i, l;
+  
+  if (!U) return 0;
+  l = lg(S);
+  for (i=1; i<l; i++)
+    if (isprimeidealconj(nfz, (GEN)S[i],pr,U)) return 1;
+  return 0;
+}
+
+typedef struct {
+  GEN polnf, U, invexpoteta1;
+  long m;
+} toK_t;
+
+static GEN
+downtoK(toK_t *T, GEN x)
+{
+  long degKz = lg(T->invexpoteta1) - 1;
+  GEN y = gmul(T->invexpoteta1, pol_to_vec(lift(x), degKz));
+  return gmodulcp(gtopolyrev(y,varn(T->polnf)), T->polnf);
+}
+
+static GEN
+tracetoK(toK_t *T, GEN x)
+{
+  GEN p1 = x;
+  long i;
+  for (i=1; i < T->m; i++) p1 = gadd(x, tauofalg(p1,T->U));
+  return downtoK(T, p1);
+}
+
+static GEN
+normtoK(toK_t *T, GEN x)
+{
+  GEN p1 = x;
+  long i;
+  for (i=1; i < T->m; i++) p1 = gmul(x, tauofalg(p1,T->U));
+  return downtoK(T, p1);
+}
+
 static GEN
 no_sol(long all, int i)
 {
@@ -142,43 +228,108 @@ no_sol(long all, int i)
   return cgetg(1,t_VEC);
 }
 
+static GEN
+get_gell(GEN bnr, GEN subgp, long all)
+{
+  GEN gell;
+  if (all)        gell = stoi(all);
+  else if (subgp) gell = det(subgp);
+  else            gell = det(diagonal(gmael(bnr,5,2)));
+  if (typ(gell) != t_INT) err(arither1);
+  return gell;
+}
+
+typedef struct {
+  GEN Sm, Sml1, Sml2, Sl, ESml2;
+} primlist;
+
 static void
 _append(GEN x, GEN t)
 {
   long l = lg(x); x[l] = (long)t; setlg(x, l+1);
 }
 
-/* si all!=0, donne toutes les equations correspondant a un sousgroupe
-   de determinant all (i.e. de degre all) */
+static GEN
+cget1(long l, long t)
+{
+  GEN z = cgetg(l, t); setlg(z, 1); return z;
+}
+
+static int
+build_list_Hecke(primlist *L, GEN nfz, GEN fa, GEN gothf, GEN gell, GEN U)
+{
+  GEN listpr, listex, pr, p, factell;
+  long vd, vp, e, i, l, ell = itos(gell), degKz = degpol(nfz[1]);
+
+  listpr = (GEN)fa[1];
+  listex = (GEN)fa[2]; l = lg(listpr);
+  L->Sm  = cget1(l,t_VEC);
+  L->Sml1= cget1(l,t_VEC);
+  L->Sml2= cget1(l,t_VEC);
+  L->Sl  = cget1(l+degKz,t_VEC);
+  L->ESml2=cget1(l,t_VECSMALL);
+  for (i=1; i<l; i++)
+  {
+    pr = (GEN)listpr[i]; p = (GEN)pr[1]; e = itos((GEN)pr[3]);
+    vp = itos((GEN)listex[i]);
+    if (!egalii(p,gell))
+    {
+      if (vp != 1) return 1;
+      if (!isconjinprimelist(nfz, L->Sm,pr,U)) _append(L->Sm,pr);
+    }
+    else
+    {
+      vd = (vp-1)*(ell-1)-ell*e;
+      if (vd > 0) return 4;
+      if (vd==0)
+      {
+	if (!isconjinprimelist(nfz, L->Sml1,pr,U)) _append(L->Sml1, pr);
+      }
+      else
+      {
+	if (vp==1) return 2;
+        if (!isconjinprimelist(nfz, L->Sml2,pr,U))
+        {
+          _append(L->Sml2, pr);
+          _append(L->ESml2,(GEN)vp);
+        }
+      }
+    }
+  }
+  factell = primedec(nfz,gell); l = lg(factell);
+  for (i=1; i<l; i++)
+  {
+    pr = (GEN)factell[i];
+    if (!idealval(nfz,gothf,pr))
+      if (!isconjinprimelist(nfz, L->Sl,pr,U)) _append(L->Sl, pr);
+  }
+  return 0; /* OK */
+}
+
+/* if all!=0, give all equations of degree 'all'. Assume bnr modulus is the
+ * conductor */
 static GEN
 rnfkummersimple(GEN bnr, GEN subgroup, long all)
 {
-  long r1,degnf,ell,j,i,l;
-  long nbgenclK,lSml2,lSl,lSp,rc,nbvunit;
-  long lastbid,nbcol,nblig,k,llistx,e,vp,vd;
+  long r1, degnf, ell, j, i, l;
+  long nbgenclK, lSml2, lSl2, lSp, rc, nbvunit;
+  long lastbid, llistx;
 
-  GEN polnf,bnf,nf,classgroup,cyclic,bid,ideal,arch,cycgen,gell,p1,p2,p3;
-  GEN cyclicK,genK,listgamma,listalpha,fa,prm,expom,Sm,Sml1,Sml2,Sl,ESml2;
-  GEN p,factell,Sp,vecbeta,matexpo,vunit,id,pr,vsml2,vecalpha0;
-  GEN cycpro,munit,vecmunit2,msign,archartif,listx,listal,listg,listgamma0;
-  GEN vecbeta0;
+  GEN polnf,bnf,nf,bid,ideal,arch,cycgen,gell,p1,p2,p3;
+  GEN cyclicK,genK,listgamma,listalpha;
+  GEN Sp,listprSp,vecbeta,matexpo,vunit,id,vecalpha0;
+  GEN munit,vecmunit2,msign,archartif,listx,listal,listg,listgamma0;
+  GEN vecbeta0,vunit_beta,fununits,torsunit;
+  primlist L;
 
-  checkbnrgen(bnr);
   bnf = (GEN)bnr[1];
-  nf = (GEN)bnf[7]; r1 = nf_get_r1(nf);
-  polnf = (GEN)nf[1]; vnf = varn(polnf); degnf = degpol(polnf);
-  if (vnf==0) err(talker,"main variable in kummer must not be x");
-
-  p1 = conductor(bnr,all ? gzero : subgroup,2);
-  bnr = (GEN)p1[2];
-  if (!all) subgroup = (GEN)p1[3];
-  classgroup = (GEN)bnr[5];
-  cyclic = (GEN)classgroup[2];
+  nf  = (GEN)bnf[7]; r1 = nf_get_r1(nf);
+  polnf = (GEN)nf[1]; degnf = degpol(polnf);
+  gell = get_gell(bnr,subgroup,all);
+  
   bid = (GEN)bnr[2];
   ideal= gmael(bid,1,1);
   arch = gmael(bid,1,2); /* this is the conductor */
-  if (!all && gcmp0(subgroup)) subgroup = diagonal(cyclic);
-  gell = all? stoi(all): det(subgroup);
   ell = itos(gell);
 
   cyclicK= gmael3(bnf,8,1,2); rc = prank(cyclicK,ell);
@@ -191,128 +342,96 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
   p1 = gmul(gell,ideal);
   for (i=1; i<=rc; i++)
   {
-    p3 = basistoalg(nf,idealcoprime(nf,(GEN)genK[i],p1));
+    p3 = basistoalg(nf, idealcoprime(nf,(GEN)genK[i],p1));
     p2 = basistoalg(nf, famat_to_nf(nf, (GEN)cycgen[i]));
     listgamma[i] = listgamma0[i] = linv(p3);
     vecalpha0[i] = (long)p2;
-    listalpha[i] = lmul((GEN)vecalpha0[i], powgi(p3,(GEN)cyclicK[i]));
+    listalpha[i] = lmul(p2, powgi(p3, (GEN)cyclicK[i]));
   }
   for (   ; i<=nbgenclK; i++)
   {
     long k;
-    p3 = basistoalg(nf,idealcoprime(nf,(GEN)genK[i],p1));
+    p3 = basistoalg(nf, idealcoprime(nf,(GEN)genK[i],p1));
     p2 = basistoalg(nf, famat_to_nf(nf, (GEN)cycgen[i]));
     k = itos(mpinvmod((GEN)cyclicK[i], gell));
     p2 = gpowgs(p2,k);
     listgamma0[i]= (long)p2;
     listgamma[i] = lmul(p2, gpowgs(p3, k * itos((GEN)cyclicK[i]) - 1));
   }
-  fa = (GEN)bid[3];
-  prm   = (GEN)fa[1];
-  expom = (GEN)fa[2]; l = lg(prm);
-  Sm  = cgetg(l,t_VEC); setlg(Sm,1);
-  Sml1= cgetg(l,t_VEC); setlg(Sml1,1);
-  Sml2= cgetg(l,t_VEC); setlg(Sml2,1);
-  Sl  = cgetg(l+degnf,t_VEC); setlg(Sl,1);
-  ESml2=cgetg(l,t_VECSMALL); setlg(ESml2,1);
-  for (i=1; i<l; i++)
-  {
-    pr = (GEN)prm[i]; p = (GEN)pr[1]; e = itos((GEN)pr[3]);
-    vp = itos((GEN)expom[i]);
-    if (!egalii(p,gell))
-    {
-      if (vp != 1) return no_sol(all,1);
-      _append(Sm, pr);
-    }
-    else
-    {
-      vd = (vp-1)*(ell-1)-ell*e;
-      if (vd > 0) return no_sol(all,4);
-      if (vd==0)
-        _append(Sml1, pr);
-      else
-      {
-	if (vp==1) return no_sol(all,2);
-        _append(Sml2, pr);
-        _append(ESml2,(GEN)vp);
-      }
-    }
-  }
-  factell = primedec(nf,gell); l = lg(factell);
-  for (i=1; i<l; i++)
-  {
-    pr = (GEN)factell[i];
-    if (!idealval(nf,ideal,pr)) _append(Sl, pr);
-  }
-  lSml2=lg(Sml2)-1; lSl=lg(Sl)-1;
-  Sp = concatsp(Sm,Sml1); lSp = lg(Sp)-1;
-  vecbeta = cgetg(lSp+1,t_VEC); matexpo=cgetg(lSp+1,t_MAT);
+  i = build_list_Hecke(&L, nf, (GEN)bid[3], ideal, gell, NULL);
+  if (i) return no_sol(all,i);
+
+  lSml2 = lg(L.Sml2)-1;
+  Sp = concatsp(L.Sm, L.Sml1); lSp = lg(Sp)-1;
+  listprSp = concatsp(L.Sml2, L.Sl); lSl2 = lg(listprSp)-1;
+
+  vecbeta = cgetg(lSp+1,t_VEC);
   vecbeta0= cgetg(lSp+1,t_VEC);
+  matexpo = cgetg(lSp+1,t_MAT);
   for (j=1; j<=lSp; j++)
   {
-    p1=isprincipalgenforce(bnf,(GEN)Sp[j]);
-    p2=basistoalg(nf,(GEN)p1[2]);
+    p1 = isprincipalgenforce(bnf,(GEN)Sp[j]);
+    p2 = basistoalg(nf,(GEN)p1[2]);
+    p1 = (GEN)p1[1];
     for (i=1; i<=rc; i++)
-      p2 = gmul(p2, powgi((GEN)listgamma[i],gmael(p1,1,i)));
+      p2 = gmul(p2, powgi((GEN)listgamma[i], (GEN)p1[i]));
     p3 = p2;
     for (   ; i<=nbgenclK; i++)
     {
-      p2 = gmul(p2, powgi((GEN)listgamma[i],gmael(p1,1,i)));
-      p3 = gmul(p3, powgi((GEN)listgamma0[i],gmael(p1,1,i)));
+      p2 = gmul(p2, powgi((GEN)listgamma[i], (GEN)p1[i]));
+      p3 = gmul(p3, powgi((GEN)listgamma0[i],(GEN)p1[i]));
     }
-    matexpo[j]=(long)p1[1];
-    vecbeta[j]=(long)p2; /* attention, ceci sont les beta modifies */
-    vecbeta0[j]=(long)p3;
+    matexpo[j] = (long)p1; setlg(p1, rc+1);
+    vecbeta[j] = (long)p2; /* attention, ceci sont les beta modifies */
+    vecbeta0[j]= (long)p3;
   }
-  listg = gmodulcp(concatsp(gmael(bnf,8,5),gmael3(bnf,8,4,2)),polnf);
+  fununits = check_units(bnf,"rnfkummer");
+  torsunit = gmael3(bnf,8,4,2);
+  listg = gmodulcp(concatsp(fununits,torsunit),polnf);
   vunit = concatsp(listg, listalpha);
-  nbvunit=lg(vunit)-1;
-  id=idmat(degnf);
-  for (i=1; i<=lSl; i++)
+
+  vunit_beta = algtobasis(nf, concatsp(vunit, vecbeta));
+  l = lg(vunit_beta);
+{
+  long prec = DEFAULTPREC +
+    ((gexpo(vunit_beta) + gexpo(gmael(nf,5,1))) >> TWOPOTBYTES_IN_LONG);
+  if (nfgetprec(nf) < prec) nf = nfnewprec(nf, prec);
+}
+
+  vecmunit2 = cgetg(lSml2+1,t_VEC);
+
+  id = idmat(degnf);
+  for (i=1; i<=lSl2; i++)
   {
-    pr = (GEN)Sl[i]; e = itos((GEN)pr[3]);
-    id = idealmul(nf, id, idealpows(nf,pr,(ell*e)/(ell-1)));
+    GEN pr = (GEN)listprSp[i];
+    long e = itos((GEN)pr[3]), z = ell * (e / (ell-1));
+
+    if (i <= lSml2)
+    {
+      GEN c;
+      z += 1 - L.ESml2[i];
+      bid = zidealstarinitgen(nf, idealpows(nf, pr, z+1));
+      c = cgetg(l,t_MAT); vecmunit2[i] = (long)c;
+      for (j=1; j<l; j++) c[j] = (long)zideallog(nf,(GEN)vunit_beta[j],bid);
+    }
+    id = idealmulpowprime(nf, id, pr, stoi(z));
   }
-  vsml2=cgetg(lSml2+1,t_VEC);
-  for (i=1; i<=lSml2; i++)
-  {
-    pr = (GEN)Sml2[i]; e = itos((GEN)pr[3]);
-    p1=idealpows(nf,pr, (e*ell)/(ell-1) + 1 - ESml2[i]);
-    id = idealmul(nf,id,p1);
-    p2 = idealmul(nf,p1,pr);
-    p3 = zidealstarinitgen(nf,p2);
-    vsml2[i] = (long)p3;
-    cycpro = gmael(p3,2,2); l = lg(cycpro)-1;
-    if (! gdivise((GEN)cycpro[l],gell)) err(talker,"bug5 in kummer");
-  }
-  bid = zidealstarinitgen(nf,id);
+  nbvunit = lg(vunit)-1;
+  matexpo = concatsp(zeromat(rc,nbvunit), matexpo);
+  archartif = cgetg(r1+1,t_VEC); for (j=1; j<=r1; j++) archartif[j] = un;
+  munit = cgetg(l, t_MAT);
+  msign = cgetg(l, t_MAT);
+  bid = zidealstarinitgen(nf, id);
   lastbid = prank(gmael(bid,2,2), ell);
-  nbcol=nbvunit+lSp; nblig=lastbid+rc;
-  munit=cgetg(nbcol+1,t_MAT); vecmunit2=cgetg(lSml2+1,t_VEC);
-  msign=cgetg(nbcol+1,t_MAT);
-  for (k=1; k<=lSml2; k++) vecmunit2[k]=lgetg(nbcol+1,t_MAT);
-  archartif=cgetg(r1+1,t_VEC); for (j=1; j<=r1; j++) archartif[j]=un;
-  for (j=1; j<=nbvunit; j++)
+  for (j=1; j<l; j++)
   {
-    p1 = zideallog(nf,(GEN)vunit[j],bid);
-    p2 = cgetg(nblig+1,t_COL); munit[j]=(long)p2;
-    for (i=1; i<=lastbid; i++) p2[i]=p1[i];
-    for (   ; i<=nblig; i++) p2[i]=zero;
-    for (k=1; k<=lSml2; k++)
-      mael(vecmunit2,k,j) = (long)zideallog(nf,(GEN)vunit[j],(GEN)vsml2[k]);
-    msign[j] = (long)zsigne(nf,(GEN)vunit[j],archartif);
+    GEN z = (GEN)vunit_beta[j];
+    p1 = zideallog(nf,z,bid); setlg(p1, lastbid+1);
+    munit[j] = (long)concatsp(p1, (GEN)matexpo[j]);
+    msign[j] = (long)zsigne(nf, z, archartif);
   }
-  for (j=1; j<=lSp; j++)
-  {
-    p1 = zideallog(nf,(GEN)vecbeta[j],bid);
-    p2 = cgetg(nblig+1,t_COL); munit[j+nbvunit]=(long)p2;
-    for (i=1; i<=lastbid; i++) p2[i]=p1[i];
-    for (   ; i<=nblig; i++) p2[i]=coeff(matexpo,i-lastbid,j);
-    for (k=1; k<=lSml2; k++)
-      mael(vecmunit2,k,j+nbvunit) = (long)zideallog(nf,(GEN)vecbeta[j],(GEN)vsml2[k]);
-    msign[j+nbvunit] = (long)zsigne(nf,(GEN)vecbeta[j],archartif);
-  }
-  listx = get_listx(arch,msign,munit,vecmunit2,ell,lSp,nbvunit);
+
+  listx = get_listx(arch,msign,munit,vecmunit2,gell,lSp,nbvunit);
   llistx= lg(listx);
   listal= cgetg(llistx,t_VEC);
   listg = concatsp(listg, concatsp(vecalpha0,vecbeta0));
@@ -324,10 +443,9 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
       p1 = gmul(p1, powgi((GEN)listg[j],(GEN)p2[j]));
     listal[i] = (long)reducealpha(nf,p1,gell);
   }
- /* A ce stade, tous les alpha dans la liste listal statisfont les
-  * congruences, noncongruences et signatures, et ne sont pas des puissances
-  * l-iemes donc x^l-alpha est irreductible, sa signature est correcte ainsi
-  * que son discriminant relatif. Reste a determiner son groupe de normes. */
+ /* Now, alpha in listal satisfies all congruences, non-congruences,
+  * x^l - alpha is irreducible, signature and relative disciminant are
+  * correct. Remains to check its norm group. */
   if (DEBUGLEVEL) fprintferr("listalpha = %Z\n",listal);
   p2 = cgetg(1,t_VEC);
   for (i=1; i<llistx; i++)
@@ -335,117 +453,34 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
     p1 = gsub(gpuigs(polx[0],ell), (GEN)listal[i]);
     if (all || gegal(rnfnormgroup(bnr,p1),subgroup)) p2 = concatsp(p2,p1);
   }
-  if (all) return gcopy(p2);
+  if (all) return p2;
   switch(lg(p2)-1)
   {
     case 0: err(talker,"bug 6: no equation found in kummer");
     case 1: break; /* OK */
-    default: fprintferr("equations = \n%Z",p2);
-      err(talker,"bug 7: more than one equation found in kummer");
+    default:err(talker,"bug 7: more than one equation found in kummer: %Z",p2);
   }
-  return gcopy((GEN)p2[1]);
+  return (GEN)p2[1];
 }
 
 static GEN
-tauofelt(GEN nf, GEN x, GEN U)
-{
-  return algtobasis(nfz, gsubst(gmul((GEN)nf[7],x),vnf,U));
-}
-static GEN
-tauofalg(GEN x, GEN U)
-{
-  return gsubst(lift(x),vnf,U);
-}
-
-static GEN
-lifttoKz(GEN nfz, GEN nf, GEN id, GEN A1)
-{
-  GEN I = ideal_two_elt(nf,id);
-  I[2] = (long)tauofelt(nf,(GEN)I[2],A1);
-  return prime_to_ideal(nfz,I);
-}
-
-
-static GEN
-tauofideal(GEN id, GEN U)
-{
-  GEN I = ideal_two_elt(nfz,id);
-  I[2] = (long)tauofelt(nfz,(GEN)I[2],U);
-  return prime_to_ideal(nfz,I);
-}
-
-static int
-isprimeidealconj(GEN pr1, GEN pr2, GEN U)
-{
-  GEN p = (GEN)pr1[1];
-  GEN x = (GEN)pr1[2];
-  GEN b1= (GEN)pr1[5];
-  GEN b2= (GEN)pr2[5];
-  if (!egalii(p, (GEN)pr2[1])
-   || !egalii((GEN)pr1[3], (GEN)pr2[3])
-   || !egalii((GEN)pr1[4], (GEN)pr2[4])) return 0;
-  if (gegal(x,(GEN)pr2[2])) return 1;
-  for(;;)
-  {
-    if (int_elt_val(nfz,x,p,b2,NULL,0)) return 1;
-    x = FpV_red(tauofelt(nfz,x,U), p);
-    if (int_elt_val(nfz,x,p,b1,NULL,0)) return 0;
-  }
-}
-
-static int
-isconjinprimelist(GEN S, GEN pr2, GEN U)
-{
-  long lS=lg(S),i;
-
-  for (i=1; i<lS; i++)
-    if (isprimeidealconj((GEN)S[i],pr2,U)) return 1;
-  return 0;
-}
-
-static GEN
-downtoK(GEN polnf, GEN x)
-{
-  GEN y = gmul(invexpoteta1, pol_to_vec(lift(x), degKz));
-  return gmodulcp(gtopolyrev(y,vnf),polnf);
-}
-
-static GEN
-tracetoK(GEN polnf, GEN x, GEN U)
-{
-  GEN p1 = x;
-  long i;
-  for (i=1; i<m; i++) p1 = gadd(x, tauofalg(p1,U));
-  return downtoK(polnf,p1);
-}
-
-static GEN
-normtoK(GEN polnf, GEN x, GEN U)
-{
-  GEN p1 = x;
-  long i;
-  for (i=1; i<m; i++) p1 = gmul(x, tauofalg(p1,U));
-  return downtoK(polnf,p1);
-}
-
-static GEN
-computepolrel(GEN polnf, GEN be, GEN U)
+computepolrel(toK_t *T, GEN be)
 {
   GEN p1 = gun, p2 = be;
   long i,j;
 
-  for (i=0; i<m; i++)
+  for (i=0; i<T->m; i++)
   {
-    p1 = gmul(p1,gsub(polx[0],p2));
-    if (i<m-1) p2 = tauofalg(p2,U);
+    p1 = gmul(p1, gsub(polx[0],p2));
+    if (i < T->m-1) p2 = tauofalg(p2, T->U);
   }
-  for (j=2; j<=m+2; j++) p1[j]=(long)downtoK(polnf,(GEN)p1[j]);
+  for (j=2; j<=T->m+2; j++) p1[j] = (long)downtoK(T, (GEN)p1[j]);
   return p1;
 }
 
 /* alg. 5.2.15. with remark */
 static GEN
-isprincipalell(GEN bnfz, GEN id, GEN vecalpha, GEN uu)
+isprincipalell(GEN bnfz, GEN id, GEN vecalpha, GEN uu, GEN gell, long rc)
 {
   long i, l = lg(vecalpha);
   GEN y,logdisc,be;
@@ -463,10 +498,10 @@ isprincipalell(GEN bnfz, GEN id, GEN vecalpha, GEN uu)
 
 /* alg. 5.3.11. */
 static GEN
-isvirtualunit(GEN bnfz, GEN v, GEN vecalpha, GEN cyc)
+isvirtualunit(GEN bnfz, GEN v, GEN vecalpha, GEN cyc, GEN gell, long rc)
 {
-  long llist,i,j,ex, l = lg(vecalpha);
-  GEN p1,listex,listpr,q,ga,eps,vecy,logdisc;
+  long llist,i,j,ex, l = lg(vecalpha), ell = itos(gell);
+  GEN p1,listex,listpr,q,ga,eps,vecy,logdisc, nfz = checknf(bnfz);
 
   p1=idealfactor(nfz,v);
   listpr = (GEN)p1[1]; llist = lg(listpr)-1;
@@ -498,21 +533,21 @@ isvirtualunit(GEN bnfz, GEN v, GEN vecalpha, GEN cyc)
 }
 
 static GEN
-steinitzaux(GEN nf, GEN id, GEN polrel)
+steinitzaux(GEN nfz, GEN nf, GEN id, GEN polrel, long m)
 {
-  long i,j;
+  long i,j, degKz = degpol(nfz[1]), degK = degpol(nf[1]);
   GEN p1,p2,p3,vecid,matid,pseudomat,pid;
 
-  p1=gsubst(gmul((GEN)nfz[7],id),vnf,polx[0]);
+  p1 = gsubst(gmul((GEN)nfz[7],id), varn(nfz[1]), polx[0]);
   for (j=1; j<=degKz; j++)
-    p1[j]=(long)gmod((GEN)p1[j],polrel);
+    p1[j] = (long)gmod((GEN)p1[j], polrel);
   p2=cgetg(degKz+1,t_MAT);
   for (j=1; j<=degKz; j++)
   {
     p3=cgetg(m+1,t_COL); p2[j]=(long)p3;
     for (i=1; i<=m; i++) p3[i]=(long)algtobasis(nf,truecoeff((GEN)p1[j],i-1));
   }
-  vecid=cgetg(degKz+1,t_VEC); matid=idmat(degK);
+  vecid=cgetg(degKz+1,t_VEC); matid = idmat(degK);
   for (j=1; j<=degKz; j++) vecid[j]=(long)matid;
   pseudomat=cgetg(3,t_VEC);
   pseudomat[1]=(long)p2; pseudomat[2]=(long)vecid;
@@ -523,28 +558,28 @@ steinitzaux(GEN nf, GEN id, GEN polrel)
 }
 
 static GEN
-normrelz(GEN nf, GEN id, GEN polrel, GEN steinitzZk)
+normrelz(GEN nfz, GEN nf, GEN id, GEN polrel, GEN steinitzZk, long m)
 {
-  GEN p1 = steinitzaux(nf,idealhermite(nfz, id), polrel);
+  GEN p1 = steinitzaux(nfz,nf,idealhermite(nfz, id), polrel, m);
   return idealdiv(nf,p1,steinitzZk);
 }
 
 static GEN
-invimsubgroup(GEN bnrz, GEN bnr, GEN subgroup, GEN tau)
+invimsubgroup(toK_t *T, GEN bnrz, GEN bnr, GEN subgroup)
 {
   long l,j;
   GEN g,Plog,raycycz,rayclgpz,genraycycz,U,polrel,steinitzZk;
-  GEN nf = checknf(bnr);
+  GEN nf = checknf(bnr), nfz = checknf(bnrz), polz = (GEN)nfz[1];
 
-  polrel = computepolrel((GEN)nf[1], polx[vnf], tau);
-  steinitzZk = steinitzaux(nf,idmat(degKz), polrel); 
+  polrel = computepolrel(T, polx[varn(polz)]);
+  steinitzZk = steinitzaux(nfz,nf,idmat(degpol(polz)), polrel, T->m); 
   rayclgpz = (GEN)bnrz[5];
-  raycycz   = (GEN)rayclgpz[2]; l=lg(raycycz);
+  raycycz   = (GEN)rayclgpz[2]; l = lg(raycycz);
   genraycycz= (GEN)rayclgpz[3];
   Plog = cgetg(l,t_MAT);
   for (j=1; j<l; j++)
   {
-    g = normrelz(nf,(GEN)genraycycz[j],polrel,steinitzZk);
+    g = normrelz(nfz,nf,(GEN)genraycycz[j],polrel,steinitzZk,T->m);
     Plog[j] = (long)isprincipalray(bnr, g);
   }
   U = (GEN)hnfall(concatsp(Plog, subgroup))[2];
@@ -639,69 +674,73 @@ mulpoltau(GEN poltau, GEN list)
 
 /* th. 5.3.5. and prop. 5.3.9. */
 static GEN
-computepolrelbeta(GEN polnf, GEN be, long g, GEN U)
+computepolrelbeta(toK_t *T, GEN be, long g, long ell)
 {
-  long i,a,b,j;
-  GEN e,u,u1,u2,u3,p1,p2,p3,p4,zet,be1,be2,listr,s,veczi,vecci,powtaubet;
+  long i, a, b, j, m, vnf;
+  GEN e,u,u1,u2,u3,p1,p2,zet,be1,be2,listr,s,veczi,vecci,powtaubet;
 
   switch (ell)
   {
     case 2: err(talker,"you should not be here in rnfkummer !!"); break;
-    case 3: e=normtoK(polnf,be,U); u=tracetoK(polnf,be,U);
+    case 3: e=normtoK(T,be); u=tracetoK(T,be);
       return gsub(gmul(polx[0],gsub(gsqr(polx[0]),gmulsg(3,e))),gmul(e,u));
-    case 5: e=normtoK(polnf,be,U);
-      if (d==2)
+    case 5: e=normtoK(T,be);
+      if (ell-1 == 2*T->m) /* d == 2 */
       {
-	u=tracetoK(polnf,gpuigs(be,3),U);
+	u=tracetoK(T,gpuigs(be,3));
 	p1=gadd(gmulsg(5,gsqr(e)), gmul(gsqr(polx[0]), gsub(gsqr(polx[0]),gmulsg(5,e))));
 	return gsub(gmul(polx[0],p1),gmul(e,u));
       }
       else
       {
-	be1=tauofalg(be,U);
-        be2=tauofalg(be1,U);
-	u1=tracetoK(polnf,gmul(be,be1),U);
-        u2=tracetoK(polnf,gmul(gmul(be,be2),gsqr(be1)),U);
-	u3=tracetoK(polnf,gmul(gmul(gsqr(be),gpuigs(be1,3)),be2),U);
+	be1=tauofalg(be,T->U);
+        be2=tauofalg(be1,T->U);
+	u1=tracetoK(T,gmul(be,be1));
+        u2=tracetoK(T,gmul(gmul(be,be2),gsqr(be1)));
+	u3=tracetoK(T,gmul(gmul(gsqr(be),gpuigs(be1,3)),be2));
 	p1=gsub(gsqr(polx[0]),gmulsg(10,e));
 	p1=gsub(gmul(polx[0],p1),gmulsg(5,gmul(e,u1)));
 	p1=gadd(gmul(polx[0],p1),gmul(gmulsg(5,e),gsub(e,u2)));
 	p1=gsub(gmul(polx[0],p1),gmul(e,u3));
 	return p1;
       }
-    default: p1=cgetg(2,t_VEC); p2=cgetg(3,t_VEC); p3=cgetg(2,t_VEC);
-      p4=cgetg(2,t_VEC); p3[1]=zero; p4[1]=un;
-      p2[1]=(long)p3; p2[2]=(long)p4; p1[1]=(long)p2;
-      zet=gmodulcp(polx[vnf], cyclo(ell,vnf));
-      listr=cgetg(m+1,t_VEC);
-      listr[1]=un;
-      for (i=2; i<=m; i++) listr[i]=(long)modii(mulis((GEN)listr[i-1],g),gell);
+    default: p2 = cgetg(3,t_VEC);
+      p2[1] = (long)_vec(gzero);
+      p2[2] = (long)_vec(gun); p1 = _vec(p2);
+      m = T->m;
+      vnf = varn(T->polnf);
+      zet = gmodulcp(polx[vnf], cyclo(ell,vnf));
+      listr = cgetg(m+1,t_VECSMALL);
+      listr[1] = 1;
+      for (i=2; i<=m; i++) listr[i] = (listr[i-1]*g) % ell;
       veczi=cgetg(m+1,t_VEC);
       for (b=0; b<m; b++)
       {
-	s=gzero;
+	s = gzero;
 	for (a=0; a<m; a++)
-	  s=gadd(gmul(polx[0],s),modii(mulii((GEN)listr[b+1],(GEN)listr[a+1]),gell));
+	  s = gaddgs(gmul(polx[0],s), (listr[b+1] * listr[a+1]) % ell);
 	veczi[b+1]=(long)s;
       }
       for (j=0; j<ell; j++)
       {
-	vecci=cgetg(m+1,t_VEC);
-	for (b=0; b<m; b++) vecci[b+1]=(long)gpui(zet,mulsi(j,(GEN)listr[b+1]),0);
-	p4=cgetg(3,t_VEC);
-	p4[1]=(long)veczi; p4[2]=(long)vecci;
-	p1=mulpoltau(p1,p4);
+	GEN p3 = cgetg(3,t_VEC);
+	vecci = cgetg(m+1,t_VEC);
+	for (b=0; b<m; b++) vecci[b+1] = (long)gpowgs(zet, j * listr[b+1]);
+	p3[1] = (long)veczi;
+        p3[2] = (long)vecci;
+	p1 = mulpoltau(p1,p3);
       }
       powtaubet=cgetg(m+1,t_VEC);
       powtaubet[1]=(long)be;
-      for (i=2; i<=m; i++) powtaubet[i]=(long)tauofalg((GEN)powtaubet[i-1],U);
-      err(impl,"difficult Kummer for ell>=7"); return gzero;
+      for (i=2; i<=m; i++)
+        powtaubet[i] = (long)tauofalg((GEN)powtaubet[i-1],T->U);
+      err(impl,"difficult Kummer for ell>=7"); /* FIXME */
   }
   return NULL; /* not reached */
 }
 
 static GEN
-fix_be(GEN bnfz, GEN be, GEN u)
+fix_be(GEN bnfz, GEN be, GEN u, long ell)
 {
   GEN e,g, nf = checknf(bnfz), fu = gmael(bnfz,8,5);
   long i,lu;
@@ -735,7 +774,7 @@ logarch2arch(GEN x, long r1, long prec)
 
 /* multiply be by ell-th powers of units as to find small L2-norm for new be */
 static GEN
-reducebetanaive(GEN bnfz, GEN be, GEN b)
+reducebetanaive(GEN bnfz, GEN be, GEN b, long ell)
 {
   long i,k,n,ru,r1, prec = nfgetprec(bnfz);
   GEN z,p1,p2,nmax,c, nf = checknf(bnfz);
@@ -769,16 +808,16 @@ reducebetanaive(GEN bnfz, GEN be, GEN b)
     b = B; c[besti] = laddis((GEN)c[besti], bestk);
   }
   if (DEBUGLEVEL) fprintferr("unit exponents = %Z\n",c);
-  return fix_be(bnfz,be,c);
+  return fix_be(bnfz,be,c,ell);
 }
 
 static GEN
-reducebeta(GEN bnfz, GEN be)
+reducebeta(GEN bnfz, GEN be, long ell)
 {
   long j,ru, prec = nfgetprec(bnfz);
   GEN emb,z,u,matunit, nf = checknf(bnfz);
 
-  if (gcmp1(gnorm(be))) return reducebetanaive(bnfz,be,NULL);
+  if (gcmp1(gnorm(be))) return reducebetanaive(bnfz,be,NULL,ell);
   matunit = gmulgs(greal((GEN)bnfz[3]), ell); /* log. embeddings of fu^ell */
   for (;;)
   {
@@ -789,40 +828,38 @@ reducebeta(GEN bnfz, GEN be)
     nf = nfnewprec(nf,prec);
   }
   z = concatsp(matunit, z);
-  u = lllintern(z, 1, prec);
-  if (!u) return reducebetanaive(bnfz,be,emb); /* shouldn't occur */
+  u = lllintern(z, 100, 1, prec);
+  if (!u) return reducebetanaive(bnfz,be,emb,ell); /* shouldn't occur */
   ru = lg(u);
   for (j=1; j<ru; j++)
     if (smodis(gcoeff(u,ru-1,j), ell)) break; /* prime to ell */
   u = (GEN)u[j]; /* coords on (fu^ell, be) of a small generator */
   ru--; setlg(u, ru);
   be = powgi(be, (GEN)u[ru]);
-  return reducebetanaive(bnfz,fix_be(bnfz,be,u), NULL);
+  return reducebetanaive(bnfz,fix_be(bnfz,be,u,ell), NULL, ell);
 }
 
 /* cf. algo 5.3.18 */
 static GEN
-testx(GEN bnfz, GEN bnf, GEN X, GEN module, GEN subgroup, GEN vecMsup,
-      GEN vecWB, long g, GEN U)
+testx(toK_t *T, GEN bnfz, GEN bnr, GEN X, GEN subgroup, GEN vecMsup,
+      GEN vecWB, long g, GEN gell, long lW)
 {
-  long i,l,lX;
+  long i,l,lX, ell = itos(gell);
   GEN be,polrelbe,p1,nf;
 
   if (gcmp0(X)) return NULL;
   lX = lg(X);
-  for (i=dv+1; i<lX; i++)
+  for (i=lW; i<lX; i++)
     if (gcmp0((GEN)X[i])) return NULL;
   l = lg(vecMsup);
   for (i=1; i<l; i++)
     if (gcmp0(FpV_red(gmul((GEN)vecMsup[i],X), gell))) return NULL;
-  be = gun;
-  for (i=1; i<lX; i++)
-    be = gmul(be, powgi((GEN)vecWB[i], (GEN)X[i]));
+  be = factorback(vecWB, X);
   if (DEBUGLEVEL>1) fprintferr("reducing beta = %Z\n",be);
-  be = reducebeta(bnfz, be);
+  be = reducebeta(bnfz, be, ell);
   if (DEBUGLEVEL>1) fprintferr("beta reduced = %Z\n",be);
-  nf = (GEN)bnf[7];
-  polrelbe = computepolrelbeta((GEN)nf[1],be,g,U);
+  nf = checknf(bnr);
+  polrelbe = computepolrelbeta(T, be, g, ell);
  
   p1 = unifpol(nf,polrelbe,0);
   l = lg(p1); /* lift to Q rational coeffs */
@@ -832,25 +869,33 @@ testx(GEN bnfz, GEN bnf, GEN X, GEN module, GEN subgroup, GEN vecMsup,
   p1 = denom(gtovec(p1));
   polrelbe = rescale_pol(polrelbe,p1);
   if (DEBUGLEVEL>1) fprintferr("polrelbe = %Z\n",polrelbe);
-  p1 = rnfconductor(bnf,polrelbe,0);
-  if (!gegal((GEN)p1[1],module) || !gegal((GEN)p1[3],subgroup)) return NULL;
+
+  p1 = rnfnormgroup(bnr,polrelbe);
+  if (!gegal(p1,subgroup)) return NULL;
   return polrelbe;
 }
 
 static GEN
-logall(long dv, GEN mod, GEN vecWA, long mginv, GEN pr, GEN unif)
+logall(GEN nfz, GEN vecWA, long lW, long mginv, long ell, GEN pr, long ex)
 {
-  GEN p1,al,M,bid = zidealstarinitgen(nfz,mod);
-  long ellrank = prank(gmael(bid,2,2), ell);
-  long val,i, l = lg(vecWA);
+  GEN p1, bid, al, M;
+  long ellrank, val, i, l = lg(vecWA);
 
+  bid = zidealstarinitgen(nfz, idealpows(nfz, pr, ex));
+  ellrank = prank(gmael(bid,2,2), ell);
   M = cgetg(l,t_MAT);
   for (i=1; i<l; i++)
   {
-    al = (GEN)vecWA[i]; val = element_val(nfz,al,pr);
-    if (val) al = gmul(al, gpuigs(unif,val));
+    al = algtobasis(nfz, (GEN)vecWA[i]);
+    val = element_val(nfz,al,pr);
+    if (val)
+    {
+      GEN p = (GEN)pr[1], tau = (GEN)pr[5];
+      al = element_mul(nfz, al, element_pow(nfz, tau, stoi(val)));
+      al = Q_div_to_int(al, gpowgs(p, val));
+    }
     p1 = zideallog(nfz, al, bid); setlg(p1,ellrank+1);
-    if (i <= dv) p1 = gmulsg(mginv, p1);
+    if (i < lW) p1 = gmulsg(mginv, p1);
     M[i] = (long)p1;
   }
   return M;
@@ -859,44 +904,39 @@ logall(long dv, GEN mod, GEN vecWA, long mginv, GEN pr, GEN unif)
 GEN
 rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
 {
-  long i,j,l,av=avma,e,vp,vd,dK,lSl,lSp,lSl2,lSml2,dc,ru,rv,nbcol,g,mginv;
-  GEN p1,p2,p3,p4,wk,pr,U,R;
-  GEN polnf,nf,bnf,bnfz,rayclgp,bid,ideal,cycgen,vselmer;
+  long i, j, l, m, d, dK, dc, rc, ru, rv, g, mginv, degK, degKz, ell;
+  long lSp, lSl2, lSml2, lW, vnf;
+  gpmem_t av = avma;
+  GEN p1,p2,p3,p4,wk,U,R,gell;
+  GEN polnf,nf,bnf,bnfz,bid,ideal,cycgen,vselmer;
   GEN kk,clgp,fununits,torsunit,vecB,vecC,Tc,Tv,P;
-  GEN Q,idealz,gothf,factgothf,factell,p;
-  GEN listprSp,listpr,listex,vecw,vecWA,vecWB;
-  GEN M,K,X,finalresult,y,module,A1,A2,vecMsup;
-  GEN uu,gencyc,cyc,vecalpha,vecalphap,vecbetap,matP,ESml2,Sp,Sm,Sml1,Sml2,Sl;
+  GEN Q,idealz,gothf,factgothf,nfz;
+  GEN listprSp,vecw,vecWA,vecWB;
+  GEN M,K,X,finalresult,y,A1,A2,A3,A3rev,vecMsup;
+  GEN uu,gencyc,cyc,vecalpha,vecalphap,vecbetap,matP,Sp;
+  primlist L;
+  toK_t T;
 
   checkbnrgen(bnr);
-  wk = gmael4(bnr,1,8,4,1);
-  if (all) gell = stoi(all);
-  else
-  {
-    if (!gcmp0(subgroup)) gell = det(subgroup);
-    else gell = det(diagonal(gmael(bnr,5,2)));
-    if (typ(gell) != t_INT) err(arither1);
-  }
-  if (gcmp1(gell)) { avma = av; return polx[varn(gmael3(bnr,1,7,1))]; }
-  if (!isprime(gell)) err(impl,"kummer for composite relative degree");
-  if (divise(wk,gell))
-    return gerepileupto(av,rnfkummersimple(bnr,subgroup,all));
-  if (all && gcmp0(subgroup))
-    err(talker,"kummer when zeta not in K requires a specific subgroup");
   bnf = (GEN)bnr[1];
   nf  = (GEN)bnf[7];
-  polnf = (GEN)nf[1]; vnf = varn(polnf); degK = degpol(polnf);
+  polnf = (GEN)nf[1]; vnf = varn(polnf);
   if (!vnf) err(talker,"main variable in kummer must not be x");
+  wk = gmael3(bnf,8,4,1);
       /* step 7 */
-  p1 = conductor(bnr,subgroup,2);
-      /* fin step 7 */
-  bnr = (GEN)p1[2]; 
+  if (all) subgroup = NULL;
+  p1 = conductor(bnr, subgroup, 2);
+  bnr      = (GEN)p1[2]; 
   subgroup = (GEN)p1[3];
-  rayclgp = (GEN)bnr[5];
-  
-  bid = (GEN)bnr[2]; module=(GEN)bid[1];
-  ideal = (GEN)module[1];
-  if (gcmp0(subgroup)) subgroup = diagonal((GEN)rayclgp[2]);
+  gell = get_gell(bnr,subgroup,all);
+  if (gcmp1(gell)) { avma = av; return polx[vnf]; }
+  if (!isprime(gell)) err(impl,"kummer for composite relative degree");
+  if (divise(wk,gell))
+    return gerepilecopy(av, rnfkummersimple(bnr,subgroup,all));
+  if (all) err(impl,"extensions by degree in kummer when zeta not in K");
+
+  bid = (GEN)bnr[2];
+  ideal = gmael(bid,1,1);
   ell = itos(gell);
       /* step 1 of alg 5.3.5. */
   if (DEBUGLEVEL>2) fprintferr("Step 1\n");
@@ -908,27 +948,22 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   kk= (GEN)p1[4];
       /* step 2 */
   if (DEBUGLEVEL>2) fprintferr("Step 2\n");
+  degK  = degpol(polnf);
   degKz = degpol(R);
   m = degKz/degK;
   d = (ell-1)/m;
-  g = powuumod(itos(lift(gener(gell))), d, ell);
+  g = powuumod(u_gener(ell), d, ell);
   if (powuumod(g, m, ell*ell) == 1) g += ell; /* ord(g)=m in all (Z/ell^k)^* */
       /* step reduction of R */
-  if (degKz <= 20)
-  {
-    GEN A3,A3rev;
-
-    if (DEBUGLEVEL>2) fprintferr("Step reduction\n");
-    p1 = polredabs2(R,prec);
-    R = (GEN)p1[1]; if (DEBUGLEVEL>2) fprintferr("polredabs = %Z",R);
-    A3= (GEN)p1[2];
-    A1 = poleval(lift(A1), A3);
-    A2 = poleval(lift(A2), A3);
-    A3rev= polymodrecip(A3);
-    U = gadd(gpowgs(A2,g), gmul(kk,A1));
-    U = poleval(lift(A3rev), U);
-  }
-  else U = gadd(gpowgs(A2,g), gmul(kk,A1));
+  if (DEBUGLEVEL>2) fprintferr("Step reduction\n");
+  p1 = polredabs0(R, nf_ORIG|nf_PARTIALFACT);
+  R = (GEN)p1[1]; if (DEBUGLEVEL>2) fprintferr("polredabs = %Z",R);
+  A3= (GEN)p1[2];
+  A1 = poleval(lift(A1), A3);
+  A2 = poleval(lift(A2), A3);
+  A3rev= modreverse_i((GEN)A3[2], (GEN)A3[1]);
+  U = gadd(gpowgs(A2,g), gmul(kk,A1));
+  U = poleval(A3rev, U);
       /* step 3 */
       /* one could factor disc(R) using th. 2.1.6. */
   if (DEBUGLEVEL>2) fprintferr("Step 3\n");
@@ -960,7 +995,8 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   Tc=cgetg(rc+1,t_MAT);
   for (j=1; j<=rc; j++)
   {
-    p1 = isprincipalell(bnfz, tauofideal((GEN)gencyc[j],U), vecalpha,uu);
+    p1 = tauofideal(nfz,(GEN)gencyc[j],U);
+    p1 = isprincipalell(bnfz, p1, vecalpha,uu,gell,rc);
     Tc[j]  = p1[1];
     vecB[j]= p1[2];
   }
@@ -982,12 +1018,13 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   Tv = cgetg(rv+1,t_MAT);
   for (j=1; j<=rv; j++)
   {
-    Tv[j] = isvirtualunit(bnfz, tauofalg((GEN)vselmer[j],U), vecalpha,cyc)[1];
+    p1 = tauofalg((GEN)vselmer[j],U);
+    Tv[j] = isvirtualunit(bnfz, p1, vecalpha,cyc,gell,rc)[1];
     if (DEBUGLEVEL>2) fprintferr("   %ld\n",j);
   }
   P = FpM_ker(gsubgs(Tv, g), gell);
-  dv= lg(P)-1; vecw = cgetg(dv+1,t_VEC);
-  for (j=1; j<=dv; j++)
+  lW= lg(P); vecw = cgetg(lW,t_VEC);
+  for (j=1; j<lW; j++)
   {
     p1 = gun;
     for (i=1; i<=rv; i++) p1 = gmul(p1, powgi((GEN)vselmer[i],gcoeff(P,i,j)));
@@ -1009,91 +1046,57 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
     p2[j] = (long)pol_to_vec(p1, degKz);
     if (j<degK) p1 = gmod(gmul(p1,A1), R);
   }
-  invexpoteta1 = invmat(p2); /* left inverse */
+  T.invexpoteta1 = invmat(p2); /* left inverse */
+  T.polnf = polnf;
+  T.U = U;
+  T.m = m;
  
   if (!divise(idealnorm(nf,ideal),gell)) gothf = idealz;
   else
   {
-    GEN bnrz = buchrayinitgen(bnfz,idealz);
-    GEN subgroupz = invimsubgroup(bnrz,bnr,subgroup,U);
+    GEN subgroupz, bnrz = buchrayinitgen(bnfz,idealz);
+    subgroupz = invimsubgroup(&T, bnrz,bnr,subgroup);
     gothf = conductor(bnrz,subgroupz,0);
   }
       /* step 9 */
   if (DEBUGLEVEL>2) fprintferr("Step 9\n");
   factgothf=idealfactor(nfz,gothf);
-  listpr = (GEN)factgothf[1];
-  listex = (GEN)factgothf[2]; l = lg(listpr);
       /* step 10 and step 11 */
   if (DEBUGLEVEL>2) fprintferr("Step 10 and 11\n");
-  Sm  = cgetg(l,t_VEC); setlg(Sm,1);
-  Sml1= cgetg(l,t_VEC); setlg(Sml1,1);
-  Sml2= cgetg(l,t_VEC); setlg(Sml2,1);
-  Sl  = cgetg(l+degKz,t_VEC); setlg(Sl,1);
-  ESml2=cgetg(l,t_VECSMALL); setlg(ESml2,1);
-  for (i=1; i<l; i++)
-  {
-    pr = (GEN)listpr[i]; p = (GEN)pr[1]; e = itos((GEN)pr[3]);
-    vp = itos((GEN)listex[i]);
-    if (!egalii(p,gell))
-    {
-      if (vp != 1) { avma = av; return gzero; }
-      if (!isconjinprimelist(Sm,pr,U)) _append(Sm,pr);
-    }
-    else
-    {
-      vd = (vp-1)*(ell-1)-ell*e;
-      if (vd > 0) { avma = av; return gzero; }
-      if (vd==0)
-      {
-	if (!isconjinprimelist(Sml1,pr,U)) _append(Sml1, pr);
-      }
-      else
-      {
-	if (vp==1) { avma = av; return gzero; }
-        if (!isconjinprimelist(Sml2,pr,U))
-        {
-          _append(Sml2, pr);
-          _append(ESml2,(GEN)vp);
-        }
-      }
-    }
-  }
-  factell = primedec(nfz,gell); l = lg(factell);
-  for (i=1; i<l; i++)
-  {
-    pr = (GEN)factell[i];
-    if (!idealval(nfz,gothf,pr))
-      if (!isconjinprimelist(Sl,pr,U)) _append(Sl, pr);
-  }
-  lSml2=lg(Sml2)-1; lSl=lg(Sl)-1; lSl2=lSml2+lSl;
-  Sp=concatsp(Sm,Sml1); lSp=lg(Sp)-1;
+  i = build_list_Hecke(&L, nfz, factgothf, gothf, gell, U);
+  if (i) return no_sol(all,i);
+
+  lSml2 = lg(L.Sml2)-1;
+  Sp = concatsp(L.Sm, L.Sml1); lSp = lg(Sp)-1;
+  listprSp = concatsp(L.Sml2, L.Sl); lSl2 = lg(listprSp)-1;
+
       /* step 12 */
   if (DEBUGLEVEL>2) fprintferr("Step 12\n");
-  vecbetap=cgetg(lSp+1,t_VEC);
-  vecalphap=cgetg(lSp+1,t_VEC);
-  matP=cgetg(lSp+1,t_MAT);
+  vecbetap = cgetg(lSp+1,t_VEC);
+  vecalphap= cgetg(lSp+1,t_VEC);
+  matP = cgetg(lSp+1,t_MAT);
   for (j=1; j<=lSp; j++)
   {
-    p1 = isprincipalell(bnfz, (GEN)Sp[j], vecalpha,uu);
+    p1 = isprincipalell(bnfz, (GEN)Sp[j], vecalpha,uu,gell,rc);
     matP[j] = p1[1];
     p2 = gun;
     for (i=1; i<=rc; i++)
       p2 = gmul(p2, powgi((GEN)vecC[i],gmael(p1,1,i)));
-    p3=gdiv((GEN)p1[2],p2); vecbetap[j]=(long)p3;
-    p2=gun;
+    p3 = gdiv((GEN)p1[2],p2);
+    vecbetap[j] = (long)p3;
+    p2 = gun;
     for (i=0; i<m; i++)
     {
       p2 = gmul(p2, gpowgs(p3, powuumod(g,m-1-i,ell)));
       if (i<m-1) p3 = tauofalg(p3,U);
     }
-    vecalphap[j]=(long)p2;
+    vecalphap[j] = (long)p2;
   }
       /* step 13 */
   if (DEBUGLEVEL>2) fprintferr("Step 13\n");
-  nbcol= lSp+dv;
   vecWB = concatsp(vecw, vecbetap);
   vecWA = concatsp(vecw, vecalphap);
-  listprSp = concatsp(Sml2, Sl);
+
       /* step 14, 15, and 17 */
   if (DEBUGLEVEL>2) fprintferr("Step 14, 15 and 17\n");
   mginv = (m * u_invmod(g,ell)) % ell;
@@ -1103,24 +1106,20 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   {
     GEN pr = (GEN)listprSp[i];
     long e = itos((GEN)pr[3]), z = ell * (e / (ell-1));
-    GEN mod, unif = basistoalg(nfz,gdiv((GEN)pr[5],(GEN)pr[1]));
 
-    if (i <= lSml2) z += 1 - ESml2[i];
-    mod  = idealpows(nfz,pr, z);
-
-    M = vconcat(M, logall(dv,mod,vecWA,mginv,pr,unif));
     if (i <= lSml2)
     {
-      GEN modsup = idealmul(nfz,pr,mod);
-      vecMsup[i] = (long)logall(dv,modsup,vecWA,mginv,pr,unif);
+      z += 1 - L.ESml2[i];
+      vecMsup[i] = (long)logall(nfz, vecWA,lW,mginv,ell,pr, z+1);
     }
+    M = vconcat(M, logall(nfz, vecWA,lW,mginv,ell,pr, z));
   }
   if (dc)
   {
     GEN QtP = gmul(gtrans_i(Q),matP);
-    M = vconcat(M, concatsp(zeromat(dc,dv), QtP));
+    M = vconcat(M, concatsp(zeromat(dc,lW-1), QtP));
   }
-  if (!M) M = zeromat(1,nbcol);
+  if (!M) M = zeromat(1, lSp + lW - 1);
       /* step 16 */
   if (DEBUGLEVEL>2) fprintferr("Step 16\n");
   K = FpM_ker(M, gell);
@@ -1137,7 +1136,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
       X = (GEN)K[dK];
       for (j=1; j<dK; j++) X = gadd(X, gmulsg(y[j],(GEN)K[j]));
       X = FpV_red(X, gell);
-      finalresult = testx(bnfz,bnf,X,module,subgroup,vecMsup,vecWB,g,U);
+      finalresult = testx(&T,bnfz,bnr,X,subgroup,vecMsup,vecWB,g,gell,lW);
       if (finalresult) return gerepilecopy(av, finalresult);
         /* step 20,21,22 */
       i = dK;
