@@ -1,4 +1,4 @@
-/* $Id: subfield.c,v 1.34 2001/11/23 22:17:42 karim Exp $
+/* $Id: subfield.c,v 1.51 2002/06/09 20:18:18 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -24,6 +24,10 @@ extern GEN matrixpow(long n, long m, GEN y, GEN P,GEN l);
 extern GEN Fp_factor_irred(GEN P,GEN l, GEN Q);
 extern GEN FpX_rand(long d1, long v, GEN p);
 extern GEN hensel_lift_fact(GEN pol, GEN Q, GEN T, GEN p, GEN pe, long e);
+extern GEN ZX_disc_all(GEN x, ulong bound);
+extern GEN indexpartial(GEN P, GEN DP);
+extern GEN initgaloisborne(GEN T, GEN dn, long prec, GEN *ptL, GEN *ptprep, GEN *ptdis);
+extern long ZX_get_prec(GEN x);
 
 static GEN print_block_system(long N,GEN Y,long d,GEN vbs,long maxl);
 
@@ -83,7 +87,7 @@ calc_block(long N,GEN Z,long d,GEN Y,GEN vbs,ulong maxl)
       }
       if (dk == nn)
       {
-	long av=avma;
+	gpmem_t av=avma;
         int Z_equal_Zp = 1;
 
         for (j=1; j<lnon; j++) non[j]=0;
@@ -119,7 +123,8 @@ calc_block(long N,GEN Z,long d,GEN Y,GEN vbs,ulong maxl)
 static GEN
 potential_block_systems(long N, long d, GEN n, ulong maxl)
 {
-  long av=avma,r,i,j,k;
+  long r, i, j, k;
+  gpmem_t av=avma;
   GEN p1,vbs,Z;
 
   r=lg(n); Z=cgetg(r,t_VEC);
@@ -134,9 +139,10 @@ potential_block_systems(long N, long d, GEN n, ulong maxl)
 
 /* product of permutations. Put the result in perm1. */
 static void
-perm_mul(GEN perm1,GEN perm2)
+perm_mul_i(GEN perm1,GEN perm2)
 {
-  long av = avma,i, N = lg(perm1);
+  long i, N = lg(perm1);
+  gpmem_t av = avma;
   GEN perm=new_chunk(N);
   for (i=1; i<N; i++) perm[i]=perm1[perm2[i]];
   for (i=1; i<N; i++) perm1[i]=perm[i];
@@ -153,14 +159,14 @@ cycle_power_to_perm(GEN perm,GEN cy,long l)
   for (i=1; i<N; i++) perm[i] = i;
   if (lp)
   {
-    long av = avma;
+    gpmem_t av = avma;
     GEN p1 = new_chunk(N);
     b = cy[1];
     for (i=1; i<lcy; i++) b = (perm[b] = cy[i+1]);
     perm[b] = cy[1];
     for (i=1; i<N; i++) p1[i] = perm[i];
 
-    for (j=2; j<=lp; j++) perm_mul(perm,p1);
+    for (j=2; j<=lp; j++) perm_mul_i(perm,p1);
     avma = av;
   }
   return perm;
@@ -298,7 +304,7 @@ print_block_system(long N,GEN Y,long d,GEN vbs,long maxl)
   perm = cgetg(N+1,t_VEC); i=ns;
   do
   {
-    long av = avma;
+    gpmem_t av = avma;
     if (DEBUGLEVEL>5)
     {
       for (l=1; l<=ns; l++)
@@ -312,7 +318,7 @@ print_block_system(long N,GEN Y,long d,GEN vbs,long maxl)
     for (u=1; u<=N; u++) perm[u]=u;
     for (u=1; u<=ns; u++)
       for (v=1; v<=t[u]; v++)
-	perm_mul(perm, cycle_power_to_perm(cyperm,gmael(Z,u,v),e[u][v]));
+	perm_mul_i(perm, cycle_power_to_perm(cyperm,gmael(Z,u,v),e[u][v]));
     vbs = append_vbs(vbs, im_block_by_perm(D,perm));
     if (lg(vbs) > maxl) return vbs;
     avma = av;
@@ -375,7 +381,7 @@ cand_for_subfields(GEN A,GEN DATA,GEN *ptlistdelta)
 {
   long N,m,i,j,d,lf;
   GEN M,T,pe,p,pol,fhk,g;
-  GEN _d_1_term,delta,listdelta,whichdelta,firstroot;
+  GEN d_1_term,delta,listdelta,whichdelta,firstroot;
 
   pol=(GEN)DATA[1];
   p = (GEN)DATA[2];
@@ -389,12 +395,12 @@ cand_for_subfields(GEN A,GEN DATA,GEN *ptlistdelta)
   lf = lg(firstroot);
   listdelta = cgetg(lf, t_VEC);
   whichdelta = cgetg(N+1, t_VECSMALL);
-  _d_1_term = gzero;
+  d_1_term = gzero;
   for (i=1; i<=m; i++)
   {
     GEN Ai = (GEN)A[i], p1 = gun;
     for (j=1; j<=d; j++)
-      p1 = FpXQ_mul(p1, (GEN)fhk[Ai[j]], T,pe);
+      p1 = Fq_mul(p1, (GEN)fhk[Ai[j]], T,pe);
     delta[i] = (long)p1;
     if (DEBUGLEVEL>2) fprintferr("delta[%ld] = %Z\n",i,p1);
     /* g = prod (X - delta[i])
@@ -402,10 +408,10 @@ cand_for_subfields(GEN A,GEN DATA,GEN *ptlistdelta)
     /* fk[k] belongs to block number whichdelta[k] */
     for (j=1; j<=d; j++) whichdelta[Ai[j]] = i;
     if (typ(p1) == t_POL) p1 = constant_term(p1);
-    _d_1_term = addii(_d_1_term, p1);
+    d_1_term = addii(d_1_term, p1);
   }
-  _d_1_term = centermod(_d_1_term, pe); /* Tr(g) */
-  if (absi_cmp(_d_1_term, (GEN)M[3]) > 0) return gdeux; /* d-1 test */
+  d_1_term = centermod(d_1_term, pe); /* Tr(g) */
+  if (absi_cmp(d_1_term, (GEN)M[3]) > 0) return gdeux; /* d-1 test */
   g = FqV_roots_to_pol(delta, T, pe, 0);
   g = centermod(polsimplify(g), pe); /* assume g in Z[X] */
   if (DEBUGLEVEL>2) fprintferr("pol. found = %Z\n",g);
@@ -535,7 +541,7 @@ embedding_of_potential_subfields(GEN g,GEN DATA,GEN listdelta)
 {
   GEN TR,w0_Q,w0,w1_Q,w1,wpow,h0,gp,T,q2,q,p,ind,maxp,a;
   long rt;
-  ulong av;
+  gpmem_t av;
 
   T = (GEN)DATA[1]; rt = brent_kung_optpow(degpol(T), 2);
   p = (GEN)DATA[2];
@@ -596,14 +602,15 @@ embedding_of_potential_subfields(GEN g,GEN DATA,GEN listdelta)
 static GEN
 choose_prime(GEN pol,GEN dpol,long d,GEN *ptff,GEN *ptlistpotbl, long *ptlcm)
 {
-  ulong av;
-  long j,k,oldllist,llist,r,lcm,oldlcm,N,pp;
-  GEN p,listpotbl,oldlistpotbl,ff,oldff,n,oldn;
+  gpmem_t av;
   byteptr di=diffptr;
-
-  if (DEBUGLEVEL) timer2();
-  di++; p = stoi(2); N = degpol(pol);
-  while (p[2]<=N) p[2] += *di++;
+  long j,k,oldllist,llist,r,lcm,oldlcm,pp,minp, N = degpol(pol), m = N/d;
+  GEN p,listpotbl,oldlistpotbl,ff,oldff,n,oldn;
+ 
+  minp = N*(m-1)/2;
+  if (DEBUGLEVEL) (void)timer2();
+  di++; p = stoi(2);
+  while (p[2]<=minp) p[2] += *di++;
   oldllist = 100000;
   oldlcm = 0;
   oldlistpotbl = oldff = oldn = NULL; pp = 0; /* gcc -Wall */
@@ -637,13 +644,7 @@ choose_prime(GEN pol,GEN dpol,long d,GEN *ptff,GEN *ptlistpotbl, long *ptlcm)
     if (DEBUGLEVEL) msgtimer("#pbs = %ld",llist);
     av = avma;
   }
-  if (DEBUGLEVEL)
-  {
-    fprintferr("Chosen prime: p = %ld\n",pp);
-    if (DEBUGLEVEL>2)
-      fprintferr("Potential block systems of size %ld: %Z\n", d,oldlistpotbl);
-    flusherr();
-  }
+  if (DEBUGLEVEL) fprintferr("Chosen prime: p = %ld\n",pp);
   *ptlistpotbl=oldlistpotbl; *ptff=oldff; *ptlcm=oldlcm; return stoi(pp);
 }
 
@@ -719,7 +720,7 @@ interpol(GEN H, GEN T, GEN p)
   {
     d = constant_term(H[i]); /* -D[i] */
     p1 = FpXQX_mul(p1,gadd(X,d), T,p);
-    p2 = FpXQ_mul(p2, gadd(a,d), T,p);
+    p2 = Fq_mul(p2, gadd(a,d), T,p);
   }
   p2 = FpXQ_inv(p2, T,p);
   return FpXQX_FpXQ_mul(p1,p2, T,p);
@@ -880,7 +881,7 @@ _subfield(GEN g, GEN h)
 static GEN
 subfields_of_given_degree(struct poldata PD,long d)
 {
-  ulong av,av2;
+  gpmem_t av, av2;
   long llist,i,nn,v;
   GEN listpotbl,ff,A,CSF,ESF,LSB,p,T,DATA,listdelta;
   GEN pol = PD.pol, dpol = PD.dis;
@@ -921,7 +922,7 @@ CHANGE:
       LSB = gerepileupto(av2, concat(LSB, _subfield(CSF,ESF)));
     }
   }
-  delete_var();
+  (void)delete_var();
   for (i=1; i<llist; i++) free((void*)listpotbl[i]);
   free((void*)(listpotbl-1));
   if (DEBUGLEVEL) fprintferr("\nSubfields of degree %ld: %Z\n",d, LSB);
@@ -937,11 +938,6 @@ fix_var(GEN x, long v)
   return x;
 }
 
-extern GEN vandermondeinverseprep(GEN T, GEN L);
-extern GEN ZX_disc_all(GEN x, ulong bound);
-extern GEN indexpartial(GEN P, GEN DP);
-extern GEN initgaloisborne(GEN T, GEN dn, GEN *ptL, GEN *ptprep, GEN *ptdis, long *ptprec);
-
 void
 subfields_poldata(GEN T, struct poldata *PD)
 {
@@ -951,24 +947,33 @@ subfields_poldata(GEN T, struct poldata *PD)
   PD->pol = dummycopy(T); /* may change variable number later */
   if (nf)
   {
-    PD->den = (GEN)nf[4];
+    PD->den = Q_denom((GEN)nf[7]);
     PD->roo = (GEN)nf[6];
     PD->dis = mulii(absi((GEN)nf[3]), sqri((GEN)nf[4]));
   }
   else
   {
-    PD->den = initgaloisborne(T,NULL, &L,NULL,&dis,NULL);
+    PD->den = initgaloisborne(T,NULL,ZX_get_prec(T), &L,NULL,&dis);
     PD->roo = L;
     PD->dis = absi(dis);
   }
 }
 
+static long
+grouporder(GEN G)
+{
+  GEN Go = (GEN)G[2];
+  long i, l = lg(Go), o = 1;
+  for (i=1; i<l; i++) o *= Go[i];
+  return o;
+}
+
 GEN
 subfields(GEN nf,GEN d)
 {
-  ulong av = avma;
-  long di,N,v0;
-  GEN LSB,pol;
+  gpmem_t av = avma;
+  long di, N, v0;
+  GEN LSB, pol, G;
   struct poldata PD;
 
   pol = get_nfpol(nf, &nf); /* in order to treat trivial cases */
@@ -977,6 +982,24 @@ subfields(GEN nf,GEN d)
   if (di==1) return gerepilecopy(av, _subfield(polx[v0], pol));
   if (di < 1 || di > N || N % di) return cgetg(1,t_VEC);
 
+  /* much easier if nf is Galois (WSS) */
+  G = galoisconj4(nf? nf: pol, NULL, 1, 0);
+  if (typ(G) != t_INT)
+  { /* Bingo */
+    GEN L = galoissubgroups(G), F;
+    long k,i, l = lg(L), o = N/di;
+    F = cgetg(l, t_VEC);
+    k = 1;
+    for (i=1; i<l; i++)
+    {
+      GEN H = (GEN)L[i];
+      if (grouporder(H) == o)
+        F[k++] = (long) galoisfixedfield(G, (GEN)H[1], 0, v0);
+    }
+    setlg(F, k);
+    return gerepilecopy(av, F);
+  }
+
   subfields_poldata(nf? nf:pol, &PD);
   pol = PD.pol;
   setvarn(pol, 0);
@@ -987,11 +1010,19 @@ subfields(GEN nf,GEN d)
 static GEN
 subfieldsall(GEN nf)
 {
-  ulong av = avma;
-  long N,ld,i,v0;
-  GEN pol,dg,LSB,NLSB;
+  gpmem_t av = avma;
+  long N, ld, i, v0;
+  GEN G, pol, dg, LSB, NLSB;
   struct poldata PD;
 
+  /* much easier if nf is Galois (WSS) */
+  G = galoisconj4(nf, NULL, 1, 0);
+  if (typ(G) != t_INT)
+  {
+    pol = get_nfpol(nf, &nf);
+    return gerepileupto(av, galoissubfields(G, 0, varn(pol)));
+  }
+
   subfields_poldata(nf, &PD);
   pol = PD.pol;
 
@@ -1005,7 +1036,7 @@ subfieldsall(GEN nf)
     setvarn(pol, 0);
     for (i=2; i<ld; i++)
     {
-      ulong av1 = avma;
+      gpmem_t av1 = avma;
       NLSB = subfields_of_given_degree(PD, N / itos((GEN)dg[i]));
       if (lg(NLSB) > 1) LSB = concatsp(LSB,NLSB); else avma = av1;
     }
@@ -1021,20 +1052,3 @@ subfields0(GEN nf,GEN d)
   return d? subfields(nf,d): subfieldsall(nf);
 }
 
-/* irreducible (unitary) polynomial of degree n over Fp[v] */
-GEN
-ffinit(GEN p,long n,long v)
-{
-  long av = avma;
-  GEN pol;
-
-  if (n<=0) err(talker,"non positive degree in ffinit");
-  if (typ(p) != t_INT) err(typeer,"ffinit");
-  if (v<0) v = 0;
-  for(;; avma = av)
-  {
-    pol = gadd(gpowgs(polx[v],n), FpX_rand(n-1,v, p));
-    if (FpX_is_irred(pol, p)) break;
-  }
-  return gerepileupto(av, FpX(pol,p));
-}
